;Revision history
;ASB 9/14/93  Wrote first version, adapted from earlier version of omega_recon
;ASB 3/7/95   Added more options, made 1-d version
;ASB 6/27/96  Added keywords for centering
;ASB 6/28/96  Added keywords for zero-filling
;ASB 8/13/96  Added circularly symmetrical windows 
;ASB 2/27/98  Fixed bug in definition of type 3
;ASB 4/28/99  Fixed bug in generation of 2D filter 8
;             Added apd typ 9.
;ASB 11/18/99 Added code to shift center of echo. (Type = 5)
;             
;This program computes a filtr for image processing
;The input parameters are
;  apd   		filtr type
;  xres			size in x-direction
;  yres			size in y-direction
;  xsize                zero fill size in x-direction
;  ysize                zero fill size in y-direction
;The input keywords are:
;  type                
;                      0-normal,center at point xres/2,
;                      1-symmetrical, 
;                      2-single sided,
;                      3-normal, place center at first point
;                      4-circular symmetry
;                      5-Center at echo_pt
;The function returns a xres x yres floating point array.

function mkfilter,apd,xres,yrs,type=type,xsize=xsize,ysize=ysize,echo_pt=echo_pt

;Set default values for keywords
IF NOT keyword_set(type) THEN type=0
IF n_elements(yrs) NE 0 AND type NE 4 THEN yres=yrs
IF n_elements(xsize) EQ 0 THEN xsize=xres
IF n_elements(ysize) EQ 0 AND n_elements(yres) NE 0 THEN ysize=yres

CASE type OF
    0: BEGIN
        filtrx = (2.*findgen(xres)-xres)/float(xres)        
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = (2.*findgen(yres)-yres)/float(yres)        
        ENDIF
    END
    1: BEGIN
        filtrx = (2.*findgen(xres)-xres+1)/(xres-1.0)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = (2.*findgen(yres)-yres+1)/(yres-1.0)
        ENDIF
    END
    2: BEGIN
        filtrx = findgen(xres)/(xres-1)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = findgen(yres)/(yres-1)
        ENDIF
    END
    3: BEGIN
        filtrx = findgen(xres)/float(xres)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = findgen(yres)/float(yres)        
        ENDIF
    END
    4: BEGIN
        foo=lindgen(xres,xres)
        x=( (foo mod xres) - (xres-1)/2.)/(xres/2.)
        y=((foo / xres) - (xres-1)/2.)/(xres/2.)
        filtrx=sqrt(x^2+y^2)
        mask = where(filtrx GT 1.0)
    END 
    5: BEGIN
;I'm not sure the best way to handle the general case.  This code
;reduces to case 0 if echo_pt=xres/2 and to case 2 for echo_pt=0.
        IF n_elements(echo_pt) EQ 0 THEN BEGIN
            print,'echo_pt must be defined for type=5'
            RETURN,0
        ENDIF
        mx=max([echo_pt,xres-1-echo_pt])
        filtrx = (findgen(xres)-echo_pt)/mx
    END                         ;This option is for 1D filters only
ENDCASE
; make filter
IF n_elements(apd) GT 1 THEN test=apd(0) ELSE test=apd
CASE test OF
    0 : BEGIN                   ; rectangular window (This code does nothing )
        IF n_elements(yres) EQ 1 THEN BEGIN
            tmp = fltarr(xres,yres) 
            tmp(*,*) = 1
        ENDIF ELSE tmp=replicate(1,xres)
    END

    1 : BEGIN                   ; Bartlett (triangle) window
        filtrx = 1.0-abs(filtrx)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry=1.0-abs(filtry)
            tmp = filtrx # filtry
        ENDIF ELSE tmp=filtrx
    END

    2 : BEGIN                   ; Hann (squared cosine) window
        filtrx = 0.5*(1.0 + cos(!pi*filtrx))
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = 0.5*(1.0 + cos(!pi*filtry))
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    3 : BEGIN                   ; Hamming (raised cosine) window
        filtrx = 0.54 + 0.46*cos(!pi*filtrx)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = 0.54 + 0.46*cos(!pi*filtry)
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    4 : BEGIN                   ; Truncated Gaussian window (default, alpha=2.5)
        IF n_elements(apd) EQ 1 THEN bet = 3.125 ELSE bet = apd(1)*apd(1)/2.
        filtrx = exp(-bet*filtrx*filtrx)
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = exp(-bet*filtry*filtry)
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    5 : BEGIN                   ; Fermi window
        IF n_elements(apd) EQ 1 THEN BEGIN
            mu = 0.85
            bet = 0.05/alog(9)
        ENDIF ELSE BEGIN
            mu = apd(1)
            bet=(apd(2)>.01)/(2.*alog(9))
        ENDELSE
        ag = ((abs(filtrx)-mu)/bet) < 20.0
        filtrx = 1/(exp(ag)+1)
        IF n_elements(yres) EQ 1 THEN BEGIN
            ag = ((abs(filtry)-mu)/bet) < 20.0
            filtry = 1/(exp(ag)+1)
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    6 : BEGIN                   ; Chebyshev window, type = 0 only
        IF type NE 0 THEN BEGIN
            IF n_elements(filtry) EQ 0 THEN BEGIN
                RETURN,filtrx
            ENDIF ELSE RETRUN,filtrx#filtry
        ENDIF
        IF n_elements(apd) EQ 1 THEN beta = 50.0 ELSE beta=apd(1)
        alpha_x=cosh(invcosh(10^(beta/20.0))/(xres-1))

        filtrx=filtrx/2
        indexl=where(abs(alpha_x*cos(!pi*filtrx)) gt 1.0 ) ;central peak
        indexs=where(abs(alpha_x*cos(!pi*filtrx)) le 1.0 ) ;side lobes

        filtrx(indexl)=$        ;central peak
          cosh((xres-1)*invcosh(abs(alpha_x*cos(!pi*filtrx(indexl)))))/  $
          cosh((xres-1)*invcosh(alpha_x)) 

        filtrx(indexs)=$
          cos((xres-1)*acos(abs(alpha_x*cos(!pi*filtrx(indexs)))))/  $
          cosh((xres-1)*invcosh(alpha_x)) 

        filtrx=shift(fft(shift(filtrx,xres/2),-1),xres/2)

        IF n_elements(yres) EQ 1 THEN BEGIN
            alpha_y=cosh(invcosh(10^(beta/20.0))/(yres-1))

            filtry=filtry/2
            indexl=where(alpha_y*cos(!pi*filtry) GT 1.0 ) ;central peak
            indexs=where(alpha_y*cos(!pi*filtry) le 1.0 ) ;side lobes

            filtry(indexl)=$    ;central peak
              cosh( (yres-1)*invcosh( alpha_y*cos(!pi*filtry(index1)) ) )/  $
              cosh( (yres-1)*invcosh(alpha_y) ) 

            filtry(indexs)=$
              cos( (yres-1)*acos(alpha_y*cos(!pi*filtry(indexs)) ) )/  $
              cosh( (yres-1)*invcosh(alpha_y) ) 

            filtry=fft(filtry,-1)
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    7 : BEGIN                   ; exponential apodization
        IF n_elements(apd) EQ 1 THEN alpha=1 ELSE alpha=apd(1)
        filtrx=exp(-abs(filtrx*alpha))
        IF n_elements(yres) EQ 1 THEN BEGIN
            filtry = exp(-abs(filtry*alpha))
            tmp = filtrx # filtry 
        ENDIF ELSE tmp=filtrx
    END

    8 : BEGIN                   ;cosine window used 
        IF n_elements(apd) GT 1 THEN wb=apd(1) ELSE wb = 0.5
        ndex=where(abs(filtrx) LT wb)
        mdex = where(abs(filtrx) GE wb)
        IF ndex(0) NE -1 THEN filtrx(ndex)=1
        IF mdex(0) NE -1 THEN $
          filtrx(mdex)=abs(cos((filtrx(mdex)-wb)/(1.0-wb)*!pi/2))
        IF n_elements(yres) NE 0 AND type NE 4 THEN BEGIN
            ndex=where(abs(filtry) LT wb)
            mdex = where(abs(filtry) GE wb)
            IF ndex(0) NE -1 THEN filtry(ndex)=1
            IF mdex(0) NE -1 THEN $
              filtry(mdex)=abs(cos((filtry(mdex)-wb)/(1.0-wb)*!pi/2))
            tmp=filtrx#filtry
        ENDIF ELSE tmp=filtrx
    END

    9 : BEGIN                   ;cosine window used in csi reconstruction
        IF n_elements(apd) GT 1 THEN wb=apd(1) ELSE wb = 0.5
        ndex=where(abs(filtrx) LT wb)
        mdex = where(abs(filtrx) GE wb)
        IF ndex(0) NE -1 THEN filtrx(ndex)=1
        IF mdex(0) NE -1 THEN $
          filtrx(mdex)=0.5*(1+cos((filtrx(mdex)-wb)/(1.0-wb)*!pi))
        IF n_elements(yres) NE 0 AND type NE 4 THEN BEGIN
            ndex=where(abs(filtry) LT wb)
            mdex = where(abs(filtry) GE wb)
            IF ndex(0) NE -1 THEN filtry(ndex)=1
            IF mdex(0) NE -1 THEN $
              filtry(mdex)=0.5*(1+cos((filtry(mdex)-wb)/(1.0-wb)*!pi))
            tmp=filtrx#filtry
        ENDIF ELSE tmp=filtrx
    END

ENDCASE
IF n_elements(mask) GE 1 THEN tmp(mask)=0
xlim = min([xres,xsize])
sftx=((xres-xsize)/2) > 0
IF type EQ 4 THEN BEGIN
    filtr=fltarr(xsize,xsize)
    ylim=xlim
    sfty=sftx
    filtr(0:xlim-1,0:xlim-1)=tmp(sftx:sftx+xlim-1,sftx:sftx+xlim-1)
    return,filtr
ENDIF
IF n_elements(yres) EQ 0 THEN BEGIN ;1D
    filtr=fltarr(xsize)
;zero out first point to make psf real if zerofilling
    IF xsize GT xres AND type NE 1 THEN tmp(0)=0
    filtr(0:xlim-1)=tmp(sftx:sftx+xlim-1)
    IF type eq 3 THEN filtr=shift(filtr,-xlim/2)
ENDIF ELSE BEGIN                ;2D
    filtr=fltarr(xsize,ysize)
;zero out first point to make psf real if zerofilling
    IF xsize GT xres AND type NE 1 THEN tmp(0,*)=0
    IF ysize GT yres AND type NE 1 THEN tmp(*,0)=0
    ylim = min([yres,ysize])
    sfty=((yres-ysize)/2) > 0
    filtr(0:xlim-1,0:ylim-1)=tmp(sftx:sftx+xlim-1,sfty:sfty+ylim-1)
    IF type eq 3 THEN filtr=shift(filtr,-xlim/2,-ylim/2)
ENDELSE
RETURN,filtr
END
