/* EPIRRI.C                                                                 */
/*****************************************************************************
  This program is copyrighted worldwide by the John P. Robarts Research
  Institute.  Any distribution, copying or redistribution is expressly
  forbidden.

  This program converts a compressed multislice epi data set into a standard
  Varian data set by doing flashc, tabc, and epi processing all on the
  same file. The procpar is set right so that Vnmr and Stimulate can read
  the data, acquired with or without navigators. A file with info on the 
  navigators is saved for the specified slice if desired.

  The phase encode table must be of the form
  66alt2k, where the 2 denotes 2 segments and nv=66.

  To compile:
  gcc -O3 -o epirri epirri.c fourn.c shiftc.c  -lm

  History
  23/7/97 Written by Ravi S. Menon
  17/9/97 modified by Brad Goodyear to do interleaved acquisitions
  2/2/98  modified by Chris Thomas to write out nav-echo data for single
          slice experiments and the addition of flag to write out nav-echo data
  2/11/98 modified by Ravi Menon to explain arguments for argc < 2

******************************************************************************/

/* include library files for the program these allow compilation in RRI environment */
#include <sys/fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/file.h>
#include <sys/param.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#define pi 3.1415927

#define EPI_DONE   8		  /* set in main_header.spare1
				   * means already converted */
int nav_echo_data;
char output_nav[128];

struct datafilehead
{
   long    nblocks;		  /* number of blocks in file     */
   long    ntraces;		  /* number of traces per block   */
   long    np;			  /* number of elements per
				   * trace */
   long    ebytes;		  /* number of bytes per element  */
   long    tbytes;		  /* number of bytes per trace    */
   long    bbytes;		  /* number of bytes per block    */
   short   transf;		  /* transposed storage flag      */
   short   status;		  /* status of whole file         */
   long    spare1;		  /* reserved for future use      */
}       main_header;

struct datablockhead
{
   short   scale;		  /* scaling factor               */
   short   status;		  /* status of data in block      */
   short   index;		  /* block index                  */
   short   spare3;		  /* reserved for future use      */
   long    ctcount;		  /* completed transients in
				   * fids */
   float   lpval;		  /* left phase in phasefile      */
   float   rpval;		  /* right phase in phasefile     */
   float   lvl;			  /* level drift correction       */
   float   tlt;			  /* tilt drift correction        */
}       block_header;

union precision
{
   short  *sh;
   long   *lo;
   char   *ch;
}      *in_data,*out_data;

char    proc_string[5000];

int     nv,
        np,
        ns,
        nseg,
        nblocks,
        ntraces,
        ebytes,
        bbytes;

struct passwd *userpw;

void
main(argc, argv)
   int     argc;
   char   *argv[];
{

   extern int nav_echo_data;
   extern char output_nav[128];

   void    nav_echo_cor(),
           epi_phase_cor();

   int     n_nblocks,
           n_ntraces,
           n_bbytes;

   int     nav_echo,
           nav_echo_no,
           nav_echo_mag_cor,
           nav_echo_ph_cor,
           dc_cor,
           tmin,
           in_file,
           out_file,
           i, j, k, kk, l, m, n,
           malloc_size;

   int    *petable;

   char    string[100],
           string1[100];

   char    petable_name[128],
           input_fid[128],
           seqcon[128],
           output_fid[128],
           orig_fid[128],
           curexpdir[128],
	   intlv[128];

   char   *indata;

   long   *temp;

   FILE   *in_ptr,
          *out_ptr,
          *fopen();

   /* Parse the input line and make decisions or use defaults */
   if (argc < 2)
   {

      printf("\nUSAGE: epirri Input_file [Mag_cor Phase_cor DC_cor Print_nav_echo_data Slice_#] \n\n");
      printf("Mag_cor:     controls whether the navigator echo magnitude is used to correct amplitudes between segments. Default 1 (=YES) \n\n");
      printf("Phase_cor:     controls whether the navigator echo is used to adjust the phase between segments. Default 1 (=YES) \n\n");
      printf("DC_cor:     controls whether DC correction is done on the echos. Default 0 (=NO) \n\n");
      printf("Print_nav_echo_data:     controls whether a navigator echo file is written out. Default 0 (=NO) \n\n");
      printf("Slice#:     is the slice from which the navigator echo info is written out. Default 1 (=slice number 1) \n\n");
      exit(1);
   }
   printf("\nEPIRRI: Starting DATA INPUT...\n");

   strcpy(curexpdir, argv[1]);
   if (argc == 2)
   {
   nav_echo_mag_cor = 1; 
   nav_echo_ph_cor = 1; 
   dc_cor = 0;
   nav_echo_data = 0; 
   nav_echo_no = 1; 
   }
   if (argc > 2)
   {
   nav_echo_mag_cor = atoi(argv[2]); 
   nav_echo_ph_cor = atoi(argv[3]); 
   dc_cor = atoi(argv[4]); 
   nav_echo_data = atoi(argv[5]);
   nav_echo_no = atoi(argv[6]); 
   }
   printf("EPIRRI: processing with Mag_cor=%d, Phase_cor=%d, DC_cor=%d, Print_nav_echo_data=%d, Slice_#=%d \n",nav_echo_mag_cor, nav_echo_ph_cor, dc_cor, nav_echo_data, nav_echo_no);

   /* Generate path to the input and output fid files */
   strcpy(input_fid, curexpdir);
   strcpy(output_fid, curexpdir);
   strcpy(output_nav, curexpdir);
   strcat(input_fid, "/fid");
   strcat(output_fid, "/epirri.temp");
   strcat(output_nav, "/nav");

   /* Open input fid file */
   if ((in_file = open(input_fid, O_RDONLY)) < 0)
   {
      perror("Open");
      printf("EPIRRI: can't open input file %s.\n", input_fid);
      exit(2);
   }

   /* Read main_header */
   malloc_size = sizeof(main_header);
   if (read(in_file, &main_header, malloc_size) != malloc_size)
   {
      perror("read");
      printf("EPIRRI: can't open input file %s header.\n", input_fid);
      exit(4);
   }

   /* See if data has already been processed */
   if (main_header.spare1 & EPI_DONE)
   {
      printf("EPIRRI: File already processed by EPI\n");
      exit(1);
   }

   /* assign the important header values */
   np = main_header.np;
   nblocks = main_header.nblocks;
   ntraces =  main_header.ntraces;
   ebytes = main_header.ebytes;
   bbytes = main_header.bbytes;

   /* read some parameters from procpar file */
   strcpy(input_fid, curexpdir);
   strcat(input_fid, "/procpar");
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("EPIRRI: Can't open input procpar file %s/procpar.\n",input_fid);
      exit(8);
   }

   while (fgets(proc_string, 5000, in_ptr) != NULL)
   {
      if (!strncmp(proc_string, "nv ", 3))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %d\n", &nv);
      }
      if (!strncmp(proc_string, "intlv ", 6))
      { 
         fgets(proc_string, 5000, in_ptr);
         sscanf(proc_string, "1 %s\n", intlv);
         for (i = 0; i < strlen(intlv) - 2; i++)
            intlv[i] = intlv[i + 1];
         intlv[i] = '\0';
      } 
      if (!strncmp(proc_string, "seqcon", 6))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %s\n", seqcon);
	 for (i = 0; i < strlen(seqcon) - 2; i++)
	    seqcon[i] = seqcon[i + 1];
	 seqcon[i] = '\0';
      }
      if (!strncmp(proc_string, "ns", 2))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %d\n", &ns);
      }
      if (!strncmp(proc_string, "petable", 5))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %s\n", petable_name);
	 for (i = 0; i < strlen(petable_name) - 2; i++)
	    petable_name[i] = petable_name[i + 1];
	 petable_name[i] = '\0';
	 j = strlen(petable_name);
	 if (petable_name[j - 1] == 'k')
	    nseg = petable_name[j - 2] - '0';
	 else
	    nseg = 1;
	 printf("EPIRRI: Data acquired in %d segment(s)\n", nseg);
      }
   }
   fclose(in_ptr);

   /* Figure out if a navigator echo was used */
   if ((nv % 32) == 0)
       nav_echo = 0;
   else
       nav_echo = 1;
   printf("EPIRRI: Data acquired with %d navigator(s)\n", (nv % 32));

   /* Read in the phase-encoding table */
   petable = (int *) malloc(nv * sizeof(int));
   userpw = (struct passwd *) getpwuid(getuid());
   strcpy(input_fid, userpw->pw_dir);
   strcat(input_fid, "/vnmrsys/tablib/");
   strcat(input_fid, petable_name);
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("EPIRRI: Can't open petable %s. Check your tablib.\n", petable_name);
      exit(8);
   }
   fgets(string, 100, in_ptr);
   tmin = 0;
   for (l = 0; l < nv; l++)
   {
      if (fscanf(in_ptr, "%d", petable + l) != 1)
      {
	 perror("fscanf");
	 printf("EPIRRI: %s length inconsistent with data.\n", petable_name);
	 exit(6);
      }
      if (tmin > petable[l])
	 tmin = petable[l];
   }
   fclose(in_ptr);

   /* Arrange petable to be all positive  */ 
   if (nav_echo == 1)
   {
   tmin = 0;
      for (i=0; i<nseg; i++)
      {
         for (k=1; k < nv/nseg; k++)
         {
            l = k + i * nv / nseg;
            if (tmin > petable[l])
	       tmin = petable[l];
         }
      }       
   }
   for (l = 0; l < nv; l++)
      petable[l] -= tmin;

   /* memory allocation for the data */
   out_data = (union precision *) malloc(nblocks * ns * nv * sizeof(union precision));
   in_data = (union precision *) malloc(nblocks * ns * nv * sizeof(union precision));
   for (i = 0; i < nblocks * nv * ns; i++)
   {
      out_data[i].ch = (char *) malloc(np * sizeof(long));
      in_data[i].ch = (char *) malloc(np * sizeof(long));
   }
   temp = (long *) malloc(np * sizeof(long));
  
   /* check for the proper data acquisition mode */
   if ((seqcon[1] == 's' && ns > 1) || seqcon[2] != 'c')
   {
      printf("EPIRRI: Data acquired without compression...can't process \n");
      exit(1);
   }

   /* Define new header values */
   bbytes = main_header.bbytes;
   n_ntraces = 1;
   n_nblocks = nblocks * ns * nv;
   n_bbytes = np * main_header.ebytes + 28;

   /* Allocate space for one input trace */
   if ((indata = (char *) malloc(ntraces * np * main_header.ebytes)) == NULL)
   {
      perror("malloc");
      printf("EPIRRI: uncompression routine: unable to malloc space for input data.\n");
      exit(8);
   }

   /* Uncompress data and time reverse */
   m = 0;
   printf("EPIRRI: Uncompressing and time reversing \n");
   for (i = 0; i < nblocks; i++)
   {
      /* Read in data and block_headers */
      if (read(in_file, &block_header, 28) != 28)
      {
	 perror("read");
	 printf("EPIRRI: unable to read data block headers.\n");
	 exit(10);
      }
/*
      printf("DATA INPUT> Reading image %d/%d\n", i + 1, nblocks);
*/
      if ((j = read(in_file, indata, bbytes - 28)) != bbytes - 28)
      {
	 perror("read");
	 printf("EPIRRI: uncompression routine: Error reading original data block #%d! \n", j);
	 abort(1);
      }
      for (k = 0; k < ns; k++)
      {
	 for (l = 0; l < nseg; l++)
	 {
	    for (j = 0; j < nv / nseg; j++)
	    {
               m = j + l * nv / nseg + k * nv + i * ns * nv;
	       for (n = 0; n < main_header.np * main_header.ebytes; n++)
               {
                 if (intlv[0] == 'y') {
		  in_data[m].ch[n] = indata[(j + l*nv*ns/nseg + k*nv/nseg) * main_header.tbytes + n];
                  }
                 else
		  {
                  in_data[m].ch[n] = indata[(j + l*nv/nseg + k*nv) * main_header.tbytes + n];
                  }
               }

   	       if (main_header.ebytes == 2)
   	       {
      			for (kk = main_header.np - 1; kk >= 0; kk--)
                        {
	    	   	   in_data[m].lo[kk] = in_data[m].sh[kk];
                        }
   	       }
               
                /*  time reversal  */
		if ((j % 2) == 1)
               	{
               		for (kk= 0; kk< main_header.np; kk += 2)
               		{
                  	temp[kk] = in_data[m].lo[kk];
                  	temp[kk+1] = in_data[m].lo[kk+1];
               		}
               		for (kk = 0; kk < main_header.np; kk += 2)
               		{
                  	in_data[m].lo[np - 2 - kk] = temp[kk];
                  	in_data[m].lo[np - 1 - kk] = temp[kk+1];
               		}
           	}
	    }
	 }
      }
   }

   /* Update the main header with the new values */
   main_header.bbytes = n_bbytes;
   main_header.ntraces = n_ntraces;
   main_header.nblocks = n_nblocks;

   /* Update procpar with the new values */
   strcpy(input_fid, curexpdir);
   strcat(input_fid, "/procpar");
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("EPIRRI: Can't open input procpar file %s.\n", input_fid);
      exit(8);
   }

   /* Open "new" procpar file for writing */
   strcpy(input_fid, curexpdir);
   strcat(input_fid, "/procpar.new");
   if ((out_ptr = fopen(input_fid, "w")) == NULL)
   {
      perror("Open");
      printf("EPIRRI: Can't open output procpar file %s.\n", input_fid);
      exit(8);
   }

   /* Read in the "old" procpar file, line by line. copy to and 
      update "new" procpar as necessary */ 
   while (fgets(proc_string, 5000, in_ptr) != NULL)
   {
      if (!strncmp(proc_string, "seqcon", 6))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 fprintf(out_ptr, "1 \"nscnn\" \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "ns", 2))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 1 \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "pss", 3))
      {
	 j = 0;
	 for (i = 0; i < strlen(proc_string); i++)
	 {
	    if (j != 6)
	    {
	       if (proc_string[i] == ' ')
		  j++;
	    } else
	    {
	       proc_string[i] = '2';
	       break;
	    }
	 }
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "ni", 2))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", nv - (nav_echo*nseg));
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "nf", 2))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 1 \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "celem", 5))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j * ns * nv);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "arrayelem", 9))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 999, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j + 1);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "arraydim", 8))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j*ns*(nv - (nav_echo*nseg)));
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "array ", 6))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %s", &i, string);
	 k = strlen(string);
	 if (k < 3)
	    fprintf(out_ptr, "1 \"\"\n");
	 else
	 {
	    for (i = 0; i < k - 1; i++)
	       string1[i] = string[i];
	    string1[k - 1] = '\0';
	    strcat(string1, ",pss\"");
	    fprintf(out_ptr, "1 %s \n", string1);
	 }
	 fgets(proc_string, 5000, in_ptr);
      }
      fputs(proc_string, out_ptr);
   }
   fclose(in_ptr);
   fclose(out_ptr);

   /* Now replace the unmodified procpar with the new procpar */
   sprintf(string, "cp %s/procpar %s/procpar.orig", curexpdir, curexpdir);
   system(string);
   sprintf(string, "mv -f %s/procpar.new %s/procpar", curexpdir, curexpdir);
   system(string);

   printf("EPIRRI: DATA INPUT done !\n");

   /* phase  correction */
   epi_phase_cor();

   /* nav echo data correction */
   if (nav_echo_mag_cor == 1)  
      nav_echo_cor(nav_echo_no,nav_echo_ph_cor,dc_cor);

   /* tabc done here  */
   printf("\nEPIRRI: RESORTING PE ORDER (tabc)...\n");
   for (i = 0; i < nblocks; i++)
   {
	for (l = 0; l < ns; l++)
	{
            for (kk = 0; kk < nseg; kk ++)
            {
		for (k = nav_echo; k < nv/nseg; k ++)
		{
			m = l + i * ns + petable[k + kk*nv/nseg] * ns * nblocks;
			for (n = 0; n < main_header.np; n++)
			  out_data[m].lo[n] = in_data[k + kk*nv/nseg + l*nv
+ i*ns*nv].lo[n];
		}
             }
          }
   }
   printf("EPIRRI: RESORTING PE ORDER done !\n");

   /* Reset nv to be the true number of phase encodes */ 
   if (nav_echo == 1) 
   {  
     nv -= nseg;
     main_header.ntraces = 1;
     main_header.nblocks = nblocks*nv*ns;
   }  
 
   /* if single precision, shift the data into shorts */ 
   printf("EPIRRI: Shifting data into short words.\n");
   if (main_header.ebytes == 2)
   {
      for (k = 0; k < nblocks * nv * ns; k++)
      {
         for (i = 0; i < np; i++)
         {
            out_data[k].sh[i] = out_data[k].lo[i];
         }  
      }   
   } 

   /* Open output fid file */
   if ((out_file = open(output_fid, O_CREAT | O_TRUNC | O_WRONLY, 0755)) < 0)
   {
      perror("open");
      printf("EPIRRI: unable to open output file %s.\n",output_fid);
      exit(14);
   }
   main_header.spare1 |= EPI_DONE;
   if (write(out_file, &main_header, sizeof(main_header)) < 0)
   {
      perror("write");
      printf("EPIRRI: unable to write main output data header.\n");
      exit(15);
   }
   block_header.index = 0;
   for (i = 0; i < ns * nblocks * nv; i++)
   {
      block_header.index++;
      if (write(out_file, &block_header, 28) != 28)
      {
	 perror("write");
	 printf("EPIRRI: unable to write output data block header.\n");
	 exit(10);
      }
      if (write(out_file, out_data[i].ch, main_header.tbytes) != main_header.tbytes)
      {
	 perror("write");
	 printf("EPIRRI: unable to write data block.\n");
	 exit(10);
      }
   }

   /* Close input fid file */
   if (close(in_file))
   {
      perror("close");
      printf("EPIRRI: problems closing input file.\n");
      exit(13);
   }
   /* Close output fid file */
   if (close(out_file))
   {
      perror("close");
      printf("EPIRRI: problems closing output file.\n");
      exit(18);
   }
  
   /* move fid file to fid.orig and move epirri.temp file to fid file */ 
   strcpy(orig_fid, curexpdir);
   strcpy(input_fid, curexpdir);
   strcat(input_fid, "/fid");
   strcat(orig_fid, "/fid.orig");
   if (rename(input_fid, orig_fid) != 0)
   {
      perror("rename");
      printf("EPIRRI: unable to rename original fid file.\n");
      exit(19);
   }
   if (rename(output_fid, input_fid) != 0)
   {
      perror("rename");
      printf("EPIRRI: unable to rename new fid file.\n");
      exit(20);
   }
   sprintf(string, "rm %s/procpar.orig", curexpdir);
   system(string);
   sprintf(string, "rm %s/fid.orig", curexpdir);
   system(string);
   printf("\nEPI finished successfully\n");
   exit(0);
}


void
epi_phase_cor()
{
   register int i, j, k, m, l;
   int     nimages;
   int     nn[3];
   float  *ftdata,
          *ftdata1;
   float   abs_temp,
          *sin_cphase,
          *cos_cphase;
   long   *temp;
   float   slpos;
   FILE   *in_ptr,
          *out_ptr;

   char    string[100];
   void    fourn(),
           shiftc();

   printf("\nEPI_PHASE_COR: Starting PHASE CORRECTION...\n");
/*
   printf("EPI_PHASE_COR: nblock: %d ntrace: %d  np: %d \n", nblocks, ntraces, np);
*/

   /* Decide how many images and slices were in the series
      and decide how many output fids will be written */
   nimages = ntraces * nblocks / (nv * ns);
/*
   printf("EPI_PHASE_COR: nv: %d nimages: %d  nslices: %d \n", nv, nimages, ns);
*/
   printf("EPI_PHASE_COR: %d slices to process\n", ns);

   /* Allocate space for the processing */

   ftdata = (float *) malloc(np * sizeof(float));
   ftdata1 = (float *) malloc(np * sizeof(float));
   sin_cphase = (float *) malloc(np * sizeof(float) * nv);
   cos_cphase = (float *) malloc(np * sizeof(float) * nv);

   temp = (long *) malloc(np * sizeof(long));

   for (l = 0; l < ns; l++)
   {
   printf("EPI_PHASE_COR: Correcting slice %d \n",l+1);

      /****
      * Cycle through each of the input reference fids.
      * and calculate the correction phase.
      ****/
      nn[0] = np / 2;
	 for (i = 0; i < nv; i++)
	 {
	    for (j = 0; j < np; j += 2)
	    {
	       ftdata[j] = in_data[i + l * nv].lo[j];
	       ftdata[j + 1] = in_data[i + l * nv].lo[j + 1];
	    }

	    shiftc(ftdata, nn, 1);
	    fourn(ftdata - 1, nn - 1, 1, 1);
	    shiftc(ftdata, nn, 1);

	    for (j = 0; j < np; j += 2)
	    {
	       abs_temp = (float) sqrt((double) ftdata[j + 1] * (double) ftdata[j + 1] + (double) ftdata[j] * (double) ftdata[j]);
	       if (abs_temp > 0.0)
	       {
		  sin_cphase[i * np + j] = -ftdata[j + 1] / abs_temp;
		  cos_cphase[i * np + j] = ftdata[j] / abs_temp;
	       } else
	       {
		  sin_cphase[i * np + j] = 0.0;
		  cos_cphase[i * np + j] = 0.0;
	       }
	    }
	 }

      /* Now fill the output buffer */
      for (k = 0; k < nimages; k++)
      {
	 for (i = 0; i < nv; i++)
	 {
	    for (j = 0; j < np; j += 2)
	    {
	       ftdata[j] = in_data[i + k * ns * nv + l * nv].lo[j];
	       ftdata[j + 1] = in_data[i + k * ns * nv + l * nv].lo[j + 1];
	    }

	    shiftc(ftdata, nn, 1);
	    fourn(ftdata - 1, nn - 1, 1, 1);
	    shiftc(ftdata, nn, 1);
	    for (j = 0; j < np; j += 2)
	    {
	       ftdata1[j] = cos_cphase[i * np + j] * ftdata[j] - sin_cphase[i * np + j] * ftdata[j + 1];
	       ftdata1[j + 1] = sin_cphase[i * np + j] * ftdata[j] + cos_cphase[i * np + j] * ftdata[j + 1];
	    }

	    shiftc(ftdata1, nn, 1);
	    fourn(ftdata1 - 1, nn - 1, 1, -1);
	    shiftc(ftdata1, nn, 1);

	    for (j = 0; j < np; j += 2)
	    {
	       in_data[i + k * ns * nv + l * nv].lo[j] = ftdata1[j];
	       in_data[i + k * ns * nv + l * nv].lo[j + 1] = ftdata1[j + 1];
	    }
	 }
      }
   }

   free(ftdata);
   free(ftdata1);
   free(sin_cphase);
   free(cos_cphase);

   printf("EPI_PHASE_COR: PHASE CORRECTION DONE \n");
}

void
nav_echo_cor(no_echo,ph_cor,dc_cor)
   int    ph_cor,dc_cor,no_echo;
{
   register int i, j, k, m, l;
   int     nimages;
   int     nn[3];
   float  *ftdata,
          *ftdata1;
   float  *mag_temp,
          *phase_temp,
          *mag_ref,
           phase_diff,
           mag_temp_av,
           mag_ratio_av,
           mag_ref_av,
           dc_real,
           dc_imag,
          *phase_ref,
          *sin_cphase,
          *cos_cphase;
   long   *temp;
   float   slpos;
   FILE   *in_ptr,
          *out_ptr,
          *fp1;

   char    string[100];
   void    fourn(),
           shiftc();

   extern int nav_echo_data;
   extern char output_nav[128];

   printf("\nNAV_ECHO_COR: Starting NAVIGATOR CORRECTION...\n");

   /* Decide how many images and slices were in the series
      and decide how many output fids will be written */

   nimages = ntraces * nblocks / (nv * ns);

   /* Allocate space for the processing */

   ftdata = (float *) malloc(np * sizeof(float));
   ftdata1 = (float *) malloc(np * sizeof(float));
   sin_cphase = (float *) malloc(np * sizeof(float));
   cos_cphase = (float *) malloc(np * sizeof(float));

   mag_ref = (float *) malloc(np * sizeof(float));
   phase_ref = (float *) malloc(np * sizeof(float));

   mag_temp = (float *) malloc(np * sizeof(float));
   phase_temp = (float *) malloc(np * sizeof(float));

   temp = (long *) malloc(np * sizeof(long));
   dc_real = 0.;
   dc_imag = 0.;
   if (dc_cor == 1)
   {
     printf("NAV_ECHO_COR: Starting DC CORRECTION...\n");
     k = 0;
     for (l = 0; l < nv*ns*nimages; l++)
     {
	    for (j = 0; j < 8; j += 2)
	    {
	        dc_real += in_data[l].lo[j];
	        dc_imag += in_data[l].lo[j + 1];
                k += 1;
            }
	    for (j = np-8; j < np-1; j += 2)
	    {
	        dc_real += in_data[l].lo[j];
	        dc_imag += in_data[l].lo[j + 1];
                k += 1;
            }
     }
     dc_real /= (float)k;
     dc_imag /= (float)k;
     for (l = 0; l < nv*ns*nimages; l++)
     {
	    for (j = 0; j < np; j += 2)
	    {
	        in_data[l].lo[j] -= dc_real;
	        in_data[l].lo[j + 1] -= dc_imag;
            }
     }
   printf("NAV_ECHO_COR: DC CORRECTION done !\n");
   } 
   if (nav_echo_data == 1)
   {
      sprintf(string, "%d.dat",no_echo);
      strcat(output_nav, string);
      fp1 = fopen(output_nav, "w+");
   }
   nn[0] = np / 2;
   for (l = 0; l < ns*nimages; l++)
   {
/*
      printf("\nNAV_ECHO_COR> Working on Image %d/%d\n", l + 1, ns*nimages);
*/

      /****
      * Cycle through each of the input reference fids.
      * and calculate the correction phase.
      ****/

          /* get reference  magnitude and phase  */
         mag_ref_av = 0.;
	 for (j = 0; j < np; j += 2)  mag_ref[j] = 0.0;

	 for (i = 0; i < nseg; i++)
	 {
	    for (j = 0; j < np; j += 2)
	    {
	       ftdata[j] = in_data[i*nv/nseg + l*nv].lo[j];
	       ftdata[j + 1] = in_data[i*nv/nseg + l*nv].lo[j + 1];
            }

	    shiftc(ftdata, nn, 1);
	    fourn(ftdata - 1, nn - 1, 1, 1);
	    shiftc(ftdata, nn, 1);
           
            /* calculate  magnitude  */ 
	    for (j = 0; j < np; j += 2)
	    {
	       mag_ref[j] += (float) sqrt((double) ftdata[j + 1] * (double) ftdata[j + 1] + (double) ftdata[j] * (double) ftdata[j]);
	    }
 
            mag_ref_av += mag_ref_av;
         }
            /* calculate  avearage magnitude  */ 
	    for (j = 0; j < np; j += 2)
	    {
	       mag_ref[j] /= nseg;
               mag_ref_av += mag_ref[j];
	    }
            mag_ref_av /= (np/2.);

       /* reference for each segment  */
	 for (i = 0; i < nseg; i++)
	 {
            mag_temp_av = 0.;
	    for (j = 0; j < np; j += 2)
	    {
	       ftdata[j] = in_data[i*nv/nseg + l*nv].lo[j];
	       ftdata[j + 1] = in_data[i*nv/nseg + l*nv].lo[j + 1];
            }

	    shiftc(ftdata, nn, 1);
	    fourn(ftdata - 1, nn - 1, 1, 1);
	    shiftc(ftdata, nn, 1);

	    for (j = 0; j < np; j += 2)
	    {
	       mag_temp[j] = (float) sqrt((double) ftdata[j + 1] * (double) ftdata[j + 1] + (double) ftdata[j] * (double) ftdata[j]);

               mag_temp_av += mag_temp[j];

               if (mag_temp[j] > 0.0)
	         phase_temp[j] = atan2((double)ftdata[j + 1],(double)ftdata[j]);
               else
                 phase_temp[j] = 0.;
            }

            mag_temp_av /= (np/2.);
            if ( ((l % ns) == (no_echo-1)) && (nav_echo_data == 1) )
	    {
               fprintf(fp1,"%d \t %d \t %f \t %f \t %f \t %f\n",(int)(l/ns),i+1,mag_temp[np/2],phase_temp[np/2],ftdata[np/2],ftdata[np/2+1]);
            }

            mag_ratio_av = mag_ref_av/mag_temp_av;

         /* calculate magnitude and phase difference if average SNR is greater than 5  */
            for (j = 0; j < np; j += 2)
            {
              phase_diff = 0.;
              if (mag_temp[j] >= (mag_temp_av/1.1))
                 mag_temp[j] = mag_ref[j]/mag_temp[j];
              else
                 mag_temp[j] = mag_ratio_av;

              /* correct phase to 0 degree   */
	         phase_diff = -phase_temp[j];  

              sin_cphase[j] = (float) sin((double)phase_diff); 
              cos_cphase[j] = (float) cos((double)phase_diff); 
            }

           for (k = 0; k < nv/nseg; k++)
           {
	      for (j = 0; j < np; j += 2)
	      {
	       ftdata[j] = in_data[i*nv/nseg + k + l*nv].lo[j];
	       ftdata[j + 1] = in_data[i*nv/nseg + k + l*nv].lo[j + 1];
               }

	    shiftc(ftdata, nn, 1);
	    fourn(ftdata - 1, nn - 1, 1, 1);
	    shiftc(ftdata, nn, 1);

	    for (j = 0; j < np; j += 2)
	    {

             /* phase correction or not   */
              if (ph_cor == 1)
              {
	       ftdata1[j] = (cos_cphase[j] * ftdata[j] - sin_cphase[j] * ftdata[j + 1])*mag_temp[j];
	       ftdata1[j + 1] = (sin_cphase[j] * ftdata[j] + cos_cphase[j] * ftdata[j + 1])*mag_temp[j];
              }
              else
              {
               ftdata1[j] = ftdata[j]*mag_temp[j];
               ftdata1[j+1] = ftdata[j+1]*mag_temp[j];
              }
            } 
	    shiftc(ftdata1, nn, 1);
	    fourn(ftdata1 - 1, nn - 1, 1, -1);
	    shiftc(ftdata1, nn, 1);

	    for (j = 0; j < np; j += 2)
	    {
	       in_data[i * nv / nseg + k + l*nv].lo[j] = ftdata1[j];
	       in_data[i * nv / nseg + k + l*nv].lo[j+1] = ftdata1[j+1];
	    }
	   }
        }
   }
   if (nav_echo_data == 1)
   {
      fclose(fp1);
   }
   free(ftdata);
   free(ftdata1);
   free(mag_temp);
   free(phase_temp);
   free(mag_ref);
   free(phase_ref);
   free(sin_cphase);
   free(cos_cphase);
   printf("NAV_ECHO_COR: NAVIGATOR CORRECTION Done !\n");
}
