; This program performs a linear fit to the phase of the echo
; data. It uses a robust algorithm for resolving branch cuts of the
; arctan function.

; The input parameter is:
; echoes(npts,nviews)   complex array containing the navigator data

; The output parameters are:
; phase0(nviews)        floating point array of zero order phase shifts
; phase1(nviews)        floating point array of first order phase shifts

pro phase_unwrap, echoes, phase0, phase1, phase_echoes

debug_flag=0

;Compute zero order correction
sz=size(echoes)
xres=sz(1)
middle=xres/2
nviews=sz(2)
phase0=fltarr(nviews)
phase1=fltarr(nviews)
ft_echoes=complexarr(xres,nviews)
del_phase_max=!pi/4

;Take FT of echoes
FOR kk=0,nviews-1 DO BEGIN ;compare other view to reference view
    ft_echoes(*,kk)=shift(fft(shift(echoes(*,kk),middle),1),middle)
ENDFOR

;unwrap phase
xt=indgen(xres)-middle		; x index for fitting
yfit=fltarr(xres,nviews)	; array to hold the fitted curve

ref=nviews-1  ; define the reference echo
tmp=reform(ft_echoes(*,ref))#replicate(1,nviews) ; array replicating ref echo
phase_echoes=ft_echoes/tmp	; calculate phase difference wrt ref echo

;find phase at location of maximum magnitude
mx=max(abs(ft_echoes(*,ref)), ind)	; find maximum amplitude and its index
ind=ind(0) ; in case it finds more than one max
poff=replicate(1,xres)#(ft_echoes(ind,*)/ft_echoes(ind,ref))
phase_echoes=phase_echoes/poff          ; subtract off the phase of the most
                                        ; intense point 
; take the arctan
tmp_phase=acos(float(phase_echoes)/abs(phase_echoes))
phase_echoes=tmp_phase*2.0*((imaginary(phase_echoes) gt 0.0)-0.5)  

; calculate the phase derivative
del_phase=shift(phase_echoes,-1, ref)-phase_echoes
wt0=abs(ft_echoes(*,ref))^2	; weights for phases
wt1=shift(wt0,-1)		; weights for phase differences is
index=where(wt0 LT wt1)         ; the smaller of the weights of the terms
if (index(0) NE -1) THEN wt1(index)=wt0(index)           ; in the difference

;Don't include points where the phase varies too quickly
wt1=wt1#replicate(1,nviews)
index=where(abs(del_phase) GT del_phase_max)
if(index(0) NE -1 ) THEN wt1(index)=0  ; set weighting to zero in places where
                                       ; the phase derrivative is very large

;Compute the 1st order phase shift via the mean of the phase diff
av_phd=total(wt1*del_phase,1)/total(wt1,1)
av_phd=reform(av_phd)
yfit0=(xt-xt(ind))#av_phd

index=where(phase_echoes-yfit0 GT !pi)
WHILE index(0) NE -1 DO BEGIN
    phase_echoes(index)=phase_echoes(index)-2*!pi
    index=where(phase_echoes-yfit0 GT !pi)
ENDWHILE

index=where(phase_echoes-yfit0 LT -!pi)
WHILE index(0) NE -1 DO BEGIN
    phase_echoes(index)=phase_echoes(index)+2*!pi
    index=where(phase_echoes-yfit0 LT -!pi)
ENDWHILE

; put back on the offset phase (which must be first calculated)
tmp_poff=acos(float(poff)/abs(poff))
poff=tmp_poff*2.0*((imaginary(poff) gt 0.0)-0.5)  
phase_echoes=phase_echoes+poff  

;Do weighted linear fit to phases
FOR kk=0, nviews-1 DO BEGIN
    ww=wt0
    ;Make sure residual never exceeds pi
    tmp=polyfitw(xt,phase_echoes(*,kk),ww,1,yf)
    ig=where(phase_echoes(1:xres-1,kk)-yf GT !pi)
    il=where(phase_echoes(1:xres-1,kk)-yf LT -!pi)
    WHILE (ig(0) NE -1) OR (il(0) NE -1) DO BEGIN
        IF (ig(0) NE -1) THEN phase_echoes(ig+1,kk)=phase_echoes(ig+1,kk)-2*!pi
        IF (il(0) NE -1) THEN phase_echoes(il+1,kk)=phase_echoes(il+1,kk)+2*!pi
        tmp=polyfitw(xt,phase_echoes(*,kk),ww,1,yf)
        ig=where(phase_echoes(1:xres-1,kk)-yf GT !pi)
        il=where(phase_echoes(1:xres-1,kk)-yf LT -!pi)
    ENDWHILE
    yfit(*,kk)=yf
    phase0(kk)=tmp(0)
    phase1(kk)=tmp(1)
    IF (debug_flag eq 1) THEN begin
        plot, phase_echoes(*,kk)
        oplot, yfit(*,kk)
        wait, 1
    endif
ENDFOR

yfit=0
END    
