
#define MAX_SINE_RES 20000  /* maximum resolution of a sine wave */

/*==========================================================================*/
double create_sine_gradient(grad_strength,fstem,cycle_duration,grate,
   num_cycles,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate a sinusoid waveform shape, given a desired period,
     gradient slew rate (grate in units of usec per g/cm) and a target
     gradient strength (grad_strength in units of g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD.

     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gro")

     cycle_duration     desired duration of the sinusoid

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)

     num_cycles         number of sinusoid cycles
 
     write_flag         only write waveform to disk if write_flag = 1
 
     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the entire sinusoidal waveform shape.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, September 2000
                                                                            */
/*==========================================================================*/
double grad_strength,grate;
char fstem[];
double cycle_duration, num_cycles;
unsigned long write_flag;
{
  double sine_res, num_sine_steps, num_cycle_steps, temp_dac;
  double actual_update;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i;
  FILE *fp=NULL;
     
  /* first check that a sine of stated amplitude and duration can be
     achieved */
  if (grad_strength*2.0*PI/cycle_duration > 1.0/grate) {
    printf("Error: gradient slew rate exceeded in create_sine_gradient()\n");
    abort(1);
  }

  num_sine_steps=rint(cycle_duration*num_cycles*1.0e6/GRAD_UPDATE_TIME);
  num_sine_steps=(num_sine_steps > MAX_SINE_RES) ? 
             MAX_SINE_RES : num_sine_steps;

  num_cycle_steps=rint(num_sine_steps/num_cycles);

  sine_res=rint(cycle_duration*num_cycles*1.0e6/WAVEFORM_RESOL/num_sine_steps);
  /* calculate the actual ramp resolution in us */
  actual_update=sine_res*WAVEFORM_RESOL/1.e6;
  if (actual_update < GRAD_UPDATE_TIME/1.0e6) {
    actual_update = GRAD_UPDATE_TIME/1.0e6;
  }

  if (cycle_duration > MAX_GRAD_DURATION) {
    printf("Error: Gradient duration too long in create_sine_gradient()\n");
    exit(0);
  }

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }  
  }

  temp_dac=0.0;

  /* sinusoid waveform */
  for (i=0; i<(int)(num_sine_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=sin((float)(i+1)/num_cycle_steps*2.0*PI)*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
  return(num_sine_steps*actual_update);
 
}
