/*
  ss_asl_epi.c
  adapted from 'myepi.c' 

  FMRIB Centre, University of Oxford

  myepi.c by Stuart Clare, April 1999
  added to by Patricia, 27.4.99  
  further adapted by PJ, March 2000
  hideously mutated by karla, spring 2001

*/

/*

  { [ASL] [crush] [Fatsat] [Excitation] [flowsuppress] [Spinecho]  [EPI] [crush] } 

  |___tasl___|__________ti__________|__________te__________|
  
  |____________________________________tr______________________________________|

*/

/*
  Real-time variables
  v1: phase cycling
  v2: phase cycling
  v3: phase cycling
  v4: phase cycling
  v5: number of pe steps
  v6: pe step number
  v7: pe increment
  v8: number of segments
  v9: segment number
  v10:
  v11:
  v12: rfmod power
  v13: number of slices
  v14: slice number
*/ 

/*
  RF Pulses
  p1   : tpwr1   : p1pat  : gss  - Excitation
  p2   : tpwr2   : p2pat  : gss2 - Spin echo
  p3   : tpwr3   : p3pat  : gss3 - Inversion
  p4   : tpwr4   : p4pat         - Fatsat
  pasl : tpwrasl : CW     : gasl - ASL
*/

/*
  Tables
  t1 : tpwr1 + rfmod
  t2 : tpwr2 + rfmod
  t3 : tpwr3 + rfmod
  t4 : tpwr4 + rfmod
  t5 : tpwr5 + rfmod
  t6 : linear modulation
*/

#include "standard.h"
#include "group.h"
#include "sequence_utilities.h"
#define SAMPPTS 8192

/* Global variables*/
double grise,grate,tep,del_seq,gbdf;
double epigrof,epigroa,episwing,epigssf;
double pw_gss,pw_gss2;
double gssr,pw_gssr,tref;
double gss3,pw_gss3;
double pw_gcrush,pw_gfat;
double pw_gasl;
double num_ints,num_segs;
double pw_gror,tref2,gbd,sgbd;
double pw_gro,sgbpe;
double etl,eff_echo,echo_del;
double sampwait;
char   gcoil[MAXSTR];
double samptime[SAMPPTS];

/* Flags */
char spinecho[MAXSTR],fat_sat[MAXSTR],half_kspace[MAXSTR];
char inversion[MAXSTR],crushers[MAXSTR],nonlinsamp[MAXSTR];
char rfmod[MAXSTR];

/* CASL params */
double asldelay;
/* double switchdelay; */
char asl[MAXSTR];
char aslconcat[MAXSTR];
char aslcrush[MAXSTR],flowsuppress[MAXSTR];

#include "globals.h"
#include "playout.h"
#include "calculate.h"
#include "tweekers.h"
#include "casl_utilities.h"
#include "ss_asl_epi_safety_calc.h"

pulsesequence()
{
  double t_inv_b,t_inv_a;
  double t_presat_b,t_presat_a;
  double t_cr,t_fat;
  double t_ex_b,t_ex_a;
  double t_se_b,t_se_a;
  double t_epi_b,t_epi_a;
  double t_flowsup,t_delay,t_aslcr;
  double t_asl;
  double min_te,min_tr,min_ti;
  
  double tr_delay,ti_delay,te_delay,se_delay_b,se_delay_a;

  double prescan_flag,pown;
  int image;
  char err_string[MAXSTR];

  initparms_sis();
  get_tweekers();

  /* 180 degree pulse gradient strength */
  gss2 = getval("gss2");
  /* Inversion recovery gradient strength */
  gss3 = getval("gss3");

  /* Trigger and counter bit pattern */
  image = (int)(getval("image"));
  /* bit 1 = wait for ticks triggers before beginning scan */
  /* bit 2 = puts out high on sp3 for duration of volume */
  /* bit 3 = epi blips off (set bit on), otherwise blips on */
  /* bit 4 = make inversion non slice selective (on) or slice selective (off)*/

  /* Get flags 'y/n' */
  getstr("spinecho",spinecho);  
  getstr("inversion",inversion);  
  getstr("crushers",crushers);  
  getstr("fat_sat",fat_sat);
  getstr("half_kspace",half_kspace);
  getstr("nonlinsamp",nonlinsamp);
  getstr("rfmod",rfmod);
  getstr("asl",asl); /* CASL (CW) inversion pulse */
  getstr("aslcrush",aslcrush); /* post-CW crusher */
  getstr("aslconcat",aslconcat); /* flag for concatenated asl pulse */
  getstr("flowsuppress",flowsuppress); /* diffusion grad for flow suppression */

  /* CASL associated delays */
  /* time to switch between pulsed & CW mode */
  switchdelay = getval("switchdelay");
  /* delay between end of CW pulse & beginning of excitation pulse */
  asldelay = getval("asldelay"); 

  if (presat) {
    presat_off = getval("presat_off");
    gsssat = getval("gsssat");
    sat_psi = getval("sat_psi");
    sat_phi = getval("sat_phi");
    sat_theta = getval("sat_theta");
    sat_thk = getval("sat_thk");
    calculate_spatial_saturation_gradients(&t_presat_b,&t_presat_a);
  }

  /* Number of interleaves */
  num_ints = getval("num_ints");  
  if (num_ints == 0) num_ints = 1;

  /* Half k-space stuff */
  if (half_kspace[0] == 'y') {
    
    if (num_ints > 1) {
      sprintf(err_string,"%s: Half K space acquisition requires 1 interleaf\n"
	      ,seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    
    pown = 2;
    do{
      if(pown == nv){
        sprintf(err_string,"%s: nv must be a few lines greater than n^2\n"
	        ,seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      eff_echo = nv - pown;
      pown *= 2;
    }while(pown < nv);
    etl = nv;
  }
  else {
    etl = (double)((int)(nv/num_ints));  /* echo train length/interleaf */
    eff_echo = etl/2.0;  /* echo position of TE */
  }

  /* Crusher gradient strength */
  gcrush=gmax/2.0;

  /* Gradient calculations */
  if(inversion[0]=='y')calculate_inversion_gradients(&t_inv_b,&t_inv_a);
  else { t_inv_b=0.0; t_inv_a=0.0; }

  if(crushers[0]=='y')calculate_crusher_gradients(&t_cr);
  else t_cr=0.0;

  if(fat_sat[0]=='y')calculate_fatsat_gradients(&t_fat);
  else t_fat=0.0;

  calculate_excitation_gradients(&t_ex_b,&t_ex_a);

  if(spinecho[0]=='y')calculate_spinecho_gradients(&t_se_b,&t_se_a);
  else { t_se_b=0.0; t_se_a=0.0; }

  calculate_epi_gradients(&t_epi_b,&t_epi_a);

  if(rfmod[0]=='y') calculate_rf_modulation();

  if(asl[0]=='y') calculate_casl_gradients(&t_asl);
  else { t_asl=0.0; switchdelay=0.0; }

  if(aslcrush[0]=='y') calculate_crusher_gradients(&t_aslcr);
  else t_aslcr=0.0;

  if (aslconcat[0]=='n' && asldelay<switchdelay) {
    t_delay=switchdelay-t_aslcr;
    printf("%s: asldelay (%f) shorter than switchdelay (%f). Delaying %f.\n",
	   seqfil,asldelay,switchdelay,switchdelay);
  }
  else t_delay=asldelay-t_aslcr;

  if(flowsuppress[0]=='y') calculate_flowsuppress(&t_flowsup);
  else t_flowsup=0.0;


  /* Timing calculations */

  /* Nonlinear sampling */
  if(nonlinsamp[0]=='y') calculate_nonlinear_sampling();

  /* Spinecho delay */
  se_delay_b = se_delay_a = 0.0;
  if(spinecho[0]=='y'){
    if( (t_ex_a+t_se_b) > (t_se_a+t_epi_b) )
      se_delay_a=(t_ex_a+t_se_b)-(t_se_a+t_epi_b);
    if( (t_ex_a+t_se_b) < (t_se_a+t_epi_b) )
      se_delay_b=(t_se_a+t_epi_b)-(t_ex_a+t_se_b);
  }

  /* te delay */
  min_te = t_ex_a + t_flowsup + se_delay_b + t_se_b + t_se_a + se_delay_a + t_epi_b;
  if(min_te>te){
    sprintf(err_string,"%s: te too short. Min te = %f\n",seqfil,min_te);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  te_delay = te - min_te;

  /* ti delay */
  if(inversion[0]=='y'){
    min_ti = t_inv_a + t_fat + t_ex_b;
    if(min_ti>ti){
      sprintf(err_string,"%s: ti too short. Min ti = %f\n",seqfil,min_ti);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    ti_delay = ti - min_ti;
  }
  else ti_delay=0;

  /* tr delay */
  min_tr = t_asl + t_delay + (t_inv_b + t_inv_a + ti_delay + t_flowsup + 
			      t_fat + t_ex_b + te + t_epi_a + t_cr ) * ns; 
  if (presat) min_tr = min_tr + (t_presat_b + t_presat_a) * ns;
  if(min_tr>tr){
    sprintf(err_string,"%s: tr too short. Min tr = %f\n",seqfil,min_tr);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  tr_delay = (tr - min_tr)/ns;

  /* Prescan flag (1=yes, 0=no) */
  prescan_flag = getval("prescan_flag"); 
  if (prescan_flag == 1)num_segs=1;
  else num_segs=(int)(nv/etl);
  
  /* Turn blips off */
  if ((image & 4) == 4) {
    sgbpe = 0.0;
    gbd=0.0;
    sgbd=0.0;
  }

  /* Safety check */
  if (ss_asl_epi_safety_calc() == -1) {
    sprintf(err_string,"%s: Failed duty cycle safety check\n",
	    seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
 
  /**********************************************************************/
  /* PULSE SEQUENCE */
  /**********************************************************************/

  /* Phase cycling */
  mod2(ct,v3);                    /* v3 = 0101 0101 */
  dbl(v3,v3);                     /* v3 = 0202 0202 */
  hlv(ct,v4);                     /* v4 = 0011 2233 */
  mod2(v4,v4);                    /* v4 = 0011 0011 */
  add(v3,v4,v1);                  /* v1 = 0213 0213 */
  assign(v1,oph);                 /* oph = 0213 0213 */
  add(one,v1,v2);

  /* phase encode increment size */
  assign(one,v7);  
  
  /* Receiver setup */
  status(A);
  rcvroff();
  rcvron();

  initval(num_segs,v8);  /* set v8 to the num of echoes/shot */

  /* wait for start scan button */
  if ((image & 1) == 1) xgate(ticks);
  if ((image & 2) == 2) sp3on(); else sp3off();
  delay(del_seq);
  
  /* loop v8 (num_ints) times using counter v9 */
  loop(v8,v9);

  /* CASL pulse */
  if(asl[0] == 'y') {
    if (aslconcat[0] == 'y') casl_pulse_concat();
    else casl_pulse_cw();
  }

  /* post-CW crusher */
  if(aslcrush[0] == 'y') crusher_gradients();

  delay(t_delay);
  
  /* sets v13 to ns and incs v14 up to it */
  msloop(seqcon[1],ns,v13,v14);

  /* Inversion pulse */
  if(inversion[0] == 'y') {
    if((image & 8) == 8) {
      nonselective_inversion_pulse();
    }
    else {
      selective_inversion_pulse(v14);
    }
    delay(ti_delay);
  }

  if (fat_sat[0] == 'y') fatsat_pulse();

  if (presat) spatial_saturation_pulse();

  excitation_pulse(v14);

  if (flowsuppress[0] == 'y' && spinecho[0] == 'y')
       flowsuppress_gradients(1.0);
  if (flowsuppress[0] == 'y' && spinecho[0] != 'y') {
       flowsuppress_gradients(1.0);
       flowsuppress_gradients(-1.0);
  }

  delay(te_delay/2.0);

  if (spinecho[0] == 'y'){
    delay(se_delay_b);
    spinecho_pulse(v14);
    delay(se_delay_a);
  }

  delay(te_delay/2.0);

  if (flowsuppress[0] == 'y' && spinecho[0] == 'y')
       flowsuppress_gradients(1.0);

  epi_gradients_and_acquire();

  crusher_gradients();

  /* double check that all gradients are zero */
  zero_all_gradients();

  delay(tr_delay);
  
  /* increment the msloop counter v14 */
  endmsloop(seqcon[1],v14);

  /* increment the interleaves loop counter v9 */
  endloop(v9);
}
