/***********************************************************************
    EDDY_TEST.C

    eddy current gradient test sequence.
    Peter Jezzard, Oxford Sept 1997

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

#include "standard.h"
#include "group.h"
#include "sequence_utilities.h"
extern double rint();

pulsesequence()
{
    /* INTERNAL VARIABLE DECLARATIONS *********************/
    int     image=0;
    double  nucgamma, grad_del, pre_acq_del, seqtime, postdelay;
    double  gread, agread, gphase, agphase, gslice, agslice, grise;
    double  pw_ramp, max_grad, grate, gdp_del, gread_rd, gread_pe, gread_sl;
    double  pw_gread;
    double  create_ramp_gradient();
    char shape_name[MAXSTR];

    initparms_sis();

    /*************************************************************************/
    /* get the global parameters */
    /*************************************************************************/
    P_getreal(GLOBAL,"grise",&grise,1); /* grad risetime to full strength */
    if ((grise < 0.0) || (grise > 0.002)) {
      printf("OX_EPI: impropper or no grise time defined in global file\n");
      abort(1);
    }
    /*************************************************************************/

    grad_del = getval("grad_del"); /* duration of gradient pulse */
    pre_acq_del = getval("pre_acq_del"); /* pre acquisition delay */
    gread = getval("gread"); /* gradient strength in read direction */
    gphase = getval("gphase"); /* gradient strength in phase direction */
    gslice = getval("gslice"); /* gradient strength in slice direction */
    image = getval("image"); /* useful counter */

    grate = grise/gmax; /* switching rate of gradients [sec/(gauss/cm)] */

    nucgamma = 1e6*sfrq/B0;  /* nucleus gamma in Hz/gauss */
                             /* derived from sfrq in MHz and B0 in gauss */

    agread = fabs(gread); /* absolute read gradient strength */
    agphase = fabs(gphase); /* absolute phase encode gradient strength */
    agslice = fabs(gslice); /* absolute slice select gradient strength */

    gread_rd=0.0;
    gread_pe=0.0;
    gread_sl=0.0;

    if (fabs(gread) > 0.0) {
      gread_rd = np/2.0/at/32.0/nucgamma;  /* FOV fixed at 32 cm */
    }
    if (fabs(gphase) > 0.0) {
      gread_pe = np/2.0/at/32.0/nucgamma;  /* FOV fixed at 32 cm */
    }
    if (fabs(gslice) > 0.0) {
      gread_sl = np/2.0/at/32.0/nucgamma;  /* FOV fixed at 32 cm */
    }

    if ((fabs(gread) > 0.0) && (fabs(gphase) > 0.0)) {
      printf("EDDY_TEST: Can't have more than one gradient on\n");
      abort(1);
    }

    if ((fabs(gread) > 0.0) && (fabs(gslice) > 0.0)) {
      printf("EDDY_TEST: Can't have more than one gradient on\n");
      abort(1);
    }

    if ((fabs(gphase) > 0.0) && (fabs(gslice) > 0.0)) {
      printf("EDDY_TEST: Can't have more than one gradient on\n");
      abort(1);
    }

    max_grad = fabs(gread) > fabs(gphase) ? fabs(gread) : fabs(gphase);
    max_grad = max_grad > fabs(gslice) ? max_grad : fabs(gslice);

    sprintf(shape_name,"gramp");
    pw_ramp=create_ramp_gradient(max_grad,shape_name,1,grate,ix);
    pw_ramp=create_ramp_gradient(max_grad,shape_name,-1,grate,ix);

    max_grad = fabs(gread_rd)>fabs(gread_pe) ? fabs(gread_rd) : fabs(gread_pe);
    max_grad = max_grad>fabs(gread_sl) ? max_grad : fabs(gread_sl);

    sprintf(shape_name,"gread");
    pw_gread=create_ramp_gradient(max_grad,shape_name,1,grate,ix);
    pw_gread=create_ramp_gradient(max_grad,shape_name,-1,grate,ix);

    gdp_del=at/2.0-pw_gread/2.0;

  /* CALCULATE RELAXATION DELAY FOR SEQUENCE */
    seqtime = (0.0005 + 2*pw_ramp + grad_del + pre_acq_del +
         pw + rof1 + rof2 + 2*pw_gread + gdp_del + 2*pw_gread + at); 
    if (tr - seqtime < 0.0) {
          printf("EDDY_TEST: Requested tr too short.  Min tr = %f\n",seqtime);
          abort(1);
    }
    postdelay = (tr - seqtime);

    status(A);
    sp4off();
    rcvroff();
    rcvron();

    /* PULSE SEQUENCE FOR CHECKING EDDY CURRENTS */
    sp1on();
    delay(0.0005);   /* 500 us scope trigger */
    sp1off();

    /* induce the eddy current using a large amplitude gradient pulse */

    if ((fabs(gread) > 0.0) || (fabs(gphase) > 0.0) || (fabs(gslice) > 0.0)) {
      obl_shapedgradient("gramp_up","gramp_up","gramp_up",pw_ramp,
          gread,gphase,gslice,1,WAIT);
      delay(grad_del);
      obl_shapedgradient("gramp_dn","gramp_dn","gramp_dn",pw_ramp,
          gread,gphase,gslice,1,WAIT);
    }

    delay(pre_acq_del);

    /* 90 degree pulse ************************************/ 
    sp4on();
    pulse(pw,oph);  /* implicitly uses tpwr for power level */
    sp4off();

    if (fabs(gread_rd) > 0.0) {
      obl_shapedgradient("gread_up","","",pw_gread,-gread_rd,0.0,0.0,1,WAIT);
      delay(gdp_del);
      obl_shapedgradient("gread_dn","","",pw_gread,-gread_rd,0.0,0.0,1,WAIT);
      obl_shapedgradient("gread_up","","",pw_gread,gread_rd,0.0,0.0,1,WAIT);
    }
    if (fabs(gread_pe) > 0.0) {
      obl_shapedgradient("","gread_up","",pw_gread,0.0,-gread_pe,0.0,1,WAIT);
      delay(gdp_del);
      obl_shapedgradient("","gread_dn","",pw_gread,0.0,-gread_pe,0.0,1,WAIT);
      obl_shapedgradient("","gread_up","",pw_gread,0.0,gread_pe,0.0,1,WAIT);
    }
    if (fabs(gread_sl) > 0.0) {
      obl_shapedgradient("","","gread_up",pw_gread,0.0,0.0,-gread_sl,1,WAIT);
      delay(gdp_del);
      obl_shapedgradient("","","gread_dn",pw_gread,0.0,0.0,-gread_sl,1,WAIT);
      obl_shapedgradient("","","gread_up",pw_gread,0.0,0.0,gread_sl,1,WAIT);
    }
    acquire(np, 1.0/sw);
    if (fabs(gread_rd) > 0.0) {
      obl_shapedgradient("gread_dn","","",pw_gread,gread_rd,0.0,0.0,1,WAIT);
    }
    if (fabs(gread_pe) > 0.0) {
      obl_shapedgradient("","gread_dn","",pw_gread,0.0,gread_pe,0.0,1,WAIT);
    }
    if (fabs(gread_sl) > 0.0) {
      obl_shapedgradient("","","gread_dn",pw_gread,0.0,0.0,gread_sl,1,WAIT);
    }

    zero_all_gradients();  /* double check that all gradients are off */

    delay(postdelay);

}
