/***********************************************************************
    EPISS.C

    Peter Jezzard, National Institues of Health

    Blipped spin/gradient echo-planar imaging sequence.
    Simplified for single shot epi

    Modified from EPISS.C written by S. Sukumar


*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
     _                       __                                 __     _
    |                       |                                     |     |
RF  | -----[alpha]----------|---------------- |** ACQ **|---------|     |
    |                       |                                     |     |
    |          < - - - - - -|- -esp - - - - - - - >               |     |
    |        ____           |                                     |     |
GSS | ______/////\     _____|_____________________________________|     |
    | <---->      \\\\/ gssr|                                     |     |
    |   predelay  <--> tref |                                     | etl | ns
    |                       |                                     |     |
    |                       |                                     |     |
GPE |  _____________________|______________/\_____________________|     |
    |                       |              \/                     |     |
    |                       |                                     |     |
    |         gror          |                  _________          |     |
GRO |  ___________     _____|_________________//////////\_________|     |
    |             \\\\/     |                                     |     |
    |_                      |__                                 __|    _|

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

#include "standard.h"
#include "group.h"

pulsesequence()
{
    /* INTERNAL VARIABLE DECLARATIONS *********************/
    int     image;
    double  sedelay,etl,predelay,seqtime;
    double  espmin,tblipc;
    double  agss,agro1,gpes,grate,gssrint,grorint,gpeint;
    double  nucgamma;
    double  grora;
    double  eff_echo,gpeb,rgss,rgro,rgpes,rgpeb,rgssr,rgror;
    double  ted,ted1,ted2,tevs,tods,tep;

    double tssr,trod,tpedc;
    double groa, gro1,agror,agssr,agpes,agpeb;
    double invtime;
    char   spinecho[MAXSTR];

    initparms_sis();
    tpedc = getval("tped");     /* duration of phase encode dephase grad */
    tep = getval("tep")/1e6;    /* correction for echo position , usec->sec */
    tblipc = getval("tblip");   /* This value is not used by the sequence */
    eff_echo = getval("eff_echo"); /* echo position num for effective echo */
    getstr("spinecho",spinecho);  /* spinecho flag 'y'=se. 'n'=ge. */
    image = getval("image"); /* image is a flag that collects a ref when 0 */
    groa = getval("groa");  /* read dephase tweaker */
    grora = getval("grora");  /* epi read gradient tweaker */

    if(tep <0)    /* timing correction for echo positioning */
    {
	tods = -tep;
	tevs = 0;
    }
    else
    {
	tevs = tep;
        tods = 0;
    }

    /* grate is the maximum slew rate of the gradients */
    grate = trise/gmax;

    /* temporarily set the read dephase time to zero */
    trod = 0;   /* time for read out dephaser */

    /* set the echo train length to be equal to the number of veiws, nv */
    etl = nv;   /* echo train length */

    /* grof is a fractional scaling factor that can be used to tweak the read
       gradient dephaser gradient strength (cf gssf for slice direction) */
    if (grof == 0.0  ||  !var_active("grof",CURRENT) )
      grof = 1.0;

    nucgamma = 1e6*sfrq/B0;  /* nucleus gamma in Hz/gauss */
                             /* derived from sfrq in MHz and B0 in gauss */

    /* dwell = (at)/(np/2) ; thus sw(Hz) = (np)/(2*at) */
    /* read gradient strength (Hz/cm) = sw/lro = (np)/(2*at*lro) */
    gro1 = np/2.0/at/lro/nucgamma;  /* note read gradient calculated here */

    /* read dephasing gradient strength */
    gror = gro1;

    agss = fabs(gss); /* absolute slice select gradient strength */
    agro1 = fabs(gro1); /* absolute read gradient strength */
    gror = gror*grof;  /* multiple by any scaling correction factor */
    agror = fabs(gror); /* absolute read dephaser gradient strength */

    rgro = grate*agro1; /* time to reach read grad strength */
    rgror = grate*agror; /* time to reach read dephaser grad strength */
    rgss = grate*agss;  /* time to reach slice select gradient strength */

  /* GRADIENT*TIME INTEGRALS FOR SLICE SELECTION, PHASE AND READOUT */
    grorint = agro1*(at/2.0 + rgro/2.0);  /* 0.5 * epi read dephase integral */
    gssrint = agss*(p1/2.0 + rof2 + rgss/2.0); /* p1 slice select integral */
    gpeint = 1.0 / (lpe*nucgamma); /* phase encode integral */

    trod = grorint/agror;   /* assume constant gradient */
    trod = trod - (rgror);  /* correct for ramps */

    gssr = gss;  /* set the slice refocus gradient to be the same as gss */
    gssr = gssr*gssf; /* multiply by any tweaker scaling factor gssf */
    agssr = fabs(gssr); /* absolute value of slice refocus gradient */
    rgssr = grate*agssr; /* time to reach slice refocus gradient strength */

    tssr = gssrint/agssr;  /* slice refocussing time for constant pulse */
    tssr = tssr - rgssr;  /* correct for ramps */

    tblipc = rgro; /* temporarily set tblipc (time for the blips) to rgro */
    /* the blips are set up to be not quite triangular waveforms. The
       attack duration is rgpes up to an amplitude of gpes, followed by
       a flat portion of duration tblipc (to take the blip to a time
       coincident with the top of the read gradient) and then a return to
       zero amplitude taking a duration rgpes */
    gpes = gpeint/tblipc;   /* for constant pulse */
    agpes = fabs(gpes); /* absolute amplitude of phase encode blips */
    rgpes = grate*agpes; /* time to reach blip amplitude (should be rgro/2) */
    tblipc = tblipc - rgpes;  /* correct tblipc time for read attack ramps */
    if (tblipc < 0.0) {
        printf("SS_EPI: Blip gradients too severe. increase phase FOV\n");
        abort(1);
    }
    
    /* calculate the values for the phase encode dephaser */
    gpeb = gpeint*eff_echo/tpedc;  /* assume constant pulse */
    agpeb = fabs(gpeb); /* absolute amplitude of phase encode dephaser grad */
    rgpeb = grate*agpeb;  /* time to reach phase encode dephaser amplitude */
    tpedc = tpedc-rgpeb;   /* correct timing for ramps */

  /* CALCULATE PHASE REFOCUSSING TO CORRECT FOR FREQUENCY JUMPS */

  /* RELAXATION DELAY */
    invtime = ti + rgss + tspoil + p2/2;

    /* calculate the minimum echo to echo interval possible */
    espmin = ((tevs+tods)*0.5) + rgro + at + rgro;  

    /* calculate the sequence time (from beginning of 90 to end of seq) */
    seqtime = rgss + rof1 + p1 + rof2 + rgss +
	        rgssr + tssr + rgssr +
	        rgpeb + tpedc + rgpeb +
	        rgror + trod + rgror + d1 +
		espmin*etl;

    if (spinecho[0] == 'y') 
    {
        /* calc second half echo time (from middle of 180 to max echo) */
        ted2 = p2/2.0 + rof2 + tspoil + rgss + (espmin*eff_echo)+ d1;
        /* calc first half before a filler has been added */
        ted1 = p1/2.0 + rof2 + rgss + 
                  rgssr + tssr + rgssr +
                  rgpeb + tpedc + rgpeb +
                  rgror + trod + rgror; 
        /* ted is the filler time added to give TE/2 at either side */
        ted = ted2 - ted1;  /* buffer delay if spin echo */
	sedelay = ted + rgss + tspoil + rof1 + p2 + rof2 + tspoil + rgss; 
	seqtime = seqtime + sedelay;

    }
    else
    {
	ted = 0;
        /* invert the phase encode and read dephaser gradients if s.e. */
	gpeb = -gpeb;
	gror = -gror;
    }

    if (ir[0] == 'y') 
	seqtime = seqtime + invtime;

    if (tr - seqtime < 0.0) {
        printf("SS_EPI: Requested tr too short.  Min tr = %f\n",seqtime);
        abort(1);
    }
    predelay = (tr - seqtime);

    /* if image = 0 then collect a reference image with the p.e. grads off */
    if (image == 0)
    {
      gpeb = 0.0;
      gpes = 0.0;
    }
    
/***
        printf("gro1: %7.4f gss: %7.4f gpes: %7.4f\n",gro1,gss,agpes);
        printf("gror: %7.4f gssr: %7.4f gpeb: %7.4f\n",gror,gssr,agpeb);
        printf("trod: %7.4f tssr: %7.4f tpe: %f \n",trod,tssr,tpe);
***/

  /* PULSE SEQUENCE *************************************/
    /* Relaxation delay ***********************************/       
    status(A);

    /* Phase cycle: Part I ********************************/
    mod2(ct,v3);                    /*0101 0101*/
    dbl(v3,v3);                     /*0202 0202*/
    hlv(ct,v4);                     /*0011 2233*/
    mod2(v4,v4);                    /*0011 0011*/
    add(v3,v4,v1);                  /*0213 0213*/
    assign(v1,oph);
    add(one,v1,v2);

      assign(zero,v10);  /* set v10 to zero to start */
      assign(zero,v12);  /* set v12 to zero to start */
      initval(etl,v11);  /* set v11 to etl to start */

        delay(predelay);
        poffset(pss[0],gss);

        /* Gating ***************************************/
        xgate(ticks);

	if (ir[0] == 'y')
	{
        	/* 180 degree RF pulse ********************************/
        	obspower(tpwr2);
        	obl_gradient(0.0,0.0,gss);
        	delay(rgss);
		delay(tspoil);
        	shapedpulse(p2pat,p2,v2,rof1,rof2);
        	delay(tspoil);
        	zero_all_gradients();
        	delay(rgss);
		delay(ti-tspoil-rgss-p2/2);
	}
       
        /* 90 degree pulse ************************************/ 
        observepower(tpwr1);
        obl_gradient(0.0,0.0,gss);
        delay(rgss);
        shapedpulse(p1pat,p1,v1,rof1,rof2);
        zero_all_gradients();
	delay(rgss);

        /* Read & Slice refocus and phase encode pulse ********/
        obl_gradient(0.0,0.0,-gssr);
        delay(rgssr);
	delay(tssr);
        zero_all_gradients();
        delay(rgssr);

        obl_gradient(0.0,gpeb,0.0);
        delay(rgpeb);
	delay(tpedc);
        zero_all_gradients();
        delay(rgpeb);

        obl_gradient(-gror+grora,0.0,0.0);
        delay(rgror);
	delay(trod);
        zero_all_gradients();
        delay(rgror);

	if (spinecho[0] == 'y')
	{
		delay(ted);
        	/* 180 degree RF pulse ********************************/
        	obspower(tpwr2);
        	obl_gradient(0.0,0.0,gss);
        	delay(rgss);
		delay(tspoil);
        	shapedpulse(p2pat,p2,v2,rof1,rof2);
        	delay(tspoil);
        	zero_all_gradients();
        	delay(rgss);
	}

	delay(d1);
        peloop(seqcon[2],etl,v5,v6);
	  incr(v10);
          mod2(v10,v12);

          obsoffset(resto);

         /* Acquire echo ***************************************/
          ifzero(v12);
            obl_gradient(gro1+groa,gpes,0.0);
	    delay(rgpes+tblipc);  /* rgpes+tblipc=rgro */
            obl_gradient(gro1+groa,0.0,0.0); /* remain at gro1, ramp down gpes */
	    delay(tods);  /* odd echo tweaker */
          elsenz(v12);
            obl_gradient(-gro1,gpes,0.0);
	    delay(rgpes+tblipc);  /* rgpes+tblipc=rgro */
            obl_gradient(-gro1,0.0,0.0); /* remain at -gro1, ramp down gpes */
	    delay(tevs);  /* even echo tweaker */
          endif(v12);
          acquire(np,1.0/sw);  /* collect np points */
          zero_all_gradients();
          delay(rgro);  /* delay required to zero the read gradient */

        endpeloop(seqcon[2],v6);
}
/***********************************************************************
                         MODIFICATION HISTORY

970320(ss): Single shot, spin echo or gradient echo version
970327(ss): tblip included in readout ramp. Timing corrected.
970327(ss): inversion pulse added
970509(ss): Renamed, episs.c. tev, tod replaced by tep.
            groa and grora adjusters added.

Changes made at NIH:
970725(pj):corrected the odd echo read amplitude to gro1+groa rather than gro1
970725(pj):line: invtime = ti + rgss*2 + tspoil*2 + p2;
	corrected to: invtime = ti + rgss + tspoil + p2/2;
970725(pj):line: espmin = ((tevs+tods)*0.5) + rgror*2 + at;  
	corrected to: espmin = ((tevs+tods)*0.5) + rgro*2 + at;
970725(pj):tidied up various mistakes in the calculation of seqtime, 
	ted1 and ted2
970725(pj):negation of gror and gpeb corrected to when spinecho[0]='n'

Notes:
	suspect that slice selection gradient during the 180 pulse is
        wrong, depending on what the pattern looks like - check this out

***********************************************************************/
