/*
  fastir_epi.c

  based on ...

  myepi.c
  
  Stuart Clare
  FMRIB Centre, University of Oxford
*/

/*

 num               ns                                                  ns  num
 ints                                                                      ints
  {    [Inversion] {  [Fatsat]  [Excitation]  [Spinecho]  [EPI][Crush] }    } 

            |____________ti__________|__________te__________|

                   |_________________________tr/ns_____________________|

  |___________________________________TR____________________________________|

*/

/*
  Real-time variables
  v1: phase cycling
  v2: phase cycling
  v3: phase cycling
  v4: phase cycling
  v5: number of pe steps
  v6: pe step number
  v7: pe increment
  v8: number of segments
  v9: segment number
  v10:
  v11:
  v12: rfmod power
  v13: number of slices
  v14: slice number
*/

/*
  RF Pulses
  p1 : tpwr1 : p1pat : gss  - Excitation
  p2 : tpwr2 : p2pat : gss2 - Spin echo
  p3 : tpwr3 : p3pat : gss3 - Inversion
  p4 : tpwr4 : p4pat        - Fatsat
  p5 : tpwr5 : p5pat : gss5 - Inversion slab
  psat : satpwr : satpat : gsssat - Spatial Saturation (QUIPSS, regular EPI)
*/

/*
  Tables
  t1 : tpwr1 + rfmod  [table element for each slice]
  t2 : tpwr2 + rfmod  [table element for each slice]
  t3 : tpwr3 + rfmod  [table element for each slice]
  t4 : tpwr4 + rfmod  [table element for each slice]
  t5 : tpwr5 + rfmod  [table element for each slice]
  t58 : slab_off    [dB correction]
  t59 : slab_off    [tpwrf correction]
  t60 : linear modulation     [tpwrf correction for t1-t5]
*/

#include "standard.h"
#include "group.h"
#include "sequence_utilities.h"

/* global */
extern codeint v15,v16,v17,v18,v19,v20;
double TR;

#include "globals.h"
#include "playout.h"
#include "calculate.h"
#include "tweekers.h"
#include "fastir_epi_safety_calc.h"
#include "fastir_epi_utilities.h"

pulsesequence()
{	
  double t_inv_b,t_inv_a;
  double t_cr,t_fat;
  double t_ex_b,t_ex_a;
  double t_se_b,t_se_a;
  double t_epi_b,t_epi_a;
  double min_te,min_tr,min_ti;
  
  double tr_delay,ti_delay,te_delay,se_delay_b,se_delay_a;

  double prescan_flag,pown,loc_tep;
  int image;
  char err_string[MAXSTR];

  double looppss,TR_delay,min_TR,blocks;
  int i;

  initparms_sis();
  get_tweekers();

  /* 180 degree pulse gradient strength */
  gss2 = getval("gss2");
  /* Inversion recovery gradient strength */
  gss3 = getval("gss3");

  /* Trigger and counter bit pattern */
  image = (int)(getval("image"));
  /* bit 1 = wait for ticks triggers before beginning scan */
  /* bit 2 = puts out high on sp3 for duration of volume */
  /* bit 3 = epi blips off (set bit on), otherwise blips on */
  /* bit 4 = make inversion non slice selective (on) or slice selective (off)*/

  /* Get flags 'y/n' */
  getstr("spinecho",spinecho);  
  getstr("inversion",inversion);  
  getstr("crushers",crushers);  
  getstr("fat_sat",fat_sat);
  getstr("half_kspace",half_kspace);
  getstr("nonlinsamp",nonlinsamp);
  getstr("rfmod",rfmod);

  /* Prescan flag (1=yes, 0=no) */
  prescan_flag = getval("prescan_flag");

  if(prescan_flag==1){
    if ((P_getreal(CURRENT,"loc_tep",&loc_tep,1))>=0)
      tep=loc_tep/1e6; /* convert from us */
  }

  /* looppss */
  looppss = getval("looppss");
  TR = getval("TR");
  blocks = getval("blocks");
  if (blocks == 0) blocks = 1;
  if(ns==1)blocks=1;
  if((int)ns%(int)blocks){
    sprintf(err_string,"%s: ns not an integral number of blocks\n"
	    ,seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  

  /* Number of interleaves */
  num_ints = getval("num_ints");  
  if (num_ints == 0) num_ints = 1;

  /* Half k-space stuff */
  if (half_kspace[0] == 'y') {
    
    if (num_ints > 1) {
      sprintf(err_string,"%s: Half K space acquisition requires 1 interleaf\n"
	      ,seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    
    pown = 2;
    do{
      if(pown == nv){
        sprintf(err_string,"%s: nv must be a few lines greater than n^2\n"
	        ,seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      eff_echo = nv - pown;
      pown *= 2;
    }while(pown < nv);
    etl = nv;
  }
  else {
    etl = (double)((int)(nv/num_ints));  /* echo train length/interleaf */
    eff_echo = etl/2.0;  /* echo position of TE */
  }

  /* Crusher gradient strength */
  gcrush=gmax/2.0;

  /* Gradient calculations */
  if(inversion[0]=='y'){
    calculate_slab_inversion_gradients(&t_inv_b,&t_inv_a);
    calculate_slab_offsets(blocks);
  }
  else { t_inv_b=0.0; t_inv_a=0.0; }

  if(crushers[0]=='y')calculate_crusher_gradients(&t_cr);
  else t_cr=0.0;

  if(fat_sat[0]=='y')calculate_fatsat_gradients(&t_fat);
  else t_fat=0.0;

  calculate_excitation_gradients(&t_ex_b,&t_ex_a);

  if(spinecho[0]=='y')calculate_spinecho_gradients(&t_se_b,&t_se_a);
  else { t_se_b=0.0; t_se_a=0.0; }

  calculate_epi_gradients(&t_epi_b,&t_epi_a);

  if(rfmod[0]=='y'){
    calculate_rf_modulation();
    calculate_slab_rf_modulation();
  }

  /* Timing calculations */

  /* Nonlinear sampling */
  if(nonlinsamp[0]=='y')calculate_nonlinear_sampling();

  /* Spinecho delay */
  se_delay_b = se_delay_a = 0.0;
  if(spinecho[0]=='y'){
    if( (t_ex_a+t_se_b) > (t_se_a+t_epi_b) )
      se_delay_a=(t_ex_a+t_se_b)-(t_se_a+t_epi_b);
    if( (t_ex_a+t_se_b) < (t_se_a+t_epi_b) )
      se_delay_b=(t_se_a+t_epi_b)-(t_ex_a+t_se_b);
  }

  /* te delay */
  min_te = t_ex_a + se_delay_b + t_se_b + t_se_a + se_delay_a + t_epi_b;
  if(min_te>te){
    sprintf(err_string,"%s: te too short. Min te = %f\n",seqfil,min_te);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  te_delay = te - min_te;

  /* ti delay */
  if(inversion[0]=='y'){
    min_ti = t_inv_a + t_fat + t_ex_b;
    if(min_ti>ti){
      sprintf(err_string,"%s: ti too short. Min ti = %f\n",seqfil,min_ti);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    ti_delay = ti - min_ti;
  }
  else ti_delay=0;

  /* tr delay */
  min_tr = (t_fat + t_ex_b + te + t_epi_a + t_cr) * (ns / blocks);
  if(min_tr>tr){
    sprintf(err_string,"%s: tr too short. Min tr = %f\n",seqfil,min_tr);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  tr_delay = (tr - min_tr) / (ns / blocks);

  /* TR delay */
  min_TR = t_inv_b + t_inv_a + ti_delay + tr;
  if(min_TR>TR){
    sprintf(err_string,"%s: TR too short. Reset to %f\n",seqfil,min_TR);
    printf(err_string);
    log_error(err_string);
    TR_delay = 0;
  }
  else TR_delay = TR - min_TR;

  if (prescan_flag == 1)num_segs=1;
  else num_segs=(int)(nv/etl);

  if((prescan_flag == 1)&&(ns == 1))TR_delay=0;

  /* Turn blips off */
  if ((image & 4) == 4) {
    sgbpe = 0.0;
    gbd=0.0;
    sgbd=0.0;
  }

  /* Saftey check */
  
  ns/=blocks;

  if (myepi_safety_calc() == -1) {
    sprintf(err_string,"%s: Failed duty cycle safety check\n",
	    seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  
  ns*=blocks;
 
  /**********************************************************************/
  /* PULSE SEQUENCE */
  /**********************************************************************/

  /* Phase cycling */
  mod2(ct,v3);                    /* v3 = 0101 0101 */
  dbl(v3,v3);                     /* v3 = 0202 0202 */
  hlv(ct,v4);                     /* v4 = 0011 2233 */
  mod2(v4,v4);                    /* v4 = 0011 0011 */
  add(v3,v4,v1);                  /* v1 = 0213 0213 */
  assign(v1,oph);                 /* oph = 0213 0213 */
  add(one,v1,v2);

  /* phase encode increment size */
  assign(one,v7);  
  
  /* Receiver setup */
  status(A);
  rcvroff();
  rcvron();

  initval(num_segs,v8);  /* set v8 to the num of echoes/shot */
  initval(blocks,v15);
  initval(ns/blocks,v17);
  initval(0,v20);

  /* wait for start scan button */
  if ((image & 1) == 1) xgate(1);
  if ((image & 2) == 2) sp3on(); else sp3off();
  delay(del_seq);
  
  /* loop v15 (blocks) times using counter v16 */
  loop(v15,v16);
  
  /* loop v8 (num_ints) times using counter v9 */
  loop(v8,v9);

  /* Inversion pulse */
  if(inversion[0] == 'y') {
      slab_inversion_pulse(v16);
      delay(ti_delay);
  }

  /* sets v13 to ns and incs v14 up to it */
  msloop(seqcon[1],ns/blocks,v13,v14);
  
  /* looppss and block */
  assign(v14,v10);
  for(i=0;i<looppss;i++)incr(v10);
  modn(v10,v13,v11);
  
  mult(v16,v17,v18);
  add(v11,v18,v19);

  if (fat_sat[0] == 'y') {
    fatsat_pulse();
  }

  excitation_pulse(v19);

  delay(te_delay/2.0);

  if (spinecho[0] == 'y'){
    delay(se_delay_b);
    spinecho_pulse(v11);
    delay(se_delay_a);
  }
  
  delay(te_delay/2.0);

  epi_gradients_and_acquire();

  crusher_gradients();

  /* double check that all gradients are zero */
  zero_all_gradients();

  delay(tr_delay);
  
  /* increment the msloop counter v14 */
  endmsloop(seqcon[1],v14);

  delay(TR_delay);
 
  /* increment the interleaves loop counter v9 */
  endloop(v9);
  
  /* increment the blocks loop counter v16 */
  endloop(v16);
}
