/* FSE2FID.C */
/*****************************************************************************
  This program is copyrighted worldwide by the John P. Robarts Research
  Institute.  Any distribution, copying or redistribution is expressly
  forbidden.

  This program converts a compressed multislice fse data set into a standard
  Varian data set by doing flashc, tabc, and fse processing all on the
  same file. The procpar is set right so that Vnmr and Stimulate can read
  the data, acquired with or without navigators.

  The phase encode table must be of the form
  fsems260_4_1, where 260 is the number of pe's + phase ref,
		4 is the echo train length (etl) and
		1 is the effective echo in the train (etl).

  To compile:
  gcc -O3 -o fse2fid fse2fid.c fourn.c shiftc.c -lm

  History
  23.07.97 Written by Ravi S. Menon
  11.19.98 modified by Ravi Menon to do phase and navigator correction together 
  03.16.99 modified by Joe Gati for Fast Spin Echo (FSE) data recon
  04.14.99 modified by Joe Gati for single shot phase reference
  04.18.00 modified by Joe Gati - added Wiener demodulation filter
  10.23.01 modified by Peter Jezzard - added a DC offset correction, and
           added capability to process non navigated acquisitions
******************************************************************************/

/* include library files for the program these allow compilation in RRI environment */
#include <sys/fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/file.h>
#include <sys/param.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#define pi 3.14159265359
#define F_FRAC 0.9
#define MAX_NV 512
#define MAX_NP 1024
#define T2 0.06

#define FSE_DONE   8		  /* set in main_header.spare1
				   * means already converted */
char output_file[128];

struct datafilehead
{
   long    nblocks;		  /* number of blocks in file     */
   long    ntraces;		  /* number of traces per block   */
   long    np;			  /* number of elements per
				   * trace */
   long    ebytes;		  /* number of bytes per element  */
   long    tbytes;		  /* number of bytes per trace    */
   long    bbytes;		  /* number of bytes per block    */
   short   transf;		  /* transposed storage flag      */
   short   status;		  /* status of whole file         */
   long    spare1;		  /* reserved for future use      */
}       main_header;

struct datablockhead
{
   short   scale;		  /* scaling factor               */
   short   status;		  /* status of data in block      */
   short   index;		  /* block index                  */
   short   spare3;		  /* reserved for future use      */
   long    ctcount;		  /* completed transients in
				   * fids */
   float   lpval;		  /* left phase in phasefile      */
   float   rpval;		  /* right phase in phasefile     */
   float   lvl;			  /* level drift correction       */
   float   tlt;			  /* tilt drift correction        */
}       block_header;

union precision
{
   short  *sh;
   long   *lo;
   char   *ch;
}      *in_data,*out_data;

char    proc_string[5000];

int     nv,
        np,
        ns,
        etl,
        nseg,
        eff_echo,
        nblocks,
        ntraces,
        ebytes,
        bbytes,
        nav_flag;	/* PJ */

double  esp,
        alpha,
        re_dc_corr,	/* PJ */
        im_dc_corr;	/* PJ */

double  time1;

struct passwd *userpw;

main(argc, argv)
   int     argc;
   char   *argv[];
{

   extern char output_file[128];

   void    fse_cor();

   int     n_nblocks,
           n_ntraces,
           n_bbytes;

   double  r2,r3,r,filter,cutoff;
   double  fermi[MAX_NV][MAX_NP];

   int     w_filter,
           f_filter,
           tmin,
           in_file,
           out_file,
           i, j, k, kk, l, m, n,
           malloc_size;

   int    *petable;

   char    string[100],
           string1[100];

   char    petable_name[128],
           input_fid[128],
           output_fid[128],
           seqcon[128],
           orig_fid[128],
           fiddir[128];

   char   *indata;

   long   *temp;

   FILE   *in_ptr,
          *out_ptr,
          *fopen();

   int fid_exists(char *infile);

   /* Parse the input line and make decisions or use defaults */
   if (argc < 2)
   {
      printf("\nUsage: fse2fid [Input_file fermi wiener alpha] \n\n");
      printf("Input_file: FSE fid file to be corrected (without .fid) \n");
      printf("fermi_filter: Default 1 (=Yes) \n");
      printf("wiener_filter: Default 0 (=No) \n");
      printf("alpha (noise variance) est: Default alpha = 0.10 \n");
      exit(1);
   }
   printf("\nFSE2FID: Reading procpar\n");

   strcpy(fiddir, argv[1]);

   if (argc == 2)
   {
     f_filter = 0;
     w_filter = 0;
     alpha = 0.10;
   }

   if ((argc > 2)) 
   {
     f_filter = atoi(argv[2]); 
     w_filter = atoi(argv[3]); 
     alpha = atof(argv[4]);
   }

   printf("FSE2FID: processing with Fermi Filter=%d \n", f_filter);
   if (w_filter == 1) {
     printf("FSE2FID: processing with Weiner Filter \n");
     printf("FSE2FID: alpha set to %lf \n", alpha);
   }

   /* Generate path to the input and output fid files */
   if(!fid_exists(fiddir)){
     strcat(fiddir,".fid");
     if(!fid_exists(fiddir)){
       printf("Cannot open fid file %s\n",fiddir);
       return;
     }
   }

   strcpy(input_fid, fiddir);
   strcat(input_fid, "/fid");
   strcpy(output_fid, fiddir);
   strcat(output_fid, "/fse2fid.temp");
   strcpy(output_file, fiddir);
   strcat(output_file, "/nav");

   /* Open input fid file */
   if ((in_file = open(input_fid, O_RDONLY)) < 0)
   {
      perror("Open");
      printf("FSE2FID: can't open input file %s.\n", input_fid);
      exit(2);
   }

   /* Read main_header */
   malloc_size = sizeof(main_header);
   if (read(in_file, &main_header, malloc_size) != malloc_size)
   {
      perror("read");
      printf("FSE2FID: can't open input file %s header.\n", input_fid);
      exit(4);
   }

   /* See if data has already been processed */
   if (main_header.spare1 & FSE_DONE)
   {
      printf("FSE2FID: File already processed by FSE2FID\n");
      exit(1);
   }

   /* assign the important header values */
   np = main_header.np;
   nblocks = main_header.nblocks;
   ntraces =  main_header.ntraces;
   ebytes = main_header.ebytes;
   bbytes = main_header.bbytes;

   /* read some parameters from procpar file */
   strcpy(input_fid, fiddir);
   strcat(input_fid, "/procpar");
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("FSE2FID: Can't open input procpar file %s/procpar.\n",input_fid);
      exit(8);
   }

   while (fgets(proc_string, 5000, in_ptr) != NULL)
   {
      if (!strncmp(proc_string, "esp 3", 5))
      {
         fgets(proc_string, 5000, in_ptr);
         sscanf(proc_string, "%d %lf\n", &i, &esp);
      }
      if (!strncmp(proc_string, "nv ", 3))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %d\n", &nv);
      }
      if (!strncmp(proc_string, "seqcon", 6))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %s\n", seqcon);
	 for (i = 0; i < strlen(seqcon) - 2; i++)
	    seqcon[i] = seqcon[i + 1];
	 seqcon[i] = '\0';
      }
      if (!strncmp(proc_string, "ns 1", 4))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %d\n", &ns);
      }
      if (!strncmp(proc_string, "petable ", 8))
      {
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "1 %s\n", petable_name);
	 for (i = 0; i < strlen(petable_name) - 2; i++)
	    petable_name[i] = petable_name[i + 1];
	 petable_name[i] = '\0';
	 sscanf(strchr(petable_name,'_')+1,"%d",&etl);
	 sscanf(strrchr(petable_name,'_')+1,"%d",&eff_echo);
	 nseg = nv/etl;
	 printf("FSE2FID: The # of PE's are %d \n", nv);
	 printf("FSE2FID: Data acquired with etl of %d \n", etl);
	 printf("FSE2FID: Effective echo is %d \n", eff_echo);
      }
   }
   fclose(in_ptr);

   esp = esp;
   printf("FSE2FID: Data acquired with esp = %lf\n", esp);
   printf("FSE2FID: Data acquired with te = %lf\n", esp*eff_echo);
   
   printf("\nFSE2FID: Reading petable\n");
   /* Read in the phase-encoding table */
   petable = (int *) malloc(nv * sizeof(int));
   strcpy(input_fid, "/vnmr/tablib/");
   strcat(input_fid, petable_name);
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("FSE2FID: Can't open petable %s. Check your tablib.\n", petable_name);
      exit(8);
   }
   fgets(string, 100, in_ptr);
   tmin = 0;
   for (l = 0; l < nv; l++)
   {
      if (fscanf(in_ptr, "%d", petable + l) != 1)
      {
	 perror("fscanf");
	 printf("FSE2FID: %s length inconsistent with data.\n", petable_name);
	 exit(6);
      }
      if (tmin > petable[l])
	 tmin = petable[l];
   }
   fclose(in_ptr);

   /* check to see if navigator lines have been collected (PJ) */
   /* note nav_flag added as multiplier at various other points in prog */
   nav_flag=0;
   for (l = 0; l < etl; l++) nav_flag+=petable[l]*petable[l];
   if (nav_flag > 0) nav_flag=0; else nav_flag=1;
   printf("FSE2FID: Navigator flag = %d.\n", nav_flag);
   /* end PJ */

   for (l = 0; l < nv; l++)
      petable[l] -= tmin;

   /* memory allocation for the data */
   out_data = (union precision *) malloc(nblocks * ns * (nv-etl*nav_flag) * sizeof(union precision));
   in_data = (union precision *) malloc(nblocks * ns * nv * sizeof(union precision));
   for (i = 0; i < nblocks * nv * ns; i++)
   {
      out_data[i].ch = (char *) malloc(np * sizeof(long));
      in_data[i].ch = (char *) malloc(np * sizeof(long));
   }
   temp = (long *) malloc(np * sizeof(long));
  
   /* check for the proper data acquisition mode */
   if ((seqcon[1] == 's' && ns > 1) || seqcon[2] != 'c')
   {
      printf("FSE2FID: Data acquired without compression...can't process \n");
      exit(1);
   }

   printf("\nFSE2FID: Reading data\n");
   /* Define new header values */
   bbytes = main_header.bbytes;
   n_ntraces = 1;
   n_nblocks = nblocks * ns * nv;
   n_bbytes = np * main_header.ebytes + 28;

   /* Allocate space for one input trace */
   if ((indata = (char *) malloc(ntraces * np * main_header.ebytes)) == NULL)
   {
      perror("malloc");
      printf("FSE2FID: uncompression routine: unable to malloc space for input data.\n");
      exit(8);
   }

   /* Uncompress data *****************/
   m = 0;
   printf("FSE2FID: Uncompressing data \n");
   for (i = 0; i < nblocks; i++)
   {
      /* Read in data and block_headers */
      if (read(in_file, &block_header, 28) != 28)
      {
	 perror("read");
	 printf("FSE2FID: unable to read data block headers.\n");
	 exit(10);
      }

      /* DC offset velues read here (PJ) */
      re_dc_corr=block_header.lvl;
      im_dc_corr=block_header.tlt;
      if (i == 0) printf("FSE2FID: DC offsets %f, %f\n",re_dc_corr,im_dc_corr);
      /* end added section (PJ) */

      if ((j = read(in_file, indata, bbytes - 28)) != bbytes - 28)
      {
	 perror("read");
	 printf("FSE2FID: uncompression routine: Error reading original data block #%d! \n", j);
	 abort(1);
      }
      for (k = 0; k < ns; k++)
      {
	 for (l = 0; l < nseg; l++)
	 {
	    for (j = 0; j < etl; j++)
	    {
               m = j + l * etl + k * nv + i * ns * nv;
	       for (n = 0; n < main_header.np * main_header.ebytes; n++)
		  in_data[m].ch[n] = indata[(j + l*ns*etl + k*etl) * main_header.tbytes + n];

   	       if (main_header.ebytes == 2)
   	       {
      			for (kk = main_header.np - 1; kk >= 0; kk--)
                        {
	    	   	   in_data[m].lo[kk] = in_data[m].sh[kk];
                        }
   	       }
	    }
	 }
      }
   }

   /* Update the main header with the new values */
   main_header.bbytes = n_bbytes;
   main_header.ntraces = n_ntraces;
   main_header.nblocks = n_nblocks;

   /* Update procpar with the new values */
   strcpy(input_fid, fiddir);
   strcat(input_fid, "/procpar");
   if ((in_ptr = fopen(input_fid, "r")) == NULL)
   {
      perror("Open");
      printf("FSE2FID: Can't open input procpar file %s.\n", input_fid);
      exit(8);
   }

   /* Open "new" procpar file for writing */
   strcpy(input_fid, fiddir);
   strcat(input_fid, "/procpar.new");
   if ((out_ptr = fopen(input_fid, "w")) == NULL)
   {
      perror("Open");
      printf("FSE2FID: Can't open output procpar file %s.\n", input_fid);
      exit(8);
   }

   /* Read in the "old" procpar file, line by line. copy to and 
      update "new" procpar as necessary */ 
   while (fgets(proc_string, 5000, in_ptr) != NULL)
   {
      if (!strncmp(proc_string, "seqcon ", 7))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 fprintf(out_ptr, "1 \"nscnn\" \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "ns ", 3))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 1 \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "pss ", 4))
      {
	 j = 0;
	 for (i = 0; i < strlen(proc_string); i++)
	 {
	    if (j != 6)
	    {
	       if (proc_string[i] == ' ')
		  j++;
	    } else
	    {
	       proc_string[i] = '2';
	       break;
	    }
	 }
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
      }
      /*
      if (!strncmp(proc_string, "nv ", 3))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", nv - etl*nav_flag);
	 fgets(proc_string, 5000, in_ptr);
      }
      */
      if (!strncmp(proc_string, "ni ", 3))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", nv - etl*nav_flag);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "nf ", 3))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 1 \n");
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "celem ", 6))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j * ns * nv);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "arrayelem ", 10))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 999, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j + 1);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "arraydim ", 9))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %d", &i, &j);
	 fprintf(out_ptr, "1 %d \n", j*ns*nv);
	 fgets(proc_string, 5000, in_ptr);
      }
      if (!strncmp(proc_string, "array ", 6))
      {
	 fputs(proc_string, out_ptr);
	 fgets(proc_string, 5000, in_ptr);
	 sscanf(proc_string, "%d %s", &i, string);
	 k = strlen(string);
	 if (k < 3)
	    fprintf(out_ptr, "1 \"\"\n");
	 else
	 {
	    for (i = 0; i < k - 1; i++)
	       string1[i] = string[i];
	    string1[k - 1] = '\0';
	    strcat(string1, ",pss\"");
	    fprintf(out_ptr, "1 %s \n", string1);
	 }
	 fgets(proc_string, 5000, in_ptr);
      }
      fputs(proc_string, out_ptr);
   }
   fclose(in_ptr);
   fclose(out_ptr);

   /* Now replace the unmodified procpar with the new procpar */
   sprintf(string, "cp %s/procpar %s/procpar.orig", fiddir, fiddir);
   system(string);
   sprintf(string, "mv -f %s/procpar.new %s/procpar", fiddir, fiddir);
   system(string);

   /* phase reference and navigator correction */
   if (nav_flag) fse_cor(w_filter);

   /* tabc done here  */
   printf("\nFSE2FID: RESORTING PE ORDER (tabc)...\n");
   for (i = 0; i < nblocks; i++)
   {
	for (l = 0; l < ns; l++)
	{
            for (kk = nav_flag; kk < nseg; kk ++) /* changed counter */
            {
		for (k = 0; k < etl; k ++)
		{
			m = l + i * ns + petable[k + kk*etl] * ns * nblocks;
			for (n = 0; n < main_header.np; n++)
			  out_data[m].lo[n] = in_data[k + kk*etl + l*nv + i*ns*nv].lo[n];
		}
             }
          }
   }

   /* DC offset correction done here  (PJ) */
   printf("\nFSE2FID: DOING A DC CORRECTION...\n");
   for (i = 0; i < nblocks; i++)
   {
	for (l = 0; l < ns; l++)
	{
            for (k = 0; k < nv/2; k ++)
            {
		m= i*ns*nv + (k*2+1)*ns + l;
		for (n = 0; n < main_header.np/2; n++)
                {
		  out_data[m].lo[2*n] = (int)((float)out_data[m].lo[2*n] + 2.0*re_dc_corr);
		  out_data[m].lo[2*n+1] = (int)((float)out_data[m].lo[2*n+1] + 2.0*im_dc_corr);
                }
             }
          }
   }
   /* end of DC Corr (PJ) */

/* JG - Insert some stuff here */
   nv -= etl*nav_flag;
   main_header.ntraces = 1;
   main_header.nblocks = nblocks*nv*ns;
/* End insert */


   /* fermi filter done here  */
   if (f_filter == 1)  {
      printf("\nFSE2FID: Generating FERMI FILTER...\n");
      cutoff = (double)F_FRAC;
      for (j = 1; j < (1+nv/2); j++) {
        for (k = 1; k < (1+np/4); k++) {
                r2 = (double)j/(double)(nv/2.);
                r3 = (double)k/(double)(np/4.);
                r = r2*r2 + r3*r3;
                filter =  1.0/(1.0+exp((r-cutoff)*50.0));
                fermi[(nv/2)-j][((np/4)-1)+k] = filter;
                fermi[((nv/2)-1)+j][(np/4)-k] = filter;
                fermi[(nv/2)-j][(np/4)-k] = filter;
                fermi[((nv/2)-1)+j][((np/4)-1)+k] = filter;
        }
      }  

   printf("\nFSE2FID: Applying FERMI FILTER...\n");
   for (i = 0; i < nv; i++)
   { 
       for (j = 0; j < nblocks; j++)
       {
          for (k = 0; k < ns; k++)
          {
              for (l = 0; l < np/2; l++)
              {
                  out_data[k+ns*j+ns*nblocks*i].lo[2*l] = out_data[k+ns*j+ns*nblocks*i].lo[2*l]*fermi[i][l];
                  out_data[k+ns*j+ns*nblocks*i].lo[2*l+1] = out_data[k+ns*j+ns*nblocks*i].lo[2*l+1]*fermi[i][l];
              }
           }
        }
    }
   }

   /* if single precision, shift the data into shorts */ 
   printf("FSE2FID: Shifting data into short words.\n");
   if (main_header.ebytes == 2)
   {
      for (k = 0; k < nblocks * nv * ns; k++)
      {
         for (i = 0; i < np; i++)
         {
            out_data[k].sh[i] = out_data[k].lo[i];
         }  
      }   
   } 
   /* Open output fid file */
   printf("\nFSE2FID: Writing data to file \n");
   if ((out_file = open(output_fid, O_CREAT | O_TRUNC | O_WRONLY, 0755)) < 0)
   {
      perror("open");
      printf("FSE2FID: unable to open output file %s.\n",output_fid);
      exit(14);
   }
   main_header.spare1 |= FSE_DONE;
   if (write(out_file, &main_header, sizeof(main_header)) < 0)
   {
      perror("write");
      printf("FSE2FID: unable to write main output data header.\n");
      exit(15);
   }
   block_header.index = 0;
   block_header.lvl=0.0;
   block_header.tlt=0.0;
   for (i = 0; i < ns * nblocks * nv; i++)
   {
      block_header.index++;
      if (write(out_file, &block_header, 28) != 28)
      {
	 perror("write");
	 printf("FSE2FID: unable to write output data block header.\n");
	 exit(10);
      }
      if (write(out_file, out_data[i].ch, main_header.tbytes) != main_header.tbytes)
      {
	 perror("write");
	 printf("FSE2FID: unable to write data block.\n");
	 exit(10);
      }
   }

   /* Close input fid file */
   if (close(in_file))
   {
      perror("close");
      printf("FSE2FID: problems closing input file.\n");
      exit(13);
   }
   /* Close output fid file */
   if (close(out_file))
   {
      perror("close");
      printf("FSE2FID: problems closing output file.\n");
      exit(18);
   }
  
   /* move fid file to fid.orig and move fse2fid.temp file to fid file */ 
   strcpy(orig_fid, fiddir);
   strcpy(input_fid, fiddir);
   strcat(input_fid, "/fid");
   strcat(orig_fid, "/fid.orig");
   if (rename(input_fid, orig_fid) != 0)
   {
      perror("rename");
      printf("FSE2FID: unable to rename original fid file.\n");
      exit(19);
   }
   if (rename(output_fid, input_fid) != 0)
   {
      perror("rename");
      printf("FSE2FID: unable to rename new fid file.\n");
      exit(20);
   }
   /*
   sprintf(string, "rm %s/procpar.orig", fiddir);
   system(string);
   sprintf(string, "rm %s/fid.orig", fiddir);
   system(string);
   */
   printf("\nFSE2FID finished successfully\n");
   exit(0);
}


void
fse_cor(wiener)
   int wiener;
{
   register int i, j, k, l, m, n;
   int     nvol;
   int     nn[3];
   float  *ftdata,
          *ftdata1;
   double  mag,
           mag_ratio,
           r2,
           t,
           sxoss,
           st2,
           sx,
           deltaB,
           theta,
           coscor,
           sincor,
          *echo,
          *mag_max,
          *phasemag_ref,
          *mtf,
          *w_ky;
   FILE   *in_ptr,
          *out_ptr,
          *fp1;

   char    string[100];
   void    fourn(),
           shiftc();

   extern char output_file[128];

   printf("\nFSE_COR: Starting PHASE CORRECTION...\n");
   printf("FSE_COR: nblock: %d ntrace: %d  np: %d \n", nblocks, ntraces, np);

   /*Decide how many volumes are present */
   nvol = ntraces * nblocks / (nv * ns);
   printf("FSE_PHASE_COR: nv: %d nvol: %d  nslices: %d \n", nv, nvol, ns);
   printf("FSE_PHASE_COR: %d slices to process\n", ns);

   /* Allocate space for the processing */

   ftdata = (float *) malloc(np * sizeof(float));
   ftdata1 = (float *) malloc(np * sizeof(float));
   phasemag_ref = (double *) malloc(np * nv * sizeof(double));
   mag_max = (double *) malloc(np * nv * sizeof(double));
   mtf = (double *) malloc((nvol+1) * etl * sizeof(double));
   w_ky = (double *) malloc((nvol+1) * etl * sizeof(double));
   echo = (double *) malloc(etl * sizeof(double));

   nn[0] = np/2;

   /* Cycle through each of the reference fids in the first shot of
   each image and calculate the correction phase for each of 
   these k-space lines. */

   for (k = 0; k < nvol; k++)
   {
      printf("FSE_COR: Correcting volume %d \n",k+1);
      for (l = 0; l < ns; l++)
      {
         mag_max[0] = 0.0;
         for (i = 0; i < etl; i++) /* changed to etl because only have ref shot */
         {
            for (j = 0; j < np; j += 2)
            {
               ftdata[j] = in_data[i + l*nv + k*ns*nv].lo[j];
               ftdata[j + 1] = in_data[i + l*nv + k*ns*nv].lo[j + 1];
            }
   
            shiftc(ftdata, nn, 1);
            fourn(ftdata - 1, nn - 1, 1, 1);
            shiftc(ftdata, nn, 1);

            for (j = 0; j < np; j += 2)
            {
               /* only calculate phase if signal magnitude > 100. */
               mag = sqrt((double)(ftdata[j+1]*ftdata[j+1]+ftdata[j]*ftdata[j]));
               phasemag_ref[i * np + j + 1] = mag;
               if (mag > 100.0)
                  phasemag_ref[i * np + j] = atan2((double)ftdata[j+1],(double)ftdata[j]);
               else
                  phasemag_ref[i * np + j] = 0.0;

               if ( (mag > mag_max[i]) )
                  mag_max[i] = mag;
            }
            if ( (wiener == 1) )
               mag_max[i] = log(mag_max[i]);
         }

         /* Do a least squares calculation to determine the T2 of the
         entire data set. Use this value to determine the MTF for
         the Wiener filter in the FSE_COR routine. */

         if (wiener == 1) {

            sx = 0.0;
            for (n=0 ; n < etl; n++) {
               echo[n] = (n+1)*esp;
               sx += echo[n];
            }
            sxoss = sx/etl;
         
            st2 = 0.0;
            r2 = 0.0;
            for (n=0; n < etl; n++) {
              t = echo[n] - sxoss;
              st2 += t*t;
              r2 += t*mag_max[n];
            }
            r2 /= -st2;
            printf("FSE_COR: T2 = %lf ms for slice %d. \n",1.0/r2*1000,l+1);

            for (n=1; n < etl+1; n++) {
                  mtf[n] = exp((eff_echo*esp-n*esp)*r2);
                  w_ky[n] = mtf[n]/(mtf[n]*mtf[n]+alpha);
            }
  
         }

         /* start phase correcting every segment of every volume */

         for (i = 1; i < nseg; i++)
         {
            for (m = 0; m < etl; m++)
            {
	       for (j = 0; j < np; j += 2) 
	       {
                  if (wiener == 1) {
	             ftdata[j] = w_ky[m+1]*in_data[m + i*etl + l*nv + k*ns*nv].lo[j];
	             ftdata[j + 1] = w_ky[m+1]*in_data[m + i*etl + l*nv + k*ns*nv].lo[j + 1]; 
                  }
                  else {
	             ftdata[j] = in_data[m + i*etl + l*nv + k*ns*nv].lo[j];
	             ftdata[j + 1] = in_data[m + i*etl + l*nv + k*ns*nv].lo[j + 1]; 
                  }
               }

   	       shiftc(ftdata, nn, 1);
	       fourn(ftdata - 1, nn - 1, 1, 1);
	       shiftc(ftdata, nn, 1);

 	       for (j = 0; j < np; j += 2)
	       {
                  deltaB = phasemag_ref[m*np + j];
                  if (deltaB < -pi)
                     deltaB = deltaB + 2.*pi;
                  if (deltaB > pi) 
                     deltaB = deltaB - 2.*pi;
                  theta = -deltaB;
                  /* make sure theta goes between 0 and 2 pi*/
                  if (theta < 0) 
                     theta = theta + 2.*pi;

	          coscor = cos(theta);
	          sincor = sin(theta);

	          ftdata1[j] = (float) coscor * ftdata[j] - (float) sincor * ftdata[j + 1];
	          ftdata1[j + 1] = (float) sincor * ftdata[j] + (float) coscor * ftdata[j + 1];

                  /* output navigator info (volume, angle, real, imaginary, magnitude) */

	       }

               shiftc(ftdata1, nn, 1);
	       fourn(ftdata1 - 1, nn - 1, 1, -1);
	       shiftc(ftdata1, nn, 1);

	       for (j = 0; j < np; j += 2)
	       {
	          in_data[m + i*etl + l*nv + k*ns*nv].lo[j] = ftdata1[j];
	          in_data[m + i*etl + l*nv + k*ns*nv].lo[j + 1] = ftdata1[j + 1]; 
	       }
            }
         }
      }
   }

   fclose(fp1);
   free(ftdata);
   free(ftdata1);
   free(phasemag_ref);
   free(mag_max);
   free(mtf);
   free(w_ky);
   free(echo);
}
int fid_exists(char *infile)
{
  char tmp[128];
  sprintf(tmp,"%s/fid",infile);
  if(fopen(tmp,"r")==NULL){
    return(0);
  }
  else {
    return(1);
  }
}
