
/*****************************************************************************
*  Safety checking routine to check for excess duty cycle values on the      *
*  gradients coil. This file should be adapted for each individual pulse     *
*  sequence so that a correct duty cycle on the logical read, phase encode   *
*  and slice axes can be calculated. The system file safety_check() is then  *
*  called to rotate the logical axes on to their respective physical axes    *
*  (according to the values of psi, phi and theta), and then calculate       *
*  whether an excessive duty cycle has been requested for the gradient coil  *
*  in use. An error code is returned to the pulse sequence if the duty       *
*  cycle is exceeded. Note that in order for this code to work, the pulse    *
*  sequence variable declarations must be outside the main pulsesequence()   *
*  routine. This allows them to be "seen" by the duty cycle calculation      *
*  routine.                                                                  *
*                                                                            *
*  This file must be correctly written for each pulse sequence so that the   *
*  RMS gradient amplitude on each logical axis can be calculated. This is    *
*  done by performing the following summation on each logical axis:          *
*                                                                            *
*           GRAD_i_RMS=sqrt[ Sum{(grad_ij^2)*grad_ij_duration} / tr ]        * 
*                                                                            *
*           where i=read, phase, slice                                       *
*                 j=1,2,3,4,5...  (gradient pulse index)                     *
*                                                                            *
*  Peter Jezzard, University of Oxford, 1998                                 *
*  This version of the routine suitable for: ox_sems.c                       *
*  Modification history:                                                     *
*       written Sept 1998 PJ                                                 *
*****************************************************************************/

/* safety_check.h must be included. This routine contains the gradient
   characteristics and does the actual safety check                         */
#include "safety_check.h"

int fse_safety_calc()
{

  /* variables to hold read, phase and slice rms gradients */
  double grad_r_rms, grad_p_rms, grad_s_rms; 
  /* declare a dummy variable that allows for a weighting to be applied
     to a calculation based on the frequency that the relevant pulse unit
     occurs per net tr time */
  double tr_wt;

  /* error value returned to main code. -1=failure, 1=success */
  int err;

  int i;

  /* initiate the rms gradient values to zero on all axes */
  grad_r_rms=0.0;
  grad_p_rms=0.0;
  grad_s_rms=0.0;

  /* initialize tr_wt value */
  tr_wt=1.0;

  /**************************************************************************/
  /* must edit the code from this point on to suit the sequence being run   */
  /**************************************************************************/

  /* calculate the contribution from the prep pulses */
  if (prep[0]=='i') {
    grad_s_rms=grad_s_rms+2*(pow(gss2,2.0)*(agss2*grate/2.0+p2
					    +agss2*grate/2.0)*tr_wt);
  }
  if (prep[0]=='s') {
    grad_s_rms=grad_s_rms+2*(pow(gss_sat,2.0)*(agss_sat*grate/2.0+ps
					       +agss_sat*grate/2.0)*tr_wt);
    grad_s_rms=grad_s_rms+2*0.64*1.0*0.01;
    grad_r_rms=grad_r_rms+2*0.64*1.0*0.01;
    grad_p_rms=grad_p_rms+2*0.64*1.0*0.01;
  }

  /* calculate the contribution from the pulse sequence itself */
  grad_s_rms=grad_s_rms+pow(gss,2.0)*(agss*grate/2.0+p1+agss*grate/2.0)
    *tr_wt; /* 90 */

  /* refocus */
  grad_r_rms=grad_r_rms+pow(gror,2.0)*tref;
  grad_s_rms=grad_s_rms+pow(gssr,2.0)*tref;

  for(i=0;i<etl;i++){
    /* 180 degree pulse */
    grad_s_rms=grad_s_rms+pow(agcrush,2.0)*(agcrush*grate/2+tcrush+
					    agcrush*grate/2)*tr_wt;
    grad_s_rms=grad_s_rms+pow(gss2,2.0)*p2*tr_wt;
    grad_s_rms=grad_s_rms+pow(agcrush,2.0)*(agcrush*grate/2+tcrush+
					    agcrush*grate/2)*tr_wt;
    
    /* phase encode */
    grad_p_rms=grad_p_rms+pow(sgpe,2.0)*(fabs(sgpe)*grate/2+tpe+
					 fabs(sgpe)*grate/2);
    grad_p_rms=grad_p_rms+pow(sgpe,2.0)*(fabs(sgpe)*grate/2+tpe+
					 fabs(sgpe)*grate/2);
    

    /* read out */
    grad_r_rms=grad_r_rms+pow(gro,2.0)*(pw_gro/2.0+at
					+pw_gro/2.0)*tr_wt;
  }

  /* factor in the number of slices per tr period, and divide by the tr */
  /* NOTE: only one slice per tr for this sequence */
  grad_r_rms=grad_r_rms*ns/tr;
  grad_p_rms=grad_p_rms*ns/tr;
  grad_s_rms=grad_s_rms*ns/tr;

  /* calculate the RMS value by taking sqrt */
  grad_r_rms=sqrt(grad_r_rms);
  grad_p_rms=sqrt(grad_p_rms);
  grad_s_rms=sqrt(grad_s_rms);

  err=-1;  /* default=-1 (1=pass, -1=fail) */
  err=safety_check(grad_r_rms, grad_p_rms, grad_s_rms);

  return err;

}
