/*
  invepi.c
  
  Stuart Clare
  FMRIB Centre, University of Oxford
  April 1999

  Adapted by Peter Jezzard from myepi.c (Aug 1999) for inversion recovery

*/

/*

 num   ns                                                              ns  num
 ints                                                                      ints
  {    { [Inversion]  [Fatsat]  [Excitation]  [Spinecho]  [EPI][crush] }    } 

              |__________ti__________|__________te__________|

       |___________________________tr/ns_______________________________|    

*/

/*
  Real-time variables
  v1: phase cycling
  v2: phase cycling
  v3: phase cycling
  v4: phase cycling
  v5: number of pe steps
  v6: pe step number
  v7: pe increment
  v8: number of segments
  v9: segment number
  v10: 
  v11: 
  v12: rfmod power
  v13: number of slices
  v14: slice number
*/ 

/*
  RF Pulses
  p1 : tpwr1 : p1pat : gss  - Excitation
  p2 : tpwr2 : p2pat : gss2 - Spin echo
  p3 : tpwr3 : p3pat : gss3 - Inversion
  p4 : tpwr4 : p4pat        - Fatsat
  psat : satpwr : satpat : gsssat - Spatial Sat
*/

/*
  Tables
  t1 : tpwr1 + rfmod
  t2 : tpwr2 + rfmod
  t3 : tpwr3 + rfmod
  t4 : tpwr4 + rfmod
  t5 : tpwr5 + rfmod
  t60 : linear modulation
*/

#include "standard.h"
#include "group.h"
#include "globals.h"
#include "sequence_utilities.h"
#define SAMPPTS 8192

/* Global variables*/
#include "playout.h"
#include "calculate.h"
#include "tweekers.h"
#include "invepi_safety_calc.h"
#include "multi_inversion_pulses.h"

pulsesequence()
{
  extern codeint v15,v16,v17,v19,v24;
  double t_inv_b,t_inv_a;
  double t_presat_b,t_presat_a;
  double t_cr,t_fat;
  double t_ex_b,t_ex_a;
  double t_se_b,t_se_a;
  double t_epi_b,t_epi_a;
  double min_te,min_tr,min_ti;
  double t_nav;
  double seqtime, max_ti_slices, temp_ns;
  double grad_settle_del, tmp_del;
  double echo_shift=1.0;

  double subset_ns[512];
  int num_ti_blocks, n_ti_slices, i;
  int last_ti_list=-1, last_tr_list=-1;
  
  double tr_delay,last_tr_delay[512],te_delay,se_delay_b,se_delay_a;
  double ti_delay,last_ti_delay[512];

  double prescan_flag,pown;
  int image;
  char err_string[MAXSTR],full_fname[MAXSTR],*home_dir;
  FILE *table;

  initparms_sis();
  get_tweekers();

  /* 180 degree pulse gradient strength */
  gss2 = getval("gss2");
  /* Inversion recovery gradient strength */
  gss3 = getval("gss3");
  grad_settle_del = getval("grad_settle_del");

  /* Trigger and counter bit pattern */
  image = (int)(getval("image"));
  /* bit 1 = wait for ticks triggers before beginning scan */
  /* bit 2 = puts out high on sp3 for duration of volume */
  /* bit 3 = epi blips off (set bit on), otherwise blips on */
  /* bit 4 = make inversion non slice selective (on) or slice selective (off)*/
  /* bit 5 = turns off the acquire */

  /* Get flags 'y/n' */
  getstr("spinecho",spinecho);  
  getstr("inversion",inversion);  
  getstr("crushers",crushers);  
  getstr("fat_sat",fat_sat);
  getstr("half_kspace",half_kspace);
  getstr("nonlinsamp",nonlinsamp);
  getstr("rfmod",rfmod);
  getstr("nav",nav);

  echo_shift=getval("echo_shift");

  gsssat = getval("gsssat");
  presat=getval("presat");
  presat_off=getval("presat_off");
  sat_psi=getval("sat_psi");
  sat_phi=getval("sat_phi");
  sat_theta=getval("sat_theta");
  sat_thk=getval("sat_thk");


  if (presat) presat_off=getval("presat_off");

  /* Number of interleaves */
  num_ints = getval("num_ints");  
  if (num_ints == 0) num_ints = 1;

  /* Half k-space stuff */
  if (half_kspace[0] == 'y') {
    
    if (num_ints > 1) {
      sprintf(err_string,"%s: Half K space acquisition requires 1 interleaf\n"
	      ,seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    
    pown = 2;
    do{
      if(pown == nv){
        sprintf(err_string,"%s: nv must be a few lines greater than n^2\n"
	        ,seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      eff_echo = nv - pown;
      pown *= 2;
    }while(pown < nv);
    etl = nv;
  }
  else {
    etl = (double)((int)(nv/num_ints));  /* echo train length/interleaf */
    eff_echo = (etl/2.0)/echo_shift;  /* echo position of TE */
    /* echo_shift shifts the effective echo position */
    /* echo_shift = 1.0 (default) : eff_echo centred in echo train */
    /* echo_shift = 2.0 : eff_echo half way beteen start and centre of train */
    /* etc. */
  }

  /* Crusher gradient strength */
  gcrush=gmax/2.0;

  if(crushers[0]=='y')calculate_crusher_gradients(&t_cr);
  else t_cr=0.0;

  if (presat) calculate_spatial_saturation_gradients(&t_presat_b,&t_presat_a);

  if(fat_sat[0]=='y')calculate_fatsat_gradients(&t_fat);
  else t_fat=0.0;

  calculate_excitation_gradients(&t_ex_b,&t_ex_a);

  if(spinecho[0]=='y')calculate_spinecho_gradients(&t_se_b,&t_se_a);
  else { t_se_b=0.0; t_se_a=0.0; }

  calculate_epi_gradients(&t_epi_b,&t_epi_a);
  echo_del=ceil(echo_del/epi_clock_time)*epi_clock_time;

  if(rfmod[0]=='y'){
    calculate_rf_modulation();
  }

  if(nav[0]=='y'){
    calculate_navigator_gradients(&t_nav);
  }
  else t_nav=0;

  /* Timing calculations */

  /* Nonlinear sampling */
  if(nonlinsamp[0]=='y')calculate_nonlinear_sampling();

  /* Spinecho delay */
  se_delay_b = se_delay_a = 0.0;
  if(spinecho[0]=='y'){
    if( (t_ex_a+t_se_b) > (t_se_a+t_epi_b+t_nav) )
      se_delay_a=(t_ex_a+t_se_b)-(t_se_a+t_epi_b+t_nav);
    if( (t_ex_a+t_se_b) < (t_se_a+t_epi_b+t_nav) )
      se_delay_b=(t_se_a+t_epi_b+t_nav)-(t_ex_a+t_se_b);
  }

  /* te delay */
  min_te = t_ex_a + se_delay_b + t_se_b + t_se_a + se_delay_a + t_nav + t_epi_b;
  if(min_te>te){
    sprintf(err_string,"%s: te too short. Min te = %f\n",seqfil,min_te);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  te_delay = te - min_te;

  seqtime = (t_fat + t_ex_b + te + t_epi_a + t_cr + grad_settle_del);
  if (presat) seqtime = seqtime + t_presat_b + t_presat_a;

  /* inversion recovery calculations. Try to fit as many slices in to one
     inversion time (ti) to save time */
  ti_delay=0; n_ti_slices=0; num_ti_blocks=1;
  max_ti_slices=ns; last_ti_delay[0]=0.0;
  subset_ns[0]=ns;   /* default */
  if(ti > 0.0){
    t_inv_b=0.0; t_inv_a=0.0;
    tmp_del=t_ex_b+t_fat+grate*fabs(gss3)+p3/2+grate*fabs(gcrush)*2+tcrush;
    n_ti_slices = (int)(floor((ti-tmp_del)/seqtime)+1.01);  
                                                /* .01 for round off error */
    if (n_ti_slices > ns) n_ti_slices = (int)ns;
    if (n_ti_slices > 8) n_ti_slices = 8;  /* to avoid SAR breech */
    calculate_multi_inversion_gradients(&t_inv_b,&t_inv_a);
    if (n_ti_slices == 1) {
      subset_ns[0]=1.0;
      min_ti = t_inv_a + t_fat + t_ex_b;
      if (presat) min_ti = min_ti - (t_presat_b + t_presat_a);
      if(min_ti>ti){
        sprintf(err_string,"%s: ti too short. Min ti = %f\n",seqfil,min_ti);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      ti_delay = 0.0;
      last_ti_delay[0] = ti - min_ti;
    } else {    /* n_ti_slices greater than 1 */
      num_ti_blocks = (int)(ceil(ns/n_ti_slices)+0.01);
      temp_ns=ns;
      for (i = 0; i < num_ti_blocks; i++) {
        if ((temp_ns - (double)n_ti_slices) > 0.0) {
          subset_ns[i]=(double)n_ti_slices;
          temp_ns=temp_ns-(double)n_ti_slices;
          max_ti_slices=(double)n_ti_slices;
        } else {
          subset_ns[i]=temp_ns;
        }
      }
      ti_delay = seqtime - t_inv_b - t_inv_a;
      if(ti_delay < 0.0) {
        sprintf(err_string,"%s: Inversion module too long. Reduce p3 or tcrush",
                 seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      for (i = 0; i < num_ti_blocks; i++) {
        last_ti_delay[i]=ti-((subset_ns[i]-1)*seqtime+t_inv_a+t_fat+t_ex_b);
        if (presat) last_ti_delay[i] = last_ti_delay[i] - 
                 (t_presat_b + t_presat_a);
        if(last_ti_delay[i] < 0.0) {
          sprintf(err_string,"%s: ti too short", seqfil);
          printf(err_string);
          log_error(err_string);
          abort(1);
        }
      }
    }
  }

  /* tr delay */
  min_tr = seqtime * max_ti_slices;
  if (ti > 0.0) {
    min_tr = min_tr + (ti + t_inv_b - t_fat - t_ex_b);
    if (presat) min_tr = min_tr - t_presat_b - t_presat_a;
  }
  if(min_tr>tr){
    sprintf(err_string,"%s: tr too short. Min tr = %f\n",seqfil,min_tr);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  if(ti > 0.0) {   
    tr_delay = 0.0;
    for (i = 0; i < num_ti_blocks; i++) {
      last_tr_delay[i] = tr - (subset_ns[i]*seqtime+ti+t_inv_b
             - t_fat - t_ex_b);
      if (presat) last_tr_delay[i]=last_tr_delay[i]-t_presat_b-t_presat_a;
    }
  }
  else {
    tr_delay = (tr - min_tr)/ns;
    last_tr_delay[0]=0.0;
  }

  /* Now create the lists in memory */
  if (ti > 0.0) {
    if (last_ti_list<0) last_ti_list=curlist++;
    create_delay_list(last_ti_delay,num_ti_blocks,last_ti_list);
  }
  if (last_tr_list<0) last_tr_list=curlist++;
  create_delay_list(last_tr_delay,num_ti_blocks,last_tr_list);
  home_dir=(char *)getenv("HOME");
  sprintf(full_fname,"%s/vnmrsys/tablib/ns_table",home_dir);
  table=fopen(full_fname,"w");
  if(table==NULL){
    sprintf(err_string,"%s: Unable to open ~/vnmrsys/tablib/ns_table\n",
            seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  fprintf(table,"t32 = ");
  for(i=0;i<num_ti_blocks;i++){
    fprintf(table,"%d ",(int)subset_ns[i]);
  }
  fprintf(table,"\n");
  fclose(table);
  loadtable("ns_table");

  /* Prescan flag (1=yes, 0=no) */
  prescan_flag = getval("prescan_flag"); 
  if (prescan_flag == 1)num_segs=1;
  else num_segs=(int)(nv/etl);
  
  /* Turn blips off */
  if ((image & 4) == 4) {
    sgbpe = 0.0;
    gbd=0.0;
    sgbd=0.0;
  }

  /* Saftey check */
  if (invepi_safety_calc() == -1) {
    sprintf(err_string,"%s: Failed duty cycle safety check\n",
	    seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }

  if (ix == 1) {
    printf("achieved tau_pe = %f us\n",(2.0*pw_gro+echo_del+at)/num_ints*1.0e6);
  }

/* debug stuff */
/* 
printf("\n\n debug\n");
printf("num_ti_blocks = %d\n",num_ti_blocks);
printf("t_inv_b = %f\n",t_inv_b*1000.0);
printf("t_inv_a = %f\n",t_inv_a*1000.0);
printf("n_ti_slices = %d\n",n_ti_slices);
printf("tr_delay=%f\n",tr_delay*1000.0);
printf("ti_delay=%f\n",ti_delay*1000.0);
printf("seqtime=%f\n",seqtime*1000.0);
for (i=0; i<num_ti_blocks; i++) {
  printf("subset_ns[%d]=%f\n",i,subset_ns[i]);
  printf("last_tr_delay[%d]=%f\n",i,last_tr_delay[i]*1000.0);
  printf("last_ti_delay[%d]=%f\n",i,last_ti_delay[i]*1000.0);
}
*/
 
  /**********************************************************************/
  /* PULSE SEQUENCE */
  /**********************************************************************/

  /* Phase cycling */
  mod2(ct,v3);                    /* v3 = 0101 0101 */
  dbl(v3,v3);                     /* v3 = 0202 0202 */
  hlv(ct,v4);                     /* v4 = 0011 2233 */
  mod2(v4,v4);                    /* v4 = 0011 0011 */
  add(v3,v4,v1);                  /* v1 = 0213 0213 */
  assign(v1,oph);                 /* oph = 0213 0213 */
  add(one,v1,v2);

  /* phase encode increment size */
  assign(one,v7);  
  
  /* Receiver setup */
  status(A);
  sp4off();
  rcvroff();
  rcvron();
  sp2off();

  initval(num_segs,v8);  /* set v8 to the num of echoes/shot */
  initval(num_ti_blocks,v15);  /* set v15 to the num of ti blocks */
  initval(0.0,v10);   /* v10 holds a zero for assignment */

  /* wait for start scan button */
  if ((image & 1) == 1) xgate(ticks);
  if ((image & 2) == 2) sp3on(); else sp3off();
  delay(del_seq);
  
  /* loop v8 (num_ints) times using counter v9 */
  loop(v8,v9);

  assign(v10,v17);   /* "current slice" counter in multislice loop back to 0 */
  assign(v10,v24);   /* "current slice" counter in inversion loop back to 0 */

  /* loop v15 times (num_ti_blocks) using counter v16 */
  loop(v15,v16);

  getelem(t32,v16,v19);

  /* do the set of inversions */
  if (ti > 0.0) {
    /* v24 holds the index of the first slice to be inverted */
    multi_inversion_pulses(ti_delay);
    vdelay_list(last_ti_list,v16);
  }

  /* replaces msloop, subset_ns slices using counter v14 */
  loop(v19,v14);

  /*xgate(ticks);*/

  if (presat) spatial_saturation_pulse();

  if (fat_sat[0] == 'y') {
    fatsat_pulse();
  }

  excitation_pulse(v17);

  delay(te_delay/2.0);

  if (spinecho[0] == 'y'){
    delay(se_delay_b);
    spinecho_pulse(v17);
    delay(se_delay_a);
  }
  
  delay(te_delay/2.0);

  if(nav[0]=='y'){
  	navigator_acquire();
  }

  if((image & 16) == 16){
	epi_gradients_no_acquire();
  }
  else epi_gradients_and_acquire();

  crusher_gradients();

  /* double check that all gradients are zero */
  zero_all_gradients();

  delay(tr_delay);

  delay(grad_settle_del);

  incr(v17);  /* increment the v17 counter since a slice has been collected */
  
  /* increment the msloop counter v14 */
  endloop(v14);
 
  vdelay_list(last_tr_list,v16);

  /* increment the ti_block loop counter v16 */
  endloop(v16);

  /* increment the interleaves loop counter v9 */
  endloop(v9);
}
