
/*****************************************************************************
*  Safety checking routine to check for excess duty cycle values on the      *
*  gradients coil. This file should be adapted for each individual pulse     *
*  sequence so that a correct duty cycle on the logical read, phase encode   *
*  and slice axes can be calculated. The system file safety_check() is then  *
*  called to rotate the logical axes on to their respective physical axes    *
*  (according to the values of psi, phi and theta), and then calculate       *
*  whether an excessive duty cycle has been requested for the gradient coil  *
*  in use. An error code is returned to the pulse sequence if the duty       *
*  cycle is exceeded. Note that in order for this code to work, the pulse    *
*  sequence variable declarations must be outside the main pulsesequence()   *
*  routine. This allows them to be "seen" by the duty cycle calculation      *
*  routine.                                                                  *
*                                                                            *
*  This file must be correctly written for each pulse sequence so that the   *
*  RMS gradient amplitude on each logical axis can be calculated. This is    *
*  done by performing the following summation on each logical axis:          *
*                                                                            *
*           GRAD_i_RMS=sqrt[ Sum{(grad_ij^2)*grad_ij_duration} / tr ]        * 
*                                                                            *
*           where i=read, phase, slice                                       *
*                 j=1,2,3,4,5...  (gradient pulse index)                     *
*                                                                            *
*  Peter Jezzard, University of Oxford, 1998                                 *
*  This version of the routine suitable for: ox_epi.c                        *
*  Modification history:                                                     *
*       written Aug 1998 PJ                                                  *
*****************************************************************************/

/* safety_check.h must be included. This routine contains the gradient
   characteristics and does the actual safety check                         */
#include "safety_check.h"

int invepi_safety_calc()
{

  /* variables to hold read, phase and slice rms gradients */
  double grad_r_rms, grad_p_rms, grad_s_rms; 
  /* declare a dummy variable that allows for a weighting to be applied
     to a calculation based on the frequency that the relevant pulse unit
     occurs per net tr time */
  double tr_wt;

  /* error value returned to main code. -1=failure, 1=success */
  int err;

  /* initiate the rms gradient values to zero on all axes */
  grad_r_rms=0.0;
  grad_p_rms=0.0;
  grad_s_rms=0.0;

  /* initialize tr_wt value */
  tr_wt=1.0;

  /**************************************************************************/
  /* must edit the code from this point on to suit the sequence being run   */
  /**************************************************************************/

  /* calculate the contribution from the inversion */
  if (inversion[0] == 'y') {
    /* slice select itself */
    grad_s_rms=grad_s_rms+pow(gss3,2.0)*(pw_gss3/2.0+p3+pw_gss3/2.0)*tr_wt;
    /* crusher gradient */
    grad_r_rms=grad_r_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
    grad_p_rms=grad_p_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
    grad_s_rms=grad_s_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
  }

  /* calculate the contribution from the fat saturation pulse */
  if (fat_sat[0] == 'y') {
    grad_p_rms=grad_p_rms+pow(gmax/2.0,2.0)*(grise/2.0+(pw_gfat-grise)+
        grise/2.0)*tr_wt;
  }

  /* calculate the contribution from the pulse sequence itself */
  grad_s_rms=grad_s_rms+pow(gss,2.0)*(fabs(gss)*grate/2.0+p1+fabs(gss)*grate/2.0)
        *tr_wt; /* 90 */
  grad_s_rms=grad_s_rms+pow(gssr,2.0)*(pw_gssr/2.0+(tref-pw_gssr)+
        pw_gssr/2.0)*tr_wt;  /* 90 deg refocus */
  if (spinecho[0] == 'y') {
    grad_s_rms=grad_s_rms+pow(gss2,2.0)*(fabs(gss2)*grate/2.0+(tspoil+p2+tspoil)+
        fabs(gss2)*grate/2.0)*tr_wt;  /* 180 deg */
  }
  grad_r_rms=grad_r_rms+pow(gror,2.0)*(pw_gror/2.0+(tref2-pw_gror)+
        pw_gror/2.0)*tr_wt;
  grad_p_rms=grad_p_rms+pow(gbd,2.0)*(pw_gror/2.0+(tref2-pw_gror)+
        pw_gror/2.0)*tr_wt;
  grad_r_rms=grad_r_rms+pow(gro,2.0)*(pw_gro/2.0+at+pw_gro/2.0)*etl*tr_wt;
  grad_p_rms=grad_p_rms+pow(sgbpe,2.0)*(pw_gro/2.0+echo_del+pw_gro/2.0)*
        etl*tr_wt;
  grad_r_rms=grad_r_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
  grad_p_rms=grad_p_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
  grad_s_rms=grad_s_rms+pow(gcrush,2.0)*(pw_gcrush/2.0+(tcrush-pw_gcrush)+
        pw_gcrush/2.0)*tr_wt;
  
  
  /* factor in the number of slices per tr period, and divide by the tr */
  grad_r_rms=grad_r_rms*ns/tr;
  grad_p_rms=grad_p_rms*ns/tr;
  grad_s_rms=grad_s_rms*ns/tr;

  /* calculate the RMS value by taking sqrt */
  grad_r_rms=sqrt(grad_r_rms);
  grad_p_rms=sqrt(grad_p_rms);
  grad_s_rms=sqrt(grad_s_rms);
  
  err=-1;  /* default=-1 (1=pass, -1=fail) */
  err=safety_check(grad_r_rms, grad_p_rms, grad_s_rms);

  return err;

}
