; ==========================================================================
; Spiral Reconstruction: Various tools
; ==========================================================================
; written in IDL by Markus Weiger, FMRIB Centre, University OF Oxford, 2001


;--------- utilities -----------

FUNCTION Phase, var
   rvar = FLOAT(var)
   ivar = IMAGINARY(var)
   rvar = rvar + ((rvar+ivar) EQ 0)
   RETURN,ATAN(ivar,rvar)
END

PRO Logstring, text, level
   COMMON Logging, logging_level
   IF level LE logging_level THEN PRINT, text
END

FUNCTION Trimstr, a
   RETURN, STRTRIM(STRING(a),2)
END

FUNCTION Unwrap, ph
   RETURN, ph-2.0*!pi*(ph GT !pi)+2.0*!pi*(ph LT -!pi)
END

PRO Exchange, a, b
   dum = a
   a = b
   b = dum
END

FUNCTION Swap, A
   s = SIZE(A)
   nx = s(1)
   ny = s(2)
   dum = A
   dum(0:nx/2-1,0:ny/2-1) = A(nx/2:nx-1,ny/2:ny-1)
   dum(0:nx/2-1,ny/2:ny-1) = A(nx/2:nx-1,0:ny/2-1)
   dum(nx/2:nx-1,0:ny/2-1) = A(0:nx/2-1,ny/2:ny-1)
   dum(nx/2:nx-1,ny/2:ny-1) = A(0:nx/2-1,0:ny/2-1)
   RETURN,dum
END

FUNCTION Max_diff, arr
   s = SIZE(arr)
   nx = s(1)
   ny = s(2)
   arr1 = arr
   arr1(*,*) = 0
   arr1(1:nx-1,1:ny-1) = arr(0:nx-2,0:ny-2)
   arr2 = arr-arr1
   mask = arr1 NE 0 ; exclude differences to exact zero
   arr2 = arr2 * mask    
   RETURN, MAX(ABS(arr2))
END

FUNCTION Zerofill, arr, res
; only for quadratic arrays
; arr = k-space array
; res = returned matrix size

   s = SIZE(arr)
   n = s(1)
   IF s(0) GT 2 THEN ns = s(3) ELSE ns = 1
   IF res EQ n THEN result = arr
   IF res LT n THEN result = arr(n/2-res/2:n/2+res/2-1, n/2-res/2:n/2+res/2-1,*)
   IF res GT n THEN BEGIN
      result = COMPLEXARR(res, res, ns)
      result(*,*,*) = 0
      result(res/2-n/2:res/2+n/2-1, res/2-n/2:res/2+n/2-1,*) = arr
   END
   RETURN, REFORM(result)
END

FUNCTION reduce_fourier, arr, res
; only for quadratic complex arrays
; multiple slices, volumes possible
   s = SIZE(arr)
   n = s(1)
   IF s(0) GT 2 THEN ns = s(3) ELSE ns = 1  
   IF s(0) GT 3 THEN nv = s(4) ELSE nv = 1  
   result = COMPLEXARR(res,res,ns,nv)
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO BEGIN
      dum = swap(fft((arr(*,*,is,iv)),1))
      result(*,*,is,iv) = dum(n/2-res/2:n/2+res/2-1,n/2-res/2:n/2+res/2-1)
   END
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO result(*,*,is,iv) = (fft(swap(result(*,*,is,iv)),-1))
   RETURN, result
END

FUNCTION Smooth_fourier, arr, frac
; only for quadratic arrays
; multiple slices, volumes possible
   result = arr
   s = SIZE(arr)
   n = s(1)
   nf = FIX(n*frac)
   IF s(0) GT 2 THEN ns = s(3) ELSE ns = 1  
   IF s(0) GT 3 THEN nv = s(4) ELSE nv = 1  
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO result(*,*,is,iv) = swap(fft(arr(*,*,is,iv),1))
   result(0:n/2-nf/2-1,*,*,*) = 0.0
   result(n/2+nf/2:n-1,*,*,*) = 0.0
   result(*,0:n/2-nf/2-1,*,*) = 0.0
   result(*,n/2+nf/2:n-1,*,*) = 0.0
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO result(*,*,is,iv) = fft(swap(result(*,*,is,iv)),-1)
   RETURN, result
END

FUNCTION Mask_arr, type, nx, ny, sx, sy, px, py
   mask = FLTARR(nx, ny)
   mask(*, *) = 0.0
   CASE type OF
      1: mask(px-sx/2:px+sx/2-1, py-sy/2:py+sy/2-1) = 1.0
      2: FOR x = px-sx/2, px+sx/2-1 DO FOR y = py-sy/2, py+sy/2-1 DO $
        IF ((x-px)/(sx/2.0))^2+((y-py)/(sy/2.0))^2 LE 1.0 THEN mask(x, y) = 1.0
   ENDCASE
   RETURN, mask
END

FUNCTION Shift_arr, arr, n
   ; shifts in first dimension of an array and zeros
   ; n < 0 : shift left
   ; n > 0 : shift right
   IF n NE 0 THEN BEGIN
      s = SIZE(arr)
      ndim = s(0)
      CASE ndim OF
         1: newarr = SHIFT(arr, n)
         2: newarr = SHIFT(arr, n, 0)
         3: newarr = SHIFT(arr, n, 0, 0)
         4: newarr = SHIFT(arr, n, 0, 0, 0)
         5: newarr = SHIFT(arr, n, 0, 0, 0, 0)
      END
      IF n GT 0 THEN newarr(0:n-1,*,*,*,*) = 0 ELSE newarr(s(1)+n:s(1)-1,*,*,*,*) = 0        
   END ELSE newarr = arr
   RETURN, newarr
END


FUNCTION Maxi, a, b
   IF a GE b THEN result = a ELSE result = b
   RETURN, result
END

FUNCTION Mini, a, b
   IF a LT b THEN result = a ELSE result = b
   RETURN, result
END

FUNCTION Numstr, i, n
; string fuer n-stellige Zahlen
   nullstr = ''
   FOR inull = 0, n-2 DO nullstr = nullstr + '0'
   IF i EQ 0 THEN ext = nullstr+'0' ELSE BEGIN
      ext = nullstr+STRMID(STRING(i), 7-BYTE(ALOG10(i)), ALOG10(i)+1)
      ext = STRMID(ext, STRLEN(ext)-n, n)
   END
   RETURN, ext
END


;----------- Various IO functions ------------------

FUNCTION File_exists, fname
   GET_LUN, unit
   OPENR, unit, fname, error = err
   CLOSE, unit
   FREE_LUN, unit
   RETURN, err EQ 0
END

PRO To_Tiff_8Bits, image, filename
   ;tvlct,r,g,b,/get
   ;s = size(r)
   ;maxcol = s(1)-1	
   ;image = byte(image*255.0/maxcol)
   image = BYTE(image/FLOAT(MAX(image))*255.0)
   image = REVERSE(rotate(image,2))
   write_tiff, filename+'.tif', image
END

PRO window_to_tif, name
   picture = TVRD(0, 0, !D.X_VSIZE, !D.Y_VSIZE)
   To_Tiff_8Bits, picture, name
END


;----------- AVW read + write  ------------------

PRO Avw_header, avw_key, avw_dim, avw_his
   avw_key = { $
               sizeof_hdr       : LONG(0), $
               data_type        : BYTARR(10), $
               db_name          : BYTARR(18), $
               extents          : LONG(0), $
               session_error    : FIX(0), $
               regular          : BYTE(0), $
               hkey_un0         : BYTE(0) $
             }
   avw_dim = { $
               dim              : INTARR(8), $
               unused8_14       : INTARR(7), $
               datatype         : FIX(0), $
               bitpix           : FIX(0), $
               dim_un0          : FIX(0), $
               pixdim           : FLTARR(8), $
               funused8_13      : FLTARR(6), $
               compressed       : FLOAT(0), $
               verified         : FLOAT(0), $
               glmax            : LONG(0), $
               glmin            : LONG(0) $
             }
   avw_his = { $
               descrip          : BYTARR(80), $
               aux_file         : BYTARR(24), $
               orient           : BYTE(0), $
               originator       : BYTARR(10), $
               generated        : BYTARR(10), $
               scannum          : BYTARR(10), $
               patient_id       : BYTARR(10), $
               exp_date         : BYTARR(10), $
               exp_time         : BYTARR(10), $
               hist_un0         : BYTARR(3), $
               views            : LONG(0), $
               vols_added       : LONG(0), $
               start_field      : LONG(0), $
               field_skip       : LONG(0), $
               omax             : LONG(0), $
               omin             : LONG(0), $
               smax             : LONG(0), $
               smin             : LONG(0) $
             }
   dumstruct = {dum:FIX(0)}
   dumstruct.dum(*) = 0
   struct_assign, dumstruct, avw_key
   struct_assign, dumstruct, avw_dim
   struct_assign, dumstruct, avw_his
END

FUNCTION Read_avw, fname, verbose

   ;PRINT, "------ read AVM data ------"
   
   ; read header
   avw_header, avw_key, avw_dim, avw_his
   GET_LUN, unit
   OPENR, unit, fname+'.hdr'
   READU, unit, avw_key, avw_dim, avw_his
   CLOSE, unit
   FREE_LUN, unit
   
   swap_bytes = avw_key.sizeof_hdr NE 348

   IF swap_bytes THEN BEGIN
      avw_key.sizeof_hdr       = swap_endian(avw_key.sizeof_hdr)
      avw_key.extents          = swap_endian(avw_key.extents)
      avw_key.session_error    = swap_endian(avw_key.session_error)
      avw_dim.dim              = swap_endian(avw_dim.dim)
      avw_dim.unused8_14       = swap_endian(avw_dim.unused8_14)
      avw_dim.datatype         = swap_endian(avw_dim.datatype)
      avw_dim.bitpix           = swap_endian(avw_dim.bitpix)
      avw_dim.dim_un0          = swap_endian(avw_dim.dim_un0)
      avw_dim.pixdim           = swap_endian(avw_dim.pixdim)
      avw_dim.funused8_13      = swap_endian(avw_dim.funused8_13)
      avw_dim.compressed       = swap_endian(avw_dim.compressed)
      avw_dim.verified         = swap_endian(avw_dim.verified)
      avw_dim.glmax            = swap_endian(avw_dim.glmax)
      avw_dim.glmin            = swap_endian(avw_dim.glmin)
      avw_his.views            = swap_endian(avw_his.views)
      avw_his.vols_added       = swap_endian(avw_his.vols_added)
      avw_his.start_field      = swap_endian(avw_his.start_field)
      avw_his.field_skip       = swap_endian(avw_his.field_skip)
      avw_his.omax             = swap_endian(avw_his.omax)
      avw_his.omin             = swap_endian(avw_his.omin)
      avw_his.smax             = swap_endian(avw_his.smax)
      avw_his.smin             = swap_endian(avw_his.smin)
   END
   
   IF verbose THEN BEGIN 
      PRINT, 'sizeof_hdr ', avw_key.sizeof_hdr
      PRINT, 'data_type ', STRING(avw_key.data_type)
      PRINT, 'db_name ', STRING(avw_key.db_name)    
      PRINT, 'extents ', avw_key.extents        
      PRINT, 'session_error ', avw_key.session_error 
      PRINT, 'regular ', STRING(avw_key.regular)         
      PRINT, 'hkey_un0 ', STRING(avw_key.hkey_un0)       
      PRINT, 'dim ', avw_dim.dim            
      PRINT, 'unused8_14 ', avw_dim.unused8_14      
      PRINT, 'datatype ', avw_dim.datatype     
      PRINT, 'bitpix ', avw_dim.bitpix       
      PRINT, 'dim_un0 ', avw_dim.dim_un0        
      PRINT, 'pixdim ', avw_dim.pixdim         
      PRINT, 'funused8_13 ', avw_dim.funused8_13    
      PRINT, 'compressed ', avw_dim.compressed     
      PRINT, 'verified ', avw_dim.verified        
      PRINT, 'glmax ', avw_dim.glmax           
      PRINT, 'glmin ', avw_dim.glmin            
      PRINT, 'descrip ', STRING(avw_his.descrip)        
      PRINT, 'aux_file ', STRING(avw_his.aux_file)         
      PRINT, 'orient ', STRING(avw_his.orient)         
      PRINT, 'originator ', STRING(avw_his.originator)     
      PRINT, 'generated ', STRING(avw_his.generated)       
      PRINT, 'scannum ', STRING(avw_his.scannum)         
      PRINT, 'patient_id ', STRING(avw_his.patient_id)      
      PRINT, 'exp_date ', STRING(avw_his.exp_date)        
      PRINT, 'exp_time ', STRING(avw_his.exp_time)        
      PRINT, 'hist_un0 ', STRING(avw_his.hist_un0)       
      PRINT, 'views ', avw_his.views           
      PRINT, 'vols_added ', avw_his.vols_added     
      PRINT, 'start_field ', avw_his.start_field     
      PRINT, 'field_skip ', avw_his.field_skip      
      PRINT, 'omax ', avw_his.omax          
      PRINT, 'omin ', avw_his.omin            
      PRINT, 'smax ', avw_his.smax            
      PRINT, 'smin ', avw_his.smin            
   END 

   ; read file
   n_dims = avw_dim.dim(0)
   n_cols = avw_dim.dim(1)
   n_rows = avw_dim.dim(2)
   n_slices = avw_dim.dim(3)
   n_vols = avw_dim.dim(4)

   CASE avw_dim.datatype(0) OF 
      2: BEGIN
         tmp_im=REFORM(BYTARR(n_cols, n_rows, n_slices))
         im=REFORM(BYTARR(n_cols, n_rows, n_slices, n_vols))
      END
      4: BEGIN
         tmp_im=REFORM(INTARR(n_cols, n_rows, n_slices))
         im=REFORM(INTARR(n_cols, n_rows, n_slices, n_vols))
      END
      8: BEGIN
         tmp_im=REFORM(LONARR(n_cols, n_rows, n_slices))
         im=REFORM(LONARR(n_cols, n_rows, n_slices, n_vols))
      END
      16: BEGIN
         tmp_im=REFORM(FLTARR(n_cols, n_rows, n_slices))
         im=REFORM(FLTARR(n_cols, n_rows, n_slices, n_vols))
      END
      32: BEGIN
         tmp_im=REFORM(COMPLEXARR(n_cols, n_rows, n_slices))
         im=REFORM(COMPLEXARR(n_cols, n_rows, n_slices, n_vols))
      END
      64: BEGIN
         tmp_im=REFORM(DBLARR(n_cols, n_rows, n_slices))
         im=REFORM(DBLARR(n_cols, n_rows, n_slices, n_vols))
      END
   ENDCASE

   GET_LUN, unit
   OPENR, unit, fname+'.img'
   FOR i=0, n_vols-1 DO BEGIN
      READU, unit, tmp_im
      IF swap_bytes THEN tmp_im = swap_endian(tmp_im)
      IF (n_slices EQ 1) THEN im(*,*,i)=tmp_im ELSE im(*,*,*,i)=tmp_im
   ENDFOR
   CLOSE, unit
   FREE_LUN, unit

   RETURN, im
END

FUNCTION Openw_avw, name
   avw_info = { name : BYTARR(128), lun : FIX(0), datatype : FIX(0), glmin : LONG(0), glmax : LONG(0), $
                ncols : LONG(0), nrows : LONG(0), nslices : LONG(0), nvols : LONG(0) }
   GET_LUN, lun
   OPENW, lun, name+'.img'
   avw_info.name = BYTE(name)
   avw_info.lun = lun
   RETURN, avw_info
END

PRO Write_avw_next, avw_info, data
   ; data has to arrive in packages of volumes
   s = SIZE(data)

   ; only first time
   IF avw_info.ncols EQ 0 THEN BEGIN
      avw_info.ncols = s(1)
      avw_info.nrows = s(2)
      avw_info.nslices = s(3)*(s(0) EQ 3) + (s(0) NE 3)
      help, data, output = h
      data_types = ['BYTE', 'INT', 'LONG', 'FLOAT', 'COMPLEX', 'DOUBLE']
      FOR i = 0, 5 DO BEGIN
         compare = stregex(h, data_types(i))
         IF compare(0) GT 0 THEN avw_info.datatype = i
      END
   END

   ; always
   avw_info.nvols = avw_info.nvols+1
   IF avw_info.datatype EQ 4 THEN BEGIN
      max_data = maxi(avw_info.glmax,MAX(ABS(data))) 
      min_data = mini(avw_info.glmin,MIN(ABS(data))) 
   END ELSE BEGIN
      max_data = maxi(avw_info.glmax,MAX(data)) 
      min_data = mini(avw_info.glmin,MIN(data)) 
   END
   avw_info.glmax = mini(max_data,2e32-1)
   avw_info.glmin = maxi(min_data,-2e32+1)

   WRITEU, avw_info.lun, data
END

PRO Closew_avw, avw_info

   ; close data file
   
   CLOSE, avw_info.lun
   FREE_LUN, avw_info.lun

   ; create and fill header
   
   avw_header, avw_key, avw_dim, avw_his
   
   avw_key.sizeof_hdr       = 348
   avw_key.extents          = 16384
   avw_key.regular          = 114
   avw_dim.dim(0)           = 2+(avw_info.nslices GT 1)+(avw_info.nvols GT 1)
   avw_dim.dim(1)           = avw_info.ncols
   avw_dim.dim(2)           = avw_info.nrows
   avw_dim.dim(3)           = avw_info.nslices
   avw_dim.dim(4)           = avw_info.nvols
   CASE avw_info.datatype OF
      0: BEGIN
         avw_dim.datatype(0) = 2
         avw_dim.bitpix = 8
      END
      1: BEGIN
         avw_dim.datatype(0) = 4
         avw_dim.bitpix = 16
      END
      2: BEGIN
         avw_dim.datatype(0) = 8
         avw_dim.bitpix = 32
      END
      3: BEGIN
         avw_dim.datatype(0) = 16
         avw_dim.bitpix = 32
      END
      4: BEGIN
         avw_dim.datatype(0) = 32
         avw_dim.bitpix = 64
      END
      5: BEGIN
         avw_dim.datatype(0) = 64
         avw_dim.bitpix = 64
      END
   ENDCASE
   avw_dim.pixdim(*)        = 1.0
   avw_dim.glmax = avw_info.glmax
   avw_dim.glmin = avw_info.glmin

   ; write header file
   GET_LUN, lun
   OPENW, lun, STRING(avw_info.name)+'.hdr'
   WRITEU, lun, avw_key, avw_dim, avw_his
   CLOSE, lun
   FREE_LUN, lun
END

PRO Write_avw, data, name
   avw_info = openw_avw(name)
   s = SIZE(data)
   nvols = s(4)*(s(0) EQ 4) + (s(0) LT 4)
   FOR iv = 0, nvols-1 DO write_avw_next, avw_info, data(*,*,*,iv)
   closew_avw, avw_info
END

PRO Delete_avw, fname
   GET_LUN, unit
   OPENR, unit, fname+'.hdr', /delete
   CLOSE, unit
   OPENR, unit, fname+'.img', /delete
   CLOSE, unit
   FREE_LUN, unit
END

;----------- VARIAN read + write  ------------------

FUNCTION Rd_varian_bin, fname, ntraces, verbose=verbose

   IF (N_ELEMENTS(verbose) EQ 0) THEN verbose=0
   verbose=0

   data_file_header_size=LONG(32) ; number of bytes in each data file header
   block_header_size=LONG(28)   ; number of bytes in each block header

; Reads a Varian binary file
   
   ;PRINT,'----- read_varian ------'

; definition of data file header
; ==============================
   nblocks=LONG(0)              ; number of blocks in file
   ntraces=LONG(0)              ; number of traces per block
   np=LONG(0)                   ; number of elements per trace
   ebytes=LONG(0)               ; number of bytes per element
   tbytes=LONG(0)               ; number of bytes per trace
   bbytes=LONG(0)               ; number of bytes per block
   vers_id=FIX(0)               ; software version, file_id status bits
   status=FIX(0)                ; status of whole file
   nbheaders=LONG(0)            ; number of block headers per block

   GET_LUN, unit
   OPENR, unit, fname+'fid'

   READU, unit, nblocks
   READU, unit, ntraces
   READU, unit, np
   READU, unit, ebytes
   READU, unit, tbytes
   READU, unit, bbytes
   READU, unit, vers_id
   READU, unit, status
   READU, unit, nbheaders

   ; use ebytes as marker to discover platform dependent data format
   IF ebytes GT 4 THEN swap_bytes = 1 ELSE swap_bytes = 0

   IF swap_bytes THEN BEGIN
      nblocks   = swap_endian(nblocks)
      ntraces   = swap_endian(ntraces)
      np        = swap_endian(np)
      ebytes    = swap_endian(ebytes)
      tbytes    = swap_endian(tbytes)
      bbytes    = swap_endian(bbytes)
      vers_id   = swap_endian(vers_id)
      status    = swap_endian(status)
      nbheaders = swap_endian(nbheaders)
   END

   IF (verbose GT 0) THEN BEGIN
      PRINT, "data file header"
      PRINT, "================"
      PRINT, ""
      PRINT, "nblocks = "+STRTRIM(STRING(nblocks),2)+" (# blocks/file)"
      PRINT, "ntraces = "+STRTRIM(STRING(ntraces),2)+" (# traces/block)"
      PRINT, "np = "+STRTRIM(STRING(np),2)+" (# elements/trace)"
      PRINT, "ebytes = "+STRTRIM(STRING(ebytes),2)+" (# bytes/element)"
      PRINT, "tbytes = "+STRTRIM(STRING(tbytes),2)+" (# bytes/trace)"
      PRINT, "bbytes = "+STRTRIM(STRING(bbytes),2)+" (# bytes/block)"
      PRINT, "vers_id = "+STRTRIM(STRING(vers_id),2)
      PRINT, "status = "+STRTRIM(STRING(status),2)
      PRINT, "nbheaders = "+STRTRIM(STRING(nbheaders),2)+" (# blocks headers/block)"
      PRINT, ""
      PRINT, "data summary:"
      PRINT, "============="
      PRINT, ""
      PRINT, "data file header at start of file : "+ $
        STRTRIM(STRING(data_file_header_size),2)+" bytes"
      PRINT, "   block header at start of each block : "+ $
        STRTRIM(STRING(block_header_size),2)+" bytes"
      PRINT, "   data size of each block : "+ $
        STRTRIM(STRING(np*ebytes*ntraces),2)+" bytes"
      PRINT, "   total size of each block : "+ $
        STRTRIM(STRING(block_header_size+np*ebytes),2)+" bytes"
      PRINT, "   number of blocks : "+STRTRIM(STRING(nblocks),2)
      PRINT, "data size : "+STRTRIM(STRING(np*ebytes*ntraces*nblocks),2)
      PRINT, "file size : "+STRTRIM(STRING((np*ebytes*ntraces+ $
                                            block_header_size)*nblocks+data_file_header_size),2)
      PRINT, ""
   ENDIF

; set the staus bits to zero as default
   S_NI2=0
   S_NI=0
   S_NF=0
   S_NP=0
   S_NP=0
   S_TRANSF=0
   S_SECND=0
   S_ACQPAR=0
   S_HYPERCOMPLEX=0
   S_COMPLEX=0
   S_FLOAT=0
   S_32=0
   S_SPEC=0
   S_DATA=0

; interpret the status bits:
; ==========================

   IF (status GT 32767) THEN BEGIN
      ; not used
      status=status-32768
   ENDIF
   IF (status GT 16383) THEN BEGIN
      S_NI2=1                   ; ni2 dimension is active
      status=status-16384
   ENDIF
   IF (status GT 8191) THEN BEGIN
      S_NI=1                    ; ni dimension is active
      status=status-8192
   ENDIF
   IF (status GT 4095) THEN BEGIN
      S_NF=1                    ; nf dimension is active
      status=status-4096
   ENDIF
   IF (status GT 2047) THEN BEGIN
      S_NP=1                    ; np dimension is active
      status=status-2048
   ENDIF
   IF (status GT 1023) THEN BEGIN
      ; not used
      status=status-1024
   ENDIF
   IF (status GT 511) THEN BEGIN
      S_TRANSF=1                ; 0=regular, 1=transposed
      status=status-512
   ENDIF
   IF (status GT 255) THEN BEGIN
      S_SECND=1                 ; 0=first FT, 1=second FT
      status=status-256
   ENDIF
   IF (status GT 127) THEN BEGIN
      S_ACQPAR=1                ; 0=not Acqpar, 1=Acqpar
      status=status-128
   ENDIF
   IF (status GT 63) THEN BEGIN
      ; not_used
      status=status-64
   ENDIF
   IF (status GT 31) THEN BEGIN
      S_HYPERCOMPLEX=1          ; 1=hypercomplex
      status=status-32
   ENDIF
   IF (status GT 15) THEN BEGIN
      S_COMPLEX=1               ; 0=real, 1=complex
      status=status-16
   ENDIF
   IF (status GT 7) THEN BEGIN
      S_FLOAT=1                 ; 0=integer, 1=floating point
      status=status-8
   ENDIF
   IF (status GT 3) THEN BEGIN
      S_32=1                    ; if integer (i.e., S_FLOAT=0) then 0=16 bit, 1=32 bit
      status=status-4
   ENDIF
   IF (status GT 1) THEN BEGIN
      S_SPEC=1                  ; 0=FID, 1=spectrum
      status=status-2
   ENDIF
   IF (status GT 0) THEN BEGIN
      S_DATA=1                  ; 0=no data, 1=data
      status=status-1
   ENDIF

; ***status should now be zero...***
   IF (status NE 0) THEN BEGIN
      PRINT, "Error: could not interpret the status bits"
   ENDIF

   IF (verbose GT 0) THEN BEGIN
      PRINT, "interpretation of status bits:"
      PRINT, "=============================="
      PRINT, ""
      IF (S_DATA NE 1) THEN BEGIN
         PRINT, "no data"
      ENDIF
      IF (S_SPEC EQ 0) THEN BEGIN
         PRINT, "FID"
      ENDIF ELSE BEGIN
         PRINT, "spectrum"
      ENDELSE
      IF (S_FLOAT EQ 0) THEN BEGIN
         IF (S_32 EQ 0) THEN BEGIN
            PRINT, '16 bit integer data'
         ENDIF ELSE BEGIN
            PRINT, '32 bit integer data'
         ENDELSE
      ENDIF ELSE BEGIN
         PRINT, "floating point data"
      ENDELSE
      IF (S_COMPLEX EQ 0) THEN BEGIN
         PRINT, "real data"
      ENDIF ELSE BEGIN
         PRINT, "complex data"
      ENDELSE
      IF (S_HYPERCOMPLEX EQ 1) THEN PRINT, "hypercomplex data"
      IF (S_ACQPAR EQ 0) THEN BEGIN
         PRINT, "not Acqpar"
      ENDIF ELSE BEGIN
         PRINT, "Acqpar"
      ENDELSE
      IF (S_SECND EQ 0) THEN BEGIN
         PRINT, "ready for first FT"
      ENDIF ELSE BEGIN
         PRINT, "ready for second FT"
      ENDELSE
      IF (S_TRANSF EQ 1) THEN PRINT, "transposed"
      IF (S_NP EQ 1) THEN PRINT, "np dimension is active"
      IF (S_NF EQ 1) THEN PRINT, "nf dimension is active"
      IF (S_NI EQ 1) THEN PRINT, "ni dimension is active"
      IF (S_NI2 EQ 1) THEN PRINT, "ni2 dimension is active"
   ENDIF

   IF (S_FLOAT EQ 1) THEN BEGIN
      IF (S_COMPLEX EQ 1) THEN BEGIN
         tmp_block_arr=FLTARR(2,np/2*ntraces)
         main_data_arr=FLTARR(2,np/2*ntraces,nblocks)
      ENDIF
      IF (S_COMPLEX EQ 0) THEN BEGIN
         tmp_block_arr=FLTARR(2,np/2*ntraces)
         main_data_arr=FLTARR(2,np/2*ntraces,nblocks)
      ENDIF
   ENDIF ELSE BEGIN
      IF (S_COMPLEX EQ 1) THEN BEGIN
         IF (S_32 EQ 0) THEN BEGIN
            tmp_block_arr=INTARR(2,np/2*ntraces)
            main_data_arr=INTARR(2,np/2*ntraces,nblocks)
         ENDIF
         IF (S_32 EQ 1) THEN BEGIN
            tmp_block_arr=LONARR(2,np/2*ntraces)
            main_data_arr=LONARR(2,np/2*ntraces,nblocks)
         ENDIF
      ENDIF
      IF (S_COMPLEX EQ 0) THEN BEGIN
         IF (S_32 EQ 0) THEN BEGIN
            tmp_block_arr=INTARR(2,np/2*ntraces)
            main_data_arr=INTARR(2,np/2*ntraces,nblocks)
         ENDIF
         IF (S_32 EQ 1) THEN BEGIN
            tmp_block_arr=LONARR(2,np/2*ntraces)
            main_data_arr=LONARR(2,np/2*ntraces,nblocks)
         ENDIF
      ENDIF
   ENDELSE


; definition of block header
; ==========================
   scale=FIX(0)                 ; scaling factor
   status=FIX(0)                ; status of data in block
   index=FIX(0)                 ; block index
   mode=FIX(0)                  ; mode of data in block
   ctcount=LONG(0)              ; ct value for fid
   lpval=FLOAT(0)               ; f2 (2D-f1) left phase in phase file
   rpval=FLOAT(0)               ; f2 (2D-f1) right phase in phase file
   lvl=FLOAT(0)                 ; level drift correction
   tlt=FLOAT(0)                 ; tilit drift correction

   FOR block_count=LONG(0), LONG(nblocks)-1 DO BEGIN

      IF (verbose GT 0) THEN BEGIN
         PRINT, ""
         PRINT, "reading block "+STRTRIM(STRING(block_count+1),2)+"..."
      ENDIF

      READU, unit, scale
      READU, unit, status
      READU, unit, index
      READU, unit, mode
      READU, unit, ctcount
      READU, unit, lpval
      READU, unit, rpval
      READU, unit, lvl
      READU, unit, tlt
      
      IF swap_bytes THEN BEGIN
         scale    = swap_endian(scale)
         status   = swap_endian(status)
         index    = swap_endian(index)
         mode     = swap_endian(mode)
         ctcount  = swap_endian(ctcount)
         lpval    = swap_endian(lpval)
         rpval    = swap_endian(rpval)
         lvl      = swap_endian(lvl)
         tlt      = swap_endian(tlt)
      END

      IF (verbose GT 0) THEN BEGIN
         PRINT, ""
         PRINT, "scale = "+STRTRIM(STRING(scale),2)
         PRINT, "status = "+STRTRIM(STRING(status),2)
         PRINT, "index = "+STRTRIM(STRING(index),2)
         PRINT, "mode = "+STRTRIM(STRING(mode),2)
         PRINT, "ctcount = "+STRTRIM(STRING(ctcount),2)
         PRINT, "lpval = "+STRTRIM(STRING(lpval),2)
         PRINT, "rpval = "+STRTRIM(STRING(rpval),2)
         PRINT, "lvl = "+STRTRIM(STRING(lvl),2)
         PRINT, "tlt = "+STRTRIM(STRING(tlt),2)
      ENDIF

      ; interpret the status bits for this block
      ; ========================================

      S_NI2=0
      S_NI=0
      S_NF=0
      S_NP=0
      S_NP=0
      S_TRANSF=0
      S_SECND=0
      MORE_BLOCKS=0
      NP_CMPLX=0
      NF_CMPLX=0
      NI_CMPLX=0
      NI2_CMPLX=0

      ; interpret the status bits:
      ; ==========================

      IF (status GT 32767) THEN BEGIN
         ; not used
         status=status-32768
      ENDIF
      IF (status GT 16383) THEN BEGIN
         ; not used
         status=status-16384
      ENDIF
      IF (status GT 8191) THEN BEGIN
         ; not used
         status=status-8192
      ENDIF
      IF (status GT 4095) THEN BEGIN
         ; not used
         status=status-4096
      ENDIF
      IF (status GT 2047) THEN BEGIN
         NI2_CMPLX=1            ; 0=real, 1=complex
         status=status-2048
      ENDIF
      IF (status GT 1023) THEN BEGIN
         NI_CMPLX=1             ; 0=real, 1=complex
         status=status-1024
      ENDIF
      IF (status GT 511) THEN BEGIN
         NF_CMPLX=1             ; 0=real, 1=complex
         status=status-512
      ENDIF
      IF (status GT 255) THEN BEGIN
         NP_CMPLX=1             ; 0=real, 1=complex
         status=status-256
      ENDIF
      IF (status GT 127) THEN BEGIN
         MORE_BLOCKS=1          ; 0=absent, 1=present
         status=status-128
      ENDIF
      IF (status GT 63) THEN BEGIN
         ; not_used
         status=status-64
      ENDIF
      IF (status GT 31) THEN BEGIN
         S_HYPERCOMPLEX=1       ; 1=hypercomplex
         status=status-32
      ENDIF
      IF (status GT 15) THEN BEGIN
         S_COMPLEX=1            ; 0=real, 1=complex
         status=status-16
      ENDIF
      IF (status GT 7) THEN BEGIN
         S_FLOAT=1              ; 0=integer, 1=floating point
         status=status-8
      ENDIF
      IF (status GT 3) THEN BEGIN
         S_32=1                 ; if integer (i.e., S_FLOAT=0) then 0=16 bit, 1=32 bit
         status=status-4
      ENDIF
      IF (status GT 1) THEN BEGIN
         S_SPEC=1               ; 0=FID, 1=spectrum
         status=status-2
      ENDIF
      IF (status GT 0) THEN BEGIN
         S_DATA=1               ; 0=no data, 1=data
         status=status-1
      ENDIF

      ; ***status should now be zero...***
      IF (status NE 0) THEN BEGIN
         PRINT, "Error: could not interpret the status bits"
      ENDIF

      IF (verbose GT 0) THEN BEGIN
         PRINT, ""
         PRINT, "interpretation of status bits:"
         PRINT, "=============================="
         PRINT, ""
         IF (S_DATA EQ 0) THEN BEGIN
            PRINT, "no data"
         ENDIF
         IF (S_SPEC EQ 0) THEN BEGIN
            PRINT, "FID"
         ENDIF ELSE BEGIN
            PRINT, "spectrum"
         ENDELSE
         IF (S_FLOAT EQ 0) THEN BEGIN
            IF (S_32 EQ 0) THEN BEGIN
               PRINT, '16 bit integer data'
            ENDIF ELSE BEGIN
               PRINT, '32 bit integer data'
            ENDELSE
         ENDIF ELSE BEGIN
            PRINT, "floating point data"
         ENDELSE
         IF (S_COMPLEX EQ 0) THEN BEGIN
            PRINT, "real data"
         ENDIF ELSE BEGIN
            PRINT, "complex data"
         ENDELSE
         IF (S_HYPERCOMPLEX EQ 1) THEN PRINT, "hypercomplex data"
         IF (MORE_BLOCKS EQ 1) THEN BEGIN
            PRINT, "more blocks"
         ENDIF ELSE BEGIN
            PRINT, "last block"
         ENDELSE
         IF (NP_CMPLX EQ 1) THEN PRINT, "np dimension is complex"
         IF (NF_CMPLX EQ 1) THEN PRINT, "nf dimension is complex"
         IF (NI_CMPLX EQ 1) THEN PRINT, "ni dimension is complex"
         IF (NI2_CMPLX EQ 1) THEN PRINT, "ni2 dimension is complex"
      ENDIF

      ; interpret the mode bits for this block
      ; ======================================

      NP_PHMODE=0
      NP_AVMODE=0
      NP_PWRMODE=0
      NF_PHMODE=0
      NF_AVMODE=0
      NF_PWRMODE=0
      NI_PHMODE=0
      NI_AVMODE=0
      NI_PWRMODE=0
      NI2_PHMODE=0
      NI2_AVMODE=0
      NI2_PWRMODE=0

      ; interpret the mode bits:
      ; ========================
      mode_start=mode

      IF (mode GT 32767) THEN BEGIN
         ; not used
         mode=mode-32768
      ENDIF
      IF (mode GT 16383) THEN BEGIN
         NI2_PWRMODE=1
         mode=mode-16384
      ENDIF
      IF (mode GT 8191) THEN BEGIN
         NI2_AVMODE=1
         mode=mode-8192
      ENDIF
      IF (mode GT 4095) THEN BEGIN
         NI2_PHMODE=1
         mode=mode-4096
      ENDIF
      IF (mode GT 2047) THEN BEGIN
         ; not used
         mode=mode-2048
      ENDIF
      IF (mode GT 1023) THEN BEGIN
         NI_PWRMODE=1
         mode=mode-1024
      ENDIF
      IF (mode GT 511) THEN BEGIN
         NI_AVMODE=1
         mode=mode-512
      ENDIF
      IF (mode GT 255) THEN BEGIN
         NI_PHMODE=1
         mode=mode-256
      ENDIF
      IF (mode GT 127) THEN BEGIN
         ; not used
         mode=mode-128
      ENDIF
      IF (mode GT 63) THEN BEGIN
         NF_PWRMODE=1
         mode=mode-64
      ENDIF
      IF (mode GT 31) THEN BEGIN
         NF_AVMODE=1
         mode=mode-32
      ENDIF
      IF (mode GT 15) THEN BEGIN
         NF_PHMODE=1
         mode=mode-16
      ENDIF
      IF (mode GT 7) THEN BEGIN
         ; not used
         mode=mode-8
      ENDIF
      IF (mode GT 3) THEN BEGIN
         NP_PWRMODE=1
         mode=mode-4
      ENDIF
      IF (mode GT 1) THEN BEGIN
         NP_AVMODE=1
         mode=mode-2
      ENDIF
      IF (mode GT 0) THEN BEGIN
         NP_PHMODE=1
         mode=mode-1
      ENDIF

      ; ***mode should now be zero...***
      IF (mode NE 0) THEN BEGIN
         PRINT, "Error: could not interpret the mode bits"
      ENDIF

      IF (verbose GT 0) THEN BEGIN
         IF (mode_start NE 0) THEN BEGIN
            PRINT, ""
            PRINT, "interpretation of mode bits:"
            PRINT, "============================"
            PRINT, ""
         ENDIF
         IF (NP_PHMODE EQ 1) THEN PRINT, "NP_PHMODE=1"
         IF (NP_AVMODE EQ 1) THEN PRINT, "NP_AVMODE=1"
         IF (NP_PWRMODE EQ 1) THEN PRINT, "NP_PWRMODE=1"
         IF (NF_PHMODE EQ 1) THEN PRINT, "NF_PHMODE=1"
         IF (NF_AVMODE EQ 1) THEN PRINT, "NF_AVMODE=1"
         IF (NF_PWRMODE EQ 1) THEN PRINT, "NF_PWRMODE=1"
         IF (NI_PHMODE EQ 1) THEN PRINT, "NI_PHMODE=1"
         IF (NI_AVMODE EQ 1) THEN PRINT, "NI_AVMODE=1"
         IF (NI_PWRMODE EQ 1) THEN PRINT, "NI_PWRMODE=1"
         IF (NI2_PHMODE EQ 1) THEN PRINT, "NI2_PHMODE=1"
         IF (NI2_AVMODE EQ 1) THEN PRINT, "NI2_AVMODE=1"
         IF (NI2_PWRMODE EQ 1) THEN PRINT, "NI2_PWRMODE=1"
      ENDIF

      READU, unit, tmp_block_arr
      
      IF swap_bytes THEN tmp_block_arr = swap_endian(tmp_block_arr)

      main_data_arr(0,*,block_count)=tmp_block_arr(0,*)-lvl
      main_data_arr(1,*,block_count)=tmp_block_arr(1,*)-tlt

   ENDFOR

   CLOSE, unit
   FREE_LUN, unit

   RETURN, REFORM(COMPLEX(main_data_arr(0,*,*),main_data_arr(1,*,*)))

END

FUNCTION Rd_varian_spiral, fname, ni, nk, ns, nv, output
   
   IF output THEN PRINT,'----- read spiral data ------'
   
   data = rd_varian_bin(fname, ntraces)
   ns = ntraces/ni
   s = SIZE(data)
   nk = s(1)/ns
   s_int = nk/ni
   IF s(0) EQ 2 THEN nv = s(2) ELSE nv = 1
   
   IF output THEN BEGIN
      PRINT, 'traces       :', ntraces
      PRINT, 'interleaves  :', ni
      PRINT, 'samples/int  :', s_int
      PRINT, 'slices       :', ns
      PRINT, 'volumes      :', nv
   END
   
   ; reorder data assuming interleaved acquisition of slices
   m = COMPLEXARR(s_int,ni,ns,nv)
   data = REFORM(data,s_int,ns,ni,nv)
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO FOR ii = 0, ni-1 DO m(*,ii,is,iv) = data(*,is,ii,iv) 
   m = REFORM(m,s_int,ni,ns,nv)
    
   
   RETURN, m
END

PRO Write_varian, data, fname, ntrac

   s = SIZE(data)
   
   GET_LUN, unit
   OPENW, unit, fname+'fid'
 
   ; definition of data file header
   ; ==============================
   nblocks=LONG(0)              ; number of blocks in file
   ntraces=LONG(0)              ; number of traces per block
   np=LONG(0)                   ; number of elements per trace
   ebytes=LONG(0)               ; number of bytes per element
   tbytes=LONG(0)               ; number of bytes per trace
   bbytes=LONG(0)               ; number of bytes per block
   vers_id=FIX(0)               ; software version, file_id status bits
   status=FIX(0)                ; status of whole file
   nbheaders=LONG(0)            ; number of block headers per block

   ; fill header
   ntraces = LONG(ntrac)
   IF s(0) GT 1 THEN nblocks = LONG(s(2)) ELSE nblocks = LONG(1)
   np = LONG(s(1)/ntrac*2)
   ebytes = LONG(4)
   status = FIX(8+16)
   
   WRITEU, unit, nblocks
   WRITEU, unit, ntraces
   WRITEU, unit, np
   WRITEU, unit, ebytes
   WRITEU, unit, tbytes
   WRITEU, unit, bbytes
   WRITEU, unit, vers_id
   WRITEU, unit, status
   WRITEU, unit, nbheaders
   
   FOR block_count=LONG(0), LONG(nblocks)-1 DO BEGIN

      ; definition of block header
      ; ==========================
      scale=FIX(0)              ; scaling factor
      status=FIX(0)             ; status of data in block
      index=FIX(0)              ; block index
      mode=FIX(0)               ; mode of data in block
      ctcount=LONG(0)           ; ct value for fid
      lpval=FLOAT(0)            ; f2 (2D-f1) left phase in phase file
      rpval=FLOAT(0)            ; f2 (2D-f1) right phase in phase file
      lvl=FLOAT(0)              ; level drift correction
      tlt=FLOAT(0)              ; tilit drift correction
      
      WRITEU, unit, scale
      WRITEU, unit, status
      WRITEU, unit, index
      WRITEU, unit, mode
      WRITEU, unit, ctcount
      WRITEU, unit, lpval
      WRITEU, unit, rpval
      WRITEU, unit, lvl
      WRITEU, unit, tlt

      ;data = REFORM(data, s(1)*nblocks)

      WRITEU, unit, data(*,block_count)
   END

   CLOSE, unit
   FREE_LUN, unit
   
END

;=============================================================================
function get_varian_par, procpar_dir, par_name
;=============================================================================
; function to return varian procpar parameters
; procpar_dir is the directrory stem containing the procpar file
; par_name is a string containing the string parameter
;=============================================================================
; Peter Jezzard, Oxford, July 2000
;=============================================================================

fname=procpar_dir+'procpar'
par_val=-9999999

get_lun, unit
openr, unit, fname

found_par=0
line=string(0)

while ((eof(unit) eq 0) and (found_par eq 0)) do begin
  readf, unit, line
  if (strpos(line, par_name+' ') eq 0) then begin
    found_par=1
    readf, unit, line
    line_len=strlen(line)

    space_positions=strpos(line, ' ')

    found_all_spaces=0
    tmp_line_len=strlen(line)
    tmp_line=strmid(line, space_positions+1, tmp_line_len-space_positions-1)
    old_pos=space_positions

    while(found_all_spaces eq 0) do begin
      tmp_pos=strpos(tmp_line, ' ')
      if (tmp_pos eq -1) then begin
        found_all_spaces=1
      endif else begin
        new_pos=old_pos+tmp_pos+1
        space_positions=[space_positions, new_pos]
        old_pos=new_pos
        tmp_line_len=strlen(line)
        tmp_line=strmid(tmp_line, tmp_pos+1, tmp_line_len-tmp_pos-1)
      endelse
    ENDWHILE
    
    num_pars=FIX(STRMID(line,0,space_positions(0)))    

    IF (num_pars EQ 1) THEN BEGIN
       fourth_letter = STRMID(line,space_positions(0)+2,1)
       IF fourth_letter EQ 'n' THEN par_val = 'n' ELSE IF $
         fourth_letter EQ 'y' THEN par_val = 'y' ELSE IF $
         fourth_letter EQ 'o' THEN par_val = 'o' ELSE IF $
         fourth_letter EQ 's' THEN par_val = 's' ELSE $
         par_val=FLOAT(STRMID(line,space_positions(0),line_len-space_positions(0)))
    ENDIF ELSE BEGIN
       par_val=FLOAT(STRMID(line,space_positions(0), $
          space_positions(1)-space_positions(0)))
    endelse

    if (num_pars gt 1) then begin
      if (n_elements(space_positions) eq num_pars) then $
         space_positions=[space_positions,line_len]
      for i=1, num_pars-1 do begin
        par_val=[par_val, float(strmid(line,space_positions(i), $
            space_positions(i+1)))]
      endfor
    endif
  endif
endwhile

close, unit
free_lun, unit

if (found_par eq 0) then begin
  print, 'string not found'
endif
return, par_val

END

PRO Write_varian_par, unit, par_name, par_val
   PRINTF, unit, par_name+' '
   PRINTF, unit, '1 '+par_val
   PRINTF, unit, '0'
END
