; ==========================================================================
; Spiral Reconstruction
; ==========================================================================
; written in IDL by Markus Weiger, FMRIB Centre, University of Oxford, 2001
; (Gridding routine by Klaas Pruessmann, ETH Zuerich)

; run spfunc, spb0 first, or type "@sp"

; all units in mm, ms, kHz, mT   


PRO Spiral
; main program with parameters
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   COMMON Logging, logging_level


   
   ; ----- PARAMETERS --------------------------------------------------------------

   
   ;path      = '/home/soy/scratch/weiger/'
   path      = '~weiger/MRdata/'
   path      = path+'005380_2001_11_23/'
   
   dname     = 'series_3.3'     ; name of spiral data file
   tname     = 'series_3.4'     ; name of trajectory data file, '' -> calculate
   B0name    = 'series_4.1'     ; name of B0-scan file
   tpath     = ''               ; path of trajetory data, '' -> same as data path

   ; ----- output -----
   isize            = 2         ; zoom factor for image display
   show_each_image  = 0         ; show each image after been reconstructed
   show_all_images  = 1         ; show all images at end of reconstruction
   show_trajectory  = 0
   show_B0map       = 1         ; 0 = no, 1 = yes, 2 = window2tif
   write_result     = 1         ; 1 = avw, 2 = tif, 3 = window2tif
   logging_level    = 3         ; 0 - 4
   
   ; ----- slice and volume loops -----
   smin      = -1               ; slice number, -1 = all
   smax      = 0
   vmin      = -1                ; volume number, -1 = all
   vmax      = 0
   
   ; ----- B0 correction -----
   B0type    = 2               ; -2 = gridding, 2 = CPR with mask, 3 = CPR with polynoms,
                                ; -1 = integration, 1 = integration with CPR
   fast      = 1                ; CPR: 0 = grid each time, 1 = grid once
   ord	     = 10               ; CPR: order/number of reconstruction steps
   dw0 	     = 0.0              ; additional frequency offset [2pi/ms]
   
   delay     = 0                ; gradient propagation delay [samples], needed for calculated trajectory

   ; ----- gridding -----
   over      = 2		; oversampling factor
   fuss      = 4		; size of convolution kernel
   breite    = 4.0              ; width of convolution kernel


   

   ; ----- MAIN ---------------------------------------------------------------------
   

   bbname = b0name
   
   ext = '.fid/'
   name = path+dname+ext
   IF tpath EQ '' THEN tpath = path
   IF tname NE '' THEN tname = tpath+tname+ext
   B0path = path+B0name+ext
   B0name = 'wmaps'

   ; procedures that have to be run only once per data set
   get_parameters, name
   init_const   
   get_data, name, tname, delay, show_trajectory
   IF ABS(B0type) GT 1 THEN make_weights
   IF B0type GT 0 THEN Make_b0map, B0path, B0name, show_B0map

   ; reconstruction with loops for slices and dynamics
   IF show_each_image THEN BEGIN
      WINDOW, xsize=isize*n, ysize=isize*n, title='EACH IMAGE', /free
      each_window = !d.window
   END
   IF smin LT 0 THEN BEGIN
      smin = 0
      smax = ns-1
   END
   IF vmin LT 0 THEN BEGIN
      vmin = 0
      vmax = nv-1
   END
   nns = smax-smin+1
   nnv = vmax-vmin+1
   result = COMPLEXARR(n,n,nns,nnv)
   maske = Mask_arr(2, n, n, n, n, n/2, n/2)
   
   IF show_each_image OR show_all_images OR write_result THEN $
     FOR iv = vmin, vmax DO BEGIN
      iiv = iv-vmin
      FOR is = smin, smax DO BEGIN
         logstring, 'volume '+trimstr(iv)+' slice '+trimstr(is), 2
         iis = is-smin
         image = reconstruct(B0type, ord, fast, dw0, is, iv, is EQ smin AND iv EQ vmin)*maske
         result(*,*,iis,iiv) = image(*,*)
         IF show_each_image THEN BEGIN
            WSET, each_window
            tvscl, (REBIN(ABS(REFORM(result(*,*,iis,iiv))),isize*n,isize*n))
         END
      END
   END

   ; display result
   IF show_all_images THEN BEGIN
      WINDOW, title='ALL IMAGES', xsize = mini(nns*nnv,4)*isize*n, $
        ysize = mini(nns*nnv/4+1*((nns*nnv MOD 4) NE 0),5)*isize*n, /free
      scal = 1.0
      REPEAT BEGIN
         ii = 0
         FOR iv = 0, nnv-1 DO FOR is = 0, nns-1 DO BEGIN
            TV,(REBIN(ABS(REFORM(result(*,*,is,iv)))/MAX(ABS(result))*255*scal,isize*n,isize*n)), ii
            ii = ii + 1
         END
         READ, 'Image Scaling (type 0 for exit): ', scal
      END UNTIL scal EQ 0
   END

   ; write result
   result_name = name+'result'+trimstr(B0type);+'_'+bbname
   CASE write_result OF
      0:
      1: write_avw, result, result_name
      2: FOR iv = 0, nnv-1 DO FOR is = 0, nns-1 DO $
        To_Tiff_8Bits, ABS(REFORM(result(*,*,is,iv))), result_name+'_'+numstr(is,2)+'_'+numstr(iv,3)
      3: window_to_tif, result_name
   END
END

PRO Get_parameters, path
; read parameters from varian parameter file
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   cm_mm  = 10.0
   s_ms   = 1000.0

   n        = get_varian_par(path, 'matrix')
   ni       = get_varian_par(path, 'nint')
   fov      = get_varian_par(path, 'lro')*cm_mm
   alpha    = get_varian_par(path, 'alpha')
   tacq     = get_varian_par(path, 'tacq')*s_ms
   reversed = get_varian_par(path, 'reversed') EQ 'y'
   xsamp    = get_varian_par(path, 'xsamp1')
   pss      = get_varian_par(path, 'pss')*cm_mm
   
   sindex = SORT(pss)
END

PRO Init_const
; initialise some constants
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   two_pi	= 2.0*!Pi
   gamma 	= two_pi*42.57493
   
   ; some numbers might become very large
   n  = LONG(n)
   ni = LONG(ni)
   ns = 1
   nv = 1
   
   !x.style=1
   !y.style=1
   !x.margin=6
   !y.margin=4
   CLOSE,/all
   DEVICE, window_state = w_open
   FOR i = 0, 64 DO IF w_open(i) THEN WDELETE, i
END

PRO Get_data, fname, tname, delay, output
; get data and trajectory from varian files

   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   logstring, '----- read data -----', 1
   
   m = rd_varian_spiral(fname, ni, nk, ns, nv, 0)
   logstring, "slices  = "+trimstr(ns), 2
   logstring, "volumes = "+trimstr(nv), 2
   make_trajectory, output
   IF tname NE '' THEN get_trajectory, tname, output      
   
   m = shift_arr(m, -delay)
   m = m/MAX(ABS(m))
   
   ; adapt to axis orientation of VARIAN scanner
   exchange, kx, ky
      
END

FUNCTION Demodulate, data, tau, dw0
; demodulate off-resonance
   
   RETURN, data*EXP(-COMPLEX(0,1)*dw0*tau)
END

FUNCTION Grid, data
; gridding routine
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   kdata = REFORM(data, nk)
   inv_dens = REFORM(inv_dens, nk)

   kspace_new = COMPLEXARR(over*(n+fuss),over*(n+fuss))     
   FOR i = LONG(0),nk-1 DO BEGIN
      von_x = k_xvon(i) + over*fuss/2
      bis_x = von_x + over*fuss-1
      von_y = k_yvon(i) + over*fuss/2
      bis_y = von_y + over*fuss-1
      IF (von_x GE 0) AND (von_y GE 0) AND (bis_x LT over*(n+fuss)) AND (bis_y LT over*(n+fuss)) THEN $
        kspace_new(von_x:bis_x,von_y:bis_y) = kspace_new(von_x:bis_x,von_y:bis_y) + inv_dens(i)*kdata(i)*weight(i,*,*) 
   END     
   kspace_new(over*fuss/2:over*fuss-1,*,*) = kspace_new(over*fuss/2:over*fuss-1,*,*) + kspace_new(over*(n+fuss/2):over*(n+fuss)-1,*,*)
   kspace_new(over*n:over*(n+fuss/2)-1,*,*) = kspace_new(over*n:over*(n+fuss/2)-1,*,*) + kspace_new(0:over*fuss/2-1,*,*)
   kspace_new(*,over*fuss/2:over*fuss-1,*) = kspace_new(*,over*fuss/2:over*fuss-1,*) + kspace_new(*,over*(n+fuss/2):over*(n+fuss)-1,*)
   kspace_new(*,over*n:over*(n+fuss/2)-1,*) = kspace_new(*,over*n:over*(n+fuss/2)-1,*) + kspace_new(*,0:over*fuss/2-1,*)
   kspace_new = kspace_new(over*fuss/2:over*(n+fuss/2)-1,over*fuss/2:over*(n+fuss/2)-1,*)
   
   RETURN, kspace_new
END

FUNCTION To_space, kdata
; do FFT to real space, undo filtering of gridding kernel, and remove oversampling
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   ; fft
   temp = swap(fft(swap(kdata),1))

   ; undo kb-filtering and oversampling
   result = temp(n*over/2-n/2:n*over/2+n/2-1,n*over/2-n/2:n*over/2+n/2-1)/kernel

   RETURN, result       
END

PRO Make_b0map, path, name, output
; get B0-map from file or make one from data
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   
   logstring, '----- B0-map -----', 1
      
   read_wmaps = file_exists(path+name+'.hdr')
   IF read_wmaps THEN BEGIN
      ; fieldmap had already been created and was saved, so just read it
      wmaps = read_avw(path+name,0)
      s = SIZE(wmaps)
      ; if the matrix size is wrong, create fieldmap again
      IF s(1) NE n THEN read_wmaps = 0
   END
   IF read_wmaps NE 1 THEN BEGIN
      ; fieldmap can be acquired with spin-echo sequence or spiral
      fm_sequence = get_varian_par(path, 'seqfil')
      IF fm_sequence EQ 'o' THEN BEGIN
         te = get_varian_par(path, 'asym_time')*1000.0
         dte = te(1)-te(0)
         wmaps = phasemap(path,n,0)/dte
      END ELSE BEGIN
         te = get_varian_par(path, 'te')*1000.0
         dte = te(1)-te(0)
         wmaps = phasemap(path+'result-2',n,1)/dte
      END
      write_avw, wmaps, path+name
   END
      
   
   IF output GT 0 THEN BEGIN
      isize = 2
      WINDOW, title='B0-MAP', xsize = mini(ns,4)*isize*n, ysize = mini(ns/4+1*((ns MOD 4) NE 0),5)*isize*n, /free
      FOR is = 0, ns-1 DO BEGIN 
         wmap = wmaps(*,*,is)
         wmean = (MIN(wmap)+MAX(wmap))/2.0
         wdelta = MAX(wmap)-MIN(wmap)
         tdelta = MAX(tau)-MIN(tau)
         logstring, 'slice, wmean, wdelta, wXtau, maxdwt: '$
           +trimstr(is)+', '+trimstr(wmean)+', '+trimstr(wdelta)+', '$
           +trimstr(wdelta*tdelta/two_pi)+', '+trimstr(max_diff(wmap)*tdelta/two_pi), 3
         scal = 255.0
         TV, (REBIN(REFORM(wmap+0.5)*scal, isize*n, isize*n)), is
      END
      IF output EQ 2 THEN window_to_tif, path+name
   END
END

FUNCTION Reconstruct, type, ord, fast, dw0, is, iv, first_time
; do spiral reconstruction by gridding or integration, with or without CPR
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex

   wmean = 0.0
   
   IF type GT 0 THEN BEGIN
      ; centre frequencies
      wmap = wmaps(*,*,is)
      wmean = (MIN(wmap)+MAX(wmap))/2.0
      wdelta = MAX(wmap)-MIN(wmap)
      wmap = wmap-wmean
      
      image = COMPLEXARR(n,n)
      image(*,*) = 0
      IF first_time AND fast AND type GT 1 THEN BEGIN
         gtau = FLOAT(grid(tau))
         ; correct for scaling of gridding
         IF NOT reversed THEN gtau = gtau/MIN(gtau)*MIN(tau) ELSE gtau = gtau/MAX(gtau)*MAX(tau)
      END
   END
   
   mm = demodulate(REFORM(m(*,*,sindex(is),iv)), tau, wmean+dw0)
   
   ; integration
   IF ABS(type) EQ 1 THEN BEGIN
      kdata = REFORM(mm, nk)
      tau = REFORM(tau, nk)
      inv_dens = REFORM(inv_dens, nk)
      kx = REFORM(kx, nk)
      ky = REFORM(ky, nk)
      image = COMPLEXARR(n,n)
      image(*,*) = 0
      IF type EQ -1 AND first_time THEN logstring, '----- Integration -----', 1
      IF type EQ  1 AND first_time THEN logstring, '----- Integration + CPR -----', 1
      FOR ix = 0, n-1 DO FOR iy = 0, n-1 DO BEGIN
         x = (ix-n/2)/FLOAT(n-1)*fov
         y = (iy-n/2)/FLOAT(n-1)*fov
         FOR i = 0, nk-1 DO BEGIN
            phase = -kx(i)*x-ky(i)*y
            IF type EQ 1 THEN phase = phase + wmap(ix,iy)*tau(i)
            image(ix,iy) = image(ix,iy)+inv_dens(i)*kdata(i)*EXP(-COMPLEX(0,1)*phase)
         END
      END
   END ELSE CASE type OF
      ; gridding
      -2: BEGIN
         IF first_time THEN logstring, '----- Gridding -----', 1
         
         image = to_space(grid(mm))
      END
      2: BEGIN
         IF first_time THEN logstring, '----- Gridding + CPR with mask -----', 1
         deltaw0 = wdelta/FLOAT(ord)
         IF fast THEN gm = grid(mm)
         FOR i = 0, ord-1 DO BEGIN
            w0i = MIN(wmap)+(i+0.5)*deltaw0
            mask = ((wmap GE w0i-deltaw0/2.0) AND (wmap LT w0i+deltaw0/2.0))
            IF fast THEN BEGIN
               dm = demodulate(gm, gtau, w0i)
               temp = to_space(dm)
            END ELSE BEGIN
               dm = demodulate(mm, tau, w0i)
               temp = to_space(grid(dm))
            END
            image = image+temp*mask
         END
      END
      3: BEGIN
         IF first_time THEN logstring, '----- Gridding + CPR with polynoms -----', 1
         IF fast THEN BEGIN
            maxx = maxi(MAX(gtau)*MAX(wmap), MIN(gtau)*MIN(wmap))
            minx = mini(MAX(gtau)*MIN(wmap), MIN(gtau)*MAX(wmap))
         END ELSE BEGIN
            maxx = maxi(MAX(tau)*MAX(wmap), MIN(tau)*MIN(wmap))
            minx = mini(MAX(tau)*MIN(wmap), MIN(tau)*MAX(wmap))
         END
         coeff = chebyshev(ord, minx, maxx)
         w = DBLARR(n,n)
         w(*,*) = 1
         IF fast THEN gm = grid(mm)
         FOR i = 0, ord-1 DO BEGIN
            IF fast THEN BEGIN
               image = image + coeff(i)*w*to_space(gm)
               gm = gm*gtau
            END ELSE BEGIN
               image = image + coeff(i)*w*to_space(grid(mm))
               mm = mm*tau
            END
            w = w*wmap
         END
      END
   END
   RETURN, image
END

PRO Make_weights
; calculate weighting coefficients for gridding using a Kaiser-Bessel kernel
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   
   logstring,'----- make_weights -----', 1
   dk_low = 2*!pi/fov
   dk = dk_low/over
   beta = 4.0*breite
   kmin = -over*0.5*n*dk                   
   x_0 = ((FINDGEN(over*fuss,over*fuss) MOD (over*fuss)) - over*fuss/2.0) * dk              
   y_0 = TRANSPOSE(x_0)                 
   weight = COMPLEXARR(nk,over*fuss,over*fuss)
   k_xvon = LONARR(nk)
   k_yvon = LONARR(nk)

   kx = REFORM(kx, nk)
   ky = REFORM(ky, nk)
   
   FOR i=LONG(0),nk-1 DO BEGIN

      ; bessel start
      sgn = 1.0
      IF (kx(i)-kmin) NE 0 THEN BEGIN
         sgn = (kx(i)-kmin)/ABS(kx(i)-kmin)
      END	
      wo_x = FIX((kx(i)-kmin)/dk+0.5*sgn)
      sgn = 1.0
      IF (ky(i)-kmin) NE 0 THEN BEGIN
         sgn = (ky(i)-kmin)/ABS(ky(i)-kmin)
      END	                                     
      wo_y = FIX((ky(i)-kmin)/dk+0.5*sgn)
      off_x = kx(i) - (kmin + dk*wo_x)
      off_y = ky(i) - (kmin + dk*wo_y)
      rad = SQRT((x_0-off_x)^2+(y_0-off_y)^2)/(over*dk)
      dum1 = 1-(2*rad/breite)^2
      sinn = dum1 GE 0
      dum1 = sinn*dum1
      dum = sinn*beta*SQRT(dum1)
      temp = sinn*(1.0/breite*beseli(dum,0))               
      ; bessel end

      k_xvon(i) = wo_x - over*fuss/2
      k_yvon(i) = wo_y - over*fuss/2
      weight(i,*,*) = temp
   END

   weight = weight/MAX(ABS(weight))

   ; build kernel for correction in real space
   x = ((FINDGEN(over*n,over*n) MOD (over*n))-over*n/2.0) * dk
   y = TRANSPOSE(x)  
   rad = SQRT(x^2+y^2)/(over*dk)
   dum1 = 1-(2*rad/breite)^2
   sinn = dum1 GE 0
   dum1 = sinn*dum1
   dum = sinn*beta*SQRT(dum1)
   temp = sinn*(1.0/breite*beseli(dum,0))
   kernel = FLOAT(swap(fft(swap(temp),1)))  
   kernel = kernel(n*over/2-n/2:n*over/2+n/2-1,n*over/2-n/2:n*over/2+n/2-1)
   kernel = kernel/MAX(kernel)
END	

PRO Make_trajectory, output
; calculate spiral trajectory identically as in the vnmr sequence
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   
   logstring,'----- create spiral trajectory -----', 1

   s_int  = nk/ni               ; # samples per interleave including extra samples
   ss_int = s_int-xsamp         ; # samples per interleave for spiral part
   
   kmax   = two_pi*n/fov/2.0    ; maximum k-value	
   nrev   = n/(2.0*ni)          ; # revolutions per interleave
   dt 	  = tacq/(ss_int-1)	; acquisition interval
   a	  = two_pi*nrev		; total angle driven per interleave
   
   kx = FLTARR(s_int, ni)
   ky = FLTARR(s_int, ni)
   inv_dens = FLTARR(s_int, ni)
   tau = FLTARR(s_int, ni)
   vkx = FLTARR(s_int, ni)
   vky = FLTARR(s_int, ni)

   mx = FLTARR(s_int, ni)       ; moments
   my = FLTARR(s_int, ni)

   max_g = 0
   max_s = 0
   
   FOR ii=0, ni-1 DO BEGIN
      psi0 = two_pi*ii/ni
      t = 0
      FOR is=LONG(0), ss_int-1 DO BEGIN

         ; k-space trajectory
         root = SQRT(alpha+(1.0-alpha)*t/tacq)
         psi = -t/tacq/root
         kx(is,ii) = kmax*psi*COS(a*psi+psi0)
         ky(is,ii) = kmax*psi*SIN(a*psi+psi0)

         ; gradients
         dpsi = t*(1.0-alpha)/tacq^2/2.0/root^3-1.0/tacq/root
         vkx(is,ii) = kmax*dpsi*(COS(a*psi+psi0)-a*psi*SIN(a*psi+psi0))
         vky(is,ii) = kmax*dpsi*(SIN(a*psi+psi0)+a*psi*COS(a*psi+psi0))
         tau(is,ii) = is*dt
         
         gx = vkx(is,ii)/gamma		
         gy = vky(is,ii)/gamma      
         max_g = maxi(max_g, ABS(gx))		
         max_g = maxi(max_g, ABS(gy))		
         IF is GT 0 THEN BEGIN
            sx = ABS(gxprev-gx)/dt
            sy = ABS(gyprev-gy)/dt
            max_s = maxi(max_s, ABS(sx))		
            max_s = maxi(max_s, ABS(sy))		
         END
         gxprev = gx
         gyprev = gy

         ; first moments
         mx(is,ii) =  t*gx
         IF is GT 0 THEN mx(is,ii) = mx(is,ii) + mx(is-1,ii)
         my(is,ii) =  t*gy
         IF is GT 0 THEN my(is,ii) = my(is,ii) + my(is-1,ii)
         
         t = t+dt
      END
   END
   tau = tau-tacq/2.0
   
   ; density correction function
   absk = SQRT(kx^2+ky^2)
   absvk = SQRT(vkx^2+vky^2)
   both_zero = kx EQ 0 AND ky EQ 0 ; workaround to avoid undefined result of atan
   angk = ATAN(ky,kx + both_zero)
   angvk = ATAN(vky,vkx)
   abscos = ABS(COS(angvk-angk))
   inv_dens = absvk*absk*abscos
   inv_dens = inv_dens/MAX(inv_dens)
   IF xsamp GT 0 THEN  inv_dens(ss_int:s_int-1,*) = 0

   ; give last turn half weight
   ;s_rev  = s_int/nrev          ; # samples per revolution
   ;   inv_dens(s_int-1-s_rev:s_int-1, *) = $
   ;     0.5*inv_dens(s_int-1-s_rev:s_int-1, *)
   ;END

   IF reversed THEN BEGIN
      ; see P. Boernert et al., Reversed Spiral MR Imaging, MRM 44:479 (2000) 
      inv_dens = REVERSE(inv_dens)
      kx = REVERSE(kx)
      ky = REVERSE(ky)
      vkx = -REVERSE(vkx)
      vky = -REVERSE(vky)
      
      inv_dens = shift_arr(inv_dens, -xsamp)
      kx = shift_arr(kx, -xsamp)
      ky = shift_arr(ky, -xsamp)
      vkx = shift_arr(vkx, -xsamp)
      vky = shift_arr(vky, -xsamp)
   END
   
   ; gradient signs as in VARIAN scanner
   kx = -kx
   ky = -ky
   vkx = -vkx
   vky = -vky

   mx = mx*dt
   my = my*dt
   max_m = maxi(MAX(ABS(mx)),MAX(ABS(my)))
   
   logstring,'Matrix:                      '+trimstr(n), 3
   logstring,'FOV:                         '+trimstr(fov), 3
   logstring,'# interleaves:               '+trimstr(ni), 3
   logstring,'# revolutions / interleave:  '+trimstr(nrev), 3
   logstring,'# samples / interleave:      '+trimstr(s_int), 3
   logstring,'sampling interval [us]:      '+trimstr(dt*1e3), 3
   logstring,'maximum gradient [mT/m]:     '+trimstr(max_g*1e3), 3
   logstring,'maximum slew rate [mT/m/ms]: '+trimstr(max_s*1e3), 3
   logstring,'maximum moment [mT/m*ms^2]:  '+trimstr(max_m*1e3), 3
   
   IF output THEN BEGIN
      WINDOW, title='trajectory', xsize = 512, ysize = 512, /free
      kk = MAX(kx)
      PLOT, kx, ky, xrange = [-kk,kk], yrange = [-kk,kk], /xstyle, /ystyle, psym=3
      ;WINDOW, title='moments', xsize = 512, ysize = 512, /free
      ;PLOT, mx
      ;OPLOT, my
   END    
END

PRO Get_trajectory, name, output
; get trajectory and 0th-order phase errors (B0-eddy currents) from measured data
   
   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   
   logstring, '----- read measured trajectory -----', 1

   ; save previously generated trajectory
   kxm = kx
   kym = ky
   vkxm = vkx
   vkym = vky

   s_int  = nk/ni               ; # samples per interleave including extra samples
   ss_int = s_int-xsamp         ; # samples per interleave for spiral part
   dt 	  = tacq/(ss_int-1)	; acquisition interval

   ; get data
   data = rd_varian_spiral(name, ni, dum, ns_ec, dum, 0)
   nexp = 2
   ndim = 2
   mec = REFORM(data, s_int, ni, ns_ec, nexp, ndim)
   d = get_varian_par(name, 'pss')*10
   
   ; -----------------------------------------------------------------------
   ; k-space data acquired as described by Duyn et al.JMR 132, 1998
   ; analysis based on Boernert et al. MAGMA 9, 1999 
   ; -----------------------------------------------------------------------

   ; get phase differences
   ph = ATAN(IMAGINARY(mec),FLOAT(mec))
   dph = ph - shift_arr(ph, 1)
   dph(0,*,*,*,*) = 0
   dph = unwrap(dph)
   
   ; correct for gradient independent errors (e.g. B0 inhomogeniety)
   ; by subtracting control experiment (spiral gradients off)
   dphc = FLTARR(s_int, ni, ns_ec, ndim)
   dphc(*,*,*,*) = dph(*,*,*,0,*)-dph(*,*,*,1,*)
   dphc = unwrap(dphc)
   
   ; get phase / k-space by integration
   phc  = FLTARR(s_int, ni, ns_ec, ndim)
   phc(0,*,*,*) = 0
   FOR is=LONG(1), s_int-1 DO phc(is,*,*,*) = phc(is-1,*,*,*)+dphc(is,*,*,*)
   
   ; calculate zero and first order term
   dd = d(0)-d(1)
   phB0x = REFORM(-(d(1)*phc(*,*,0,0)-d(0)*phc(*,*,1,0))/dd)
   phB0y = REFORM(-(d(1)*phc(*,*,0,1)-d(0)*phc(*,*,1,1))/dd)
   kx = REFORM((phc(*,*,0,0)-phc(*,*,1,0))/dd)
   ky = REFORM((phc(*,*,0,1)-phc(*,*,1,1))/dd)
   vkx = REFORM((dphc(*,*,0,0)-dphc(*,*,1,0))/dd/dt)
   vky = REFORM((dphc(*,*,0,1)-dphc(*,*,1,1))/dd/dt)
      
   ; correct data for zero order eddy currents
   FOR iv = 0, nv-1 DO FOR is = 0, ns-1 DO m(*,*,is,iv) = m(*,*,is,iv)*EXP(-COMPLEX(0,1)*(phB0x+phB0y)*1.0)

   ; find start of spiral gradients
   IF 1 THEN BEGIN
      minlsqx = 99999999.9
      minlsqy = 99999999.9
      posx = 0
      posy = 0
      nsamp = 200
      nshift = xsamp*reversed+fix(0.2/dt) ; assume max. propagation delay of 0.2 ms
      FOR j = 0, nshift-1 DO BEGIN
         lsqx = TOTAL((vkxm(0:nsamp-1,0)-vkx(j:j+nsamp-1,0))^2)
         lsqy = TOTAL((vkym(0:nsamp-1,0)-vky(j:j+nsamp-1,0))^2)
         IF lsqx LT minlsqx THEN BEGIN
            minlsqx = lsqx
            posx = j
         END
         IF lsqy LT minlsqy THEN BEGIN
            minlsqy = lsqy
            posy = j
         END
      END
      logstring, 'time shift x, y: '+trimstr(posx)+' '+trimstr(posy), 3
      pos = mini(posx,posy)
      logstring, 'time shift: '+trimstr(pos), 3
      m = shift_arr(m, -pos)
      kx = shift_arr(kx, -pos)
      ky = shift_arr(ky, -pos)
      vkx = shift_arr(vkx, -pos)
      vky = shift_arr(vky, -pos)
      phB0x = shift_arr(phB0x, -pos)
      phB0y = shift_arr(phB0y, -pos)
      t1 = s_int-maxi(xsamp,pos)
      t2 = s_int-1
      IF t1 LE t2 THEN BEGIN
         inv_dens(t1:t2, *) = 0
         kx(t1:t2, *) = 0
         ky(t1:t2, *) = 0
         vkx(t1:t2, *) = 0
         vky(t1:t2, *) = 0
         phB0x(t1:t2, *) = 0
         phB0y(t1:t2, *) = 0
         kxm(t1:t2, *) = 0
         kym(t1:t2, *) = 0
         vkxm(t1:t2, *) = 0
         vkym(t1:t2, *) = 0
      END
   END
   
   ; find trajectory shift due to EC at end of spiral
   ; can be used for correcting the trajectory
   nsamp = 200
   shkx = TOTAL(kx(ss_int-nsamp:ss_int-1, 0)-kxm(ss_int-nsamp:ss_int-1, 0))/nsamp
   shky = TOTAL(ky(ss_int-nsamp:ss_int-1, 0)-kym(ss_int-nsamp:ss_int-1, 0))/nsamp
   logstring, 'amplitude shifts kx, ky [1/mm]: '+trimstr(shkx)+', '+trimstr(shky), 3
   ;kxm = kxm + (tau-MIN(tau))/MAX((tau-MIN(tau)))*shkx
   ;kym = kym + (tau-MIN(tau))/MAX((tau-MIN(tau)))*shky

   ; use calculated trajectory
   ;kx = kxm
   ;ky = kym
     
   IF output THEN BEGIN

      WINDOW, title = 'B0 eddy currents x [rad]', xsize=400, ysize=300, /free
      PLOT, phB0x, yrange = [-3,1]
      WINDOW, title = 'B0 eddy currents y [rad]', xsize=400, ysize=300, /free
      PLOT, phB0y, yrange = [-3,1]
      
      WINDOW, title='trajectory [rad/mm]', xsize = 512, ysize = 512, /free
      kk = maxi(MAX(ABS(kx)),MAX(ABS(ky)))
      PLOT, kx, ky, xrange = [-kk,kk], yrange = [-kk,kk], /xstyle, /ystyle
      OPLOT, kxm, kym, psym = 3
      
      WINDOW, title='kx(t) [rad/mm]', xsize = 500, ysize = 300, /free
      PLOT, kx
      OPLOT, kxm, psym = 3
      WINDOW, title='ky(t) [rad/mm]', xsize = 500, ysize = 300, /free
      PLOT, ky
      OPLOT, kym, psym = 3
      
      WINDOW, title='gx(t)[mT/m]', xsize = 500, ysize = 300, /free
      PLOT, vkx/gamma*1000.0
      OPLOT, vkxm/gamma*1000.0
      WINDOW, title='gy(t)[mT/m]', xsize = 500, ysize = 300, /free
      PLOT, vky/gamma*1000.0
      OPLOT, vkym/gamma*1000.0, psym = 3
      
      WINDOW, title='dkx(t) [rad/mm]', xsize = 500, ysize = 300, /free
      PLOT, kx-kxm, yrange = [-0.01,0.06]
      WINDOW, title='dky(t) [rad/mm]', xsize = 500, ysize = 300, /free
      PLOT, ky-kym, yrange = [-0.01,0.06]
      WINDOW, title='dgx(t) [mT/m]', xsize = 500, ysize = 300, /free
      PLOT, (vkx-vkxm)/gamma*1000.0
      WINDOW, title='dgy(t) [mT/m]', xsize = 500, ysize = 300, /free
      PLOT, (vky-vkym)/gamma*1000.0
   END
END


