PRO spmake
; create spiral trajectory and data from image

   COMMON Spiral_common, n, nk, ni, fov, tacq, alpha, over, fuss, breite, $
     kx, ky, vkx, vky, inv_dens, tau, m, weight, kernel, gamma, two_pi, k_xvon, k_yvon, $
     ns, nv, wmaps, gtau, reversed, xsamp, sindex
   COMMON logging, logging_level

   path     = '~weiger/spiral/data/simu08/'
   imgfile  = 'image'
   w0file   = 'fieldmap'
   mfile    = 'spdata'
   
   ; ----- parameters -----
   
   n 	    = 80                ; image matrix
   ni       = 1                 ; # interleaves
   fov 	    = 240.0             ; FOV
   tacq     = 36.0              ; total acquisition time per interleave
   alpha    = 0.33		; parameter of pb formula
   reversed = 0

   simulation_mode  = 1         ; 0 = gridding, 1 = integration, 2 = PSF
   w0               = 0.0       ; off-resonance for PSF
   logging_level    = 3
   show_trajectory  = 1
   
   ; ----- create trajectory -----
   
   xsamp    = 0
   init_const
   
   kmax = two_pi*n/fov/2.0      ; maximum k-value			
   nrev = n/(2.0*ni)            ; # revolutions per interleave
   dk 	= two_pi/fov		; Shannon dk
   a	= two_pi*nrev		; total angle driven per interleave

   ; determine sampling interval at end of spiral: t=tacq
   t = tacq
   psi0 = 0
   root = SQRT(alpha+(1.0-alpha)*t/tacq)
   psi = -t/tacq/root
   dpsi = t*(1.0-alpha)/tacq^2/2.0/root^3-1.0/tacq/root
   vkx = kmax*dpsi*(COS(a*psi+psi0)-a*psi*SIN(a*psi+psi0))
   vky = kmax*dpsi*(SIN(a*psi+psi0)+a*psi*COS(a*psi+psi0))
   vk = SQRT(vkx^2+vky^2)		
   dt = dk/vk			; sampling interval
   s_int = LONG(tacq/dt)        ; # samples per interleave
   nk = ni*s_int                ; # samples for all interleaves

   make_trajectory, show_trajectory
   
   ; ----- create simulation data -----
   
   ; adapt to axis orientation of VARIAN scanner
   exchange, kx, ky

   IF simulation_mode LT 2 THEN BEGIN
      imgfile = path+imgfile
      image = read_avw(imgfile,0)
   END
   
   CASE simulation_mode OF
      
      0: BEGIN   
         ; create sample values on spiral trajectory 
         ; by regridding from cartesian grid to spiral k-space locations
         
         over     = 2                 
         fuss     = 4                 
         breite   = 4.0
         
         make_weights
         
         m = COMPLEXARR(nk)
         
         ; gridding
         temp = COMPLEXARR(over*n,over*n)
         kspace = COMPLEXARR(over*(n+fuss),over*(n+fuss))
         temp(n*over/2-n/2:n*over/2+n/2-1,n*over/2-n/2:n*over/2+n/2-1) = image / kernel
         kspace(over*fuss/2:over*(n+fuss/2)-1,over*fuss/2:over*(n+fuss/2)-1) = swap(fft(swap(temp),-1))
         kspace(over*(n+fuss/2):over*(n+fuss)-1,*) = kspace(over*(n+fuss/2):over*(n+fuss)-1,*) + kspace(over*fuss/2:over*fuss-1,*)
         kspace(0:over*fuss/2-1,*) = kspace(0:over*fuss/2-1,*) + kspace(over*n:over*(n+fuss/2)-1,*) 
         kspace(*,over*(n+fuss/2):over*(n+fuss)-1) = kspace(*,over*(n+fuss/2):over*(n+fuss)-1) + kspace(*,over*fuss/2:over*fuss-1)
         kspace(*,0:over*fuss/2-1) = kspace(*,0:over*fuss/2-1) + kspace(*,over*n:over*(n+fuss/2)-1)
         FOR i = 0,nk-1 DO BEGIN
            von_x = k_xvon(i) + over*fuss/2
            bis_x = von_x + over*fuss-1
            von_y = k_yvon(i) + over*fuss/2
            bis_y = von_y + over*fuss-1       
            m(i) = TOTAL(weight(i,*,*)*kspace(von_x:bis_x,von_y:bis_y))
         END
      END
      
      1: BEGIN      
         ; create sample values on spiral trajectory 
         ; by integration
         
         IF w0file NE '' THEN BEGIN
            w0file = path+w0file+'.fid/wmaps'
            wmap = read_avw(w0file,0) 
         END ELSE wmap = FLTARR(n,n)
         
         m = COMPLEXARR(nk)
         m(*) = 0
         FOR i=0, nk-1 DO BEGIN
            FOR ix=0, n-1 DO FOR iy=0, n-1 DO BEGIN
               x = (ix-n/2)/FLOAT(n-1)*fov
               y = (iy-n/2)/FLOAT(n-1)*fov
               m(i) = m(i)+image(ix,iy)*EXP(COMPLEX(0,1)*(-kx(i)*x-ky(i)*y+wmap(ix,iy)*tau(i)))
            END
         END
      END

      2: BEGIN
         ; create sample values for off-resonance PSF
         SPAWN, 'mkdir '+path
         m = COMPLEXARR(ni, s_int)
         FOR ii = 0, ni-1 DO m(ii,*) = EXP(COMPLEX(0,1)*w0*tau)
      END
   ENDCASE

   ; ----- write result -----

   m = REFORM(m,s_int*ni)
   
   mfile = path+mfile+'.fid'
   SPAWN, 'mkdir '+mfile
   write_varian, m, mfile+'/', ni
   
   ; write parameters to procpar-file
   cm_mm  = 10.0
   s_ms   = 1000.0
   OPENW, 1, mfile+'/procpar'
   write_varian_par, 1, 'matrix', trimstr(n) 
   write_varian_par, 1, 'nint', trimstr(ni) 
   write_varian_par, 1, 'lro', trimstr(fov/cm_mm) 
   write_varian_par, 1, 'alpha', trimstr(alpha) 
   write_varian_par, 1, 'tacq', trimstr(tacq/s_ms)
   IF reversed THEN reversed_string = '"y"' ELSE reversed_string = '"n"'
   write_varian_par, 1, 'reversed', reversed_string
   write_varian_par, 1, 'xsamp1', trimstr(xsamp) 
   write_varian_par, 1, 'pss', trimstr(0)
   CLOSE, 1

END


PRO Makei
; create sample image

   path    = '~weiger/spiral/data/simu08/'
   imgfile = 'image'	
   n 	   = 64
   image = FLTARR(n,n)
   image = image + mask_arr( 2, n, n, 52, 52, n/2, n/2)
   image = image - 0.5*mask_arr( 2, n, n, 20, 20, n/2+5, n/2+5)
   WINDOW, 0, xsize = 256, ysize = 256
   tvscl, REBIN(image, 256, 256)
   
   SPAWN, 'mkdir '+path
   imgfile = path + imgfile
   write_avw, image, imgfile
END

PRO Makew
; create frequency map
   
   path     = '~weiger/spiral/data/simu08/'
   imgfile  = 'fieldmap'

   n        = 64
   dw0      = 0.5
   a        = 0.5

   wmap = FLTARR(n,n)
   FOR ix = 0,n-1 DO FOR iy=0, n-1 DO BEGIN
      x = ix/FLOAT(n)-0.5
      y = iy/FLOAT(n)-0.5
      wmap(ix,iy) = dw0*EXP(-(x^2+y^2)/a^2)
   END
      
   PRINT, MIN(wmap),MAX(wmap)
   WINDOW, 0, xsize = 256, ysize = 256
   tvscl, REBIN(wmap, 256, 256)

   imgfile = path+imgfile+'.fid/'
   SPAWN, 'mkdir '+imgfile
   write_avw, wmap, imgfile+'wmaps'
END
