/* Modified by Patricia, January 2000 */
/* Spiral readout included by Markus Weiger, 2001, changes marked: mw sp */

/*
  quipss2.c
  
  Stuart Clare
  FMRIB Centre, University of Oxford
  April 1999

  Adapted for perfusion MRI : Peter Jezzard
  Aug 1999
  see Wong et.al.  MRM 39,702 (1998) for details

  Can be run as QUIPSS, QUIPSS2, FAIR or EPISTAR

*/

/*

QUIPSS2 tag:

   uses psat   uses p5       uses p6
      90         inv        90     90
     slice       tag        tag    
    |                                            |
 rf |-oOo------o0X0X0o------o0o------------------| -
          _             _       _                |  \
    |  _ |*|   _______ |*|   _ |*|               |  |
    |_/ \|*|__/       \|*|__/ \|*|_______________|  |
GSS |   crush         crush   crush              |  |
    |             |<---TI1---->|                 |  /
    |             |<-------------TI2------------>|  > EPI MODULE
    |                                            |  \
GPE |--------------------------------------------|  |
    |                                            |  |
    |                                            |  |
GRO |--------------------------------------------|  /
    |                                            | -


 num   ns                                                              ns  num
 ints                                                                      ints
  {    { [Inversion]  [Fatsat]  [Excitation]  [Spinecho]  [EPI][crush] }    } 

              |__________ti__________|__________te__________|

       |___________________________tr/ns_______________________________|    

*/

/*
	plist	patlist	pwrlist	sslist	thklist	fliplist
	p1	p1pat	tpwr1	gss	thk	90	excitation pulse
	p2	p2pat	tpwr2	gss2	thk	180	spin echo 180 pulse
	p3	p3pat	tpwr3	gss3	thk	180	inversion (not used)
	p4	p4pat	tpwr4	gss4	-	80	fat sat
	p5	p5pat	tpwr5	gss5	tag_thk	180	tagging slab inversion
	p6	p6pat	tpwr6	gss6	tag_thk	90	tagging slab saturation
	psat	satpat	satpwr	gsssat	sat_thk	90	spatial saturation
*/

/*
  Real-time variables
  v1:  phase cycling
  v2:  calculation of phase cycling
  v3:  calculation of phase cycling 
  v4:  calculation of phase cycling
  v5:
  v6:
  v7:  max # of interleaves
  v8:  # interleaves
  v9:  interleave number
  v10: interleave selection
  v11: interleave selection
  v12: rfmod power
  v13: # slices
  v14: slice number
*/ 

#include "standard.h"
#include "group.h"
#include "sequence_utilities.h"
#define SAMPPTS 8192

/* Global variables*/
#include "globals.h"
#include "playout.h"
#include "tweekers.h"
#include "tagging_utilities.h"
#include "spiral_utilities.h"    /* mw sp */
#include "sp_perf_safety_calc.h" /* mw sp */
#include "calculate.h"

pulsesequence()
{
  double t_inv_b,t_inv_a;
  double t_presat_b,t_presat_a;
  double t_taginv_b,t_taginv_a;
  double t_tagsat_b,t_tagsat_a;
  double t_macro_flow_suppress;
  double ti1=0.0,ti2=0.0,dTI1,dTI2;
  double t_cr,t_fat;
  double t_ex_b,t_ex_a;
  double t_se_b,t_se_a;
  double t_epi_b,t_epi_a;
  double min_te,min_tr,min_ti;
  double t_nav;
  double n_sat;
  double echo_shift=1.0;
  double ss_tr_min, ss_tr, ss_tr_delay;
  double foci_factor=10.0;
  
  double tr_delay,ti_delay,te_delay,se_delay_b,se_delay_a;

  double prescan_flag,pown, gss_scale;
  int image;
  char err_string[MAXSTR];

  initparms_sis();
  get_tweekers();

  gss_scale = getval("gss_scale");
  /* 180 degree pulse gradient strength */
  gss2 = getval("gss2");
  /* Inversion recovery gradient strength */
  gss3 = getval("gss3");

  /* Trigger and counter bit pattern */
  image = (int)(getval("image"));
  /* bit 1 = wait for ticks triggers before beginning scan */
  /* bit 2 = puts out high on sp3 for duration of volume */
  /* bit 3 = epi blips off (set bit on), otherwise blips on */
  /* bit 4 = make inversion non slice selective (on) or slice selective (off)*/
  /* bit 5 = turns off the acquire */

  /* Get flags 'y/n' */
  getstr("spinecho",spinecho);  
  getstr("inversion",inversion);  
  getstr("crushers",crushers);  
  getstr("fat_sat",fat_sat);
  getstr("half_kspace",half_kspace);
  getstr("nonlinsamp",nonlinsamp);
  getstr("rfmod",rfmod);
  getstr("nav",nav);
  getstr("foci",foci);

  /* mw sp start */
  /* Get spiral parameters */
  getstr("reversed",reversed);
  getstr("flow_comp",flow_comp);
  presat=getval("presat");
  type = (int)(getval("type"));
  ec_comp = (getval("ec_comp"));
  /* 0 = standard spiral */
  /* 1 = rd test */
  /* 2 = rd control */
  /* 3 = ph test */
  /* 4 = ph control */
  matrix =  getval("matrix");
  nint = getval("nint");  
  tacq = getval("tacq");  
  alpha = getval("alpha");  
  xsamp = getval("xsamp");  
  upslew = getval("upslew")*US_S;  /* convert to s/G/cm */
  dnslew = getval("dnslew")*US_S;  
  dkx = getval("dkx");  
  dky = getval("dky");  
  /* mw sp end */
  
  ss_tr = getval("ss_tr");

  gss5 = getval("gss5");
  if(foci[0] == 'y') gss5*=foci_factor;
  gss6 = getval("gss6");
  gsssat = getval("gsssat");
  p6=getval("p6");
  tpwr6=getval("tpwr6");
  getstr("p6pat",p6pat);
  echo_shift=getval("echo_shift");

  presat=getval("presat");
  tag_invert=getval("tag_invert");
  tag_sat=getval("tag_sat");
  gflow_flag=getval("gflow_flag");
  ti1=getval("ti1");
  ti2=getval("ti2");
  getstr("tag_plane",tag_plane);
  n_sat=getval("n_sat");
  tag_gap=getval("tag_gap");

  /* Number of interleaves */
  num_ints = getval("num_ints");  
  if (num_ints == 0) num_ints = 1;

  if (presat) {
    sat_thk=getval("sat_thk");
    sat_psi=getval("sat_psi");
    sat_phi=getval("sat_phi");
    sat_theta=getval("sat_theta");
    presat_off=getval("presat_off");
    post_presat_del=getval("post_presat_del");
  }
  if ((tag_invert) || (tag_sat)) {
    tag_thk=getval("tag_thk");
    /* 
      tag_direction=0: frequency=resto, non slice selective
      tag_direction=1: frequency=tag_off, slice selective
      tag_direction=-1: frequency=control_tag_off, slice selective
      tag_direction=2: frequency=tag_off, non slice selective
      tag_direction=-2: frequency=control_tag_off, non slice selective
    */
    tag_direction=getval("tag_direction");
    tag_sign=getval("tag_sign");
    tag_off=getval("tag_off");
    control_tag_off=getval("control_tag_off");
    mean_pss=getval("mean_pss");
  }
  t_macro_flow_suppress=0.0;
  if (gflow_flag) {
    gflow=getval("gflow");
    tflow=getval("tflow");
    calculate_macro_flow_suppress_gradients(&t_macro_flow_suppress);
  }

  /* Half k-space stuff */
  if (half_kspace[0] == 'y') {
    
    if (num_ints > 1) {
      sprintf(err_string,"%s: Half K space acquisition requires 1 interleaf\n"
	      ,seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    
    pown = 2;
    do{
      if(pown == nv){
        sprintf(err_string,"%s: nv must be a few lines greater than n^2\n"
	        ,seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      eff_echo = nv - pown;
      pown *= 2;
    }while(pown < nv);
    etl = nv;
  }
  else {
    etl = (double)((int)(nv/num_ints));  /* echo train length/interleaf */
    eff_echo = (etl/2.0)/echo_shift;  /* echo position of TE */
    /* echo_shift shifts the effective echo position */
    /* echo_shift = 1.0 (default) : eff_echo centred in echo train */
    /* echo_shift = 2.0 : eff_echo half way beteen start and centre of train */
    /* etc. */
  }

  /* Crusher gradient strength */
  /* gcrush = getval("gcrush"); */
  gcrush=gmax/2.0;

  /* Gradient calculations */
  if(inversion[0]=='y')calculate_inversion_gradients(&t_inv_b,&t_inv_a);
  else { t_inv_b=0.0; t_inv_a=0.0; }

  if(crushers[0]=='y')calculate_crusher_gradients(&t_cr);
  else t_cr=0.0;

  if(fat_sat[0]=='y')calculate_fatsat_gradients(&t_fat);
  else t_fat=0.0;

  if (presat) calculate_spatial_saturation_gradients(&t_presat_b,&t_presat_a);
  t_presat_b = n_sat*t_presat_b;
  t_presat_a = n_sat*t_presat_a;
  
  if (foci[0] == 'y') {
    if (tag_invert) calculate_foci_inversion_gradients(&t_taginv_b,&t_taginv_a);
  }
  else {
    if (tag_invert) calculate_tag_inversion_gradients(&t_taginv_b,&t_taginv_a);
  }
  if (tag_sat) calculate_tag_saturation_gradients(&t_tagsat_b,&t_tagsat_a);
  t_tagsat_b = t_tagsat_b * n_sat;
  t_tagsat_a = t_tagsat_a * n_sat;

  calculate_excitation_gradients(&t_ex_b,&t_ex_a);
  if (ec_comp == 1) t_ex_b += t_ex_b+t_ex_a; /* mw sp */
  gss = gss*gss_scale;
  gssr = gssr*gss_scale;

  if(spinecho[0]=='y')calculate_spinecho_gradients(&t_se_b,&t_se_a);
  else { t_se_b=0.0; t_se_a=0.0; }

  calculate_spiral_gradients(&t_epi_b,&t_epi_a); /* mw sp */
  /* calculate_epi_gradients(&t_epi_b,&t_epi_a); */

  if(rfmod[0]=='y'){
    calculate_rf_modulation();
  }

  if(nav[0]=='y'){
    calculate_navigator_gradients(&t_nav);
  }
  else t_nav=0;

  /* Timing calculations */

  /* Nonlinear sampling */
  if(nonlinsamp[0]=='y')calculate_nonlinear_sampling();

  /* Spinecho delay */
  se_delay_b = 0.0; se_delay_a = 0.0;
  if(spinecho[0]=='y'){
    if( (t_ex_a+t_se_b) > (t_se_a+t_epi_b+t_nav) )
      se_delay_a=(t_ex_a+t_se_b)-(t_se_a+t_epi_b+t_nav);
    if( (t_ex_a+t_se_b) < (t_se_a+t_epi_b+t_nav) )
      se_delay_b=(t_se_a+t_epi_b+t_nav)-(t_ex_a+t_se_b);
  }

  /* te delay */
  min_te = t_ex_a + t_macro_flow_suppress + se_delay_b + t_se_b + t_se_a + 
         se_delay_a + t_nav + t_epi_b;
  if(min_te>te){
    sprintf(err_string,"%s: te too short. Min te = %f\n",seqfil,min_te);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  te_delay = te - min_te;

  /* if (ix == 1) printf("Delta_t between flow gradients = %.3f ms\n",
     (t_se_b + se_delay_b + te_delay + t_se_a + se_delay_a)*1000.0); */

  /* ti delay */
  if(inversion[0]=='y'){
    min_ti = t_inv_a + t_fat + t_ex_b;
    if(min_ti>ti){
      sprintf(err_string,"%s: ti too short. Min ti = %f\n",seqfil,min_ti);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    ti_delay = ti - min_ti;
  }
  else ti_delay=0;

  /* tr delay */
  min_tr = (t_inv_b + t_inv_a + ti_delay +
	    (t_fat + t_ex_b + te + t_epi_a + t_cr)*ns);

  ss_tr_min= (t_fat + t_ex_b + te + t_epi_a + t_cr);

  ss_tr_delay=0.0;
  if (ss_tr > 0.0) {
    ss_tr_delay=ss_tr-ss_tr_min;
    if (ss_tr_delay < 0.0) {
      sprintf(err_string,"%s: ss_tr too short. Min ss_tr = %f\n",seqfil,
            ss_tr_min);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
  }

  min_tr = min_tr + ss_tr_delay*ns;

  dTI1=0.0; dTI2=0.0;

  if (presat) min_tr = min_tr + t_presat_b + t_presat_a;

  if (tag_invert) {
    if (tag_sat) {
      dTI2 = ti1 - (t_taginv_a + t_tagsat_b);
    } else {
      dTI2 = ti2 - (t_taginv_a + t_fat + t_ex_b);
    }
    if (dTI2 < 0.0) {
      sprintf(err_string,"%s: ti2 too short. Min ti2 = %f\n",seqfil,
              ti2-dTI2);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    min_tr = min_tr + t_taginv_b + t_taginv_a + dTI2;
  }

  if (tag_sat) {
    dTI1 = ti2 - ti1 - (t_tagsat_a + t_fat + t_ex_b);
    if (dTI1 < 0.0) {
      sprintf(err_string,"%s: ti1 too short. Min ti1 = %f\n",seqfil,
              ti1-dTI1);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    min_tr = min_tr + t_tagsat_b + t_tagsat_a + dTI1;
  }

  min_tr = min_tr;

  if(min_tr>tr){
    sprintf(err_string,"%s: tr too short. Min tr = %f\n",seqfil,min_tr);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  tr_delay = (tr - min_tr);

  if (ix == 1) printf("min. te: %7.5f, min. tr: %7.5f\n", min_te, min_tr); /* mw sp */

  /* Prescan flag (1=yes, 0=no) */
  prescan_flag = getval("prescan_flag"); 
  if (prescan_flag == 1) num_segs=1;
  else num_segs=nint; /* mw sp */
  /* else num_segs=(int)(nv/etl); */
  
  /* Turn blips off */
  if ((image & 4) == 4) {
    sgbpe = 0.0;
    gbd=0.0;
    sgbd=0.0;
  }

  /* Saftey check */
  /* if (quipss2_safety_calc() == -1) { mw sp */
  if (sp_perf_safety_calc() == -1) {
    sprintf(err_string,"%s: Failed duty cycle safety check\n",
	    seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
 
  /**********************************************************************/
  /* PULSE SEQUENCE */
  /**********************************************************************/

  /* Phase cycling */
  mod2(ct,v3);                    /* v3 = 0101 0101 */
  dbl(v3,v3);                     /* v3 = 0202 0202 */
  hlv(ct,v4);                     /* v4 = 0011 2233 */
  mod2(v4,v4);                    /* v4 = 0011 0011 */
  add(v3,v4,v1);                  /* v1 = 0213 0213 */
  assign(v1,oph);                 /* oph = 0213 0213 */
  add(one,v1,v2);

  /* phase encode increment size 
  assign(one,v7);  */
  
  /* Receiver setup */
  status(A);
  rcvroff();
  rcvron();
  sp2off();

  /* wait for start scan button */
  if ((image & 1) == 1) xgate(ticks);
  if ((image & 2) == 2) sp3on(); else sp3off();
  delay(del_seq);
  
  initval(MAXINT,v7); /* mw sp */

  /* loop interleaves, v8 times using counter v9, mw sp */
  initval(num_segs,v8);
  loop(v8,v9);
 
  /* Inversion pulse */
  if(inversion[0] == 'y') {
    if((image & 8) == 8) {
      nonselective_inversion_pulse();
    }
    else {
      selective_inversion_pulse(v14);
    }
    delay(ti_delay);
  }

  if (presat) {
    if (n_sat == 1) spatial_saturation_pulse();
    if (n_sat == 2) {
      spatial_saturation_pulse1();
      spatial_saturation_pulse2();
    }
    if (n_sat == 3) {
      spatial_saturation_pulse1();
      spatial_saturation_pulse2();
      spatial_saturation_pulse3();
    }
    if (n_sat > 3) spatial_saturation_pulse();
  }

  if (tag_invert) {
    if (foci[0] == 'y') foci_inversion_pulse();
    else tag_inversion_pulse();
    delay(dTI2);
  }

  if (tag_sat) {
    if (n_sat == 1) tag_saturation_pulse();
    if (n_sat == 2) {
      tag_saturation_pulse1();
      tag_saturation_pulse2();
    }
    if (n_sat == 3) {
      tag_saturation_pulse1();
      tag_saturation_pulse2();
      tag_saturation_pulse3();
    }
    if (n_sat > 3) tag_saturation_pulse();
    delay(dTI1);
  }

  /* sets v13 to ns and incs v14 up to it */
  msloop(seqcon[1],ns,v13,v14);

  if (fat_sat[0] == 'y') {
    fatsat_pulse();
  }

  /* quipss_excitation_pulse(v14);
  excitation_pulse(v14); mw sp */
  excitation_pulse1(v14);

  if (gflow_flag) {
    if (spinecho[0]=='n') {
      macro_flow_suppress_gradients();
    } else {
      macro_flow_suppress_gradients_se();
    }
  }

  obsoffset(resto); /* mw sp: does not work for spin-echo !? */

  delay(te_delay/2.0);

  if (spinecho[0] == 'y'){
    delay(se_delay_b);
    spinecho_pulse(v14);
    delay(se_delay_a);
  }
  
  delay(te_delay/2.0);

  if (gflow_flag) {
    if (spinecho[0]!='n') {
      macro_flow_suppress_gradients_se();
    }
  }

  if(nav[0]=='y'){
  	navigator_acquire();
  }

  /* if((image & 16) == 16){
	epi_gradients_no_acquire();
  }
  else epi_gradients_and_acquire(); mw sp */
  if (prescan_flag == 1) spiral_prescan();
  else spiral_gradients_and_acquire(); 

  crusher_gradients();

  /* double check that all gradients are zero */
  zero_all_gradients();

  delay(ss_tr_delay);

  /* increment the msloop counter v14 */
  endmsloop(seqcon[1],v14);
 
  delay(tr_delay);

  /* increment the interleaves loop counter v9 */
  endloop(v9);
}
