/*
  spiral_utilities.h
  
  Markus Weiger
  FMRIB Centre, University of Oxford
  2001
*/

/* -------------------------------------------------------- */
/* global spiral parameters and constants */
/* -------------------------------------------------------- */

/* units: G, cm, s */

/* nint:       number of interleaves */
/* tacq:       acquisition time */
/* alpha:      transition between constant linear velocity (0) and angluar velocity (1) */
/* dkx, dky:   if nonzero, offsets gradients to correct for drift (get numbers from trajectory measurement */
/* ec_comp:    if 1, adds gradient lobes before slice selection to compensate for eddy currents */

#define MAXSAMP  20000
#define MAXINT   10                /* has to match playout routine ! */
#define TWOPI    (2.0*M_PI)
#define US_S     1.0e-6
#define GAMMA    (TWOPI*4257.493)  /* [1/s/G] */
#define MAX(a,b) ((a)>(b) ? (a):(b))
#define MIN(a,b) ((a)<(b) ? (a):(b))

int    matrix, nint, xsamp, type, ec_comp;
double tacq, alpha, upslew, dnslew, dkx, dky;
char   reversed[MAXSTR], flow_comp[MAXSTR];
double pw_sp, np_sp, gsprd[MAXINT], gspph[MAXINT];
char   sprd[MAXINT][MAXSTR], spph[MAXINT][MAXSTR];
double grd_sqr, gph_sqr;


/* -------------------------------------------------------- */
/* write to files */
/* -------------------------------------------------------- */


shapelib_path(char *name)
/* create path for shapelib */
{
  char *home_dir, pname[MAXSTR], dum_str[MAXSTR];

  home_dir = (char *) getenv("HOME");
  strcpy(pname, home_dir);
  if (pname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(dum_str,"/vnmrsys/shapelib/");
  strcat(pname, dum_str);
  sprintf(name, pname);
}

write_pars_to_file(long np, long xsamp1, int write_flag)
/* write some parameters to file that will be written into the procpar file */
{
  char   fname[MAXSTR], dum_str[MAXSTR];
  FILE   *fp=NULL;

  if (write_flag == 1) { 

    shapelib_path(fname);
    strcat(fname,"temp_pars" );  

    if ((fp=fopen(fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", fname);
      exit(0);
    }
    /* printf("%ld \n %ld \n",np,xsamp1); */
    sprintf(dum_str, "%ld \n %ld \n",np,xsamp1);
    fprintf(fp, dum_str);
    fclose(fp);
  }
}

write_comp_gradient(FILE *fp, long n1, long n2, int ntrig, double maxval)
/* write the compensation gradient waveforms to file */
{
  int val, itrig=0, trig;
  long i;

  /* ramp up */
  /* start with first non-zero gradient */
  for (i=0; i<n1; i++) {
    val = floor((float)(i+1)/(float)n1*maxval*MAX_DAC_VAL);
    if (itrig++ < ntrig) trig = 0; else trig = 1;
    fprintf(fp, "%d   1   %d\n", val, trig);
  }
    
  /* flat top portion */
  for (i=0; i<n2; i++) {
    val = (int)(maxval*MAX_DAC_VAL);
    if (itrig++ < ntrig) trig = 0; else trig = 1;
    fprintf(fp, "%d   1   %d\n", val, trig);
  }
 
  /* ramp down */ 
  for (i=0; i<n1; i++) {
    val = floor((1.0-(float)(i+1)/(float)n1)*maxval*MAX_DAC_VAL);
    if (itrig++ < ntrig) trig = 0; else trig = 1;
    fprintf(fp, "%d   1   %d\n", val, trig);
  }
}

write_spiral_to_file(int iint, long nsamples, 
		     long n1cup, long n2cup, long n1cdn, long n2cdn, long n1fc, long n2fc,
		     double gcupx, double gcupy, double gcdnx, double gcdny, double gfcx, double gfcy,
		     double *gx, double *gy, int write_flag)
/* write the spiral gradient waveforms to file */
{
  char   fname_rd[MAXSTR],fname_ph[MAXSTR], dum_str[MAXSTR];
  double max_gx, max_gy, sval;
  long   j;
  int    maxtrig = 4;
  FILE   *fp_rd=NULL, *fp_ph=NULL;

  /* scaling */

  max_gx = MAX(fabs(gcupx), fabs(gcdnx));
  max_gy = MAX(fabs(gcupy), fabs(gcdny));
  max_gx = MAX(max_gx, gfcx);
  max_gy = MAX(max_gy, gfcy);
  for (j=0; j<nsamples; j++) {
    max_gx = MAX(max_gx, fabs(*(gx+j)));
    max_gy = MAX(max_gy, fabs(*(gy+j)));
  }
  gsprd[iint] = max_gx;
  gspph[iint] = max_gy;

  /* write to file */

  if (write_flag == 1) { 

    shapelib_path(dum_str);
    sprintf(fname_ph, dum_str);  
    sprintf(fname_rd, dum_str);  
    strcat(fname_rd, sprd[iint]);  
    strcat(fname_ph, spph[iint]);  
    strcat(fname_rd,".GRD" );  
    strcat(fname_ph,".GRD");  

    if ((fp_rd=fopen(fname_rd, "w")) == NULL) {
      printf("Error opening output data file %s\n", fname_rd);
      exit(0);
    }
    if ((fp_ph=fopen(fname_ph, "w")) == NULL) {
      printf("Error opening output data file %s\n", fname_ph);
      exit(0);
    }

    if (n1fc > 0) {
      write_comp_gradient(fp_rd, n1fc, n2fc, maxtrig, gfcx/max_gx);
      write_comp_gradient(fp_ph, n1fc, n2fc, maxtrig, gfcy/max_gy);
      write_comp_gradient(fp_rd, n1fc, n2fc, 0, -gfcx/max_gx);
      write_comp_gradient(fp_ph, n1fc, n2fc, 0, -gfcy/max_gy);
      maxtrig = 0;
    }

    write_comp_gradient(fp_rd, n1cup, n2cup, maxtrig, gcupx/max_gx);
    write_comp_gradient(fp_ph, n1cup, n2cup, maxtrig, gcupy/max_gy);

    for (j=0; j<nsamples; j++) {
      sval = *(gx+j)/max_gx*MAX_DAC_VAL;
      fprintf(fp_rd, "%d   1    1\n",(int)sval);
      sval = *(gy+j)/max_gy*MAX_DAC_VAL;
      fprintf(fp_ph, "%d   1    1\n",(int)sval);
      }

    write_comp_gradient(fp_rd, n1cdn, n2cdn, 0, gcdnx/max_gx);
    write_comp_gradient(fp_ph, n1cdn, n2cdn, 0, gcdny/max_gy);

    fprintf(fp_rd, "\n");
    fprintf(fp_ph, "\n");
    fclose(fp_rd);
    fclose(fp_ph);
  }
}

init_spiral()
/* initialize some variables and create dummy waveform files for all potential interleaves */
{
  int    i, j, n=4;
  double dum_arr[n];
  char   dum_str[MAXSTR];
    
  for (j=0; j<n; j++) dum_arr[j]=0;
    
  for (i=0; i<MAXINT; i++) {
      
    sprintf(sprd[i], "sprd");  
    sprintf(spph[i], "spph");
    sprintf(dum_str, "%d",i);
    strcat(sprd[i], dum_str);  
    strcat(spph[i], dum_str); 
      
    write_spiral_to_file(i, n, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, &dum_arr[0], &dum_arr[0], ix);
  }
}


/* -------------------------------------------------------- */
/* generate waveforms */
/* -------------------------------------------------------- */

double calculate_flowcomp_gradient(double m, double max_r, double max_g, double dt,
				   double *act_g, long *n_ramp, long *n_top)
{
  double g, t, t1, t2;
  long   n1, n2;

  /* try to use a triangular gradient shape */
  g = -pow(m/(2*max_r*max_r),1.0/3.0);
  if (fabs(g) <= max_g) {
    t1 = max_r*fabs(g);
    n1 = ceil(t1/dt);      /* prefer too high gradient */
    if (n1<2) n1 = 2;
    t1 = n1*dt;
    t = 2*t1;
    g = -2*m/(t*t);
    n2 = 0;
  }
  /* otherwise use trapeziodal gradient shape */
  else {
    g = -m/fabs(m)*max_g;
    t1 = max_r*fabs(g);
    n1 = ceil(t1/dt);
    t1 = n1*dt;
    t2 = sqrt(0.25*t1*t1-m/g)-1.5*t1;
    n2 = ceil(t2/dt);
    t2 = n2*dt;
    t = 2*t1+t2;
    g = m/(-t*t+t*t1);
  }

  *act_g = g;
  *n_ramp = n1;
  *n_top = n2;
  return 2*t; /* time for binomial shape */
}

double calculate_comp_gradient(double k, double max_r, double max_g, double dt,
			       double *act_g, long *n_ramp, long *n_top)
/* calculate the gradients that make spiral to start in the centre of k-space 
   i.e. compensate for ramping up to the initial spiral gradient value */
{
  double g, t, a, t1, t2;
  long   n1, n2;

  a = k/GAMMA;

  /* try to use a triangular gradient shape */
  g = sqrt(a/max_r);
  if (g <= max_g) {
    t1 = max_r*g;
    n1 = ceil(t1/dt);      /* prefer too high gradient */
    if (n1<2) n1 = 2;
    t1 = n1*dt;
    g = a/t1;
    t = 2*t1;
    n2 = 0;
  }
  /* otherwise use trapeziodal gradient shape */
  else {
    g = max_g;
    t1 = max_r*g;
    n1 = ceil(t1/dt);
    t1 = n1*dt;
    t2 = a/g-t1;
    n2 = ceil(t2/dt);
    t2 = n2*dt;
    g = a/(t1+t2);
    t = 2*t1+t2;
  }

  *act_g = g;
  *n_ramp = n1;
  *n_top = n2;
  return t;
}

spiral_gradient(double t, double kmax, double nrev, double psi0, double *gx, double *gy)
/* spiral waveform from: P. Boernert, MAGMA 9:29-41 (1999) */
{
  double a, root, psi, dpsi;

  if (t < 0) t = tacq; 
  else if (reversed[0] == 'y') {
    /* reverse gradient in time and amplitude */
    /* see P. Boernert et al., Reversed Spiral MR Imaging, MRM 44:479 (2000) */
    t = tacq-t;
    kmax = -kmax;
  }
  a = TWOPI*nrev;
  root = sqrt(alpha+(1.0-alpha)*t/tacq);
  psi = -t/tacq/root;
  dpsi = t*(1.0-alpha)/(tacq*tacq)/2.0/(root*root*root)-1.0/tacq/root;
  *gx = kmax/GAMMA*dpsi*(cos(a*psi+psi0)-a*psi*sin(a*psi+psi0));
  *gy = kmax/GAMMA*dpsi*(sin(a*psi+psi0)+a*psi*cos(a*psi+psi0));
}

calculate_spiral_gradients(double *t_sp_b, double *t_sp_a)
/* main procedure called from spiral.c */
{
  double kmax, dk, nrev, psi0; 
  double t, dt, gdt, dtres, tup, tdn, tcup, tcdn, tm, tfc;
  double gx[MAXINT][MAXSAMP], gy[MAXINT][MAXSAMP], gx0, gy0, stepx, stepy;
  double gcup, gcdn, max_g, max_s, max_g_start, max_g_end, gfc;
  double kx_sp, ky_sp, kx_b[MAXINT], ky_b[MAXINT], kx_a[MAXINT], ky_a[MAXINT], max_kb, max_ka;
  double mx[MAXINT], my[MAXINT], max_m;
  long   nsp, nup, ndn, i, j, n1cup, n2cup, n1cdn, n2cdn, n1fc, n2fc;

  init_spiral();

  /* ------------- calculate sampling timing ------------------------------ */

  kmax 	 = TWOPI*(float)matrix/lro/2.0;	     /* maximum k-value */			
  nrev 	 = (float)matrix/(float)(2*nint);    /* # revolutions per interleave */
  dk 	 = TWOPI/lro;		             /* Shannon dk */
	
  /* determine sampling interval at end of spiral: t=tacq */
  spiral_gradient(-1, kmax, nrev, 0, &gx0, &gy0); 
  dt = dk/(sqrt(gx0*gx0+gy0*gy0)*GAMMA);     /* sampling interval */
  dtres = 2.0*WAVEFORM_RESOL*US_S;
  dt = floor(dt/dtres)*dtres;                /* make dt a multiple of sampling resolution */
  tacq = floor(tacq/dt)*dt;                  /* exact acquisition time */
 
  /* make gradient sampling interval half of acquisition sampling interval */
  if (dt >= 2*GRAD_UPDATE_TIME*US_S) gdt = dt/2.0;
  else gdt = dt;
  nsp = tacq/gdt;

  /* ------------- calculate timing for ramping up and down ---------------- */

  /* get maximum gradient amplitude at start and end of spiral */
  max_g_start = 0;
  max_g_end = 0;
  for (i = 0; i<nint; i++) {
    psi0 = TWOPI*(float)i/(float)nint;
    spiral_gradient(0, kmax, nrev, psi0, &gx0, &gy0);
    max_g_start = MAX(max_g_start, fabs(gx0));
    max_g_start = MAX(max_g_start, fabs(gy0));
    spiral_gradient(tacq, kmax, nrev, psi0, &gx0, &gy0);
    max_g_end = MAX(max_g_end, fabs(gx0));
    max_g_end = MAX(max_g_end, fabs(gy0));
  }
  
  /* calculate duration for ramping up and down */
  tup = upslew*max_g_start;
  nup = ceil(tup/gdt);
  tup = nup*gdt;
  tdn = dnslew*max_g_end;
  ndn = ceil(tdn/gdt);
  tdn = ndn*gdt;

  /* ------------- calculate waveforms  ------------------------- */
                                          
  max_g = 0;
  max_s = 0;
  max_kb = 0;
  max_ka = 0;

  /* loop interleaves */
  for (i=0; i<nint; i++) {
    kx_sp = 0;
    ky_sp = 0;
    mx[i] = 0;
    my[i] = 0;
    /* calculate first gradient moment at end of spiral with respect to start of ramp. 
       For the echo time all smaller moments (the only one is k) are zero, therefore 
       the point of expansion does not matter. */
    tm = -gdt/2.0; 
    psi0 = TWOPI*(float)i/(float)nint;

    /* linear ramp to initial spiral gradient */
    /* has to be compensated for in advance */
    spiral_gradient(0, kmax, nrev, psi0, &gx0, &gy0);
    stepx = gx0/(float)nup;
    stepy = gy0/(float)nup;
    for (j=0; j<nup; j++) { 
      gx[i][j] = stepx*(j+1);
      gy[i][j] = stepy*(j+1);

      tm += gdt;
      if (j == 0) {
	mx[i] += 0.5*gx[i][j]*tm*gdt;
	my[i] += 0.5*gy[i][j]*tm*gdt;
      } else {
	mx[i] += 0.5*(gx[i][j]+gx[i][j-1])*tm*gdt;
	my[i] += 0.5*(gy[i][j]+gy[i][j-1])*tm*gdt;
      } 
    }
    kx_b[i] = GAMMA*0.5*gx0*tup;
    ky_b[i] = GAMMA*0.5*gy0*tup;

    /* actual spiral */
    t = 0;
    for (j=nup; j<nup+nsp; j++) {
      t += gdt;
      spiral_gradient(t, kmax, nrev, psi0, &gx[i][j], &gy[i][j]); 
      max_g = MAX(max_g, fabs(gx[i][j]));
      max_g = MAX(max_g, fabs(gy[i][j]));
      max_s = MAX(max_s, fabs(gx[i][j]-gx[i][j-1])/gdt);
      max_s = MAX(max_s, fabs(gy[i][j]-gy[i][j-1])/gdt);	
      kx_sp += GAMMA*0.5*(gx[i][j]+gx[i][j-1])*gdt;
      ky_sp += GAMMA*0.5*(gy[i][j]+gy[i][j-1])*gdt;
      tm += gdt;
      mx[i] += 0.5*(gx[i][j]+gx[i][j-1])*tm*gdt;
      my[i] += 0.5*(gy[i][j]+gy[i][j-1])*tm*gdt;
    }

    /* linear ramp down to G = 0 */
    /* has to be compensated for afterwards */
    gx0 = gx[i][nup+nsp-1];
    gy0 = gy[i][nup+nsp-1];
    stepx = -gx0/(float)ndn;
    stepy = -gy0/(float)ndn;
    for (j=nup+nsp; j<nup+nsp+ndn; j++) { 
      gx[i][j] = gx[i][j-1]+stepx;
      gy[i][j] = gy[i][j-1]+stepy;
    }
    kx_a[i] = GAMMA*0.5*gx0*tdn;
    ky_a[i] = GAMMA*0.5*gy0*tdn;

    if (reversed[0] == 'n') {
      kx_a[i] += kx_sp;
      ky_a[i] += ky_sp;
    } else {
      kx_b[i] += kx_sp;
      ky_b[i] += ky_sp;
    }
    max_kb = MAX(max_kb, fabs(kx_b[i]));
    max_kb = MAX(max_kb, fabs(ky_b[i]));
    max_ka = MAX(max_ka, fabs(kx_a[i]));
    max_ka = MAX(max_ka, fabs(ky_a[i]));

    /* prospective compensation of first order ECs using a constant
       gradient offset corresponding to a linear ramp in k-space */
    {
      double dgx = dkx/GAMMA/tacq;
      double dgy = dky/GAMMA/tacq;
      for (j=nup; j<nup+nsp; j++) {
	gx[i][j] += dgx;
	gy[i][j] += dgy;
      }
    }

  } /* end of interleaves loop */
        
  /* ------------- calculate compensation gradients and write all waveforms -------------- */
  
  tcup = calculate_comp_gradient(max_kb, upslew, 0.5*gmax, gdt, &gcup, &n1cup, &n2cup);
  tcdn = calculate_comp_gradient(max_ka, dnslew, 0.5*gmax, gdt, &gcdn, &n1cdn, &n2cdn);  

  if ((flow_comp[0] == 'y') && (reversed[0] == 'y')) {
    /* first moment assuming that compensation gradient ENDS at t=0 (see above) */
    double mcup = 0.5*tcup*(tcup-n1cup*gdt)*gcup;
    max_m = 0;
    for (i=0; i<nint; i++) {
      mx[i] += mcup*kx_b[i]/max_kb;
      my[i] += mcup*ky_b[i]/max_kb;
      max_m = MAX(max_m, fabs(mx[i]));
      max_m = MAX(max_m, fabs(my[i]));
    }
    tfc = calculate_flowcomp_gradient(max_m, upslew, 0.5*gmax, gdt, &gfc, &n1fc, &n2fc);
  } else {
    tfc = 0;
    n1fc = 0;
    n2fc = 0;
    gfc = 0;
    max_m = 1;
  }

  for (i=0; i<nint; i++) {
    double gcupx = -gcup*kx_b[i]/max_kb;
    double gcupy = -gcup*ky_b[i]/max_kb;
    double gcdnx = -gcdn*kx_a[i]/max_ka;
    double gcdny = -gcdn*ky_a[i]/max_ka;
    double gfcx = -gfc*mx[i]/max_m;
    double gfcy = -gfc*my[i]/max_m;

    write_spiral_to_file(i, nup+nsp+ndn, n1cup, n2cup, n1cdn, n2cdn, n1fc, n2fc,
			 gcupx, gcupy, gcdnx, gcdny, gfcx, gfcy, &gx[i][0], &gy[i][0], ix);
  }

  /* safety calculation worst case */
  grd_sqr = pow(gcup,2.0)*tcup+pow(gcdn,2.0)*tcdn+pow(gfc,2.0)*tfc;
  gph_sqr = grd_sqr;
  for (j=0; j<nup+nsp+ndn; j++) {
    grd_sqr += pow(gx[0][j],2.0)*gdt;
    gph_sqr += pow(gy[0][j],2.0)*gdt;
  }

  /* ------------- calculate timing  ----------------------------- */

  sw = 1.0/dt;
  fb = sw/2.0*1.2;
  pw_sp = tfc+tcup+tup+tacq+tdn+tcdn;
  np = floor(pw_sp/dt+1.0+(float)xsamp);
  write_pars_to_file(2*(long)np, (long)(np-(tacq/dt+1)), ix);

  *t_sp_b = tfc+tcup+tup;
  *t_sp_a = tdn+tcdn;
  if (reversed[0] == 'n') *t_sp_a += tacq; else *t_sp_b += tacq;
  
  /* ------------- user output  ---------------------------------- */
  if (ix == 1) { 
    char   err_string[MAXSTR];
    printf("\n");
    /* printf("np: %10.4f \n", np); */
    printf("maximum gradient: %6.2f (max: %6.2f) G/cm\n", max_g, gmax);
    printf("minimum slewrate: %6.2f (min: %6.2f) us/G/cm\n", (1.0/max_s)/US_S, grise/gmax/US_S);
    if(max_g>gmax){
      sprintf(err_string,"%s: gradient too large (> %4.2f G/cm\n)\n",seqfil,gmax);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    if(1.0/max_s<grise/gmax){
      sprintf(err_string,"%s: slewrate too small (< %4.2f us/G/cm\n)\n",seqfil,grise/gmax/US_S);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
  }
}


/* -------------------------------------------------------- */
/* runtime routines */
/* -------------------------------------------------------- */


spiral_gradients_and_acquire()
{
  /* if k-space trajectories are measured, scale some gradients to zero */
  int  i;

  if (type != 0) 
    for (i=0; i<MAXINT; i++) {
      if (type != 1) gsprd[i] = 0;
      if (type != 3) gspph[i] = 0;
    }

  obsoffset(resto);

  /* decide in RT which gradient waveform has to be used */
  /* each file has to be available, even if it is not used */

  /* 1) maximum possible number of interleaves (v7) is assigned to v10 */
  /* 2) interleave number to compare with is set by decrementing v10 */
  /* 3) interleave counter v9 is subtracted from v10, result stored in v11 */
  /* 4) if result is zero, perform interleave */
  /* 5) repeat from 2) */

  assign(v7,v10);
  
  /* interleave no. 9 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[9],spph[9],"",pw_sp, gsprd[9], gspph[9], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 8 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[8],spph[8],"",pw_sp, gsprd[8], gspph[8], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 7 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[7],spph[7],"",pw_sp, gsprd[7], gspph[7], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 6 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[6],spph[6],"",pw_sp, gsprd[6], gspph[6], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 5 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[5],spph[5],"",pw_sp, gsprd[5], gspph[5], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 4 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[4],spph[4],"",pw_sp, gsprd[4], gspph[4], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 3 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[3],spph[3],"",pw_sp, gsprd[3], gspph[3], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 2 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[2],spph[2],"",pw_sp, gsprd[2], gspph[2], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 1 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
    obl_shapedgradient(sprd[1],spph[1],"",pw_sp, gsprd[1], gspph[1], 0.0,1,NOWAIT);
    /* xgate(1); */
    acquire(2.0*np,1.0/sw);
  endif(v11);

  /* interleave no. 0 */
  decr(v10);
  sub(v10,v9,v11);
  ifzero(v11);
     obl_shapedgradient(sprd[0],spph[0],"",pw_sp, gsprd[0], gspph[0], 0.0,1,NOWAIT);
     /* xgate(1); */
     acquire(2.0*np,1.0/sw);
  endif(v11);
}

spiral_prescan()
{
  /* if k-space trajectories are measured, scale some gradients to zero */
  int  i;

  if (type != 0) 
    for (i=0; i<MAXINT; i++) {
      if (type != 1) gsprd[i] = 0;
      if (type != 3) gspph[i] = 0;
    }

  obsoffset(resto);

  /* interleave no. 0 */
  obl_shapedgradient(sprd[0],spph[0],"",pw_sp, gsprd[0], gspph[0], 0.0,1,NOWAIT);
  /* xgate(1); */
  acquire(2.0*np,1.0/sw);
}

excitation_pulse1(codeint vslice)
{
  if((ns>1)&&(exlist<0))exlist=curlist++;
  position_offset_list(pss,gss,ns,resto,OBSch,exlist,vslice);
  if(rfmod[0]=='y'){
    getelem(t1,vslice,v12);
    power(v12,OBSch);
    getelem(t60,vslice,v12);
  }
  else obspower(tpwr1);

  /* mw spiral */
  /* for Eddy current compensation apply inverted gradient before excitation */
  if (ec_comp == 1) {
    switch(type) {
    case 0: 
      obl_shapedgradient("","","gss",pw_gss,0.0,0.0,-gss,1,WAIT);
      break;
    case 1: case 2:
      obl_shapedgradient("gss","","",pw_gss,-gss,0.0,0.0,1,WAIT);
      break;
    case 3: case 4:
      obl_shapedgradient("","gss","",pw_gss,0.0,-gss,0.0,1,WAIT);
      break;
    }
    /* refocus pulse */
    switch(type) {
    case 0: 
      obl_shapedgradient("","","gssr_up",pw_gssr,0.0,0.0,-gssr,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("","","gssr_dn",pw_gssr,0.0,0.0,-gssr,1,WAIT);
      break;
    case 1: case 2:
      obl_shapedgradient("gssr_up","","",pw_gssr,-gssr,0.0,0.0,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("gssr_dn","","",pw_gssr,-gssr,0.0,0.0,1,WAIT);
      break;
    case 3: case 4:
      obl_shapedgradient("","gssr_up","",pw_gssr,0.0,-gssr,0.0,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("","gssr_dn","",pw_gssr,0.0,-gssr,0.0,1,WAIT);
      break;
    }
  }
  /* mw spiral */

  /* put out trigger on sp1 every spin excitation */
  sp1on();
  /* put out trigger on sp2 every volume */
  ifzero(v9);
    ifzero(v14);
       sp2on();
    endif(v14);
  endif(v9);

  switch(type) {
    case 0: 
      obl_shapedgradient("","","gss",pw_gss,0.0,0.0,gss,1,NOWAIT);
      break;
    case 1: case 2:
      obl_shapedgradient("gss","","",pw_gss,gss,0.0,0.0,1,NOWAIT);
      break;
    case 3: case 4:
      obl_shapedgradient("","gss","",pw_gss,0.0,gss,0.0,1,NOWAIT);
      break;
  }
  delay(fabs(gss)*grate+rftep);
  if(rfmod[0]=='y') {
    sp4on();
    shapedvpulse(p1pat,p1,v12,v1,rof1,rof2);
    sp4off();
  } else {
    sp4on();
    shaped_pulse(p1pat,p1,v1,rof1,rof2);
    sp4off();
  }
  if(tep<fabs(gss)*grate)delay(fabs(gss)*grate-tep);        

  sp1off();
  sp2off();

  /* refocus pulse */
  switch(type) {
    case 0: 
      obl_shapedgradient("","","gssr_up",pw_gssr,0.0,0.0,gssr,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("","","gssr_dn",pw_gssr,0.0,0.0,gssr,1,WAIT);
      break;
    case 1: case 2:
      obl_shapedgradient("gssr_up","","",pw_gssr,gssr,0.0,0.0,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("gssr_dn","","",pw_gssr,gssr,0.0,0.0,1,WAIT);
      break;
    case 3: case 4:
      obl_shapedgradient("","gssr_up","",pw_gssr,0.0,gssr,0.0,1,WAIT);
      delay(tref-pw_gssr);
      obl_shapedgradient("","gssr_dn","",pw_gssr,0.0,gssr,0.0,1,WAIT);
      break;
  }
}
