/*
  myepi.c
  
  Stuart Clare
  FMRIB Centre, University of Oxford
  April 1999
*/

/*

 num   ns                                                              ns  num
 ints                                                                      ints
  {    { [Inversion]  [Fatsat]  [Excitation]  [Spinecho]  [EPI][crush] }    } 

              |__________ti__________|__________te__________|

       |___________________________tr/ns_______________________________|    

*/

/*
  Real-time variables
  v1: phase cycling
  v2: phase cycling
  v3: phase cycling
  v4: phase cycling
  v5: number of pe steps
  v6: pe step number
  v7: pe increment
  v8: number of segments
  v9: segment number
  v10: 
  v11: 
  v12: rfmod power
  v13: number of slices
  v14: slice number
*/ 

/*
  RF Pulses
  p1 : tpwr1 : p1pat : gss  - Excitation
  p2 : tpwr2 : p2pat : gss2 - Spin echo
  p3 : tpwr3 : p3pat : gss3 - Inversion
  p4 : tpwr4 : p4pat        - Fatsat
  psat : satpwr : satpat : gsssat - Spatial Sat
*/

/*
  Tables
  t1 : tpwr1 + rfmod
  t2 : tpwr2 + rfmod
  t3 : tpwr3 + rfmod
  t4 : tpwr4 + rfmod
  t5 : tpwr5 + rfmod
  t60 : linear modulation
*/

#include "standard.h"
#include "group.h"
#include "sequence_utilities.h"
#define SAMPPTS 8192

/* Global variables*/
#include "globals.h"
#include "playout.h"
#include "calculate.h"
#include "tweekers.h"
#include "myepi_safety_calc.h"

pulsesequence()
{
  double t_inv_b,t_inv_a;
  double t_presat_b,t_presat_a;
  double t_cr,t_fat;
  double t_ex_b,t_ex_a;
  double t_se_b,t_se_a;
  double t_epi_b,t_epi_a;
  double min_te,min_tr,min_ti;
  double t_nav;
  
  double tr_delay,ti_delay,te_delay,se_delay_b,se_delay_a;

  double prescan_flag,echo_shift,pown;
  int image;
  char err_string[MAXSTR], pslabel[MAXSTR], test_string[MAXSTR];

  initparms_sis();
  get_tweekers();

  /* 180 degree pulse gradient strength */
  gss2 = getval("gss2");
  /* Inversion recovery gradient strength */
  gss3 = getval("gss3");

  /* Trigger and counter bit pattern */
  image = (int)(getval("image"));
  /* bit 1 = wait for ticks triggers before beginning scan */
  /* bit 2 = puts out high on sp3 for duration of volume */
  /* bit 3 = epi blips off (set bit on), otherwise blips on */
  /* bit 4 = make inversion non slice selective (on) or slice selective (off)*/
  /* bit 5 = turns off the acquire */

  TR = getval("TR");
  if(TR==0.0)TR=tr;

  /* Get flags 'y/n' */
  getstr("spinecho",spinecho);  
  getstr("inversion",inversion);  
  getstr("crushers",crushers);  
  getstr("fat_sat",fat_sat);
  getstr("half_kspace",half_kspace);
  getstr("nonlinsamp",nonlinsamp);
  getstr("pslabel",pslabel);
  getstr("rfmod",rfmod);
  getstr("nav",nav);

  presat=getval("presat");
  if (presat) {
    gsssat = getval("gsssat");
    presat_off=getval("presat_off");
    sat_psi=getval("sat_psi");
    sat_phi=getval("sat_phi");
    sat_theta=getval("sat_theta");
    sat_thk=getval("sat_thk");
    calculate_spatial_saturation_gradients(&t_presat_b,&t_presat_a);
  }

  /* Number of interleaves */
  num_ints = getval("num_ints");  
  if (num_ints == 0) num_ints = 1;

  echo_shift = getval("echo_shift"); 

  /* Half k-space stuff */
  if (half_kspace[0] == 'y') {
    
    if (num_ints > 1) {
      sprintf(err_string,"%s: Half K space acquisition requires 1 interleaf\n"
	      ,seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    
    pown = 2;
    do{
      if(pown == nv){
        sprintf(err_string,"%s: nv must be a few lines greater than n^2\n"
	        ,seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
      }
      eff_echo = nv - pown;
      pown *= 2;
    }while(pown < nv);
    etl = nv;
  }
  else {
    etl = (double)((int)(nv/num_ints));  /* echo train length/interleaf */
    eff_echo = etl/2.0/echo_shift;  /* echo position of TE */
    /* echo shift shifts the effective echo position */
    /* echo_shift = 1.0 (default) : eff_echo centred in echo train */
    /* echo_shift = 2.0 : eff_echo half way beteen start and centre of train */
    /* etc. */
  }

  /* Crusher gradient strength */
  gcrush=gmax/2.0;

  /* Gradient calculations */
  if(inversion[0]=='y')calculate_inversion_gradients(&t_inv_b,&t_inv_a);
  else { t_inv_b=0.0; t_inv_a=0.0; }

  if(crushers[0]=='y')calculate_crusher_gradients(&t_cr);
  else t_cr=0.0;

  if(fat_sat[0]=='y')calculate_fatsat_gradients(&t_fat);
  else t_fat=0.0;

  calculate_excitation_gradients(&t_ex_b,&t_ex_a);

  if(spinecho[0]=='y')calculate_spinecho_gradients(&t_se_b,&t_se_a);
  else { t_se_b=0.0; t_se_a=0.0; }

  calculate_epi_gradients(&t_epi_b,&t_epi_a);

  if(rfmod[0]=='y'){
    calculate_rf_modulation();
  }

  if(nav[0]=='y'){
    calculate_navigator_gradients(&t_nav);
  }
  else t_nav=0;

  /* Timing calculations */

  /* Nonlinear sampling */
  if(nonlinsamp[0]=='y')calculate_nonlinear_sampling();

  /* Spinecho delay */
  se_delay_b = se_delay_a = 0.0;
  if(spinecho[0]=='y'){
    if( (t_ex_a+t_se_b) > (t_se_a+t_epi_b+t_nav) )
      se_delay_a=(t_ex_a+t_se_b)-(t_se_a+t_epi_b+t_nav);
    if( (t_ex_a+t_se_b) < (t_se_a+t_epi_b+t_nav) )
      se_delay_b=(t_se_a+t_epi_b+t_nav)-(t_ex_a+t_se_b);
  }

  /* te delay */
  min_te = t_ex_a + se_delay_b + t_se_b + t_se_a + se_delay_a + t_nav + t_epi_b;
  if(min_te>te){
    sprintf(err_string,"%s: te too short. Min te = %f\n",seqfil,min_te);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  te_delay = te - min_te;

  /* ti delay */
  if(inversion[0]=='y'){
    min_ti = t_inv_a + t_fat + t_ex_b;
    if(min_ti>ti){
      sprintf(err_string,"%s: ti too short. Min ti = %f\n",seqfil,min_ti);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    ti_delay = ti - min_ti;
  }
  else ti_delay=0;

  /* tr delay */
  min_tr = (t_inv_b + t_inv_a + ti_delay +
	    t_fat + t_ex_b + te + t_epi_a + t_cr);
  if (presat) min_tr = min_tr + t_presat_b + t_presat_a;
  min_tr = min_tr * ns;
  if(min_tr>tr){
    sprintf(err_string,"%s: tr too short. Min tr = %f\n",seqfil,min_tr);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
  tr_delay = (tr - min_tr)/ns;

  /* Prescan flag (1=yes, 0=no) */
  prescan_flag = getval("prescan_flag"); 
  if (prescan_flag == 1)num_segs=1;
  else num_segs=(int)(nv/etl);
  
  /* Turn blips off */
  if ((image & 4) == 4) {
    sgbpe = 0.0;
    gbd=0.0;
    sgbd=0.0;
  }

  /* Saftey check */
  if (myepi_safety_calc() == -1) {
    sprintf(err_string,"%s: Failed duty cycle safety check\n",
	    seqfil);
    printf(err_string);
    log_error(err_string);
    abort(1);
  }
 
  /* set up the No r.f. epi sequence if pslabel is set to "EPI_norf" */
  sprintf(test_string,"EPI_norf");
  if (strcmp(pslabel,test_string) == 0) {
    tpwr1=-16;
    tpwr2=-16;
    tpwr3=-16;
    tpwr4=-16;
    satpwr=-16;
  }

  /**********************************************************************/
  /* PULSE SEQUENCE */
  /**********************************************************************/

  /* Phase cycling */
  mod2(ct,v3);                    /* v3 = 0101 0101 */
  dbl(v3,v3);                     /* v3 = 0202 0202 */
  hlv(ct,v4);                     /* v4 = 0011 2233 */
  mod2(v4,v4);                    /* v4 = 0011 0011 */
  add(v3,v4,v1);                  /* v1 = 0213 0213 */
  assign(v1,oph);                 /* oph = 0213 0213 */
  add(one,v1,v2);

  /* phase encode increment size */
  assign(one,v7);  
  
  /* Receiver setup */
  status(A);
  sp4off();
  rcvroff();
  rcvron();
  sp2off();

  initval(num_segs,v8);  /* set v8 to the num of segments */


  if ((image & 1) == 1) {
    xgate(ticks);  /* wait for start scan button */
    delay(1.0e-3); /* 1 ms delay to allow for trigger TTL level recovery */
  }
  if ((image & 2) == 2) sp3on(); else sp3off();

  /* xgate(ticks); */

  delay(del_seq);
  
  /* loop v8 (num_ints) times using counter v9 */
  loop(v8,v9);

  /* sets v13 to ns and incs v14 up to it */
  msloop(seqcon[1],ns,v13,v14);
  
  /* Inversion pulse */
  if(inversion[0] == 'y') {
    if((image & 8) == 8) {
      nonselective_inversion_pulse();
    }
    else {
      selective_inversion_pulse(v14);
    }
    delay(ti_delay);
  }

  if (presat) spatial_saturation_pulse();

  if (fat_sat[0] == 'y') {
    fatsat_pulse();
  }

  excitation_pulse(v14);

  delay(te_delay/2.0);

  if (spinecho[0] == 'y'){
    delay(se_delay_b);
    spinecho_pulse(v14);
    delay(se_delay_a);
  }
  
  delay(te_delay/2.0);

  if(nav[0]=='y'){
  	navigator_acquire();
  }

  if((image & 16) == 16){
	epi_gradients_no_acquire();
  }
  else epi_gradients_and_acquire();

  crusher_gradients();

  /* double check that all gradients are zero */
  zero_all_gradients();

  delay(tr_delay);
  
  /* increment the msloop counter v14 */
  endmsloop(seqcon[1],v14);
 
  /* increment the interleaves loop counter v9 */
  endloop(v9);

  delay(TR-tr);
}
