function rd_varian_bin, fname, verbose=verbose

if (n_elements(verbose) eq 0) then verbose=0

data_file_header_size=long(32)	; number of bytes in each data file header
block_header_size=long(28)	; number of bytes in each block header

; Reads a Varian binary file

; definition of data file header
; ==============================
nblocks=long(0)		; number of blocks in file
ntraces=long(0)		; number of traces per block
np=long(0)		; number of elements per trace
ebytes=long(0)		; number of bytes per element
tbytes=long(0)		; number of bytes per trace
bbytes=long(0)		; number of bytes per block
vers_id=fix(0)		; software version, file_id status bits
status=fix(0)		; status of whole file
nbheaders=long(0)	; number of block headers per block

get_lun, unit
openr, unit, fname

readu, unit, nblocks
readu, unit, ntraces
readu, unit, np
readu, unit, ebytes
readu, unit, tbytes
readu, unit, bbytes
readu, unit, vers_id
readu, unit, status
readu, unit, nbheaders

byteswap_flag=0
if (np*ebytes ne tbytes) then begin
  byteswap_flag=1
  byteorder, nblocks, /LSWAP
  byteorder, ntraces, /LSWAP
  byteorder, np, /LSWAP
  byteorder, ebytes, /LSWAP
  byteorder, tbytes, /LSWAP
  byteorder, bbytes, /LSWAP
  byteorder, vers_id
  byteorder, status
  byteorder, nbheaders, /LSWAP
endif

if (verbose gt 0) then begin
  print, "data file header"
  print, "================"
  print, ""
  print, "nblocks = "+strtrim(string(nblocks),2)+" (# blocks/file)"
  print, "ntraces = "+strtrim(string(ntraces),2)+" (# traces/block)"
  print, "np = "+strtrim(string(np),2)+" (# elements/trace)"
  print, "ebytes = "+strtrim(string(ebytes),2)+" (# bytes/element)"
  print, "tbytes = "+strtrim(string(tbytes),2)+" (# bytes/trace)"
  print, "bbytes = "+strtrim(string(bbytes),2)+" (# bytes/block)"
  print, "vers_id = "+strtrim(string(vers_id),2)
  print, "status = "+strtrim(string(status),2)
  print, "nbheaders = "+strtrim(string(nbheaders),2)+" (# blocks headers/block)"
  print, ""
  print, "data summary:"
  print, "============="
  print, ""
  print, "data file header at start of file : "+ $
                 strtrim(string(data_file_header_size),2)+" bytes"
  print, "   block header at start of each block : "+ $
                 strtrim(string(block_header_size),2)+" bytes"
  print, "   data size of each block : "+ $
                 strtrim(string(np*ebytes*ntraces),2)+" bytes"
  print, "   total size of each block : "+ $
                 strtrim(string(block_header_size+np*ebytes),2)+" bytes"
  print, "   number of blocks : "+strtrim(string(nblocks),2)
  print, "data size : "+strtrim(string(np*ebytes*ntraces*nblocks),2)
  print, "file size : "+strtrim(string((np*ebytes*ntraces+ $
                 block_header_size)*nblocks+data_file_header_size),2)
  print, ""
endif

; set the staus bits to zero as default
S_NI2=0
S_NI=0
S_NF=0
S_NP=0
S_NP=0
S_TRANSF=0
S_SECND=0
S_ACQPAR=0
S_HYPERCOMPLEX=0
S_COMPLEX=0
S_FLOAT=0
S_32=0
S_SPEC=0
S_DATA=0

; interpret the status bits:
; ==========================

if (status gt 32767) then begin
  ; not used
  status=status-32768
endif
if (status gt 16383) then begin
  S_NI2=1 ; ni2 dimension is active
  status=status-16384
endif
if (status gt 8191) then begin
  S_NI=1 ; ni dimension is active
  status=status-8192
endif
if (status gt 4095) then begin
  S_NF=1 ; nf dimension is active
  status=status-4096
endif
if (status gt 2047) then begin
  S_NP=1 ; np dimension is active
  status=status-2048
endif
if (status gt 1023) then begin
  ; not used
  status=status-1024
endif
if (status gt 511) then begin
  S_TRANSF=1 ; 0=regular, 1=transposed
  status=status-512
endif
if (status gt 255) then begin
  S_SECND=1 ; 0=first FT, 1=second FT
  status=status-256
endif
if (status gt 127) then begin
  S_ACQPAR=1 ; 0=not Acqpar, 1=Acqpar
  status=status-128
endif
if (status gt 63) then begin
  ; not_used
  status=status-64
endif
if (status gt 31) then begin
  S_HYPERCOMPLEX=1 ; 1=hypercomplex
  status=status-32
endif
if (status gt 15) then begin
  S_COMPLEX=1 ; 0=real, 1=complex
  status=status-16
endif
if (status gt 7) then begin
  S_FLOAT=1 ; 0=integer, 1=floating point
  status=status-8
endif
if (status gt 3) then begin
  S_32=1 ; if integer (i.e., S_FLOAT=0) then 0=16 bit, 1=32 bit
  status=status-4
endif
if (status gt 1) then begin
  S_SPEC=1 ; 0=FID, 1=spectrum
  status=status-2
endif
if (status gt 0) then begin
  S_DATA=1 ; 0=no data, 1=data
  status=status-1
endif

; ***status should now be zero...***
if (status ne 0) then begin
  print, "Error: could not interpret the status bits"
endif

if (verbose gt 0) then begin
  print, "interpretation of status bits:"
  print, "=============================="
  print, ""
  if (S_DATA eq 0) then begin
    print, "no data"
  endif
  if (S_SPEC eq 0) then begin
    print, "FID"
  endif else begin
    print, "spectrum"
  endelse
  if (S_FLOAT eq 0) then begin
    if (S_32 eq 0) then begin
      print, '16 bit integer data'
    endif else begin
      print, '32 bit integer data'
    endelse
  endif else begin
    print, "floating point data"
  endelse
  if (S_COMPLEX eq 0) then begin
    print, "real data"
  endif else begin
    print, "complex data"
  endelse
  if (S_HYPERCOMPLEX eq 1) then print, "hypercomplex data"
  if (S_ACQPAR eq 0) then begin
    print, "not Acqpar"
  endif else begin
    print, "Acqpar"
  endelse
  if (S_SECND eq 0) then begin
    print, "ready for first FT"
  endif else begin
    print, "ready for second FT"
  endelse
  if (S_TRANSF eq 1) then print, "transposed"
  if (S_NP eq 1) then print, "np dimension is active"
  if (S_NF eq 1) then print, "nf dimension is active"
  if (S_NI eq 1) then print, "ni dimension is active"
  if (S_NI2 eq 1) then print, "ni2 dimension is active"
endif

if (S_FLOAT eq 1) then begin
  if (S_COMPLEX eq 1) then begin
    tmp_block_arr=fltarr(2,np/2*ntraces)
    main_data_arr=fltarr(2,np/2*ntraces,nblocks)
  endif
  if (S_COMPLEX eq 0) then begin
    tmp_block_arr=fltarr(2,np/2*ntraces)
    main_data_arr=fltarr(2,np/2*ntraces,nblocks)
  endif
  byteswap_val=0
endif else begin
  if (S_COMPLEX eq 1) then begin
    if (S_32 eq 0) then begin
      tmp_block_arr=intarr(2,np/2*ntraces)
      main_data_arr=intarr(2,np/2*ntraces,nblocks)
      byteswap_val=1
    endif
    if (S_32 eq 1) then begin
      tmp_block_arr=lonarr(2,np/2*ntraces)
      main_data_arr=lonarr(2,np/2*ntraces,nblocks)
      byteswap_val=2
    endif
  endif
  if (S_COMPLEX eq 0) then begin
    if (S_32 eq 0) then begin
      tmp_block_arr=intarr(2,np/2*ntraces)
      main_data_arr=intarr(2,np/2*ntraces,nblocks)
      byteswap_val=1
    endif
    if (S_32 eq 1) then begin
      tmp_block_arr=lonarr(2,np/2*ntraces)
      main_data_arr=lonarr(2,np/2*ntraces,nblocks)
      byteswap_val=2
    endif
  endif
endelse


; definition of block header
; ==========================
scale=fix(0)		; scaling factor
status=fix(0)		; status of data in block
index=fix(0)		; block index
mode=fix(0)		; mode of data in block
ctcount=long(0)		; ct value for fid
lpval=float(0)		; f2 (2D-f1) left phase in phase file
rpval=float(0)		; f2 (2D-f1) right phase in phase file
lvl=float(0)		; level drift correction
tlt=float(0)		; tilt drift correction

for block_count=long(0), long(nblocks)-1 do begin

  if (verbose gt 0) then begin
    print, ""
    print, "reading block "+strtrim(string(block_count+1),2)+"..."
  endif

  readu, unit, scale
  readu, unit, status
  readu, unit, index
  readu, unit, mode
  readu, unit, ctcount
  readu, unit, lpval
  readu, unit, rpval
  readu, unit, lvl
  readu, unit, tlt

  if (byteswap_flag) then begin
    byteorder, scale
    byteorder, status
    byteorder, index
    byteorder, mode
    byteorder, ctcount, /LSWAP
    byteorder, lpval, /FTOXDR
    byteorder, rpval, /FTOXDR
    byteorder, lvl, /FTOXDR
    byteorder, tlt, /FTOXDR
  endif

  if (verbose gt 0) then begin
    print, ""
    print, "scale = "+strtrim(string(scale),2)
    print, "status = "+strtrim(string(status),2)
    print, "index = "+strtrim(string(index),2)
    print, "mode = "+strtrim(string(mode),2)
    print, "ctcount = "+strtrim(string(ctcount),2)
    print, "lpval = "+strtrim(string(lpval),2)
    print, "rpval = "+strtrim(string(rpval),2)
    print, "lvl = "+strtrim(string(lvl),2)
    print, "tlt = "+strtrim(string(tlt),2)
  endif

  ; interpret the status bits for this block
  ; ========================================

  S_NI2=0
  S_NI=0
  S_NF=0
  S_NP=0
  S_NP=0
  S_TRANSF=0
  S_SECND=0
  MORE_BLOCKS=0
  NP_CMPLX=0
  NF_CMPLX=0
  NI_CMPLX=0
  NI2_CMPLX=0

  ; interpret the status bits:
  ; ==========================

  if (status gt 32767) then begin
    ; not used
    status=status-32768
  endif
  if (status gt 16383) then begin
    ; not used
    status=status-16384
  endif
  if (status gt 8191) then begin
    ; not used
    status=status-8192
  endif
  if (status gt 4095) then begin
    ; not used
    status=status-4096
  endif
  if (status gt 2047) then begin
    NI2_CMPLX=1 ; 0=real, 1=complex
    status=status-2048
  endif
  if (status gt 1023) then begin
    NI_CMPLX=1 ; 0=real, 1=complex
    status=status-1024
  endif
  if (status gt 511) then begin
    NF_CMPLX=1 ; 0=real, 1=complex
    status=status-512
  endif
  if (status gt 255) then begin
    NP_CMPLX=1 ; 0=real, 1=complex
    status=status-256
  endif
  if (status gt 127) then begin
    MORE_BLOCKS=1 ; 0=absent, 1=present
    status=status-128
  endif
  if (status gt 63) then begin
    ; not_used
    status=status-64
  endif
  if (status gt 31) then begin
    S_HYPERCOMPLEX=1 ; 1=hypercomplex
    status=status-32
  endif
  if (status gt 15) then begin
    S_COMPLEX=1 ; 0=real, 1=complex
    status=status-16
  endif
  if (status gt 7) then begin
    S_FLOAT=1 ; 0=integer, 1=floating point
    status=status-8
  endif
  if (status gt 3) then begin
    S_32=1 ; if integer (i.e., S_FLOAT=0) then 0=16 bit, 1=32 bit
    status=status-4
  endif
  if (status gt 1) then begin
    S_SPEC=1 ; 0=FID, 1=spectrum
    status=status-2
  endif
  if (status gt 0) then begin
    S_DATA=1 ; 0=no data, 1=data
    status=status-1
  endif

  ; ***status should now be zero...***
  if (status ne 0) then begin
    print, "Error: could not interpret the status bits"
  endif

  if (verbose gt 0) then begin
    print, ""
    print, "interpretation of status bits:"
    print, "=============================="
    print, ""
    if (S_DATA eq 0) then begin
      print, "no data"
    endif
    if (S_SPEC eq 0) then begin
      print, "FID"
    endif else begin
      print, "spectrum"
    endelse
    if (S_FLOAT eq 0) then begin
      if (S_32 eq 0) then begin
        print, '16 bit integer data'
      endif else begin
        print, '32 bit integer data'
      endelse
    endif else begin
      print, "floating point data"
    endelse
    if (S_COMPLEX eq 0) then begin
      print, "real data"
    endif else begin
      print, "complex data"
    endelse
    if (S_HYPERCOMPLEX eq 1) then print, "hypercomplex data"
    if (MORE_BLOCKS eq 1) then begin
      print, "more blocks"
    endif else begin
      print, "last block"
    endelse
    if (NP_CMPLX eq 1) then print, "np dimension is complex"
    if (NF_CMPLX eq 1) then print, "nf dimension is complex"
    if (NI_CMPLX eq 1) then print, "ni dimension is complex"
    if (NI2_CMPLX eq 1) then print, "ni2 dimension is complex"
  endif

  ; interpret the mode bits for this block
  ; ======================================

  NP_PHMODE=0
  NP_AVMODE=0
  NP_PWRMODE=0
  NF_PHMODE=0
  NF_AVMODE=0
  NF_PWRMODE=0
  NI_PHMODE=0
  NI_AVMODE=0
  NI_PWRMODE=0
  NI2_PHMODE=0
  NI2_AVMODE=0
  NI2_PWRMODE=0

  ; interpret the mode bits:
  ; ========================
  mode_start=mode

  if (mode gt 32767) then begin
    ; not used
    mode=mode-32768
  endif
  if (mode gt 16383) then begin
    NI2_PWRMODE=1
    mode=mode-16384
  endif
  if (mode gt 8191) then begin
    NI2_AVMODE=1
    mode=mode-8192
  endif
  if (mode gt 4095) then begin
    NI2_PHMODE=1
    mode=mode-4096
  endif
  if (mode gt 2047) then begin
    ; not used
    mode=mode-2048
  endif
  if (mode gt 1023) then begin
    NI_PWRMODE=1
    mode=mode-1024
  endif
  if (mode gt 511) then begin
    NI_AVMODE=1
    mode=mode-512
  endif
  if (mode gt 255) then begin
    NI_PHMODE=1
    mode=mode-256
  endif
  if (mode gt 127) then begin
    ; not used
    mode=mode-128
  endif
  if (mode gt 63) then begin
    NF_PWRMODE=1
    mode=mode-64
  endif
  if (mode gt 31) then begin
    NF_AVMODE=1
    mode=mode-32
  endif
  if (mode gt 15) then begin
    NF_PHMODE=1
    mode=mode-16
  endif
  if (mode gt 7) then begin
    ; not used
    mode=mode-8
  endif
  if (mode gt 3) then begin
    NP_PWRMODE=1
    mode=mode-4
  endif
  if (mode gt 1) then begin
    NP_AVMODE=1
    mode=mode-2
  endif
  if (mode gt 0) then begin
    NP_PHMODE=1
    mode=mode-1
  endif

  ; ***mode should now be zero...***
  if (mode ne 0) then begin
    print, "Error: could not interpret the mode bits"
  endif

  if (verbose gt 0) then begin
    if (mode_start ne 0) then begin
      print, ""
      print, "interpretation of mode bits:"
      print, "============================"
      print, ""
    endif
    if (NP_PHMODE eq 1) then print, "NP_PHMODE=1"
    if (NP_AVMODE eq 1) then print, "NP_AVMODE=1"
    if (NP_PWRMODE eq 1) then print, "NP_PWRMODE=1"
    if (NF_PHMODE eq 1) then print, "NF_PHMODE=1"
    if (NF_AVMODE eq 1) then print, "NF_AVMODE=1"
    if (NF_PWRMODE eq 1) then print, "NF_PWRMODE=1"
    if (NI_PHMODE eq 1) then print, "NI_PHMODE=1"
    if (NI_AVMODE eq 1) then print, "NI_AVMODE=1"
    if (NI_PWRMODE eq 1) then print, "NI_PWRMODE=1"
    if (NI2_PHMODE eq 1) then print, "NI2_PHMODE=1"
    if (NI2_AVMODE eq 1) then print, "NI2_AVMODE=1"
    if (NI2_PWRMODE eq 1) then print, "NI2_PWRMODE=1"
  endif

  readu, unit, tmp_block_arr
  if (byteswap_flag) then begin
    if (byteswap_val eq 0) then byteorder, tmp_block_arr, /FTOXDR
    if (byteswap_val eq 1) then byteorder, tmp_block_arr
    if (byteswap_val eq 2) then byteorder, tmp_block_arr, /LSWAP
  endif

  main_data_arr(0,*,block_count)=tmp_block_arr(0,*)-lvl
  main_data_arr(1,*,block_count)=tmp_block_arr(1,*)-tlt

endfor

close, unit
free_lun, unit

return, reform(complex(main_data_arr(0,*,*),main_data_arr(1,*,*)))

end
