
/*****************************************************************************
*  Safety checking routine to check for excess duty cycle values on the      *
*  gradients coil. This file should be adapted for each individual pulse     *
*  sequence so that a correct duty cycle on the logical read, phase encode   *
*  and slice axes can be calculated. The system file safety_check() is then  *
*  called to rotate the logical axes on to their respective physical axes    *
*  (according to the values of psi, phi and theta), and then calculate       *
*  whether an excessive duty cycle has been requested for the gradient coil  *
*  in use. An error code is returned to the pulse sequence if the duty       *
*  cycle is exceeded. Note that in order for this code to work, the pulse    *
*  sequence variable declarations must be outside the main pulsesequence()   *
*  routine. This allows them to be "seen" by the duty cycle calculation      *
*  routine.                                                                  *
*                                                                            *
*  This file must be correctly written for each pulse sequence so that the   *
*  RMS gradient amplitude on each logical axis can be calculated. This is    *
*  done by performing the following summation on each logical axis:          *
*                                                                            *
*           GRAD_i_RMS=sqrt[ Sum{(grad_ij^2)*grad_ij_duration} / tr ]        * 
*                                                                            *
*           where i=read, phase, slice                                       *
*                 j=1,2,3,4,5...  (gradient pulse index)                     *
*                                                                            *
*  Peter Jezzard, University of Oxford, 1998                                 *
*  This version of the routine suitable for: ox_sems.c                       *
*  Modification history:                                                     *
*       written Sept 1998 PJ                                                 *
*****************************************************************************/

/* safety_check.h must be included. This routine contains the gradient
   characteristics and does the actual safety check                         */
#include "safety_check.h"

int ox_sems_safety_calc()
{

  /* variables to hold read, phase and slice rms gradients */
  double grad_r_rms, grad_p_rms, grad_s_rms; 
  /* declare a dummy variable that allows for a weighting to be applied
     to a calculation based on the frequency that the relevant pulse unit
     occurs per net tr time */
  double tr_wt;

  /* error value returned to main code. -1=failure, 1=success */
  int err;

  /* initiate the rms gradient values to zero on all axes */
  grad_r_rms=0.0;
  grad_p_rms=0.0;
  grad_s_rms=0.0;

  /* initialize tr_wt value */
  tr_wt=1.0;

  /**************************************************************************/
  /* must edit the code from this point on to suit the sequence being run   */
  /**************************************************************************/

  /* calculate the contribution from the inversion pulse, if ti > 0.0 */
  if (ti > 0.0) {
    grad_s_rms=grad_s_rms+pow(gssi,2.0)*(agssi*grate/2.0+p3+tspoil
        +agssi*grate/2.0)*tr_wt;
  }

  /* calculate the contribution from the pulse sequence itself */
  grad_s_rms=grad_s_rms+pow(gss,2.0)*(agss*grate/2.0+p1+agss*grate/2.0)
        *tr_wt; /* 90 */
  if ((flow_comp[0] != 'y') && (flow_comp[0] != 'Y')) { /* no flow comp */
    grad_r_rms=grad_r_rms+pow(gror,2.0)*(pw_gpe/2.0+(tref-pw_gpe)+
        pw_gpe/2.0)*tr_wt;  /* read dephaser */
    grad_p_rms=grad_p_rms+pow(sgpe*nv/2.0,2.0)*(pw_gpe/2.0+(tref-pw_gpe)+
        pw_gpe/2.0)*tr_wt;  /* phase encode dephaser */
  }
  if ((flow_comp[0] == 'y') || (flow_comp[0] == 'Y')) { /* flow comp */
    /* FLOW COMP STILL NEEDS TO BE IMPLEMENTED FOR OX_SEMS.C */
    /* the phase encode rms contribution is easy to calculate */
    grad_p_rms=grad_p_rms+pow(sgpe*nv/2.0,2.0)*(pw_gpe/2.0+(tref-grise)+
        pw_gpe/2.0)*tr_wt;  /* phase encode dephaser */
    /* the read and slice gradients are more difficult to calculate, because
       they are hidden in the waveform calculation algorithm. But they roughly
       follow a 1:2:1 integral, and the integral of the read out and slice
       gradients are known. So estimate the values from that... */
    grad_r_rms=grad_r_rms+pow(gro,2.0)*(grate*agro/2+tspoil+at/2-grate*agro
        +grate*agro/2+grate*agro/2+2.0*tspoil+at+grate*agro/2)*tr_wt;
    grad_s_rms=grad_s_rms+pow(gss,2.0)*(grate*agss/2+p1+2.0*rof2
        +grate*agss/2)*tr_wt;
    grad_s_rms=grad_s_rms+pow(gssr,2.0)*(pw_gpe/2.0+(tref-pw_gpe)+
        pw_gpe/2.0)*tr_wt;
  }

  /* 180 degree pulse */
  grad_s_rms=grad_s_rms+pow(agssc,2.0)*(agssc*grate/2+tspoil_se1+
      agssc*grate/2)*tr_wt;
  grad_s_rms=grad_s_rms+pow(gss2,2.0)*(p2)*tr_wt;
  grad_s_rms=grad_s_rms+pow(agssc,2.0)*(agssc*grate/2+tspoil_se2+
      agssc*grate/2)*tr_wt;

  /* final read out gradient */
  grad_r_rms=grad_r_rms+pow(gro,2.0)*(pw_gread/2.0+tspoil+at
      +pw_gread/2.0)*tr_wt;

  /* account for the crushers */
  if (pw_end_gcrush > 0.0) {
    grad_p_rms=grad_p_rms+pow(gcrush,2.0)*(agcrush*grate+tcrush)*tr_wt;
  }
  
  /* factor in the number of slices per tr period, and divide by the tr */
  /* NOTE: only one slice per tr for this sequence */
  grad_r_rms=grad_r_rms*ns/tr;
  grad_p_rms=grad_p_rms*ns/tr;
  grad_s_rms=grad_s_rms*ns/tr;

  /* calculate the RMS value by taking sqrt */
  grad_r_rms=sqrt(grad_r_rms);
  grad_p_rms=sqrt(grad_p_rms);
  grad_s_rms=sqrt(grad_s_rms);

  err=-1;  /* default=-1 (1=pass, -1=fail) */
  err=safety_check(grad_r_rms, grad_p_rms, grad_s_rms);

  return err;

}
