/*
 *  General reconstruction using a gridding algorithm.
 *
 *  Maps signa 5x data onto Fourier grid using a Kaiser-Bessel window.  
 *  The gridded data are then reconstructed using a 2D FFT.
 *
 *  The Bo correction fits a plane to the field map and does constant +
 *  linear gx and gy corrections.
 *
 *     3dpr7	9/15/93		ghg     proj recon with kernel corr	
 *     grecon	4/30/95		ghg	arb k-space recon
 *     grecon1	5/1/95		ghg	add b0 map
 *     grecon3	5/8/95		ghg	add nav correction
 *     grecon4	7/12/95		ghg	auto open of trajectory files
 *     grecon5	7/21/95			separate nviews from rhnframes
 *     grecon6	12/19/95	ghg	fix file close bug for phase imgs
 *     grecon7	1/25/96		ghg	add baselines for fast_rec
 *     grecon8	1/27/96		ghg	read freq offset from header
 *     grecon9	1/29/96		ghg	new data format: single slice
 *     grecon10	2/2/96		ghg	remove baseline code, add frec zit
 *					remover algorithm
 *     grecon11	2/8/96		ghg	back to old data format
 *     grecon12	2/28/96		ghg	can do subset of slices
 *     grecon13	3/29/96		ghg	add fermi filter (to pbuf)
 *     grecon14	8/8/96			navcor checks data for nulls and 
 *					won't activate if ampl too low
 *     grecon15	8/27/96			fix B0 map bug, sum buffers not
 *					inited
 *     grecon16	11/11/96		fix B0 bug for 1 shot, (rhnfrm=2)
 *     grecon17	11/17/96		add zooming during gridding
 *					use z(1)+z(-1) for zit remover
 *     grecon18	11/24/96		new data structure:  all
 *					frames are in rhnframes
 *		2/24/97			replace irint with int
 *     grecon19	7/25/97			navcor finds max point/view and 
 *					uses separate ref point instead 
 *					of at k = 0. Also turns navcor off for 
 *					nviews = 1 (doesn't help).
 *     grecon20	7/30/97			uses old navcor alg if data ok at origin
 *     grecon21	8/1/97			can do multi echo
 *     grecon22	2/18/98			trap divide by zero for atan2
 *     		6/29/98			trap fwrite errors
 *
 *     grecon22lx 	7/3/98		lx rdbmheader
 *     greconlx23 	10/15/98	can do just B0 map for shimming
 *     greconlx24 	10/28/98	does concat acquisition
 *     greconlx25 	10/29/98	new data format: concat acq are
 *					adjacent views
 *     grecons26 	2/4/99		calc trajectory internally	
 *     grecons27 	3/4/99		does maxwell field correction
 *     grecons28 	3/20/99		fix factor of 2 in B0 map scale
 *     grecons29 	4/2/99		if bmapflag=2, make distinctive
 *					file names for B0 maps
 *     grecons30 	4/15/99		new genspiral
 *     grecons31 	5/16/99		does real/imaginary
 *
 *  Written 1993- 1999 by Gary Glover.  Gridding loop is J. Pauly-like.
 *  (c) Board of Trustees, Leland Stanford Junior University
 * ---------------------------------------------------------------------
 *    gr	99_09_09   WML	Converted to work with spep.e with new
 *				looping structure
 *		99_10_26   WML	Changed to command line parsing
 *		99_10_28   ECW	Added 3D spiral FSE recon
 *				Recon type controlled by recotype
 *		99_11_13   WML	Changed output filename for phase imgs, etc.
 *				Added -subk option for subtraction in k-space
 *		99_12_12   WML  Added -float option to write out float format
 *				Added -scale option to scale output
 *              00_02_04   PJ   Altered code to deal with Varian data
 *
 ***** recotype values ******
 *  1  2D spiral
 *  2  3D spiral FSE
 */
 
#include <string.h>
#include <stdio.h>
#include <math.h>
#define RECON_FLAG		/*  for the next header include  */
#include "rdbmlx.h"

#define MXNIM        256      /* max image size           */
#define IMD    (2*MXNIM)      /* reconstruction size      */
#define MXSAMP     16390       /* max acquisition length   */
#define MXNVW        128       /* max num views   	*/
#define MAG            1      /* flags for write_image()  */
#define PHASE          2
#define REAL           4
#define IMAG           8
#define NOSUBSC       16
#define	NGRID		512	/* cal factor used in making traj files */
#define FERMIW   (10.0/128)	/* normalized width of fermi window  */
#define REV   	      "gr: 12/15/99"	/* rev number of this source  */
int genspiral(float D, int N, float Tmax, float dts); /* prototype  */
float bessi0();         /*  modified Bessel function  */
void fftc();
void load_view(),load_ft(),ft2d();  /* WML */
void init(), set_traj(), maxwell(), calcden() ;  /* WML */
void instructions(), parsearg(); /* WML */

float raw[IMD][IMD][2];		/*  kspace data */
float img[MXNIM][MXNIM][2];	/*  image */
float dbuf[MXSAMP][2];		/*  data buffer  */
float cosbf[MXSAMP];		/*  zit buffer  */
float pbuf[MXSAMP][2];		/*  phase vector for freq shift */
float gf[IMD], igf[IMD];	/*  gridding buffers  */
float ws[IMD][IMD];		/*  weighting buffer  */
short int ibuf[MXSAMP][2]; 	/*  read buffer  */
float fbuf[MXSAMP][2]; 		/*  float read buffer  */
float gx[2*MXSAMP], gy[2*MXSAMP];
float kx[MXNVW][MXSAMP], ky[MXNVW][MXSAMP], dens[MXNVW][MXSAMP]; /* traject */
float kx0[MXNVW][MXSAMP], ky0[MXNVW][MXSAMP];
float Kx[MXSAMP], Ky[MXSAMP], Dens[MXSAMP];
float phim[MXSAMP]; 		/* maxwell field correction */
int npix, npixo2, nsamp, nslc, necho, ngrid;
int rawimsize;
int rawnslc;
int reps, ntotal, nf2do;
int rhfrsize, rhnecho;
int nbpi;  		/* bytes per image  */
int imrot, imtran;
int offset;
int recotype;
float zoomf = 1;
int imgtyp = 1;		/* output image type bitmap  */
float xoff, yoff;	/*  offsets in pixels  */
float foffin = 0;	/* offset from command line */
float foff = 0;		/* total freq offset */
float frcvr;		/* recevier frequency offset  */
float gain = 1.0;	
float dts, mapdt;
float gfov, gtype, gamp, gslew;	/* for genspiral  */
int N, nl;			/* matrix size for genspiral */
int bline;		/* number of baselines */
int bmapflag;		/* do the b0 map if 1  */
int navc = 1;		/* do nav correction if set */
float w0 = 0.0, wx = 0.0, wy = 0.0;
float dwin = 2;		/* half length of Kaisser-Bessel window */
float b = 8.0;		/*  Kaisser-Bessel param  */
float fermir;		/* radius of fermi window */
float fermiw;		/* width of fermi window  */
int isl, ifr, iech;
int isl1, isl2;
int etl, zres, zstep, mode3d=0, fbufptr, zpat;
float *fbuf3d;
RDB_SLICE_INFO_ENTRY datab[RDB_MAX_SLICES];
float rotmat[3][2];	/* gradient rotation matrix  */
float B0;		/* field, in Guass  */
int domaxwell = 0;	/* do maxwell field correction if 1 */
FILE *fin;
char hfile[120], pfile[120], outfile[120];

float Gamma = 2*M_PI*4.257e3;

/* Prep parameters  WML */
int loopmode,nprep,tag,q,casl,vti,vti_flat,vti_ramp,fixedtr;
int prepti,preptilast,prepti1,prepti1s,psatn,qsatn;
float tagoffset,tagthick;
int pwafptot, froffset;
long mpsr1,mpsr2,mpstg,mpsfreq,apsr1,apsr2,apstg,apsfreq;
int printflag = 2;
int subkn = 0;
int outfloat = 0;
float outscale = 1.;
short int subkbuf[MXSAMP][2]; 	/*  restore k-space data for subtraction  */
void storesubk();

main(argc,argv)
int argc;
char **argv;

{
  int open_files();
  void spiral2d(), spiral3d();

  if(!open_files(argv, argc)) exit (0); /* header read in here */
  
  switch (recotype) {
   case 1: 
     printf("Starting 2d spiral recon \n");
     spiral2d();
     break;
   case 2:
     printf("Starting 3d spiral recon \n");
     spiral3d();
     break;
   default:
     printf("\nSorry, I don't recognize this data file \n");
  }
  
  printf("\nDone!!\n");
}

void spiral2d()
{
  int i,j,k;
  int goodnav = 0;

  if (!(fin=fopen(pfile,"r"))) {
    fprintf(stderr,"Can't open data file %s!\n",pfile);
    exit(0);
  }

  init();                                /* start off some things  */
  for (isl=isl1; isl<isl2; isl++) {
    offset = isl*nl*4*rhfrsize*rhnecho;
    fseek(fin, offset, 0);		/* place pointer at first data view  */
    if(domaxwell) maxwell();		/* init maxwell correction */		
    if(navc)  {
      goodnav = navcor(-1);		 /*  get the reference navigator  */
      fseek(fin, offset, 0);
    }
    if(bmapflag)  {		/* calculate field map and redo k traj */
      b0map();
      fseek(fin, offset, 0);
    }
    if(bmapflag < 2)  {		/* do normal recon  */
      for (iech=0; iech<necho; iech++) {
        if (printflag > 0) printf("Slice %d    echo %d\n", isl+1, iech+1);
        if (subkn) storesubk(iech);
	for (ifr=0; ifr<nf2do; ifr++) {
          froffset =  offset+iech*nl*4*rhfrsize
                     + ifr*nslc*nl*4*rhfrsize*rhnecho;
	  fseek(fin, froffset, 0);
          load_ft(goodnav);                     /* grid   */
	  ft2d();                                   
	  write_image(outfile, iech+1, isl+1, ifr+1, imgtyp);  
          if (printflag > 1) printf("\n");
        }  /*  end frames  */
      }    /*  end echoes  */
    }	   /*  end if bmap */
  }	 /*  end slices  */

  close(fin);

}

void spiral3d()
{
  int i,j,k,l,m,n,ii,jj,kk,ll,mm,zshots,nz,nz2,iz,iif;
  short int *ibuf3d;
  int goodnav = 0;

  if (!(fin=fopen(pfile,"r"))) {
    fprintf(stderr,"Can't open data file %s!\n",pfile);
    exit(0);
  }

  mode3d = 1;
  necho = 1;
  zshots = zres/etl;
  i = 2*etl*(reps*nl*zshots+1)*rhfrsize;
  ibuf3d = (short int *)calloc(2*i,sizeof(short int));
  fseek(fin, RDB_HEADER_SIZE_BYTES, 0); /* skip header */
  if ((fread(ibuf3d,2,i,fin))!=i)  {      /* whole enchilada */
    fprintf(stderr,"Short of 3D data!\n");
    exit(0);
  }
  for (nz=1;nz<zres;nz*=2);
  nz2 = nz >> 1;
  i = reps*nl*nz*rhfrsize; /* omit baseline views, make nz 2^n */
  fbuf3d = (float *)calloc(2*i,sizeof(float));
  for (i=0;i<reps/nprep;i++) {
    ii = i*nl*zshots*nprep + 1;
    for (j=0;j<nl;j++) {
      jj = ii + j*zshots*nprep;
      for (k=0;k<zshots;k++) {
        kk = jj + k*nprep;
	if (zpat<2) {             /* Centric encoding */
	  iz = nz2 + ((k+1)/2) * (1-2*(k%2));
	  zstep = zshots/2 * (1-2*(k%2));
	}
	else {                    /* Linear encoding */
	  iz = k;
	  zstep = zshots;
	}
        for (l=0;l<nprep;l++) {
	  ll = kk + l;
	  for (m=0;m<etl;m++) {
	    mm = (ll + m*(nl+1)) * 2 * rhfrsize;
	    iif = (((i*nprep+l)*nz + iz+m*zstep)*nl + j) * 2 * rhfrsize;
	    for (n=0;n<2*rhfrsize;n++) {
	      fbuf3d[iif+n] = ibuf3d[mm+n];
	    }
          }
	}
      }
    }
  }
  free(ibuf3d);
  
  /* Z Fourier Transform */
  for (i=0;i<reps;i++) {
    ii = i*nz*nl*rhfrsize*2;
    for (j=0;j<nl*rhfrsize;j++) {
      jj = ii + j*2;
      fftc(&fbuf3d[jj], nz, 2*nl*rhfrsize, 1);
    }
  }
  
  init();                       /* initialize spiral gridding stuff  */
  if (isl2==nslc) isl2 = nz;    /* Expand nslc to nz for 3D */
  for (isl=isl1; isl<isl2; isl++) {
    if (printflag > 0) printf("\nSlice %d\n", isl+1);
    fbufptr = 0;		/* at first data view  */
    if(navc)  goodnav = navcor(-1);	 /*  get the reference navigator  */
    for (ifr=0; ifr<nf2do; ifr++) {
      fbufptr = (ifr*nz + isl) * 2*nl*rhfrsize;
      load_ft(goodnav);                     /* grid   */
      ft2d();                                   
      write_image(outfile, 1, isl+1, ifr+1, imgtyp);  
    }  /*  end frames  */
  }	 /*  end slices  */

  close(fin);
}

void parsearg(argv, argc, read)
int argc, read;
char **argv;
{
  int nopt, i;
/* parse arguments */
	nopt = 0;
	for (i = 1; i < argc; i++) { /* ------- Options ------- */

		if (!strncmp(argv[i], "-sl1", 4)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) isl1 = atoi(argv[i])-1;
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-sl2", 4)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) isl2 = atoi(argv[i])-1;
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-reps", 5)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) nf2do = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-b0map", 6)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) bmapflag = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-foffset", 8)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) foffin = atof(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-navc", 5)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) navc = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-imgtype", 8)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) imgtyp = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-npix", 5)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) npix = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-zoomf", 6)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) zoomf = 1.0/atof(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-maxwell", 8)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) domaxwell = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-print", 6)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) printflag = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-subk", 5)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) subkn = atoi(argv[i]);
			nopt++; nopt++;
		}
		if (!strncmp(argv[i], "-float", 6)) {
			if (+i >= argc) { instructions(); exit(0); }
			if (read) outfloat = 1;
			nopt++;
		}
		if (!strncmp(argv[i], "-scale", 6)) {
			if (++i >= argc) { instructions(); exit(0); }
			if (read) outscale = atof(argv[i]);
			nopt++; nopt++;
		}		
	} /* end of options */
	if (argc-nopt < 3){
		instructions(argv);
		exit(0);
	}    return;
}

void instructions(argv)
char **argv;
{
    printf("\nGeneral reconstruction for spep spiral data.");
    printf("\nUsage: %s [options] raw_file out_file\n", argv[0]);
    printf("\nOptions: -sl1 n [1]        Start slice to reconstruct\n");
    printf("         -sl2 n [all]      End slice to reconstruct\n");
    printf("         -reps n [all]     Number of repetitions to reconstruct\n");
    printf("         -b0map n [header] 1=shim correction 0=no\n");
    printf("         -foffset n [0]    Freq_offset(Hz) [overwritten if b0map=1]\n");
    printf("         -navc n [1]       1=navigator correction 0=no\n");
    printf("         -imgtype n [1]    Bit map: 1=mag, 2=phase, 4=real(I), 8=imag(Q)\n");
    printf("         -npix n [header]  Reconstruction size [power of 2]\n");
    printf("         -zoomf n [1]      Magnification factor\n");
    printf("         -maxwell n [1]    1=maxwell field correction 0=no\n");
    printf("         -print n [2]      2=all 1=header info 0=none\n");
    printf("         -subk n [0]       Subtract #n reps in k-space 0=none\n");
    printf("         -float            Output data in float format\n");
    printf("         -scale n [1.]     Output data scale factor for imgtype:mag,real,imag\n");
    printf("This is source rev %s\n", REV);
    return;
}


int open_files(argv, argc)
int argc;
char **argv;

{
  char fname[80];
  int i;
  char line[512];
  float tmp;

  parsearg(argv, argc, 0);
  strcpy(pfile, argv[argc-3]);		/* save for later  */
  strcpy(hfile, argv[argc-2]);		/* save for later  */
  strcpy(outfile, argv[argc-1]);	/* save for later  */

  if (!(fin=fopen(hfile,"r"))) {
    fprintf(stderr,"Can't open header %s!\n",hfile);
    return(0);
  }

/*      read raw hdr            */

  /* nslc (number of slices, [ns]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  nslc=(int)tmp;

  /* necho (number of echoes per spin excitation, [ne]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  necho=(int)tmp;
  rhnecho=necho;

  /* nsamp (number of samples per echo, [np/2]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  nsamp=(int)tmp;
  rhfrsize=nsamp;

  /* nl (number of interleaves, [n_ints]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  nl=(int)tmp;
  nbpi = nl*nsamp*4;		/* num bytes per image  */

  /* reps (number of time points, [n_t_pts]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  reps=(int)tmp;

  /* rawimsize (desired reconstruction matrix size, [spmatrix]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  rawimsize=(int)tmp;

  /* imrot (image rotation value) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  imrot=(int)tmp;

  /* imtran (image transpose flag) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  imtran=(int)tmp;

  /* xoff (image x offset, pixels, [pro/lro*spmatrix) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  xoff=tmp;

  /* yoff (image y offset, pixels, [ppe/lro*spmatrix) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  yoff=tmp;

  /* dts (sampling interval, [1000000.0/sw]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  dts=tmp;

  /* bmapflag (flag for when B0 map information collected) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  bmapflag=(int)tmp;
  if (bmapflag>2) bmapflag = 0; /* ECW temporary kludge for VTE */	

  /* mapdt (delta_te if B0 map collected [delta_te]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  mapdt=tmp;

  /* gfov (gradient field of view, [lro]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  gfov=tmp;

  /* N (gradient design matrix size, same as used in genspiral [spmatrix]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  N=(int)tmp;

  /* recotype (reconstruction type, 1=2dspiral, 2=3dspiral) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  recotype=(int)tmp;

  /* gamp (maximum gradient strength, [gmax]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  gamp=tmp;

  /* gslew (maximum gradient slew rate, mT/m/ms, [gmax/(100.0*grise)]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  gslew=tmp;

  /* bline (number of basline "noise" views) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  bline=(int)tmp;

  /* frcvr (receiver center frequency, Hz) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  frcvr=tmp;

  /* etl (echo train length for 3D FSE) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  etl=(int)tmp;

  /* zres (number of Z phase encode steps for 3D FSE) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  zres=(int)tmp;

  /* zpat (Z phase encode method, 1=centric, 0=linear, for 3D FSE) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  zpat=(int)tmp;

  isl1 = 0;
  isl2 = nslc;
  if (recotype == 1 || recotype == 2) gtype = 0.;

  /* B0 (field strength in Gauss [sfrq/42.575*10000.0]) */
  fgets(line,512,fin);
  sscanf(line,"%f",&tmp);
  B0=tmp;

/* PJPJ */
printf("nsamp=%d\n",nsamp);
printf("necho=%d\n",necho);
printf("nl=%d\n",nl);
printf("rawimsize=%d\n",rawimsize);
printf("imrot=%d\n",imrot);
printf("imtran=%d\n",imtran);
printf("dts=%f\n",dts);
printf("bmapflag=%d\n",bmapflag);
printf("gfov=%f\n",gfov);
printf("N=%d\n",N);
printf("gamp=%f\n",gamp);
printf("gslew=%f\n",gslew);
printf("bline=%d\n",bline);
printf("frcvr=%f\n",frcvr);

/* Reading Prep parameters  WML */
  /* just setting these to sensible values for now (PJ). Don't
     really know what they all do */
  loopmode = 1;
  nprep = 1;
  if (nprep > reps) nprep = reps;
  tag = 0;
  q = 0;
  casl = 0;
  vti = 0;
  vti_flat = 1;
  vti_ramp = 1;
  fixedtr = 1;
  prepti = 1400000;
  preptilast = 1600000;
  prepti1 = 700000;
  prepti1s = 1050000;
  psatn = 2;
  qsatn = 2;
  tagoffset = -64.0;
  tagthick = 100.0;
  pwafptot = 1200000;
  mpsr1 = 11;
  mpsr2 = 15;
  mpstg = 66;
  mpsfreq = 638704000;
  apsr1 = 11;
  apsr2 = 15;
  apstg = 66;
  apsfreq = 638704000;

  foffin = 0.0;			/* freq offset */
  nf2do = reps;		/* num time frames to do */
  npix = rawimsize; 

  parsearg(argv, argc, 1);

  foff = foffin;
  if(reps == 1 || nl == 1) navc = 0;	/* need more than 1 time frame & view  */
  
  ngrid = npix*2;
  npixo2 = npix/2;

/*  get values for maxwell field correction  */

  /* values here are those from Gary Glover's scanner (PJ) */
  rotmat[0][0] =0.0; rotmat[0][1] = 1.0;
  rotmat[1][0] =-1.0; rotmat[1][1] = 0.0;
  rotmat[2][0] =0.0; rotmat[2][1] = 0.0;

/*   memcpy(&datab,header+RDB_DATAACQ_OFF,RDB_DATAACQ_SIZE); */

  if (printflag > 0) {
    fprintf(stdout, "\n nsamp = %d    nl = %d\n", nsamp, nl);
    fprintf(stdout, " nslices = %d   nechoes = %d   reps = %d\n", nslc,
    	necho, reps);
    fprintf(stdout, " mpsR1=%d mpsR2=%d mpsTG=%d mpsFREQ=%d\n", mpsr1, mpsr2,
  	mpstg, mpsfreq);
    fprintf(stdout, " apsR1=%d apsR2=%d apsTG=%d apsFREQ=%d\n", apsr1, apsr2,
  	apstg, apsfreq);
    if (tag) {
      fprintf(stdout, " loopmode=%i nprep=%i tag=%i q=%i casl=%i\n",loopmode,
    	nprep,tag,q,casl);
      fprintf(stdout, " psatn=%i tagoffset=%.2fmm tagthick=%.2fmm\n",psatn,
    	tagoffset,tagthick);
      if (q) fprintf(stdout, " qsatn=%i prepti1=%ims\n",qsatn,prepti1/1000);
      if (q==3) fprintf(stdout, " prepti1s=%ims\n",prepti1s/1000);
      if (casl) fprintf(stdout, " pwafptot=%ims\n",pwafptot/1000);
      fprintf(stdout, " prepti=%ims\n",prepti/1000);
      if (vti) {
       fprintf(stdout, " vti=%i vti_flat=%i vti_ramp=%i fixedtr=%i\n",vti,
     	 vti_flat,vti_ramp,fixedtr);
       fprintf(stdout, " preptilast=%ims\n",preptilast/1000);
      }
    }
  }

  close(fin);

  return(1);
}

void storesubk(iecho)
int iecho;
{
  int i;
  
  if (loopmode == 1) froffset =  offset+iecho*(nl+1)*4*rhfrsize
	  		+ ((subkn-1)/nprep)*nbpi*nprep
			+ ((subkn-1)%nprep)*4*rhfrsize; 
  else froffset =  offset+iecho*(nl+1)*4*rhfrsize+(subkn-1)*nbpi;
  fseek(fin, froffset, 0);
  for (i=0; i<nl; i++) {
    if ((fread(subkbuf,4,rhfrsize,fin))!=rhfrsize)  {  /* just one acq */
      fprintf(stderr,"Short of data!\n");
      exit(0);
    }
    if (loopmode == 1) fseek(fin, 4*rhfrsize*(nprep-1), SEEK_CUR); /* WML */
  }
}

void init()                          /* calc trajectory, other stuff */
{
  int i, j, k;
  float z, p;
  float c, s, theta;

/*  calc trajectory   */

  if (!genspiral(gfov, N, 1.0, dts*1.e-6)) {
    printf("genspiral failed!\n");
    exit(0);
  }
  calcden(nsamp, Kx, Ky, Dens);
  for (i=0; i<nl; i++) {
    theta = i*2*M_PI/nl;
    c = cos(theta);  s = sin(theta);
    for (j=0; j<nsamp; j++) {
      kx0[i][j] = c*Kx[j] - s*Ky[j];
      ky0[i][j] = s*Kx[j] + c*Ky[j];
      dens[i][j] = Dens[j];
    }
  }

/*  calculate gridding function  */

  for (i=0;i<=dwin;i++) {
    z = i*.5/dwin;
    gf[i]=bessi0(b*sqrt(1.0-z*z))/bessi0(b);
  }

/*  calculate inverse gridding function  */

  for (i=0;i<npix;i++) {
    z=(i-npix/2+.5)/npix*1.5*dwin*M_PI;
    z=z*z-b*b;
    igf[i]= (z>0)? b*sin(sqrt(z))/sqrt(z)/sinh(b) : b*sinh(sqrt(-z))/sqrt(-z)/sinh(b);
  }

/*  set up trajectory */

 set_traj();

/*  make zit remover fft buffer  */

  if(frcvr != 0)  {
    for (j=0; j<nsamp; j++) {
      theta = j*2*M_PI/nsamp;
      cosbf[j] = cos((double)theta);
    }
  }

}

void load_ft(donav)                                 /* grid the data */
int donav;
{
  int i, j, m,  lx, ly;
  float w, dkx, dky;
  float x, y, tr, ti, cz, sz;
  double z;

  if (printflag > 1) { printf("Gridding .. "); fflush(stdout); }

/* null out the data matrix  */

  for (i=0; i<ngrid; i++)
    for (j=0; j<ngrid; j++)  {
      raw[i][j][0] = 0.0;
      raw[i][j][1] = 0.0;
      ws[i][j] = .001;
    }

/*  do the gridding  */

  for (i=0; i<nl; i++) {
    load_view();                       	/* load a view    */
    if (loopmode == 1 && !mode3d) fseek(fin, 4*rhfrsize*(nprep-1), SEEK_CUR); /* WML */
    if(donav) navcor(i);		/* do nav correction */
    for (j=0; j<nsamp; j++) {
      dkx = kx[i][j]; 
      dky = ky[i][j]; 
      for (lx = ceil(dkx-dwin); lx<=floor(dkx+dwin); lx++) {
        if ((lx<0) || (lx>=ngrid)) break;        
        for (ly = ceil(dky-dwin); ly<=floor(dky+dwin); ly++) {
          if ((ly<0) || (ly>=ngrid)) break;
	  w = dens[i][j]*gf[(int) rint(fabs(dkx-lx))] * gf[(int) rint(fabs(dky-ly))];
          raw[lx][ly][0] += dbuf[j][0] * w;         /* insert data */
          raw[lx][ly][1] += dbuf[j][1] * w;
          ws[lx][ly] += w;              /* update weighting */
	}
      }
    }
  }
  for (i=0; i<ngrid; i++)                /* normalize weighting */
    for (j=0; j<ngrid; j++) {
      raw[i][j][0] /= ws[i][j];
      raw[i][j][1] /= ws[i][j];
    }

/*  offset the center of fov  */
 
  if ((xoff!=0) || (yoff!=0)) {
    x = M_PI*xoff/rawimsize;		/* (2*pi/ngrid)*(off*npix/rawimsize) */
    y = -M_PI*yoff/rawimsize;
    for (i=0;i<ngrid;i++) 
      for (j=0;j<ngrid;j++) {
        z = i*x + j*y;
	cz = cos(z);  sz = sin(z);
        tr = raw[i][j][0];
        ti = raw[i][j][1];
        raw[i][j][0] = tr*cz - ti*sz;
        raw[i][j][1] = tr*sz + ti*cz;
      }
  }
}

void ft2d()             /* 2D ft and inverse grid weighting  */
{
  int i, j, i1, fftx, ffty;
  float wi;

  if (printflag > 1) { printf("2DFT .. "); fflush(stdout); }
  
/*  do the fft  */

  fftx = 1;
  ffty = (imtran)? 1 : -1;		/* mirror */
  for(i=0; i<ngrid; i++) 
    fftc(&raw[i][0][0],ngrid,2,fftx);
  for(i=npixo2; i<3*npixo2; i++) 	/* only do center */
    fftc(&raw[0][i][0],ngrid,2*IMD,ffty);

/*  remove the gridding filter and reduce to npix  */

  for (i=0; i<npix; i++)  {
    i1 = i+npixo2;
    for (j=0; j<npix; j++) {
      wi = 1.0/(igf[i]*igf[j]);
      img[i][j][0] = raw[i1][j+npixo2][0]*wi;
      img[i][j][1] = raw[i1][j+npixo2][1]*wi;
    }
  }
}

write_image(fnm, iech, isl, imnum, mode)
char *fnm;
int iech, isl, imnum, mode;
{
  int i, j, imx, isize, fsize;
  float imi, imr, imm, *fm, fmx;
  short int *bm;
  char fname[80];
  double dmr, dmi;
  FILE *fout;

  if (printflag > 1) {
    if (necho>1) printf("Writing image %s.%02d.%02d.%04d .. ", outfile,
	    			iech, isl, imnum);
    else printf("Writing image %s.%02d.%04d .. ", outfile, isl, imnum);
    fflush(stdout);
  }
  fsize = sizeof(float);
  isize = sizeof(short int);
  if (outfloat) fm = (float *)malloc((unsigned) npix * fsize);
  else bm = (short int *)malloc((unsigned) npix * isize);
  fmx = 0.;
  imx = 0;
  if(mode & MAG) {	
    if (necho>1) sprintf(fname,"%s.%02d.%02d.%04d",fnm, iech, isl, imnum);
    else if (mode3d) sprintf(fname,"%s.%03d.%04d",fnm, isl, imnum);
    else sprintf(fname,"%s.%02d.%04d",fnm, isl, imnum);
    if (!(fout=fopen(fname,"w"))) {
      fprintf(stderr,"Can't open %s!\n",fname);
      exit(0);
    }
    for (i=0; i<npix; i++) {
      for (j=0; j<npix; j++) {
	if(imrot==3) {
          dmr = img[j][i][0];
          dmi = img[j][i][1];          
	}
	else  {
          dmr = img[npix-i][j][0];
          dmi = img[npix-i][j][1];          
	}
        if (outfloat) {
	  fm[j] = outscale * hypot(dmr, dmi);
          if (fm[j]>fmx) fmx = fm[j];
	}
	else {
	  bm[j] = outscale * hypot(dmr, dmi);
          if (bm[j]>imx) imx = bm[j];
	}
      }
      if (outfloat) {
	if(fwrite(fm,fsize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");    
	  exit(0);
	}
      }
      else {
	if(fwrite(bm,isize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");    
	  exit(0);
        }
      }
    }
    if (printflag > 1) {
      if (outfloat) printf("max = %5.1f", fmx);
      else printf("max = %3d", imx);
    }
    fclose(fout);
  }

  if(mode & PHASE) {	
    if (necho>1) sprintf(fname,"%s_p.%02d.%02d.%04d",fnm, iech, isl, imnum);
    else if (mode3d) sprintf(fname,"%s.%03d.%04dp",fnm, isl, imnum);
    else sprintf(fname,"%s.%02d.%04dp",fnm, isl, imnum);
    if (!(fout=fopen(fname,"w"))) {
      fprintf(stderr,"Can't open %s!\n",fname);
      exit(0);
    }
    for (i=0; i<npix; i++) {
      for (j=0; j<npix; j++) {
	if(imrot==3) {
          dmr = img[j][i][0];
          dmi = img[j][i][1];          
	}
	else  {
          dmr = img[npix-i][j][0];
          dmi = img[npix-i][j][1];          
	}
        if (outfloat) fm[j] = (dmr != 0)? 1000.*atan2(dmi, dmr) : 
	  (dmi >= 0.)? 3142. : -3142.;
	else bm[j] = (dmr != 0)? 1000*atan2(dmi, dmr) :
	  (dmi >= 0)? 3142 : -3142;
      }
      if (outfloat) {
	if(fwrite(fm,fsize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");    
	  exit(0);
	}
      }
      else {
	if(fwrite(bm,isize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");
	  exit(0);
        }
      }
    }
    fclose(fout);
  }

  if(mode & REAL) {	
    if(mode & NOSUBSC) {
      if (necho>1) sprintf(fname,"%s.%02d.%02d.%04d",fnm, iech, isl, imnum);
      else if (mode3d) sprintf(fname,"%s.%03d.%04d",fnm, isl, imnum);
      else sprintf(fname,"%s.%02d.%04d",fnm, isl, imnum);
    }
    else {
      if (necho>1) sprintf(fname,"%s_i.%02d.%02d.%04d",fnm, iech, isl, imnum);
      else if (mode3d) sprintf(fname,"%s_i.%03d.%04d",fnm, isl, imnum);
      else sprintf(fname,"%s_i.%02d.%04d",fnm, isl, imnum);
    }
    if (!(fout=fopen(fname,"w"))) {
      fprintf(stderr,"Can't open %s!\n",fname);
      exit(0);
    }
    for (i=0; i<npix; i++) {
      for (j=0; j<npix; j++) {
	if(imrot==3) {
          if (outfloat) fm[j] = outscale * img[j][i][0];
	  else bm[j] = outscale * img[j][i][0];
	}
	else  {
          if (outfloat) fm[j] = outscale * img[npix-i][j][0];
	  else bm[j] = outscale * img[npix-i][j][0];
	}
      }
      if (outfloat) {
	if(fwrite(fm,fsize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");    
	  exit(0);
	}
      }
      else {
	if(fwrite(bm,isize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");
	  exit(0);
        }
      }
    }
    fclose(fout);
  }

  if(mode & IMAG) {	
    if (necho>1) sprintf(fname,"%s_q.%02d.%02d.%04d",fnm, iech, isl, imnum);
    else if (mode3d) sprintf(fname,"%s_q.%03d.%04d",fnm, isl, imnum);
    else sprintf(fname,"%s_q.%02d.%04d",fnm, isl, imnum);
    if (!(fout=fopen(fname,"w"))) {
      fprintf(stderr,"Can't open %s!\n",fname);
      exit(0);
    }
    for (i=0; i<npix; i++) {
      for (j=0; j<npix; j++) {
	if(imrot==3) {
          if (outfloat) fm[j] = outscale * img[j][i][1];
	  else bm[j] = outscale * img[j][i][1];
	}
	else  {
          if (outfloat) fm[j] = outscale * img[npix-i][j][1];
	  else bm[j] = outscale * img[npix-i][j][1];
	}
      }
      if (outfloat) {
	if(fwrite(fm,fsize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");    
	  exit(0);
	}
      }
      else {
	if(fwrite(bm,isize,npix,fout) != npix)  {
	  fprintf(stderr,"Can't write file!\n");
	  exit(0);
        }
      }
    }
    fclose(fout);
  }
  
  return(1);
}

void set_traj()
{
  float g, p, scale;
  int i, j, k;
  float krad[MXSAMP];
  float fermi;
  double phi;

/*  set k-space values, using linear gradient shift  */

  scale = dts*.000001*npix/4;
  fermir = 0.;
  for (i=0;i<nl;i++) 
    for (j=0;j<nsamp;j++) {
      kx[i][j] = ngrid/2+kx0[i][j]*zoomf - wx*scale*j;
      ky[i][j] = ngrid/2-ky0[i][j]*zoomf - wy*scale*j;
      krad[j] = zoomf*hypot((double)kx0[i][j], (double)ky0[i][j]);
      if(krad[j] > fermir) fermir = krad[j];
    }

/*  calc freq offset phase shift and mult by fermi window  */

  fermiw = FERMIW*fermir;
  g = gain/64.;		
  p = -2*M_PI*(foff+frcvr)*dts*.000001;
  for (i=0; i<nsamp; i++) {
    fermi = g/(1 + exp((double)((krad[i] - fermir)/fermiw)));
    phi = p*i + domaxwell*phim[i];
    pbuf[i][0] = fermi*cos(phi);
    pbuf[i][1] = fermi*sin(phi);
  }

}

void load_view()                            /* load and float a view of data */
{
  int i, j, k, n;
  int datr, dati;
  float c, s;
  float sum0[2], sum1[2], corr[2];

  if (mode3d) {
    for (i=0;i<nsamp;i++) {
      fbuf[i][0] = fbuf3d[fbufptr++];
      fbuf[i][1] = fbuf3d[fbufptr++];
    }
  }
  else {
    if ((fread(ibuf,4,rhfrsize,fin))!=rhfrsize)  {  /* just one acq */
      fprintf(stderr,"Short of data!\n");
      exit(0);
    }
    if (subkn) {
      for (j=0; j<nsamp; j++) {
        ibuf[j][0] -= subkbuf[j][0];
        ibuf[j][1] -= subkbuf[j][1];      
      }
    }
    for (i=0; i<nsamp; i++) {
      fbuf[i][0] = ibuf[i][0];
      fbuf[i][1] = ibuf[i][1];
    }

  } /* 2D mode */

  if(frcvr != 0)  {		/* replace dc term with f1+f-1  for zit */
    sum0[0] = sum0[1] = sum1[0] = sum1[1] = 0;
    for (j=0; j<nsamp; j++) {
      datr = fbuf[j][0];
      dati = fbuf[j][1];
      c = cosbf[j];
      sum0[0] += datr;
      sum0[1] += dati;
      sum1[0] += datr*c;
      sum1[1] += dati*c;
    }
    corr[0] = (sum0[0] - sum1[0])/nsamp;
    corr[1] = (sum0[1] - sum1[1])/nsamp;
  }
  else  {
    corr[0] = corr[1] = 0;
  }
  for (j=0; j<nsamp; j++) {
    datr = fbuf[j][0] - corr[0];
    dati = fbuf[j][1] - corr[1];
    dbuf[j][0] = datr*pbuf[j][0] - dati*pbuf[j][1];
    dbuf[j][1] = datr*pbuf[j][1] + dati*pbuf[j][0];
  }
}

b0map()		/* recon first two images, calc field, fit plane */
{
  int i, j, k;
  float img1[MXNIM][MXNIM][2], img2[MXNIM][MXNIM][2]; 
  float dmi, dmr, dmr1, dmi1, dmr2, dmi2, wt;
  float x, y;
  double sxx, syy, sxy, sx, sy, sxw, syw, sw, snn, den;
  float scale, fscale;
  char str[80];

  if (printflag > 1) {
    printf("Calculating Bo map .. "); fflush(stdout);
  }
  scale = 1.0/(mapdt*.000001);		/* freq in radians/s  */
  fscale = scale/(2*M_PI);

/*  make the complex images */

  foff = wx = wy = 0;
  set_traj();		/* reset to no correction  */
  load_ft(0);		/* read data and grid  */
  ft2d();		/*  do fft */
  for (i=0; i<npix; i++)
    for (j=0; j<npix; j++)  {
      img1[i][j][0] = img[i][j][0];
      img1[i][j][1] = img[i][j][1];
    }
  if ((loopmode == 1) && (nprep > 1) && !mode3d)
  	fseek(fin, -4*rhfrsize*2*(nprep-1), SEEK_CUR); /* WML */
  load_ft(0);		/* read data and grid  */
  ft2d();		/*  do fft */
  for (i=0; i<npix; i++)
    for (j=0; j<npix; j++)  {
      dmr2 = img[i][j][0];
      dmi2 = img[i][j][1];
      dmr1 = img1[i][j][0];
      dmi1 = img1[i][j][1];
      dmr = dmr1*dmr2 + dmi1*dmi2;
      dmi = dmi1*dmr2 - dmr1*dmi2;
      img2[i][j][0] = hypot(dmi2, dmr2);
      img2[i][j][1] = (dmr != 0)? atan2(dmi, dmr) : (dmi >= 0)? M_PI : -M_PI;
      img[i][j][0] = fscale*img2[i][j][1];
    }
  if(bmapflag==2)  {	/* give P.xxxxx.B0, also save the mag images */
    sprintf(str, "%s.B0", outfile);
    write_image(str, 1, 1, isl+1, REAL+NOSUBSC);
    for (i=0; i<npix; i++)
      for (j=0; j<npix; j++)  
        img[i][j][0] = img2[i][j][0];
    write_image(outfile, 1, 1, isl+1, REAL+NOSUBSC);
  }
  else 
    write_image("B0", 1, 1, isl+1, REAL+NOSUBSC);

/*  fit a plane w = wx*x + wy*y + w0  */

  sxx = sxy = syy = sx = sy = sxw = syw = sw = snn = 0;
  for (i=0; i<npix; i++)  {
    x = i-npixo2+.5;
    for (j=0; j<npix; j++)  {
      y = j-npixo2+.5;
      dmr = img2[i][j][0];
      wt = dmr*dmr;
      dmi = img2[i][j][1];
      sx += x*wt;
      sy += y*wt;
      sxx += x*x*wt;
      syy += y*y*wt;
      sxy += x*y*wt;
      sxw += x*dmi*wt;
      syw += y*dmi*wt;
      sw += dmi*wt;
      snn += wt;
    }
  }
  den = sxx*syy*snn + 2*sxy*sx*sy - sx*sx*syy - sy*sy*sxx - sxy*sxy*snn;
  wx = (sxw*syy*snn + sxy*sy*sw + sx*syw*sy - sw*syy*sx - sy*sy*sxw - syw*sxy*snn)/den;
  wy = (sxx*syw*snn + sxw*sy*sx + sx*sxy*sw - sx*syw*sx - sw*sy*sxx - sxy*sxw*snn)/den;
  w0 = (sxx*syy*sw + sxy*syw*sx + sxw*sxy*sy - sx*syy*sxw - sy*syw*sxx - sw*sxy*sxy)/den;

  wx *= ((imtran)? 1 : -1)*scale;  	/* compensate for mirror in fft */
  wy *= scale;
  foff = foffin + w0*fscale;
  if (printflag > 1)
    printf("foff = %5.1f Hz  wx, wy = %4.1f %4.1f rad/s/pix\n", foff, wx, wy);
 
  set_traj();		/* new foff, k-space traj  */

  return(1);
}

/*  navcor takes the first several points of the second frame 
	of data and averages them to make a phase correction 
	reference.  Then each view is similarly sampled to derive
	a complex multiplier for the view.  This is a poor man's
	navigator technique.
    */

int navcor(mode)
int mode;
{
#define NAVFR   2		/* reference time frame to use  */
#define NAVPTS 	2		/* num k-space points to average  */
#define NAVFRAC 0.25            /* k=0 data must be >= this fract. of max */
  int i, j, k;
  float datr, dati;
  static float refr[MXNVW], refi[MXNVW];
  static int refpt[MXNVW];
  double corr, cori;
  float mag;
  float max, dat;
  float refr0, refi0;

  if(mode == -1)  {		/* generate the reference  */

    if (printflag > 1) printf("Generating navcor ref ");
    if (!mode3d) fseek(fin, (NAVFR-1)*nbpi, 1); /* move to this frame  */ 
    refr0 = refi0 = 0;
    for (i=0; i<nl; i++)  {
      max = refr[i] = refi[i] = 0;
      load_view();
      for (j=0; j<nsamp; j++) {
        dat = hypot((double)dbuf[j][0], (double)dbuf[j][1]);
	if(dat>max) {
	  max = dat;
	  refpt[i] = j;
        }
      }
      for (j=0; j<NAVPTS; j++) {
        refr0 += dbuf[j][0];
        refi0 += dbuf[j][1];
      }  
      k = refpt[i];
      for (j=k; j<k+NAVPTS; j++) {
        refr[i] += dbuf[j][0];
        refi[i] += dbuf[j][1];
      }
    }

/* now check to see if the origin data is ok to use  */

    dat = hypot((double)refr0, (double)refi0);
    if(dat>max*NAVPTS*nl*NAVFRAC)  {
      for (i=0; i<nl; i++)  {	/* yup, store it  */
        refr[i] = refr0;		
        refi[i] = refi0;
        refpt[i] = 0;
      }
      if (printflag > 1) printf("0\n");
    } 
    else
      if (printflag > 1) printf(">\n");
    return(1);		
  }
  else  {			/*  do the correction; mode = view  */

    k = refpt[mode];		/* reference point for this view */
    datr = dati = 0;
    for (j=k; j<k+NAVPTS; j++) {
      datr += dbuf[j][0];
      dati += dbuf[j][1];
    }
    corr = (refr[mode]*datr + refi[mode]*dati);		/*note conj  */
    cori = (refi[mode]*datr - refr[mode]*dati);
    mag = hypot(corr, cori);
    corr = corr/mag;
    cori = cori/mag;
    for (j=0; j<nsamp; j++) {
      datr = dbuf[j][0];
      dati = dbuf[j][1];
      dbuf[j][0] = datr*corr - dati*cori;
      dbuf[j][1] = datr*cori + dati*corr;
    }
    return(2);

  }

}


/*   multi- shot spiral design
*    uses Duyn's approximate slewrate limited design
*    augmented with archimedian gmax limit
*    inputs (args)
*        D = FOV, cm
*        N = matrix size
*	 Tmax = longest acquisition allowed, s
*	 dts = output sample spacing, s
*        gtype = trajectory type
*    inputs (CVs)
*        nl = number of interleaves
*        gamp = design grad max, G/cm
*        gslew = design slew rate, mT/m/ms
*	 nramp = number of rampdown points
*    outputs
*        Gx, Gy
*        grev
*    time is in sec
*
*		rev 0 12/26/98	original
*               rev 1 4/15/99   little better calc of ts
*/

int genspiral(float D, int N, float Tmax, float dts)
{

  float dt;
  float S0;
  int i, j, n;
  float Ts, T, ts, t, theta, dthdt, thetas, x, y;
  float a1, a2, beta, Gmax, bx, by, c, s;
  float gabs, gmax;
  short gxi, gyi;
  float scl;
  float q;
#define GTYPE	0

  if(gtype != GTYPE)  {
    fprintf(stderr, "wrong trajectory type");
    return 0;
  }

  q = 5;                /* nice number  */
  S0 = gslew*100;
  dt = dts*.5;

/*  slew-rate limited approximation  */

  Ts = .666667/nl*sqrt(pow((double)(M_PI*N), (double)3.0)/(Gamma*D*S0));
  if(Ts > Tmax)  {
    fprintf(stderr,"slew limited readout too long");
    return 0;
  }
  a2 = N*M_PI/(nl*pow(Ts, .666667));
  a1 = 1.5*S0/a2;
  beta = S0*Gamma*D/nl;
  Gmax = a1*pow(Ts, .333333);
  gmax = 0;
  i = 0;
  for (t = 0; t<=Ts; t+=dt)  {
    x = pow(t, 1.333333);
    theta = .5*beta*t*t/(q + .5*beta/a2*x);
    y = q+.5*beta/a2*x;
    dthdt = beta*t*(q+.166667*beta/a2*x)/(y*y);
    c = cos(theta);
    s = sin(theta);
    gx[i] = nl/(D*Gamma)*dthdt*(c - theta*s);
    gy[i] = nl/(D*Gamma)*dthdt*(s + theta*c);
    gabs = hypot(gx[i], gy[i]);
    if(gabs>=gamp)  {
      if(gmax==0)  gmax = hypot(gamp/theta, gamp);
      if(gabs>gmax) break;
    }
    ts = t;
    thetas = theta;
    i++;
  }

/*  gmax limited approximation  */

  if(Gmax > gamp)  {
    T = ((M_PI*N/nl)*(M_PI*N/nl) - thetas*thetas)/(2*Gamma*gamp*D/nl) + ts;
    if(T > Tmax)  {
      fprintf(stderr,"gmax limited readout too long");
      return 0;
    }
    for (t=ts+dt; t<=T; t+=dt)  {
      theta = sqrt(thetas*thetas + 2*Gamma*gamp*D*(t-ts)/nl);
      c = cos(theta);
      s = sin(theta);
      gx[i] = gamp*(c/theta - s);
      gy[i++] = gamp*(s/theta + c);
    }
  }

/*  calc k and decimate by 2 to get back to proper sampling */

  scl = Gamma*gfov*dts/(2*M_PI);
  n = 0;
  Kx[n] = 0;
  Ky[n++] = 0;
  x = y = 0;
  for (j=0; j<i; j+=2)  {
    x += scl*(gx[j] + gx[j+1]);
    Kx[n] = x;
    y += scl*(gy[j] + gy[j+1]);
    Ky[n] = y;
    gx[n] = gx[j];
    gy[n++] = gy[j];
  }

  return 1;

}

void maxwell()		/* maxwell field correction */
{
  float p11,p12,p13,p21,p22,p23,p31,p32,p33;
  float a,b,c,d,r;
  float x0, y0, z0;	/* maxwell field center loc  */
  float gxl, gyl, gxp, gyp, gzp;
  float bmaxw, phi, bmaxwm;
  int i;

/* corner points of image */

  p11 = datab[isl].gw_point1[0]; p12 = datab[isl].gw_point1[1]; p13 = datab[isl].gw_point1[2];
  p21 = datab[isl].gw_point2[0]; p22 = datab[isl].gw_point2[1]; p23 = datab[isl].gw_point2[2];
  p31 = datab[isl].gw_point3[0]; p32 = datab[isl].gw_point3[1]; p33 = datab[isl].gw_point3[2];
/*  printf("gw_point1= %f  %f  %f\n", p11, p12, p13);
  printf("gw_point2= %f  %f  %f\n", p21, p22, p23);
  printf("gw_point3= %f  %f  %f\n", p31, p32, p33);  */

/* solve for gradient center */

  d = p11*p22*p33 + p12*p23*p31 + p13*p21*p32 -
          p31*p22*p13 - p32*p23*p11 - p33*p21*p12;
  if(fabs((double)d) < 0.01)  {
    x0 = y0 = z0 = 0;
  } 
  else  {
    a = (p22*p33 + p12*p23 + p13*p32 - p22*p13 - p32*p23 - p33*p12)/d;
    b = (p11*p33 + p23*p31 + p13*p21 - p31*p13 - p23*p11 - p33*p21)/d;
    c = (p11*p22 + p12*p31 + p21*p32 - p31*p22 - p32*p11 - p21*p12)/d;
    r = a*a + b*b + c*c;
    x0 = .1*a/r;        /* center of normal to image plane, cm */
    y0 = .1*b/r;	/* this is where the maxwell correction is */
    z0 = .1*c/r;
  }
  /* printf("loc of imaging ctr= %f  %f  %f\n", x0, y0, z0);  */

/* calc rotated grads and then the phase correction */

  phi = 0;
  bmaxwm = 0;
  for (i = 0; i<nsamp; i++)  {		/* each time point in traj  */
    gxl = gx[i]; gyl = gy[i];
    gxp = rotmat[0][0]*gxl + rotmat[0][1]*gyl;
    gyp = rotmat[1][0]*gxl + rotmat[1][1]*gyl;
    gzp = rotmat[2][0]*gxl + rotmat[2][1]*gyl;
    bmaxw = (.25*gzp*gzp*(x0*x0+y0*y0) + (gxp*gxp +gyp*gyp)*z0*z0
	- (gxp*x0+gyp*y0)*z0*gzp)/(2*B0);
    phi += bmaxw*Gamma*dts*1.e-6;
    phim[i] = phi;		/* time dependent phase correction */
    if(bmaxw>bmaxwm) bmaxwm = bmaxw;
  }
  /* printf("B0 (G), bmaxw (Hz) = %f  %f\n", B0, bmaxwm*4257.); */
 
  return;

}	/* end maxwell  */
