
#define GRAD_UPDATE_TIME 2.0    /* minimum gradient update time of 2us */
#define MAX_RAMP_RES 50         /* maximum desired resolution for a simple */
                                /* ramp */
#define MAX_TRAP_RAMP_RES 100   /* maximum desired resolution for the ramp */
                                /* portion of a trapezoid function */
#define WAVEFORM_RESOL 5.0e-2   /* best resolution of waveforms (us) */
#define MAX_GRAD_DURATION 2.0   /* maximum allowed gradient duration in sec */
#define MAX_DAC_VAL 32767.0     /* maximum DAC value for system */

/*==========================================================================*/
int log_error(err_string)
/*==========================================================================*/
/* 

     Routine to write out a specified error message into the xvnmr error
     log file. The error log file used is $HOME/.xvnmr_pages/xvnmr_errlog
     This can be piped continuously to an xterm window using the command
     xterm -T "Error Log" -e /usr/bin/tail -f ~/.xvnmr_pages/xvnmr_errlog

     INPUTS:
     =======

     err_string      string containing the error message to be written

     OUTPUTS:
     ========
 
     The routine returns a 1 if the log file is correctly updated and a 0
     if the log file is incorrectly written
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, University of Oxford, July 1998
 
                                                                            */
/*==========================================================================*/
char err_string[];
{
  FILE *fp;
  char *home_dir, full_fname[MAXSTR], file_name[MAXSTR];
  
  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }

  sprintf(file_name,"/.xvnmr_pages/xvnmr_errlog");
  strcat(full_fname, file_name);

  if ((fp=fopen(full_fname, "r")) == NULL) {
    /* create and open file */
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  } else {
    /* append existing file */
    if ((fp=fopen(full_fname, "a")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  fprintf(fp, "%s\n", err_string);

  fclose(fp);

}

/*==========================================================================*/
double create_ramp_gradient(grad_strength,fstem,grad_direction,grate,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum ramp waveform shape, given a desired
     gradient slew rate (grate in units of usec per g/cm) and a target
     gradient strength (grad_strength in units of g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}_up.GRD for the attack portion of the waveform (if
     grad_direction=1) or {fstem}_dn.GRD for the descent portion of the
     waveform (if grad_direction=-1). If the ramp duration is less than
     MAX_RAMP_RES * GRAD_UPDATE_TIME microseconds in duration, then the
     closest integral number of GRAD_UPDATE_TIME intervals is found and
     an appropriately reduced resolution waveform is generated. Otherwise,
     a waveform of MAX_RAMP_RES points is generated, and the pulse sequence
     generator pads the timing of the waveform accordingly with 20ns
     resolution to bring it up to the correct ramp duration.

     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gro")

     grad_direction     ramp direction (1=ramp up, -1=ramp down, 0=flat)

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)

     write_flag         only write waveform to disk if write_flag = 1
 
     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the ramp.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, NIH, August 1997
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double grad_strength,grate;
char fstem[];
int grad_direction;
unsigned long write_flag;
{
  double grad_ramp_time, num_grad_steps, temp_dac;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i;
  FILE *fp=NULL;
     
  grad_ramp_time=fabs(grad_strength)*fabs(grate);
  num_grad_steps=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
  num_grad_steps=(num_grad_steps > MAX_RAMP_RES) ? MAX_RAMP_RES:num_grad_steps;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);
  if (grad_direction == 1) sprintf(fname_ext,"_up.GRD");
  if (grad_direction == 0) sprintf(fname_ext,"_flat.GRD");
  if (grad_direction == -1) sprintf(fname_ext,"_dn.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }  

  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    if (grad_direction == 1) temp_dac=(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (grad_direction == 0) temp_dac=MAX_DAC_VAL;
    if (grad_direction == -1) temp_dac=MAX_DAC_VAL-
                        (float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  grad_ramp_time=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
     
  return(grad_ramp_time*GRAD_UPDATE_TIME/1.0e6);
 
}


 
/*==========================================================================*/
double create_trap_gradient(grad_strength,fstem,grad_duration,grate,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum trapezoid waveform shape, given a desired
     gradient slew rate (grate in units of usec per g/cm) and a target
     gradient strength (grad_strength in units of g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. If the ramp duration is less than
     MAX_TRAP_RAMP_RES * GRAD_UPDATE_TIME microseconds in duration, then the
     closest integral number of GRAD_UPDATE_TIME intervals is found and
     an appropriately reduced resolution waveform is generated. Otherwise,
     a waveform of MAX_TRAP_RAMP_RES points is generated, and the pulse
     sequence generator pads the timing of the waveform accordingly with 20ns
     resolution to bring it up to the correct ramp duration.

     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gro")

     grad_duration      desired duration of flat part of trapezoid

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)
 
     write_flag         only write waveform to disk if write_flag = 1
 
     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the entire trapezoid.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
     bug in waveform resolution calculation corrected, Feb 1999
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double grad_strength,grate;
char fstem[];
double grad_duration;
unsigned long write_flag;
{
  double grad_ramp_time, grad_ramp_res, num_grad_steps, temp_dac;
  double actual_ramp_update, flat_top_points;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp=NULL;
     
  grad_ramp_time=fabs(grad_strength)*fabs(grate);
  num_grad_steps=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
  num_grad_steps=(num_grad_steps > MAX_TRAP_RAMP_RES) ? 
             MAX_TRAP_RAMP_RES : num_grad_steps;

  if (grad_duration > MAX_GRAD_DURATION) {
    printf("Error: Gradient duration too long in create_trap_gradient()\n");
    exit(0);
  }

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }  
  }

  temp_dac=0.0;

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  /* flat top portion of trapezoid */
  if (num_grad_steps > 0) {
    grad_ramp_res=ceil(grad_ramp_time*1.0e6/WAVEFORM_RESOL/num_grad_steps);
    /* calculate the actual ramp resolution in us */
    actual_ramp_update=grad_ramp_res*WAVEFORM_RESOL/1.e6;
      /* start of Feb 99 bug fix (PJ) */
      if (actual_ramp_update < GRAD_UPDATE_TIME/1.0e6) {
        actual_ramp_update = GRAD_UPDATE_TIME/1.0e6;
      }
      /* end of Feb 99 bug fix (PJ) */
    flat_top_points=ceil(grad_duration/actual_ramp_update);
  } else {
    actual_ramp_update=0.0;
    flat_top_points=0.0;
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac)); 
  }
  if (write_flag == 1) 
      fprintf(fp, "%d   %d\n",(int)(temp_dac), (int)(flat_top_points+1.e-6)-
      (num_clock_blocks-1)*255); 

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
  return((num_grad_steps*2.0+flat_top_points)*actual_ramp_update);
 
}

/*==========================================================================*/
double create_spin_echo_gradient(fstem,grad_strength,grad_duration,
  crush_strength,left_crush_time,right_crush_time,grate,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum spin echo slice select gradient
     waveform, given a desired gradient slew rate (grate in units of usec 
     per g/cm) and a target gradient strength (grad_strength in units of 
     g/cm). A crusher gradient surrounding the slice selective gradient
     can also be specified (crush_strength in g/cm), along with the
     crusher durations on either side of the 180 slice selective pulse.
     The routine writes out a waveform file into the ~/vnmrsys/shapelib
     directory with the name {fstem}.GRD. If the ramp duration is less than
     MAX_TRAP_RAMP_RES * GRAD_UPDATE_TIME microseconds in duration, then the
     closest integral number of GRAD_UPDATE_TIME intervals is found and
     an appropriately reduced resolution waveform is generated. Otherwise,
     a waveform of MAX_TRAP_RAMP_RES points is generated, and the pulse
     sequence generator pads the timing of the waveform accordingly with 20ns
     resolution to bring it up to the correct ramp duration.


           left_crush_time                  right_crush_time
                <---->                         <------->
                ______                         _________
               /      \                       /         \
              /  crush \                     /   crush   \
             / _strength\___________________/  _strength  \
            /                                              \
           /                grad_strength                   \
          /                                                  \
     ____/                                                    \________
                         <----------------->                      
                             grad_duration

     INPUTS:
     =======

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gro")

     grad_strength      target slice select gradient strength in g/cm

     grad_duration      duration of the flat portion of the slice select pulse
                        (in s)

     crush_strength     target crusher gradient strength in g/cm

     left_crush_time    desired duration of the left hand crusher pulse

     right_crush_time   desired duration of the right hand crusher pulse

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)
 
     write_flag         only write waveform to disk if write_flag = 1
 
     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the entire slice select gradient.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, Aug 1999
 
                                                                            */
/*==========================================================================*/
char fstem[];
double grad_strength,crush_strength,grate;
double grad_duration,left_crush_time,right_crush_time;
unsigned long write_flag;
{
  double grad_ramp_time, grad_ramp_res, num_grad_steps, temp_dac;
  double actual_ramp_update, flat_top_points;
  double left_flat_top_points, right_flat_top_points;
  double max_grad_strength, max_grad_change, grad_scal, start_scal, end_scal;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp=NULL;
     
  /* figure out the largest absolute ramp duration */
  max_grad_change=(fabs(crush_strength) > fabs(grad_strength)) ? 
         fabs(crush_strength) : fabs(grad_strength);
  max_grad_change=(fabs(crush_strength-grad_strength) > max_grad_change) ? 
         fabs(crush_strength-grad_strength) : max_grad_change;
  max_grad_strength=(fabs(crush_strength) > fabs(grad_strength)) ?
         fabs(crush_strength) : fabs(grad_strength);
  grad_ramp_time=max_grad_change*fabs(grate);
  num_grad_steps=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
  num_grad_steps=(num_grad_steps > MAX_TRAP_RAMP_RES) ? 
             MAX_TRAP_RAMP_RES : num_grad_steps;

  if (grad_duration > MAX_GRAD_DURATION) {
    printf("Error: Gradient duration too long: create_spin_echo_gradient()\n");
    exit(0);
  }

  if (left_crush_time > MAX_GRAD_DURATION) {
    printf("Error: Crusher duration too long: create_spin_echo_gradient()\n");
    exit(0);
  }

  if (right_crush_time > MAX_GRAD_DURATION) {
    printf("Error: Crusher duration too long: create_spin_echo_gradient()\n");
    exit(0);
  }

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }  
  }

  temp_dac=0.0;

  /* attack portion of left crusher */
  grad_scal=crush_strength/max_grad_strength*MAX_DAC_VAL;
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_grad_steps*grad_scal;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  /* flat top portion of left crusher */
  if (num_grad_steps > 0) {
    grad_ramp_res=ceil(grad_ramp_time*1.0e6/WAVEFORM_RESOL/num_grad_steps);
    /* calculate the actual ramp resolution in us */
    actual_ramp_update=grad_ramp_res*WAVEFORM_RESOL/1.e6;
      if (actual_ramp_update < GRAD_UPDATE_TIME/1.0e6) {
        actual_ramp_update = GRAD_UPDATE_TIME/1.0e6;
      }
    left_flat_top_points=ceil(left_crush_time/actual_ramp_update);
  } else {
    actual_ramp_update=0.0;
    left_flat_top_points=0.0;
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((left_flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac)); 
  }
  if ((write_flag == 1) && (left_flat_top_points > 0.0))
      fprintf(fp, "%d   %d\n",(int)(temp_dac), 
      (int)(left_flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of left crusher */
  start_scal=crush_strength/max_grad_strength*MAX_DAC_VAL;
  end_scal=grad_strength/max_grad_strength*MAX_DAC_VAL;
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=start_scal-(float)(i+1)/num_grad_steps*(start_scal-end_scal);
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  /* flat top portion of slice select itself */
  if (num_grad_steps > 0) {
    grad_ramp_res=ceil(grad_ramp_time*1.0e6/WAVEFORM_RESOL/num_grad_steps);
    /* calculate the actual ramp resolution in us */
    actual_ramp_update=grad_ramp_res*WAVEFORM_RESOL/1.e6;
      if (actual_ramp_update < GRAD_UPDATE_TIME/1.0e6) {
        actual_ramp_update = GRAD_UPDATE_TIME/1.0e6;
      }
    flat_top_points=ceil(grad_duration/actual_ramp_update);
  } else {
    actual_ramp_update=0.0;
    flat_top_points=0.0;
  }
  temp_dac=grad_strength/max_grad_strength*MAX_DAC_VAL;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac)); 
  }
  if ((write_flag == 1) && (flat_top_points > 0.0))
      fprintf(fp, "%d   %d\n",(int)(temp_dac), (int)(flat_top_points+1.e-6)-
      (num_clock_blocks-1)*255); 

  /* attack portion of right crusher */
  start_scal=grad_strength/max_grad_strength*MAX_DAC_VAL;
  end_scal=crush_strength/max_grad_strength*MAX_DAC_VAL;
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=start_scal+(float)(i+1)/num_grad_steps*(end_scal-start_scal);
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  /* flat top portion of right crusher */
  if (num_grad_steps > 0) {
    grad_ramp_res=ceil(grad_ramp_time*1.0e6/WAVEFORM_RESOL/num_grad_steps);
    /* calculate the actual ramp resolution in us */
    actual_ramp_update=grad_ramp_res*WAVEFORM_RESOL/1.e6;
      if (actual_ramp_update < GRAD_UPDATE_TIME/1.0e6) {
        actual_ramp_update = GRAD_UPDATE_TIME/1.0e6;
      }
    right_flat_top_points=ceil(right_crush_time/actual_ramp_update);
  } else {
    actual_ramp_update=0.0;
    right_flat_top_points=0.0;
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((right_flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac)); 
  }
  if ((write_flag == 1) && (right_flat_top_points > 0.0))
      fprintf(fp, "%d   %d\n",(int)(temp_dac), 
      (int)(right_flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of right crusher */
  grad_scal=crush_strength/max_grad_strength*MAX_DAC_VAL;
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=grad_scal-(float)(i+1)/num_grad_steps*grad_scal;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
  return((num_grad_steps*4.0+left_flat_top_points+flat_top_points
      +right_flat_top_points)*actual_ramp_update);
 
}

/*==========================================================================*/
double create_epiblip_gradient(grad_strength,fstem,grate,dwell_time,
   read_matrix_size,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum multiple (EPI) trapezoid waveform
     shape, given a desired gradient slew rate (grate in units of usec per
     g/cm, i.e., grise/gmax) and a target gradient strength (grad_strength
     in units of g/cm). The routine writes out a read direction waveform file
     into the ~/vnmrsys/shapelib directory with the name {fstem}_rd.GRD
     and a phase encode direction gradient waveform file into the
     ~/vnmrsys/shapelib directory with the name {fstem}_ph.GRD
     The routine returns the duration of the ramp required to reach
     grad_strength. Note that the blip is at the end of the flat top.

                _________
                         \
                          \
               zero___     \       {fstem}_rd.GRD
                            \
                             \
                              \


                           /\
                          /  \
                _________/    \    {fstem}_ph.GRD

     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm
     (double)

     fstem              string variable containing the filestem of the
     (char)             desired waveform (e.g., "epird")

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)
     (double) 

     dwell_time         dwell time of the acquisition in secs
     (double)

     read_matrix_size   number of points to be collected in read direction
     (int)              per echo
 
     write_flag         only write waveform to disk if write_flag = 1

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the ramp portion of the trapezoid.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, April 1998
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double grad_strength,grate,dwell_time;
char fstem[];
int read_matrix_size;
unsigned long write_flag;
{
  double ramp_duration, num_ramp_steps, flat_top_clock_cycles, temp_dac;
  double ramp_duration_estimate, ramp_res_estimate, approx_ramp_res;
  double flat_top_to_ramp_fact, ramp_update_time, ramp_update_clock_cycles;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir, prog_name[MAXSTR];
  int i, exit_loop, num_clock_blocks;
  FILE *fp=NULL;

  sprintf(prog_name, "create_epiblip_gradient");

  /* check that flat top period is divisible by WAVEFORM_RESOL */
  if ((float)((int)((float)(read_matrix_size)*dwell_time*1.0e6/
      WAVEFORM_RESOL+0.5)) != (float)((float)(read_matrix_size)*
      dwell_time*1.0e6/WAVEFORM_RESOL)) {
    printf("%s: Error: flat top duration of trapezoid is not a\n",prog_name);
    printf("multiple of the minimum waveform resolution (%f us)\n",
         WAVEFORM_RESOL);
    return(-1.0);
  }

  /* it is important that the flat top portion of the sequence be accurately
     set to the flat top duration requested. This is important in EPI
     sequences to keep good time. Since the ramp resolution will need to be
     an integer division of the flat top duration, need to calculate a
     reasonable value for the ramp step resolution from the flat top
     duration. */
  
  /* calculate the number of WAVEFORM_RESOL clock cycles on the flat top */
  flat_top_clock_cycles=(float)(read_matrix_size)*dwell_time*1.0e6/
           WAVEFORM_RESOL;

  /* estimate the duration of the ramp in secs */
  ramp_duration_estimate=fabs(grad_strength)*grate;

  /* estimate the resolution of the ramp in units of GRAD_UPDATE_TIME */
  ramp_res_estimate=ramp_duration_estimate*1.0e6/GRAD_UPDATE_TIME;

  /* calculate the approximate desired resolution of the ramp */
  if (ramp_res_estimate <= (float)(MAX_TRAP_RAMP_RES)) {
    approx_ramp_res=(float)((int)(ramp_res_estimate+0.5)); 
                  /* limited by GRAD_UPDATE_TIME */
  } else {
    approx_ramp_res=(float)(MAX_TRAP_RAMP_RES); /* lmtd by MAX_TRAP_RAMP_RES */
  }

  if (approx_ramp_res > 0) {
    ramp_update_clock_cycles=ceil(ramp_duration_estimate/approx_ramp_res/
       WAVEFORM_RESOL*1.0e6);
  } else {
    ramp_update_clock_cycles=0.0;
  }

  /* now make sure that the ramp_update_clock_cycles is an integral multiple
     of WAVEFORM_RESOL and also is divisible into the flat top duration to
     yield an integer */
  exit_loop=-1;
  flat_top_to_ramp_fact=1.0; /* default value */

  if (ramp_update_clock_cycles > 0) {
    while (exit_loop == -1) {
      if (fabs(flat_top_clock_cycles/ramp_update_clock_cycles - (float) ((int)
           (flat_top_clock_cycles/ramp_update_clock_cycles+0.5))) < 1.e-6) {
        /* set the flat top to ramp duration factor */
        flat_top_to_ramp_fact=flat_top_clock_cycles/ramp_update_clock_cycles;
        exit_loop=1;  /* jump out of the loop */
      } else {
        ramp_update_clock_cycles=ramp_update_clock_cycles+1;
      }
    }
  }

  ramp_update_time=(double)(flat_top_clock_cycles/flat_top_to_ramp_fact* 
         WAVEFORM_RESOL/1.0e6);
  if (ramp_update_time > 0.0) {
    num_ramp_steps=(float)((int)(ramp_duration_estimate/ramp_update_time+0.5));
    ramp_duration=num_ramp_steps*ramp_update_time;
    if ((num_ramp_steps < 10.0) && (ramp_update_time > 100.0e-6)) {
      printf("%s: Warning: ramp resolution is very low\n",prog_name);
    }
  } else {
    num_ramp_steps=0.0;
    ramp_duration=0.0;
  }

  home_dir = (char *) getenv("HOME");

  /*********************************************************************/
  /* first do the read direction gradient waveform */
  /*********************************************************************/
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,"_rd.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  temp_dac=0.0;
    
  /* flat top portion of trapezoid */
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_to_ramp_fact)/255.0)+1;
  temp_dac=MAX_DAC_VAL;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac));
  }
  if (write_flag == 1) fprintf(fp, "%d   %d\n",(int)(temp_dac), 
      (int)(flat_top_to_ramp_fact)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)num_ramp_steps; i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  /* attack portion of next trapezoid (which goes negative) */
  for (i=0; i<(int)num_ramp_steps; i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*(-MAX_DAC_VAL);
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  if (write_flag == 1) fprintf(fp, "\n"); /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);

  /*********************************************************************/
  /* second do the phase encode direction gradient waveform */
  /*********************************************************************/
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,"_ph.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  temp_dac=0.0;

  /* zero portion of waveform */
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_to_ramp_fact)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    if (write_flag == 1) fprintf(fp, "0   255\n");
  }
  if (write_flag == 1) fprintf(fp, "0   %d\n", (int)(flat_top_to_ramp_fact)
          -(num_clock_blocks-1)*255);

  /* attack portion of trianglular waveform */
  for (i=0; i<(int)num_ramp_steps; i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  /* descent portion of triangular waveform */
  for (i=0; i<(int)num_ramp_steps; i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }

  if (write_flag == 1) fprintf(fp, "\n"); /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  return((double)(ramp_duration));
 
}

/*==========================================================================*/
double create_first_epiblip_gradient(grad_strength,fstem,grate,write_flag)
/*==========================================================================*/
/*   

     Routine to calculate the phase encode blips in combination with the
     read direction ramps for an EPI sequence. The routine calculates the
     optimum ramp waveform shape, given a desired gradient slew rate (grate
     in units of usec per g/cm) and a target gradient strength (grad_strength
     in units of g/cm). The routine writes out the read direction waveform
     file into the ~/vnmrsys/shapelib directory with the name {fstem}_rd.GRD
     and the phase encode direction waveform file into the ~/vnmrsys/shapelib
     directory with the name {fstem}_ph.GRD (a further file called
     {fstem}1_rd.GRD is created with the first read gradient triangle
     interval set to zero (see below). If the ramp duration is less than
     MAX_RAMP_RES * GRAD_UPDATE_TIME microseconds in duration, then the
     closest integral number of GRAD_UPDATE_TIME intervals is found and
     an appropriately reduced resolution waveform is generated. Otherwise,
     a waveform of MAX_RAMP_RES points is generated, and the pulse sequence
     generator pads the timing of the waveform accordingly with 20ns
     resolution to bring it up to the correct ramp duration.

                    /\
                   /  \
                  /    \     {fstem}_ph.GRD
                 /      \
                /        \


                \
                 \
                  \          {fstem}_rd.GRD
                   \
         zero___    \
                     \
                      \
                       \
                        \
                         \


                         /
                        /
                       /     {fstem}1_rd.GRD
                      /
                _____/


     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "blip")

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)

     write_flag         only write waveform to disk if write_flag = 1
 
     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the ramp (i.e., half the triangle duration).
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, University of Oxford, April 1998
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double grad_strength,grate;
char fstem[];
unsigned long write_flag;
{
  double grad_ramp_time, num_grad_steps, temp_dac;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i;
  FILE *fp=NULL;
     
  grad_ramp_time=fabs(grad_strength)*grate;
  num_grad_steps=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
  num_grad_steps=(num_grad_steps > MAX_RAMP_RES) ? MAX_RAMP_RES:num_grad_steps;

  home_dir = (char *) getenv("HOME");

  /************************************************************************/
  /* first construct the read gradient waveform */
  /************************************************************************/
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);
  sprintf(fname_ext,"_rd.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_grad_steps*(-MAX_DAC_VAL);
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  /************************************************************************/
  /* next construct the phase encode gradient waveform */
  /************************************************************************/
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);
  sprintf(fname_ext,"_ph.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1)  {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  if (write_flag == 1) fprintf(fp, "\n"); /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  /************************************************************************/
  /* finally construct the read gradient waveform for the first lobe */
  /************************************************************************/
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);
  sprintf(fname_ext,"1_rd.GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
  for (i=0; i<(int)(num_grad_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_grad_steps*MAX_DAC_VAL;
    if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  grad_ramp_time=ceil(grad_ramp_time*1.0e6/GRAD_UPDATE_TIME);
     
  return(grad_ramp_time*GRAD_UPDATE_TIME/1.0e6);
 
}

/*==========================================================================*/
double create_ge_slice_flow_comp(obj_grad,fstem,obj_duration,obj_ramp,max_grad,
   max_slew,seq_time)
/*==========================================================================*/
/*   
     Routine to calculate the optimum flow compensation scheme for the
     read direction of a gradient echo sequence. The routine takes an
     object slice selection gradient (usually the second half of the slice
     select gradient waveform) as an argument, and also the maximum
     gradient strength allowed (in units of g/cm), and the maximum allowable
     slew rate (max_slew in units of usec per g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. The waveform is calculated to have an integer number
     of steps of resolution GRAD_UPDATE_TIME microseconds.

                                                  Delta2
                                                  ->  <-
                                              2*delta
                                            ->      <-
                                  obj_ramp          __
                                   -> <-           /  \
          _ _ _ _  _ _ ______________             /gmax\
     ____/              obj_grad     \           /      \_____________
                       <------------> \         /   ->   <-
                        obj_duration   \  gmax /     delta
                                        \_____/
                                    ->   <-
                                     delta
                                         <--->
                                         Delta1
     INPUTS:
     =======

     obj_grad           gradient strength of the gradient to be flow
                        compensated in g/cm (normally this would be the
                        second half of the slice select gradient).

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gssfc")

     obj_duration       duration of flat part of the gradient waveform to
                        be compensated (normally this would be the flat top
                        portion of the second half of the slice select
                        gradient).

     obj_ramp           duration of ramp down part of the gradient waveform
                        to be compensated (normally this would be the flat top
                        portion of the second half of the slice select
                        gradient).

     max_grad           maximum gradient strength allowed for the flow
                        compensation gradients.
     
     max_slew           maximum gradient slew rate of the flow compensation
                        gradient waveform in sec per g/cm (i.e., grise/gmax).

     seq_time           seq_time is the duration of the entire waveform
                        (i.e., TE - t_90/2 -pw_gss + at/2 + pw_gro). This
                        is needed to zero pad the waveform to give it the
                        same number of points as the read and phase encode
                        waveforms.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of max_grad in case it is adjusted down by the code.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp;
double max_grad,max_slew,seq_time;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points;
  double delta, delta1, delta2;
  double standard_flow_comp_duration();
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  if (obj_grad == 0.0) {
    printf("Error: zero gradient value passed to create_ge_slice_flow_comp\n");
    exit(0);
  }  

  delta2=-1.0;
  while (delta2 == -1.0) {
    delta2=standard_flow_comp_duration(obj_grad,obj_duration,obj_ramp,
      fabs(max_grad),max_slew);
    /* keep reducing the maximum gradient until delay delta2 becomes +ive */
    if (delta2 == -1.0) max_grad=max_grad*0.9;
  }
  delta1=delta2+obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  /* round delta1 and delta2 off to the nearest GRAD_UPDATE_TIME */
  delta1=(float)((int)(delta1*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;
  delta2=(float)((int)(delta2*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;

  /* round off flow comp ramp time (delta) to the nearest GRAD_UPDATE_TIME */
  delta=(float)((int)(fabs(max_grad)*max_slew*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=delta*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* firstly write out the longer negative flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta1*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac)); 
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac), 
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* secondly write out the shorter positive flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta2*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac)); 
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac), 
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* lastly write out the final zero filler points that guarantee that the
     read, slice and phase encode waveforms have the same number of points */

  flat_top_points=(float)((int)(seq_time*1.e6/GRAD_UPDATE_TIME+0.5))-
         (4.0*delta+delta1+delta2)*1.e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n"); 
  }
  fprintf(fp, "0   %d\n", 
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  return(max_grad);
 
}

/*==========================================================================*/
double create_ge_read_flow_comp(obj_grad,fstem,obj_duration,obj_duration2,
   obj_ramp,max_grad,max_slew,seq_time)
/*==========================================================================*/
/*   
     Routine to calculate the optimum flow compensation scheme for the
     read direction of a gradient echo sequence. The routine takes an
     object read direction gradient (usually the first half of the read out
     acquisition gradient waveform) as an argument, and also the maximum
     gradient strength allowed (in units of g/cm), and the maximum allowable
     slew rate (max_slew in units of usec per g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. The waveform is calculated to have an integer number
     of steps of resolution GRAD_UPDATE_TIME microseconds.

                  Delta2
                  ->  <-
                     2*delta
                    ->      <-
                    __           obj_ramp
                   /  \           -> <-
                  /gmax\             ___________ _ _ _ _ _ _
           ______/      \           /  obj_grad             \___
               ->   <-   \         / 
                delta     \  gmax /  <---------><----------> 
                           \_____/  obj_duration obj_duration2
                      ->   <-
                       delta
                            <--->
                            Delta1


     INPUTS:
     =======

     obj_grad           gradient strength of the gradient to be flow
                        compensated in g/cm (normally this would be the
                        first half of the read out gradient). The ramp
                        down portion of the object gradient is also given
                        the same ramp time.

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "grofc")

     obj_duration       duration of flat part of the gradient waveform to
                        be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     obj_duration2      duration of flat part of the second half of the
                        object gradient (normally this would be the flat top
                        portion of the second half of the read out gradient).

     obj_ramp           duration of ramp up part of the gradient waveform
                        to be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     max_grad           maximum gradient strength allowed for the flow
                        compensation gradients.
     
     max_slew           maximum gradient slew rate of the flow compensation
                        gradient waveform in sec per g/cm (i.e., grise/gmax).

     seq_time           seq_time is the duration of the entire waveform
                        (i.e., TE - t_90/2 -pw_gss + at/2 + pw_gro). This
                        is needed to zero pad the waveform to give it the
                        same number of points as the slice and phase encode
                        waveforms.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of max_grad in case it is adjusted down by the code.
     If the waveform will not fit into the requested time window then the
     negative duration by which the waveform overflows is returned.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_duration2,obj_ramp;
double max_grad,max_slew,seq_time;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points,rnd_up_duration;
  double delta, delta1, delta2;
  double standard_flow_comp_duration();
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  if (obj_grad == 0.0) {
    printf("Error: zero gradient value passed to create_ge_read_flow_comp\n");
    exit(0);
  }

  delta2=-1.0;
  while (delta2 == -1.0) {
    delta2=standard_flow_comp_duration(obj_grad,obj_duration,obj_ramp,
      fabs(max_grad),max_slew);
    /* keep reducing the maximum gradient until delay delta2 becomes +ive */
    if (delta2 == -1.0) max_grad=max_grad*0.9;
  }
  delta1=delta2+obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  /* round delta1 and delta2 off to the nearest GRAD_UPDATE_TIME */
  delta1=(float)((int)(delta1*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;
  delta2=(float)((int)(delta2*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;

  /* round off flow comp ramp time (delta) to the nearest GRAD_UPDATE_TIME */
  delta=(float)((int)(fabs(max_grad)*max_slew*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  /* round off obj_ramp time to the nearest GRAD_UPDATE_TIME */
  obj_ramp=(float)((int)(obj_ramp*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=delta*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* round off the flat top durations to give integer GRAD_UPDATE_TIME */
  rnd_up_duration=(float)((int)((obj_duration+obj_duration2)*1.0e6/
       GRAD_UPDATE_TIME+0.5))*GRAD_UPDATE_TIME/1.0e6;

  /* firstly write out the initial zero filler points that guarantee that the
     read, slice and phase encode waveforms have the same number of points */

  flat_top_points=(float)((int)(seq_time*1e6/GRAD_UPDATE_TIME+0.5))-
     (4.0*delta+delta1+delta2+2.0*obj_ramp)*1e6/GRAD_UPDATE_TIME-
     rnd_up_duration*1e6/GRAD_UPDATE_TIME;

  if (flat_top_points < 0.0) {
    printf("Error: create_ge_read_flow_comp: insufficient echo time\n");
    fclose(fp);
    return(flat_top_points*GRAD_UPDATE_TIME/1.0e6);
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n");
  }
  fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* secondly write out the shorter positive flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta2*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* thirdly write out the longer negative flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta1*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* lastly write out the read out waveform */

  num_ramp_steps=obj_ramp*1.0e6/GRAD_UPDATE_TIME;

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=rnd_up_duration*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=fabs(obj_grad)/max_grad*MAX_DAC_VAL-(float)(i+1)/num_ramp_steps* 
           fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  return(max_grad);
 
}

/*==========================================================================*/
double create_ge_phase_flow_comp(obj_grad,fstem,obj_duration,obj_ramp,
    seq_time)
/*==========================================================================*/
/*   
     Routine to calculate the phase encode gradient waveform required as
     part of a flow compensation scheme for a gradient echo sequence.
     The routine writes out a waveform file into the ~/vnmrsys/shapelib
     directory with the name {fstem}.GRD. The waveform is calculated to have
     an integer number of steps of resolution GRAD_UPDATE_TIME microseconds.

         obj_duration
            ->   <-
              ___  obj_grad
             /   \
            /     \
           /       \______________________________________
         ->   <-
         obj_ramp


     INPUTS:
     =======

     obj_grad           maximum desired gradient strength (gauss/cm) of the
                        phase encode gradient waveform (typically nv/2.0*
                        phase_encode_gradient_step_size).

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "grofc")

     obj_duration       duration (sec.) of flat part of the p.e. gradient
                        waveform to be compensated.

     obj_ramp           duration (sec.) of ramp up part of the p.e. gradient
                        waveform.

     seq_time           seq_time is the duration of the entire waveform
                        (i.e., TE - t_90/2 -pw_gss + at/2 + pw_gro). This
                        is needed to zero pad the waveform to give it the
                        same number of points as the read and slice direction
                        waveforms.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of waveform duration in case it has been adjusted by the code.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp;
double seq_time;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points, gradient_polarity;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  /* round off phase encode ramp time (obj_ramp) to nearest GRAD_UPDATE_TIME */
  obj_ramp=(float)((int)(obj_ramp*1.0e6/GRAD_UPDATE_TIME+0.5))
             *GRAD_UPDATE_TIME/1.0e6;
     
  /* round off phase encode flat top time (obj_duration) to the nearest
      GRAD_UPDATE_TIME */
  obj_duration=(float)((int)(obj_duration*1.0e6/GRAD_UPDATE_TIME+0.5))*
      GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=obj_ramp*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* firstly write out the phase encode gradient waveform pulse */

  if (obj_grad != 0.0) {
    gradient_polarity=obj_grad/fabs(obj_grad);
  } else {
    gradient_polarity=1.0;
  }

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(gradient_polarity*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=obj_duration*1e6/GRAD_UPDATE_TIME;
  if (obj_grad != 0.0) {
    /* correct for a bug in the waveform memory calculation which doesn't
       allow for more than 255 clock cycles of a given DAC level */
    num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
    for (i=0; i<(num_clock_blocks-1); i++) {
      fprintf(fp, "%d   255\n",(int)(gradient_polarity*temp_dac));
    }
    fprintf(fp, "%d   %d\n",(int)(gradient_polarity*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);
   } else {
    /* correct for a bug in the waveform memory calculation which doesn't
       allow for more than 255 clock cycles of a given DAC level */
    num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
    for (i=0; i<(num_clock_blocks-1); i++) {
      fprintf(fp, "0   255\n");
    }
    fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);
   }

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(gradient_polarity*temp_dac));
  }
    
  /* finally write out the zero filler points that guarantee that the
     read, slice and phase encode waveforms have the same number of points */

  flat_top_points=(float)((int)(seq_time*1e6/GRAD_UPDATE_TIME+0.5))-
     (2.0*obj_ramp+obj_duration)*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n");
  }
  fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  return((float)((int)(seq_time*1e6/GRAD_UPDATE_TIME+0.5))*
               GRAD_UPDATE_TIME/1.0e6);
 
}

/*==========================================================================*/
double create_se_read_flow_comp(obj_grad,fstem,obj_duration,obj_duration2,
   obj_ramp,max_grad,max_slew,TE,t_180,start_point)
/*==========================================================================*/
/*   
     Routine to calculate the optimum flow compensation scheme for the
     read direction of a spin echo sequence. The routine takes an
     object read gradient pulse (usually the first half of the read out
     acquisition gradient waveform) as an argument, and also the maximum
     gradient strength allowed (in units of g/cm), and the maximum allowable
     slew rate (max_slew in units of usec per g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. The waveform is calculated to have an integer number
     of steps of resolution GRAD_UPDATE_TIME microseconds.

                                              obj_ramp
            delta           delta              -> <-
           ->   <-         ->   <-                ___________ _ _ _ _ _ _
    _________        ________           _________/  obj_grad             \___
             \      /        \         / 
   ^          \gmax/   t_180  \  gmax /           <---------><----------> 
   |           \__/  <------>  \_____/           obj_duration obj_duration2
start_point
              ->  <-            <--->
              Delta2           Delta1

                                       te/2
                         <--------------------------------->


     INPUTS:
     =======

     obj_grad           gradient strength of the gradient to be flow
                        compensated in g/cm (normally this would be the
                        first half of the read out gradient). The ramp
                        down portion of the object gradient is also given
                        the same ramp time.

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "grofc")

     obj_duration       duration of flat part of the gradient waveform to
                        be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     obj_duration2      duration of flat part of the second half of the
                        object gradient (normally this would be the flat top
                        portion of the second half of the read out gradient).

     obj_ramp           duration of ramp up part of the gradient waveform
                        to be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     max_grad           maximum gradient strength allowed for the flow
                        compensation gradients.
     
     max_slew           maximum gradient slew rate of the flow compensation
                        gradient waveform in sec per g/cm (i.e., grise/gmax).

     t_180              t_180 is the duration of the 180 degree pulse. The
                        flow comp pulses are placed either side of the 180.

     TE                 The desired TE time of the sequence (in secs).

     start_point        The starting point of the flow comp waveform. This
                        should be the point after the 90 degree slice seletion
                        waveform has ramped down. (i.e., t_90/2 + pw_gss).
                        This is needed to zero pad the waveform to give it the
                        same number of points as the slice and phase encode
                        waveforms. Note that the centre of the 90 degree pulse
                        is defined as t=0.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of max_grad in case it is adjusted down by the code.
     If the waveform will not fit into the requested time window then the
     negative duration by which the waveform overflows is returned.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, March 1998
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_duration2,obj_ramp;
double max_grad,max_slew,TE,t_180,start_point;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points,rnd_up_duration,rnd_up_duration2;
  double rnd_up_half_t_180, rnd_up_start_point;
  double delta, delta1, delta2;
  double se_read_flow_comp_duration();
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  if (obj_grad == 0.0) {
    printf("Error: zero gradient value passed to create_se_read_flow_comp\n");
    exit(0);
  }

  delta2=-1.0;
  while (delta2 == -1.0) {
    delta2=se_read_flow_comp_duration(obj_grad,obj_duration,obj_ramp,
      fabs(max_grad),max_slew,TE,t_180);
    /* keep reducing the maximum gradient until delay delta2 becomes +ive */
    if (delta2 == -1.0) max_grad=max_grad*0.9;
  }
  delta1=delta2+obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  /* round delta1 and delta2 off to the nearest GRAD_UPDATE_TIME */
  delta1=(float)((int)(delta1*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;
  delta2=(float)((int)(delta2*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;

  /* round off flow comp ramp time (delta) to the nearest GRAD_UPDATE_TIME */
  delta=(float)((int)(fabs(max_grad)*max_slew*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  /* round off obj_ramp time to the nearest GRAD_UPDATE_TIME */
  obj_ramp=(float)((int)(obj_ramp*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=delta*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* round off the flat top durations to give integer GRAD_UPDATE_TIME */
  rnd_up_duration=(float)((int)((obj_duration)*1.0e6/
       GRAD_UPDATE_TIME+0.5))*GRAD_UPDATE_TIME/1.0e6;
  rnd_up_duration2=(float)((int)((obj_duration2)*1.0e6/
       GRAD_UPDATE_TIME+0.5))*GRAD_UPDATE_TIME/1.0e6;

  /* round off the t_180 duration to give integer GRAD_UPDATE_TIME */
  rnd_up_half_t_180=(float)((int)(t_180/2.0*1.0e6/GRAD_UPDATE_TIME+0.5))*
              GRAD_UPDATE_TIME/1.0e6;

  /* round off the start_point time marker to give integer GRAD_UPDATE_TIME */
  rnd_up_start_point=(float)((int)(start_point*1.0e6/GRAD_UPDATE_TIME+0.5))*
              GRAD_UPDATE_TIME/1.0e6;

  /* firstly write out the initial zero filler points that guarantee that the
     read, slice and phase encode waveforms have the same number of points */

  flat_top_points=(float)((int)(TE/2*1e6/GRAD_UPDATE_TIME+0.5))-
     (2.0*delta+delta2)*1e6/GRAD_UPDATE_TIME
     -rnd_up_half_t_180*1e6/GRAD_UPDATE_TIME
     -rnd_up_start_point*1e6/GRAD_UPDATE_TIME;

  if (flat_top_points < 0.0) {
    printf("Error: create_se_read_flow_comp: insufficient echo time\n");
    fclose(fp);
    return(flat_top_points*GRAD_UPDATE_TIME/1.0e6);
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n");
  }
  fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* secondly write out the pre-180 degree flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta2*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* thirdly write out the zeros that coincide with the 180 degree pulse */

  flat_top_points=2.0*(float)((int)(rnd_up_half_t_180*1e6/GRAD_UPDATE_TIME));

  if (flat_top_points < 0.0) {
    printf("Error: create_se_read_flow_comp: incorrect 180 degree pulse\n");
    fclose(fp);
    return(flat_top_points*GRAD_UPDATE_TIME/1.0e6);
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n");
  }
  fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* fourthly write out the longer negative flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta1*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* fifthly write out the zeroes between the 2nd flow comp and the read out */

  flat_top_points=(float)((int)(TE/2*1e6/GRAD_UPDATE_TIME+0.5))-
     (2.0*delta+delta1+obj_ramp+obj_duration)*1e6/GRAD_UPDATE_TIME-
     rnd_up_half_t_180*1e6/GRAD_UPDATE_TIME;

  if (flat_top_points < 0.0) {
    printf("Error: create_se_read_flow_comp: te too short\n");
    fclose(fp);
    return(flat_top_points*GRAD_UPDATE_TIME/1.0e6);
  }
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "0   255\n");
  }
  fprintf(fp, "0   %d\n",
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* lastly write out the read out waveform */

  num_ramp_steps=obj_ramp*1.0e6/GRAD_UPDATE_TIME;

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=(rnd_up_duration+rnd_up_duration2)*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=fabs(obj_grad)/max_grad*MAX_DAC_VAL-(float)(i+1)/num_ramp_steps* 
           fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  return(max_grad);
 
}

/*==========================================================================*/
double create_epi_read_gradients(grad_strength,fstem,grate,dwell_time,
   read_matrix_size,nloops,grad_offset_tweak,echo_gap_tweak,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum multiple (EPI) trapezoid waveform
     shape, given a desired gradient slew rate (grate in units of usec per
     g/cm, i.e., grise/gmax) and a target gradient strength (grad_strength
     in units of g/cm). The routine writes out a waveform file into the 
     ~/vnmrsys/shapelib directory with the name {fstem}.GRD. 
     The routine returns the duration of the ramp required to reach
     grad_strength.

     INPUTS:
     =======

     grad_strength      target gradient strength in g/cm
     (double)

     fstem              string variable containing the filestem of the
     (char)             desired waveform (e.g., "epird")

     grate              gradient slew rate in sec per g/cm (i.e., grise/gmax)
     (double) 

     dwell_time         dwell time of the acquisition in secs
     (double)

     read_matrix_size   number of points to be collected in read direction
     (int)              per echo
 
     nloops             number of trapezoid cycles (i.e., phase encode steps)
     (int)

     grad_offset_tweak  percentage value that the magnitude of the positive
     (double)           going gradients should exceed the magnitude of the
                        negative going gradients (0.0 would give the same
                        -ive and +ive magnitude, 1.0 would give a 1% increase
                        to the positive lobes relative to the negative lobes,
                        etc.)

     echo_gap_tweak     duration (secs) to be added between each gradient
     (double)           lobe to allow for excessively large phase encode
                        gradients (particularly for segmented epi sequences).
                        The default value would normally be zero.

     write_flag         only write waveform to disk if write_flag = 1

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the ramp portion of the trapezoid.
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double grad_strength,grate,grad_offset_tweak,echo_gap_tweak;
char fstem[];
double dwell_time;
int read_matrix_size, nloops;
unsigned long write_flag;
{
  double ramp_duration, num_ramp_steps, flat_top_clock_cycles, temp_dac;
  double ramp_duration_estimate, ramp_res_estimate, approx_ramp_res;
  double flat_top_to_ramp_fact, ramp_update_time, ramp_update_clock_cycles;
  double pos_lobe_factor, neg_lobe_factor, required_ramp_duration;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir, prog_name[MAXSTR];
  int i, loop_no, exit_loop, while_count, num_clock_blocks, num_gap_steps;
  FILE *fp=NULL;

  sprintf(prog_name, "create_epi_read_gradients");

  if (grad_strength == 0.0) {
    printf("Error: zero gradient value passed to %s\n", prog_name);
    exit(0);
  }

  /* check that flat top period is divisible by WAVEFORM_RESOL */
  if ((float)((int)((float)(read_matrix_size)*dwell_time*1.0e6/
      WAVEFORM_RESOL+0.5)) != (float)((float)(read_matrix_size)*
      dwell_time*1.0e6/WAVEFORM_RESOL)) {
    printf("%s: Error: flat top duration of trapezoid is not a\n",prog_name);
    printf("multiple of the minimum waveform resolution (%f us)\n",
         WAVEFORM_RESOL);
    return(-1.0);
  }

  /* it is important that the flat top portion of the sequence be accurately
     set to the flat top duration requested. This is important in EPI
     sequences to keep good time. Since the ramp resolution will need to be
     an integer division of the flat top duration, need to calculate a
     reasonable value for the ramp step resolution from the flat top
     duration. */
  
  /* calculate the number of WAVEFORM_RESOL clock cycles on the flat top */
  flat_top_clock_cycles=(float)(read_matrix_size)*dwell_time*1.0e6/
           WAVEFORM_RESOL;

  /* estimate the duration of the ramp in secs */
  ramp_duration_estimate=fabs(grad_strength)*grate;

  /* estimate the resolution of the ramp in units of GRAD_UPDATE_TIME */
  ramp_res_estimate=ramp_duration_estimate*1.0e6/GRAD_UPDATE_TIME;

  /* calculate the approximate desired resolution of the ramp */
  if (ramp_res_estimate <= (float)(MAX_TRAP_RAMP_RES)) {
    approx_ramp_res=(float)((int)(ramp_res_estimate+0.5)); 
                  /* limited by GRAD_UPDATE_TIME */
  } else {
    approx_ramp_res=(float)(MAX_TRAP_RAMP_RES); /* lmtd by MAX_TRAP_RAMP_RES */
  }

  ramp_update_clock_cycles=ceil(ramp_duration_estimate/approx_ramp_res/
       WAVEFORM_RESOL*1.0e6);

  /* now make sure that the ramp_update_clock_cycles is an integral multiple
     of WAVEFORM_RESOL and also is divisible into the flat top duration to
     yield an integer */
  exit_loop=-1;
  flat_top_to_ramp_fact=1.0; /* default value */
  while (exit_loop == -1) {
    if (fabs(flat_top_clock_cycles/ramp_update_clock_cycles - (float)
         ((int)(flat_top_clock_cycles/ramp_update_clock_cycles+0.5))) < 1.e-6) {
      /* set the flat top to ramp duration factor */
      flat_top_to_ramp_fact=flat_top_clock_cycles/ramp_update_clock_cycles;
      exit_loop=1;  /* jump out of the loop */
    } else {
      ramp_update_clock_cycles=ramp_update_clock_cycles+1;
    }
  }

  ramp_update_time=(double)(flat_top_clock_cycles/flat_top_to_ramp_fact* 
         WAVEFORM_RESOL/1.0e6);
  num_ramp_steps=(float)((int)(ramp_duration_estimate/ramp_update_time+0.5));
  ramp_duration=num_ramp_steps*ramp_update_time;

  /* now recalculate it using the following same code that is used in 
     create_epi_phase_gradients() to make sure that they give the same value */
  /* ========================================================================*/

  required_ramp_duration=ramp_duration;

  /* estimate the resolution of the ramp in units of GRAD_UPDATE_TIME */
  ramp_res_estimate=required_ramp_duration*1.0e6/GRAD_UPDATE_TIME;

  /* calculate the approximate desired resolution of the ramp */
  if (ramp_res_estimate <= (float)(MAX_TRAP_RAMP_RES)) {
    approx_ramp_res=(float)((int)(ramp_res_estimate+0.5)); 
               /* limited by GRAD_UPDATE_TIME */
  } else {
    approx_ramp_res=(float)(MAX_TRAP_RAMP_RES); /* lmtd by MAX_TRAP_RAMP_RES */
  }

  ramp_update_clock_cycles=(float)((int)(required_ramp_duration/approx_ramp_res/
       WAVEFORM_RESOL*1.0e6+0.5));

  /* now make sure that the ramp_update_clock_cycles is an integral multiple
     of WAVEFORM_RESOL and also is divisible into the flat top duration to
     yield an integer */

  exit_loop=-1;
  while_count=0;
  num_ramp_steps=0.0; /*initialization */
  ramp_update_time=0.0; /*initialization */
  flat_top_to_ramp_fact=1.0; /* initialize */
  while (exit_loop == -1) {
    if (fabs(flat_top_clock_cycles/ramp_update_clock_cycles - (float)
       ((int)(flat_top_clock_cycles/ramp_update_clock_cycles+0.5))) < 1.e-6) {
      /* set the flat top to ramp duration factor */
      flat_top_to_ramp_fact=flat_top_clock_cycles/ramp_update_clock_cycles;
      exit_loop=1;  /* jump out of the loop */

      /* double check that  the right ramp_update_time has been found */
      ramp_update_time=flat_top_clock_cycles/flat_top_to_ramp_fact* 
             WAVEFORM_RESOL/1.0e6;
      num_ramp_steps=(float)((int)(required_ramp_duration/
          ramp_update_time+0.5));
      ramp_duration=num_ramp_steps*ramp_update_time;
      if (fabs(ramp_duration-required_ramp_duration) > 1.e-6) {
        exit_loop=-1; /* carry on looking for right value */
        ramp_update_clock_cycles=ramp_update_clock_cycles+1;
        while_count=while_count+1;
      }
    } else {
      ramp_update_clock_cycles=ramp_update_clock_cycles+1;
      while_count=while_count+1;
    }
    if (while_count > 3000) {
      printf("%s: Error: Cannot match ramp durations\n", prog_name);
      return(-1.0);
    }
  }

  /* ========================================================================*/

  if ((num_ramp_steps < 10.0) && (ramp_update_time > 100.0e-6)) {
    printf("%s: Warning: ramp resolution is very low\n",prog_name);
  }

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  temp_dac=0.0;

  pos_lobe_factor=1.0;
  neg_lobe_factor=1.0;
  if (grad_offset_tweak >= 0.0) {
    neg_lobe_factor=(100.0-grad_offset_tweak)/100.0;
  }
  if (grad_offset_tweak < 0.0) {
    pos_lobe_factor=(100.0+grad_offset_tweak)/100.0;
  }

  for (loop_no=0; loop_no < nloops/2; loop_no++) {

    /* do the positive half of the trapezoid cycle */

    /* attack portion of trapezoid */
    for (i=0; i<(int)num_ramp_steps; i++) {
      temp_dac=0.0;  /* default */
      temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL*pos_lobe_factor;
      if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
    }
    
    /* flat top portion of trapezoid */
    /* correct for a bug in the waveform memory calculation which doesn't
       allow for more than 255 clock cycles of a given DAC level */
    num_clock_blocks=(int)((flat_top_to_ramp_fact)/255.0)+1;
    for (i=0; i<(num_clock_blocks-1); i++) {
      if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac));
    }
    if (write_flag == 1) fprintf(fp, "%d   %d\n",(int)(temp_dac), 
        (int)(flat_top_to_ramp_fact)-(num_clock_blocks-1)*255);

    /* descent portion of trapezoid */
    for (i=0; i<(int)num_ramp_steps; i++) {
      temp_dac=0.0;  /* default */
      temp_dac=MAX_DAC_VAL*pos_lobe_factor-(float)(i+1)/num_ramp_steps*
             MAX_DAC_VAL*pos_lobe_factor;
      if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
    }

    /* do the gap if echo_gap_tweak is set */
    if (echo_gap_tweak > 0.0) {
      num_gap_steps=ceil(echo_gap_tweak/ramp_update_time);
      /* correct for a bug in the waveform memory calculation which doesn't
         allow for more than 255 clock cycles of a given DAC level */
      num_clock_blocks=(int)((num_gap_steps)/255.0)+1;
      for (i=0; i<(num_clock_blocks-1); i++) {
        if (write_flag == 1) fprintf(fp, "0   255\n");
      }
      if (write_flag == 1) fprintf(fp, "0   %d\n", 
          (int)(num_gap_steps)-(num_clock_blocks-1)*255);
    }

    /* do the negative half of the trapezoid cycle */

    /* attack portion of trapezoid */
    for (i=0; i<(int)num_ramp_steps; i++) {
      temp_dac=0.0;  /* default */
      temp_dac=(float)(i+1)/num_ramp_steps*(-MAX_DAC_VAL)*neg_lobe_factor;
      if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
    }
    
    /* flat top portion of trapezoid */
    /* correct for a bug in the waveform memory calculation which doesn't
       allow for more than 255 clock cycles of a given DAC level */
    num_clock_blocks=(int)((flat_top_to_ramp_fact)/255.0)+1;
    for (i=0; i<(num_clock_blocks-1); i++) {
      if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac));
    }
    if (write_flag == 1) fprintf(fp, "%d   %d\n",(int)(temp_dac), 
        (int)(flat_top_to_ramp_fact)-(num_clock_blocks-1)*255);

    /* descent portion of trapezoid */
    for (i=0; i<(int)num_ramp_steps; i++) {
      temp_dac=0.0;  /* default */
      temp_dac=(-MAX_DAC_VAL)*neg_lobe_factor+(float)(i+1)/num_ramp_steps*
           (MAX_DAC_VAL)*neg_lobe_factor;
      if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
    }

    /* do the gap if echo_gap_tweak is set */
    if (echo_gap_tweak > 0.0) {
      num_gap_steps=ceil(echo_gap_tweak/ramp_update_time);
      /* correct for a bug in the waveform memory calculation which doesn't
         allow for more than 255 clock cycles of a given DAC level */
      num_clock_blocks=(int)((num_gap_steps)/255.0)+1;
      for (i=0; i<(num_clock_blocks-1); i++) {
        if (write_flag == 1) fprintf(fp, "0   255\n");
      }
      if (write_flag == 1) fprintf(fp, "0   %d\n", (int)(num_gap_steps)-
          (num_clock_blocks-1)*255);
    }

  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  return((double)(ramp_duration));
 
}

/*==========================================================================*/
double create_epi_phase_gradients(fstem,blip_duration,dwell_time,
   read_matrix_size,nloops,echo_gap_tweak,write_flag)
/*==========================================================================*/
/*   
     Routine to calculate the optimum multiple (EPI) phase encode blip
     waveform shape, given the duration of each blip as input. The duration
     of the blip should equal the output of create_epi_read_gradients, so
     that the waveforms are correctly lined up.
     The routine writes out a waveform file into the ~/vnmrsys/shapelib
     directory with the name {fstem}.GRD. The routine returns the duration
     of the ramp.

     INPUTS:
     =======

     fstem              string variable containing the filestem of the
     (char)             desired waveform (e.g., "epiph")

     blip_duration      duration of the blip (i.e. the up or down portion).
     (double)           This should typically be the value output from
                        create_epi_read_gradients so that the waveforms are
                        correctly aligned.

     dwell_time         dwell time of the acquisition in secs
     (double)

     read_matrix_size   number of points to be collected in read direction
     (int)              per echo
 
     nloops             number of trapezoid cycles (i.e., phase encode steps)
     (int)

     echo_gap_tweak     duration (secs) to be added between each gradient
     (double)           lobe to allow for excessively large phase encode
                        gradients (particularly for segmented epi sequences).
                        The default value would normally be zero.

     write_flag         only write waveform to disk if write_flag = 1

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual duration of the blip (single up or down ramp).
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
     added write_flag to select/deselect disk write, Aug 1999
 
                                                                            */
/*==========================================================================*/
double blip_duration,echo_gap_tweak;
char fstem[];
double dwell_time;
int read_matrix_size, nloops;
unsigned long write_flag;
{
  double ramp_duration, num_ramp_steps, flat_top_clock_cycles, temp_dac;
  double ramp_res_estimate, required_ramp_duration, approx_ramp_res;
  double flat_top_to_ramp_fact, ramp_update_time, ramp_update_clock_cycles;
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir, prog_name[MAXSTR];
  int i, loop_no, exit_loop, while_count, num_clock_blocks, num_gap_steps;
  FILE *fp=NULL;

  sprintf(prog_name, "create_epi_phase_gradients");

  /* check that flat top period is divisible by WAVEFORM_RESOL */
  if ((float)((int)((float)(read_matrix_size)*dwell_time*1.0e6/
      WAVEFORM_RESOL+0.5))
       != (float)((float)(read_matrix_size)*dwell_time*1.0e6/WAVEFORM_RESOL)) {
    printf("%s: Error: flat top duration of trapezoid is not a multiple\n",
      prog_name);
    printf("of the minimum waveform resolution (%f us)\n",WAVEFORM_RESOL);
    return(-1.0);
  }

  /* check that ramp period is divisible by WAVEFORM_RESOL */
  /* 1.e-9 added to avoid some wierd round off error that sometimes occurs */
  if ((float)((int)(blip_duration*1.0e6/WAVEFORM_RESOL+1.e-9)) !=
      (float)(blip_duration*1.0e6/WAVEFORM_RESOL)) {
    printf("%s: Error: blip is not a multiple of the minumum\n", prog_name);
    printf("waveform resolution (%f us)\n",WAVEFORM_RESOL);
    return(-1.0);
  }

  /* it is important that the flat top portion of the sequence be accurately
     set to the flat top duration requested. This is important in EPI
     sequences to keep good time. Since the ramp resolution will need to be
     an integer division of the flat top duration, need to calculate a
     reasonable value for the ramp step resolution from the flat top
     duration. */
  
  /* calculate the number of WAVEFORM_RESOL clock cycles on the flat top */
  flat_top_clock_cycles=(float)(read_matrix_size)*dwell_time*1.0e6/
           WAVEFORM_RESOL;

  /* this section of code should match the code in create_epi_read_gradients()
     so that the two numbers generated are the same */
  /* ========================================================================*/

  required_ramp_duration=blip_duration;

  /* estimate the resolution of the blip in units of GRAD_UPDATE_TIME */
  ramp_res_estimate=required_ramp_duration*1.0e6/GRAD_UPDATE_TIME;

  /* calculate the approximate desired resolution of the ramp */
  if (ramp_res_estimate <= (float)(MAX_TRAP_RAMP_RES)) {
    approx_ramp_res=(float)((int)(ramp_res_estimate+0.5));
               /* limited by GRAD_UPDATE_TIME */
  } else {
    approx_ramp_res=(float)(MAX_TRAP_RAMP_RES); /* lmtd by MAX_TRAP_RAMP_RES */
  }

  ramp_update_clock_cycles=(float)((int)(required_ramp_duration/approx_ramp_res/
       WAVEFORM_RESOL*1.0e6+0.5));

  /* now make sure that the ramp_update_clock_cycles is an integral multiple
     of WAVEFORM_RESOL and also is divisible into the flat top duration to
     yield an integer */

  exit_loop=-1;
  while_count=0;
  ramp_duration=blip_duration; /* initialize */
  num_ramp_steps=0.0; /*initialization */
  ramp_update_time=0.0; /*initialization */
  flat_top_to_ramp_fact=1.0; /* initialize */
  while (exit_loop == -1) {
    if (fabs(flat_top_clock_cycles/ramp_update_clock_cycles - (float)
       ((int)(flat_top_clock_cycles/ramp_update_clock_cycles+0.5))) < 1.e-6) {
      /* set the flat top to ramp duration factor */
      flat_top_to_ramp_fact=flat_top_clock_cycles/ramp_update_clock_cycles;
      exit_loop=1;  /* jump out of the loop */

      /* double check that  the right ramp_update_time has been found */
      ramp_update_time=flat_top_clock_cycles/flat_top_to_ramp_fact* 
             WAVEFORM_RESOL/1.0e6;
      num_ramp_steps=(float)((int)(required_ramp_duration/
         ramp_update_time+0.5));
      ramp_duration=num_ramp_steps*ramp_update_time;
      if (fabs(ramp_duration-required_ramp_duration) > 1.e-6) {
        exit_loop=-1; /* carry on looking for right value */
        ramp_update_clock_cycles=ramp_update_clock_cycles+1;
        while_count=while_count+1;
      }
    } else {
      ramp_update_clock_cycles=ramp_update_clock_cycles+1;
      while_count=while_count+1;
    }
    if (while_count > 3000) {
      printf("%s: Error: Cannot match ramp durations\n", prog_name);
      return(-1.0);
    }
  }

  /* ========================================================================*/

  if ((num_ramp_steps < 10.0) && (ramp_update_time > 100.0e-6)) {
    printf("%s: Warning: ramp resolution is very low\n",prog_name);
  }

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if (write_flag == 1) {
    if ((fp=fopen(full_fname, "w")) == NULL) {
      printf("Error opening output data file %s\n", full_fname);
      exit(0);
    }
  }

  temp_dac=0.0;

  for (loop_no=0; loop_no<nloops; loop_no++) {

    /* do the blip ramp down */
    if (loop_no != 0) {
      /* descent portion of trapezoid */
      for (i=0; i<(int)num_ramp_steps; i++) {
        temp_dac=0.0;  /* default */
        temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
        if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
      }
    } else {
      /* correct for a bug in the waveform memory calculation which doesn't
         allow for more than 255 clock cycles of a given DAC level */
      num_clock_blocks=(int)((num_ramp_steps)/255.0)+1;
      for (i=0; i<(num_clock_blocks-1); i++) {
        if (write_flag == 1) fprintf(fp, "0   255\n");
      }
      if (write_flag == 1) fprintf(fp, "0   %d\n",
         (int)(num_ramp_steps)-(num_clock_blocks-1)*255);
    }

    /* insert zeros for duration of flat top portion of the read trapezoid */
    /* correct for a bug in the waveform memory calculation which doesn't
       allow for more than 255 clock cycles of a given DAC level */
    num_clock_blocks=(int)((flat_top_to_ramp_fact)/255.0)+1;
    for (i=0; i<(num_clock_blocks-1); i++) {
      if (write_flag == 1) fprintf(fp, "0   255\n");
    }
    if (write_flag == 1) fprintf(fp, "0   %d\n",(int)(flat_top_to_ramp_fact)-
        (num_clock_blocks-1)*255);

    /* do the blip ramp up */
    if (loop_no != nloops-1) {
      /* attack portion of trapezoid */
      for (i=0; i<(int)num_ramp_steps; i++) {
        temp_dac=0.0;  /* default */
        temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
        if (write_flag == 1) fprintf(fp, "%d   1\n",(int)(temp_dac));
      }
    } else {
      temp_dac=0.0;
      /* correct for a bug in the waveform memory calculation which doesn't
         allow for more than 255 clock cycles of a given DAC level */
      num_clock_blocks=(int)((num_ramp_steps)/255.0)+1;
      for (i=0; i<(num_clock_blocks-1); i++) {
        if (write_flag == 1) fprintf(fp, "0   255\n");
      }
      if (write_flag == 1) fprintf(fp, "0   %d\n",
           (int)(num_ramp_steps)-(num_clock_blocks-1)*255);
    }

    /* do the gap if echo_gap_tweak is set */
    if (echo_gap_tweak > 0.0) {
      num_gap_steps=ceil(echo_gap_tweak/ramp_update_time);
      /* correct for a bug in the waveform memory calculation which doesn't
         allow for more than 255 clock cycles of a given DAC level */
      num_clock_blocks=(int)((num_gap_steps)/255.0)+1;
      for (i=0; i<(num_clock_blocks-1); i++) {
        if (write_flag == 1) fprintf(fp, "%d   255\n",(int)(temp_dac));
      }
      if (write_flag == 1) fprintf(fp, "%d   %d\n", (int)(temp_dac), 
            (int)(num_gap_steps)-(num_clock_blocks-1)*255);
    }

  }
    
  if (write_flag == 1) fprintf(fp, "\n");  /* end file with a carriage return */
    
  if (write_flag == 1) fclose(fp);
 
  return(ramp_duration);
 
}

/*==========================================================================*/
double standard_flow_comp_duration(obj_grad,obj_duration,obj_ramp,max_grad,
   max_slew)
/*==========================================================================*/
/*   
    Routine to calculate the waveform duration Delta2 (and hence) Delta1
    in order to flow compensate an object waveform (e.g. slice select
    gradient or read out gradient). The required waveform is sketched out
    below, and is placed next to the object gradient to be compensated.
    Note that for a readout gradient, the waveform as drawn is time-
    reversed, and so the Delta2 delay must preceed the Delta1 delay.

    Typically this form of flow compensation would be used in a gradient
    echo sequence.
                                                  Delta2
                                                  ->  <-
                                              2*delta
                                            ->      <- 
                                  obj_ramp          __
                                   -> <-           /  \
          _ _ _ _  _ _ ______________             /gmax\
     ____/              obj_grad     \           /      \_____________
                       <------------> \         /   ->   <-
                        obj_duration   \  gmax /     delta
                                        \_____/
                                    ->   <-
                                     delta
                                         <--->
                                         Delta1

     INPUTS:
     =======

     obj_grad           gradient strength in g/cm of the gradient to be
                        compensated.

     obj_duration       duration of flat part of trapezoid to be flow
                        compensated (note that this would typically be
                        either half the slice select time in the case of
                        a slice select gradient, or half the read out
                        duration in the case of a read out gradient).

     obj_ramp           duration of the ramp of the object gradient to be
                        flow compensated.

     max_grad           maximum allowable gradient strength for the flow
                        compensating gradients. Note that the max_grad will
                        be used in order to minimise the echo time. If the
                        max_grad is too high then Delta1 and/or Delta2 will
                        be negative, and a lower value of max_grad should
                        be used.

     max_slew           maximum allowable slew rate in sec per g/cm (i.e.,
                        grise/gmax). Note that this gives delta=max_grad*
                        max_slew.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     duration of Delta2 in seconds. Delta1 may be calculated from Delta2
     using the equation:

       Delta1 - Delta2 = obj_grad/max_grad*(obj_duration+obj_ramp/2.0)
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp,max_grad,max_slew;
{
  double A_coeff,B_coeff,C_coeff,delta,Delta1,Delta2;

  delta=max_grad*max_slew;

  /* the value of Delta1 is the solution to:
      A_coeff*Delta1^2 + B_coeff*Delta1 + C_coeff = 0            */

  A_coeff=max_grad;

  B_coeff=3.0*max_grad*delta - obj_grad*(obj_duration*2.0+obj_ramp);

  C_coeff=obj_grad*obj_grad/8.0/max_grad*(obj_duration*2.0+obj_ramp)*
    (obj_duration*2.0+obj_ramp) -
    obj_grad/2.0*(obj_duration*2.0+obj_ramp)*(obj_duration+obj_ramp+3.5*delta)
    +obj_grad/2.0*(obj_duration*obj_duration+obj_ramp*obj_ramp/3.0
    +(obj_duration*2.0*obj_ramp)/2.0)+2.0*max_grad*delta*delta;

  Delta1=(-1.0*B_coeff+sqrt(B_coeff*B_coeff-4.0*A_coeff*C_coeff))/(2.0*A_coeff);

  Delta2=Delta1-obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  if (Delta2 < 0.0) Delta2=-1.0;  /* return a minus one if the delay is -ive */

  return(Delta2);

}

/*==========================================================================*/
double se_read_flow_comp_duration(obj_grad,obj_duration,obj_ramp,max_grad,
   max_slew,TE,t_180)
/*==========================================================================*/
/*   
    Routine to calculate the waveform duration Delta2 (and hence) Delta1
    in order to flow compensate in the read direction for a spin echo
    sequence. The required compensation waveform is sketched out below, and
    is placed either side of the 180 degree pulse to provide additional
    FID and STE crushing.

    Typically this form of flow compensation would be used in the read
    direction of a spin echo sequence.

               delta       delta               obj_ramp
              ->   <-     ->   <-               -> <-
                                                   _____________ _ _ _ _ _ _
    _______        _________          ____________/   obj_grad              \
           \      /  180    \        /             <----------->
            \gmax/           \ gmax /               obj_duration
             \__/  <----->    \____/
        ->   <-     t_180         ->   <-
        delta                      delta
            ->  <-            <--->
            Delta2            Delta1
                        <------------------------------------->
                                         TE/2
     INPUTS:
     =======

     obj_grad           gradient strength in g/cm of the gradient to be
                        compensated (typically the read out gradient).

     obj_duration       duration of flat part of trapezoid to be flow
                        compensated (note that this would typically be
                        half the read out duration).

     obj_ramp           duration of the ramp of the object gradient to be
                        flow compensated.

     max_grad           maximum allowable gradient strength for the flow
                        compensating gradients. Note that the max_grad will
                        be used in order to minimise the echo time. If the
                        max_grad is too high then Delta1 and/or Delta2 will
                        be negative, and a lower value of max_grad should
                        be used.

     max_slew           maximum allowable slew rate in sec per g/cm (i.e.,
                        grise/gmax). Note that this gives delta=max_grad*
                        max_slew.

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     duration of Delta2 in seconds. Delta1 may be calculated from Delta2
     using the equation:

       Delta1 - Delta2 = obj_grad/max_grad*(obj_duration+obj_ramp/2.0)
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp,max_grad,max_slew,TE,t_180;
{
  double A_coeff,B_coeff,C_coeff,delta,Delta1,Delta2;

  delta=max_grad*max_slew;

  /* the value of Delta1 is the solution to:
      A_coeff*Delta1^2 + B_coeff*Delta1 + C_coeff = 0            */

  A_coeff=max_grad;

  B_coeff=-1.0*obj_grad*(obj_duration+obj_ramp/2.0)+max_grad*t_180+
       3.0*max_grad*delta;

  C_coeff=max_grad/2.0*(obj_grad/max_grad*(obj_duration+obj_ramp/2.0))*
    (obj_grad/max_grad*(obj_duration+obj_ramp/2.0))
  - (max_grad*TE/2.0+max_grad*t_180/2.0+1.5*max_grad*delta) 
       *(obj_grad/2.0/max_grad*(obj_duration*2.0+obj_ramp))
  + (obj_duration*2.0)*(obj_duration*2.0)*obj_grad/8.0 + 
   obj_ramp*obj_ramp*obj_grad/6.0 + obj_grad*obj_duration*obj_ramp/2.0 
  + max_grad*t_180*delta + 2.0*max_grad*delta*delta;

  Delta1=(-1.0*B_coeff+sqrt(B_coeff*B_coeff-4.0*A_coeff*C_coeff))/(2.0*A_coeff);

  Delta2=Delta1-obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  if (Delta2 < 0.0) Delta2=-1.0;  /* return a minus one if the delay is -ive */

  return(Delta2);

}

/*==========================================================================*/
double se_slice_flow_comp_duration1(obj_grad,obj_duration,obj_ramp,max_grad,
   max_slew,TE,t_180,tspoil)
/*==========================================================================*/
/*   
    Routine to calculate the waveform duration Delta2 (and hence) Delta1
    in order to flow compensate in the slice direction for a spin echo
    sequence. The required compensation waveform is sketched out below, and
    is placed on the side preceding the 180 degree pulse.

    Typically this form of flow compensation would be used in the slice
    direction of a spin echo sequence. It should be used in conjunction with
    se_slice_flow_comp_duration2()

                                                     Delta2
                                                     ->  <-
                             obj_ramp                  __|
                              -> <-                   /   \ _ _ _
       _ _ _ _ _ _ _____________                     /   |
      /               obj_grad  \        ___________/       
                   <-----------> \      /           <->  |
                    obj_duration  \gmax/           delta
                                   \__/<->               | tspoil
                                       delta               +t_180
                                  ->  <-                  <------>
                                  Delta1
                   <--------------------------------------------->
                                    TE/2
     INPUTS:
     =======

     obj_grad           gradient strength in g/cm of the gradient to be
                        compensated (typically the read out gradient).

     obj_duration       duration of flat part of trapezoid to be flow
                        compensated (note that this would typically be
                        half the slice select duration).

     obj_ramp           duration of the ramp of the object gradient to be
                        flow compensated.

     max_grad           maximum allowable gradient strength for the flow
                        compensating gradients. Note that the max_grad will
                        be used in order to minimise the echo time. If the
                        max_grad is too high then Delta1 and/or Delta2 will
                        be negative, and a lower value of max_grad should
                        be used.

     max_slew           maximum allowable slew rate in sec per g/cm (i.e.,
                        grise/gmax). Note that this gives delta=max_grad*
                        max_slew.

     t_180              the duration of the 180 degree pulse (in secs).

     TE                 the TE of the sequences (in secs).

     tspoil             the duration of the gradient spoiler time before the
                        180 degree pulse (in secs).

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     duration of Delta2 in seconds. Delta1 may be calculated from Delta2
     using the equation:

       Delta1 - Delta2 = obj_grad/max_grad*(obj_duration+obj_ramp/2.0)
                                 -gmax*grate/2.0
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, March 1998
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp,max_grad,max_slew,TE,t_180,tspoil;
{
  double A_coeff,B_coeff,C_coeff,delta,Delta1,Delta2;
  double small_a, small_b, small_c, small_d, small_e, small_f, small_g;

  delta=max_grad*max_slew;

  /* the value of Delta1 is the solution to:
      A_coeff*Delta1^2 + B_coeff*Delta1 + C_coeff = 0            */

  small_a=-max_grad/2.0;
  small_b=-max_grad/2.0;
  small_c=-1.0*max_grad*(obj_duration+obj_ramp+3.0*delta/2.0);
  small_d=max_grad*(TE/2.0-t_180/2.0-delta/2.0-tspoil);
  small_e=0.0;
  small_f=obj_grad*(obj_duration*obj_duration/2.0+obj_ramp*obj_ramp/6.0+
     obj_duration*obj_ramp/2.0)-max_grad*delta*delta/6.0-
     max_grad*(2.0*obj_duration+2.0*obj_ramp-TE/2.0+2.0*delta+t_180/2.0
     +tspoil)*delta/2.0;
  small_g=obj_grad/max_grad*(obj_duration+obj_ramp/2.0)-max_grad*max_slew/2.0;

  A_coeff=small_a+small_b+small_e;

  B_coeff=small_c+small_d-small_e*small_g-2.0*small_g*small_b;

  C_coeff=small_b*small_g*small_g-small_g*small_d+small_f;

  Delta1=(-1.0*B_coeff+sqrt(B_coeff*B_coeff-4.0*A_coeff*C_coeff))/(2.0*A_coeff);

  Delta2=Delta1-(obj_grad/max_grad*(obj_duration+obj_ramp/2.0)
                 -max_grad*max_slew/2.0);

  if (Delta2 < 0.0) Delta2=-1.0;  /* return a minus one if the delay is -ive */

  return(Delta2);

}

/* PJPJ */
/*==========================================================================*/
double se_slice_flow_comp_duration2(obj_grad,obj_duration,obj_ramp,max_grad,
   max_slew,TE,t_180,tspoil)
/*==========================================================================*/
/*   
    Routine to calculate the waveform duration Delta2 (and hence) Delta1
    in order to flow compensate in the slice direction for a spin echo
    sequence. The required compensation waveform is sketched out below, and
    is placed on the side following the 180 degree pulse.

    Typically this form of flow compensation would be used in the slice
    direction of a spin echo sequence. It should be used in conjunction with
    se_slice_flow_comp_duration1()

                    tspoil     Delta1
                     <-->       <->
               gamp1 ___    |   ___ gamp1
                        \______/   \
                  gslice <---->     \
                          t_180      \
   _ _ _ _ _ _ _ _ _ |                \_________    _______________
                            |    ->    <-       \  /
                     |             delta         \/ gamp2
                            |                 ->  <-
                                              Delta2
                            <---------------------------->
                                    TE/2
     INPUTS:
     =======

     max_slew           maximum allowable slew rate in sec per g/cm (i.e.,
                        grise/gmax). Note that this gives delta=gamp1*max_slew.

     t_180              the duration of the 180 degree pulse (in secs).

     TE                 the TE of the sequences (in secs).

     tspoil             the duration of the gradient spoiler time before the
                        180 degree pulse (in secs).

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     duration of Delta2 in seconds. Delta1 may be calculated from Delta2
     using the equation:

 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, March 1998
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp,max_grad,max_slew,TE,t_180,tspoil;
{
  double A_coeff,B_coeff,C_coeff,delta,Delta1,Delta2;
  double small_a, small_b, small_c, small_d, small_e, small_f, small_g;

  delta=obj_grad*max_slew;

  /* the value of Delta1 is the solution to:
      A_coeff*Delta1^2 + B_coeff*Delta1 + C_coeff = 0            */

  small_a=-max_grad/2.0;
  small_b=-max_grad/2.0;
  small_c=-1.0*max_grad*(obj_duration+obj_ramp+3.0*delta/2.0);
  small_d=max_grad*(TE/2.0-t_180/2.0-delta/2.0-tspoil);
  small_e=0.0;
  small_f=obj_grad*(obj_duration*obj_duration/2.0+obj_ramp*obj_ramp/6.0+
     obj_duration*obj_ramp/2.0)-max_grad*delta*delta/6.0-
     max_grad*(2.0*obj_duration+2.0*obj_ramp-TE/2.0+2.0*delta+t_180/2.0
     +tspoil)*delta/2.0;
  small_g=obj_grad/max_grad*(obj_duration+obj_ramp/2.0)-max_grad*max_slew/2.0;

  A_coeff=small_a+small_b+small_e;

  B_coeff=small_c+small_d-small_e*small_g-2.0*small_g*small_b;

  C_coeff=small_b*small_g*small_g-small_g*small_d+small_f;

  Delta1=(-1.0*B_coeff+sqrt(B_coeff*B_coeff-4.0*A_coeff*C_coeff))/(2.0*A_coeff);

  Delta2=Delta1-(obj_grad/max_grad*(obj_duration+obj_ramp/2.0)
                 -max_grad*max_slew/2.0);

  if (Delta2 < 0.0) Delta2=-1.0;  /* return a minus one if the delay is -ive */

  return(Delta2);

}
