int safety_check(grad_r_rms, grad_p_rms, grad_s_rms)
double grad_r_rms, grad_p_rms, grad_s_rms;
{

/****************************************************************************
* This section of code does a rough calculation to check that the gradient
* amplifiers and the gradient coil are not being overstressed.  It uses the
* values for the cut off safety limits as recommended by the gradient coil
* manufacturers:
*
* SIEMENS (main): values estimated.
* MAGNEX (head): Fred Goldie and Simon Pittard (01235) 534488. 
*
* Written by Peter Jezzard, Oxford Aug 1998
*****************************************************************************/

int debug_flag=0;   /* debug flag, 1=messages on, 0=messages off */
int fail_flag=1;    /* flag to indicate duty cycle failure (1 or -1) */

/*****************************************************************************
*  This section describe the variables used in the duty cycle calculation.
*****************************************************************************/

float xgrad_amp_fs=250.0; /* maximum amps at full scale for x amplifier */
float ygrad_amp_fs=250.0; /* maximum amps at full scale for y amplifier */
float zgrad_amp_fs=250.0; /* maximum amps at full scale for z amplifier */
float max_rms_x_curr=0.0; /* max continuous rms current on x gradient (amps) */
float max_rms_y_curr=0.0; /* max continuous rms current on y gradient (amps) */
float max_rms_z_curr=0.0; /* max continuous rms current on z gradient (amps) */
float max_x_grad=0.0; /* max x gradient (gauss/cm) */
float max_y_grad=0.0; /* max y gradient (gauss/cm) */
float max_z_grad=0.0; /* max z gradient (gauss/cm) */
float x_coil_resistance=1.0;  /* resistance of x windings (ohms) */
float y_coil_resistance=1.0;  /* resistance of y windings (ohms) */
float z_coil_resistance=1.0;  /* resistance of z windings (ohms) */
float x_cooling_efficiency=1.0;  /* cooling efficiency of x windings */
float y_cooling_efficiency=1.0;  /* cooling efficiency of y windings */
float z_cooling_efficiency=1.0;  /* cooling efficiency of z windings */
float irms_x=1.0;             /* rms current on gradient axis */
float irms_y=1.0;             /* rms current on gradient axis */
float irms_z=1.0;             /* rms current on gradient axis */
float grad_x_rms=1.0;         /* rms gradient in physical x direction */
float grad_y_rms=1.0;         /* rms gradient in physical y direction */
float grad_z_rms=1.0;         /* rms gradient in physical z direction */
float max_coil_heat_pwr=0.0;  /* maximum allowed heating in coil (Watts) */
float tot_power=1.0;          /* total heating power generated by sequence */
char test_string[MAXSTR];	/* temporary string used to figure out gcoil */
char prog_name[MAXSTR];   /* program name */
double m11,m12,m13,m21,m22,m23,m31,m32,m33;  /* euler rot matrix elements */
double tm11,tm12,tm13,tm21,tm22,tm23,tm31,tm32,tm33;  /* transposed elements */
double cosang1,sinang1,cosang2,sinang2,cosang3,sinang3;  /* cos, sin calcs */
double tol; /* tolerance in rotation matrix */

sprintf(prog_name,"safety_check.h");

/*****************************************************************************
*  This section sets the variables for the individual gradient coils used at 
*  FMRIB. 'head'=MAGNEX SGRAD III 595/379, 'main'=SIEMENS AS25 950/600.
*****************************************************************************/

  sprintf(test_string,"main");  /* check to see if main coil is being used */
  if (strcmp(gcoil,test_string) == 0) {
    max_x_grad=2.5;
    max_y_grad=2.5;
    max_z_grad=2.5;
    max_rms_x_curr=150.0;
    max_rms_y_curr=150.0;
    max_rms_z_curr=150.0;
    x_cooling_efficiency=1.000;
    y_cooling_efficiency=1.000;
    z_cooling_efficiency=1.000;
    x_coil_resistance=0.15;
    y_coil_resistance=0.15;
    z_coil_resistance=0.15;
    max_coil_heat_pwr=6000.0;
  }

  sprintf(test_string,"head");  /* check to see if head coil is being used */
  if (strcmp(gcoil,test_string) == 0) {
    max_x_grad=3.7;
    max_y_grad=3.7;
    max_z_grad=3.7;
    max_rms_x_curr=90.0;
    max_rms_y_curr=90.0;
    max_rms_z_curr=90.0;
    x_cooling_efficiency=0.700;
    y_cooling_efficiency=1.875;
    z_cooling_efficiency=0.395;
    x_coil_resistance=0.07;
    y_coil_resistance=0.075;
    z_coil_resistance=0.075;
    max_coil_heat_pwr=1126.0;
  }

/*****************************************************************************
*  Check that a valid gcoil value has been found
*****************************************************************************/

  if ((max_x_grad == 0.0) || (max_y_grad == 0.0) || (max_z_grad == 0.0)) {
    printf("%s: Invalid value of gcoil specified\n", prog_name);
    return -1;
  }

/*****************************************************************************
*  Tip the read, phase and slice gradients onto their correct physical axes
*****************************************************************************/

  cosang1=cos(M_PI/180.0*psi);  /* M_PI gets picked up from math.h */
  sinang1=sin(M_PI/180.0*psi);

  cosang2=cos(M_PI/180.0*phi);
  sinang2=sin(M_PI/180.0*phi);

  cosang3=cos(M_PI/180.0*theta);
  sinang3=sin(M_PI/180.0*theta);

  m11 = (sinang2*cosang1 - cosang2*cosang3*sinang1);
  m12 = (-1.0*sinang2*sinang1 - cosang2*cosang3*cosang1);
  m13 = (sinang3*cosang2);

  m21 = (-1.0*cosang2*cosang1 - sinang2*cosang3*sinang1);
  m22 = (cosang2*sinang1 - sinang2*cosang3*cosang1);
  m23 = (sinang3*sinang2);

  m31 = (sinang1*sinang3);
  m32 = (cosang1*sinang3);
  m33 = (cosang3);

  tol = 1.0e-14;

  if (fabs(m11) < tol) m11 = 0.0;
  if (fabs(m12) < tol) m12 = 0.0;
  if (fabs(m13) < tol) m13 = 0.0;
  if (fabs(m21) < tol) m21 = 0.0;
  if (fabs(m22) < tol) m22 = 0.0;
  if (fabs(m23) < tol) m23 = 0.0;
  if (fabs(m31) < tol) m31 = 0.0;
  if (fabs(m32) < tol) m32 = 0.0;
  if (fabs(m33) < tol) m33 = 0.0;

  /* Transpose matrix */

  tm11 = m11;
  tm21 = m12;
  tm31 = m13;
  tm12 = m21;
  tm22 = m22;
  tm32 = m23;
  tm13 = m31;
  tm23 = m32;
  tm33 = m33;

  /* Apply matrix elements to logical axes */

  grad_x_rms = (grad_r_rms*tm11 + grad_p_rms*tm12 + grad_s_rms*tm13);
  grad_y_rms = (grad_r_rms*tm21 + grad_p_rms*tm22 + grad_s_rms*tm23);
  grad_z_rms = (grad_r_rms*tm31 + grad_p_rms*tm32 + grad_s_rms*tm33);

  /* make sure the values are positive */
  grad_x_rms=fabs(grad_x_rms);
  grad_y_rms=fabs(grad_y_rms);
  grad_z_rms=fabs(grad_z_rms);

/*****************************************************************************
*  Convert the RMS gradients in to RMS currents
*****************************************************************************/

  irms_x=grad_x_rms/max_x_grad*xgrad_amp_fs;
  irms_y=grad_y_rms/max_y_grad*ygrad_amp_fs;
  irms_z=grad_z_rms/max_z_grad*zgrad_amp_fs;

/*****************************************************************************
*  Now calculate the heating power that is induced in the coil by the sequence
*****************************************************************************/

  tot_power=pow(irms_x,2.0)*x_coil_resistance/x_cooling_efficiency+
            pow(irms_y,2.0)*y_coil_resistance/y_cooling_efficiency+
            pow(irms_z,2.0)*z_coil_resistance/z_cooling_efficiency;

/*****************************************************************************
*  Do the safety check itself
*****************************************************************************/

  fail_flag=1;
  if (tot_power > max_coil_heat_pwr) {
    printf("%s: Excess duty cycle. Increase FOV, thk or TR\n", prog_name);
    fail_flag=-1;
  }

  if (irms_x > max_rms_x_curr) {
    printf("%s: Excess gradient duty cycle on physical x axis\n", prog_name);
    fail_flag=-1;
  }

  if (irms_y > max_rms_y_curr) {
    printf("%s: Excess gradient duty cycle on physical y axis\n", prog_name);
    fail_flag=-1;
  }

  if (irms_z > max_rms_z_curr) {
    printf("%s: Excess gradient duty cycle on physical z axis\n", prog_name);
    fail_flag=-1;
  }

/*****************************************************************************
*  Print out stuff if the debug_flag is on
*****************************************************************************/

  if ((debug_flag == 1) || (fail_flag == -1)) {
    printf("%s: gcoil = '%s'\n",prog_name,gcoil);
    printf("%s: X axis RMS current = %f A\n",prog_name,irms_x);
    printf("%s: Max allowed X axis RMS current = %f A\n",prog_name,
        max_rms_x_curr);
    printf("%s: Y axis RMS current = %f A\n",prog_name,irms_y);
    printf("%s: Max allowed Y axis RMS current = %f A\n",prog_name,
        max_rms_y_curr);
    printf("%s: Z axis RMS current = %f A\n",prog_name,irms_z);
    printf("%s: Max allowed Z axis RMS current = %f A\n",prog_name,
        max_rms_z_curr);
    printf("%s: Total power generated in coil = %f W\n",prog_name,tot_power);
    printf("%s: Max allowed power generated in coil = %f W\n",prog_name,
        max_coil_heat_pwr);
  }

/*****************************************************************************
*  Otherwise return an OK signal
*****************************************************************************/

  return fail_flag;

}
