#ifndef LINT
static char SCCSid[] = "@(#)flash.c 11.1 05/07/96 Copyright (c) 1994-1996 Varian Assoc.,Inc. All Rights Reserved";
#endif
/***********************************************************************
    VENOGRAM.C

    3D Gradient-echo fast imaging sequence.                     

    VERSION 2.0        6/94/94  
    Revisions:
      2.1   2/96    Fixed rfspoiling for nv=0
      
    Extended to 3D, by Patricia Figueiredo, May 1999
    Altered for long TE T2* weighted angiography, Peter Jezzard, March 2000
    flow compensated in all three directions

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
        __                                                        __ 
       |                                                            |
  RF   |----- [ p1 ] ----------------------- |** ACQ **| --      ---|
       |                                                            |
       |         < - - - - - - - te - - - - - - - >                 |
       |       ____       __   __  ___                          ___ |
  GSS  |______/////\     /  \_/__\/___\________________________/___\|
       |<---->      \___/ gssr\__/\___/                        \___/|
       |  predelay  <--> tref                                       | nv
       |                       __  ___                          ___ |
  GPE  |______________________/__\/___\________________________/___\|
       |                      \__/\___/                        \___/|
       |                                                            |
       |                                __       ____________       |
  GRO  |_______________________________/  \     /////////////\______|
       |                                   \___/<tspoil>            |
       |                                                            |
       |< - - - - - - - - - - - tr - - - - - - - - - - - - - - - - >|
       |__                                                        __|

*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*/

#include "standard.h"
#include "math.h"
#include "group.h"
#include "sequence_utilities.h"
#include "venogram_utilities.h"
#define OFFSETDELAY 26e-6
#define RFSPOILDELAY 50e-6

/*==========================================================================*
  define the pulse sequence variables outside the pulse sequence so that
  they can be seen as global variables by the safety monitoring code
*==========================================================================*/
    double  predelay,seqtime,tref,risetref,grise,tep,rftep,delta1,delta2;
    double  trefmin,trormin,tssrmin,tpemin,temin,tA,pe_A,pe_B,pe_C,t0;
    double  agro,agss,sgpe,grate,gssrint,grorint,gpeint;     
    double  tpe2min,sgpe2,gpe2int,temp_sgpe,temp_sgpe2;
    double  reffreq,nextphase,spoilfreqs[1024],pw_gss_d,extra_te_error;
    double  pw_gro,pw_gss,wave_time_read, wave_time_slice,true_wave_time;
    double  greadfc,gslicefc;
    int     i,image=0,dummy_ssc;
    char    shape_name[MAXSTR],flow_comp[MAXSTR],err_string[MAXSTR];
    char    gcoil[MAXSTR], pefc[MAXSTR];

#include "venogram_safety_calc.h"

pulsesequence()
{
    double  create_trap_gradient(), create_ramp_gradient();
    double  create_venogram_slice_flow_comp(),create_venogram_read_flow_comp();

    initparms_sis();
    loop_check();

    /*************************************************************************/
    /* get the global parameters */
    /*************************************************************************/
    P_getreal(GLOBAL,"grise",&grise,1); /* grad risetime to full strength */
    if ((grise < 0.0) || (grise > 0.002)) {
      sprintf(err_string,"%s: impropper grise time defined in global file\n",
                       seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }
    P_getreal(GLOBAL,"tep",&tep,1); /* small tweaker time (us) to allow for
                            the gradient propagation time, causes echoes to
                            appear symmetrically about echo center line */
    tep=tep/1e6;
    P_getreal(GLOBAL,"rftep",&rftep,1); /* small tweaker time (us) to allow for
                            the gradient propagation time, w.r.t. RF pulses */
    rftep=rftep/1e6;

    /*************************************************************************/
    /* Get local parameters */
    /*************************************************************************/
    getstr("gcoil",gcoil);          /* gradient coil in use */
    getstr("flow_comp",flow_comp);  /* flow comp flag 'y'=fc. 'n'=no fc. */
    getstr("pefc",pefc);  /* p.e. flow comp flag 'y'=fc. 'n'=no fc. */
    image = (int)(getval("image")); /* trigger and counter bit pattern */
    /* bit 1 = wait for ticks triggers before beginning scan */
    /* bit 2 = puts out high on sp3 for duration of volume */
    /* bit 3 = reserved for epi blips on/off */

    /*************************************************************************/  
    /* Calculate parameters */
    /*************************************************************************/
    gcrush=gmax*0.5; /* hard wire a crusher pulse in to the sequence */   
    grate = grise/gmax;
    agro = fabs(gro);
    agss = fabs(gss);
    if (agss == 0.0) {
        sprintf(err_string,"%s: Zero slice select gradient\n",seqfil);
        printf(err_string);
        log_error(err_string);
        abort(1);
    }

    /* figure out a sensible number of dummy scans */
    dummy_ssc=0;  /* default */
    if (nv > 0.0) {
      dummy_ssc=(int)((10.0*exp(-tr/0.1)+6.0*exp(-tr/0.6))+0.5)*2;
    }

    /*******************************************************/
    /* create the shaped gradient waveforms                */
    /*******************************************************/
    sprintf(shape_name,"gss");
    pw_gss=create_trap_gradient(fabs(gss),shape_name,rof1+p1+rof2,grate,ix);

    pw_gro=0.0; /* initialization value */
    wave_time_read=0.0; /* initialization value */
    wave_time_slice=0.0; /* initialization value */
    greadfc=0.0; /* initialization value */
    gslicefc=0.0; /* initialization value */

    pw_gro=grate*fabs(gro);
    sprintf(shape_name,"read");
    greadfc=create_venogram_read_flow_comp(fabs(gro),shape_name,tspoil+at/2,
           at/2,pw_gro,0.8*gmax,grate,&wave_time_read);
    if (greadfc < 0.0) {
      sprintf(err_string,"%s: TE too short. Min TE = %f ms\n",seqfil,
                  (te-greadfc)*1e3);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }

    sprintf(shape_name,"slice");
    gslicefc=create_venogram_slice_flow_comp(fabs(gss),shape_name,p1/2,
           agss*grate,0.8*gmax,grate,&wave_time_slice);

    /*******************************************************/    
    /* GRADIENT*TIME integrals for slice selection, readout and phase encoding */
    /*******************************************************/
    grorint = agro*(at/2.0 + tspoil + pw_gro/2.0);
    gssrint = agss*(p1/2.0 + rof1 + agss*grate/2.0);
    gpeint = B0/(lpe*sfrq*1e6);
    gpe2int = B0/(lpe2*sfrq*1e6);

    /*******************************************************
    * Calculate minimum time for phase-encoding, slice refocussing
    * and readout prefocussing block.  Trapezoidal if GMAX is exceeded.
    * RISETREF is the gradient fall time for the PE block.
    *******************************************************/
    if (grise*gmax >= gssrint)
	tssrmin = sqrt(grate*gssrint);
    else
	tssrmin = gssrint/gmax;

    if (grise*gmax >= grorint)
        trormin = sqrt(grate*grorint);
    else
	trormin = grorint/gmax;

    if (grise*gmax >= gpeint*nv/2.0)
        tpemin = sqrt(grate*gpeint*nv/2.0);
    else
	tpemin = gpeint*nv/2.0/gmax;

    if (grise*gmax >= (gssrint + gpe2int*nv2/2.0) )
        tpe2min = sqrt(grate*(gssrint + gpe2int*nv2/2.0) );
    else
        tpe2min = (gssrint + gpe2int*nv2/2.0)/gmax;

    /* Minimum refocussing block time is the longest of calculated minima */
    trefmin = trormin;
    trefmin = (trefmin > tpemin) ? trefmin : tpemin;
    trefmin = (trefmin > tpe2min) ? trefmin : tpe2min;
    
    /* If TREFMIN implies trapezoidal PE block, mimimum RISETREF
     is RISETIME, otherwise minimum RISETREF is TREFMIN */
    risetref = (trefmin > grise) ? grise : trefmin;

    /*******************************************************/
    /* Check minimum TE ***********************************/
    /*******************************************************/
    temin=0.0; /* default value */

    /*******************************************************
    * Calculate TREF
    * TREF is the refocussing block length, not including gradient fall
    * time.  TREF is scaled up to a maximum value consistent with TE
    *******************************************************/
    tref = trefmin + 0.5*(te - temin);
    tref = (tref > 2.0*trefmin) ? 2.0*trefmin : tref;

    /*******************************************************/
    /* Calculate phase encode gradients, adjusting for integral dac value */
    /*******************************************************/
    sgpe = gpeint/tref;
    sgpe = gmax/gradstepsz*floor(sgpe*gradstepsz/gmax + 0.5);
    sgpe2 = gpe2int/tref;
    sgpe2 = gmax/gradstepsz*floor(sgpe2*gradstepsz/gmax + 0.5);
    /* If nv is 0 we want a read profile so turn off all phase-encode ***/
    temp_sgpe=sgpe;
    temp_sgpe2=sgpe2;
    if (nv == 0) {
      sgpe = 0.0;
      sgpe2 = 0.0;
    }

    /*******************************************************
    * Calculate new TREF
    * A single solution for tref is not guaranteed to exist for the two
    * values of sgpe and sgpe2.  A good compromise is the geometric mean.
    * If this tref results in a distortion in either PE direction of more
    * than 1%, issue an advisory message.
    *******************************************************/
    tref = sqrt((gpeint/temp_sgpe)*(gpe2int/temp_sgpe2));    
    if (nth2D == 1  &&  (fabs(1.0 - temp_sgpe*tref/gpeint) > 0.01  ||
      fabs(1.0 - temp_sgpe2*tref/gpe2int) > 0.01)) {
        sprintf(err_string,"%s: NOTE: lpe distortion =%6.2f%% lpe2 =%6.2f%%\n",
         seqfil,100.0-temp_sgpe*tref/gpeint*100,100.0
         -temp_sgpe2*tref/gpe2int*100.0);
        printf(err_string);
        log_error(err_string);
    }

    /*******************************************************/
    /* Calculate refocussing gradients with new TREF *******/
    /*******************************************************/
    gror = -grorint/tref*gro/fabs(gro);
    gssr = -gssrint/tref*gss/fabs(gss);
 
    /*******************************************************/
    /* Calculate extra_te_error: */
    /* account for case of rftep being longer than the gradient ramp time */
    /*******************************************************/
    pw_gss_d = agss*grate - rftep;
    extra_te_error = 0.0;
    if (pw_gss_d < 0.0) {
      pw_gss_d = 0.0;
      extra_te_error = rftep-agss*grate;
    }

    /*******************************************************/
    /* Determine PE block gradient fall time with new TREF */    
    /*******************************************************/
    risetref = fabs(gror);
    risetref = risetref > fabs(temp_sgpe*nv/2.0) ? 
                      risetref : fabs(temp_sgpe*nv/2.0);
    risetref = risetref > fabs(gssr) + fabs(temp_sgpe2*nv2/2.0) ?
               risetref : fabs(gssr) + fabs(temp_sgpe2*nv2/2.0);
    risetref *= grate;
 
    sprintf(shape_name,"gdephase");
    risetref=create_ramp_gradient(risetref/grise*gmax,shape_name,1,grate,ix);
    risetref=create_ramp_gradient(risetref/grise*gmax,shape_name,-1,grate,ix);

    /* solve equations for p.e. direction flow comp */
    t0=te-(p1/2.0+rof2+pw_gss_d+wave_time_slice);
    pe_A=1.0;
    pe_B=5.0*risetref + 2.0*(tref-risetref);
    pe_C=6.0*risetref*risetref+9.0/2.0*(tref-risetref)*risetref+
           (tref-risetref)*(tref-risetref)/2.0-t0*((tref-risetref)+risetref);

    delta1=(-pe_B+sqrt(pe_B*pe_B-4.0*pe_A*pe_C))/(2.0*pe_A);
    delta2=delta1+risetref+(tref-risetref);

    /*******************************************************/
    /* Calculate tA, the delay between refocussing and readout */    
    /*******************************************************/
    tA=0.0; /* default value */
    tA=te-(p1/2+rof2+pw_gss_d+wave_time_slice+(wave_time_read-at/2));
    if (pefc[0] == 'y') {
      tA=tA-(4.0*risetref+delta1+delta2);
    } else {
      tA=tA-(tref+risetref);
    }
    if (tA < 0.0) {
        sprintf(err_string,"%s: TE too short. Min TE = %f\n",seqfil,te-tA);
        printf(err_string);
        log_error(err_string);
        abort(1);
    }

    /*******************************************************/
    /* Relaxation delay and minimum TR */
    /*******************************************************/
    seqtime = te + p1/2.0 + rof1 + tep +extra_te_error + at/2.0 + agss*grate  
            + pw_gro + tref + risetref;
    if (rfspoil[0] == 'y'  ||  rfspoil[0] == 'Y')
	seqtime += RFSPOILDELAY;
    if (tr < seqtime) {
        sprintf(err_string,"%s: TR too short. Min TR = %f\n",seqfil,seqtime);
        printf(err_string);
        log_error(err_string);
        abort(1);
    }
    predelay = tr - seqtime;

    /*******************************************************
    * Calculation of frequencies to get phase shifts for optional
    * RF spoiling.  Phase shifts are obtained by hopping to variable
    * frequency for fixed time.
    * Ref:  Zur, Y., Magn. Res. Med., 21, 251, (1991).
    *******************************************************/
    if ((rfspoil[0] == 'y' || rfspoil[0] == 'Y') && nv > 0) {
	reffreq = resto + sfrq*1e6/B0*gro*pro;
        for (i=1; i<= (int)(nv+0.5); i++) {
	    nextphase = (double)i/2.0*((double)i-1.0)*rfphase;
	    nextphase = nextphase - 360.0*floor(nextphase/360.0);
	    spoilfreqs[i] = reffreq + nextphase/RFSPOILDELAY/360.0;
        }
        create_offset_list(spoilfreqs,nv,TODEV,0);
    }

    if (venogram_safety_calc() == -1) {
      sprintf(err_string,"%s: Failed duty cycle safety check\n",seqfil);
      printf(err_string);
      log_error(err_string);
      abort(1);
    }

    if (nv2 == 1) sgpe2=0.0;  /* to allow single slice (2D) mode */

    /* PULSE SEQUENCE *************************************/

    /* Phase cycle ****************************************/
    mod2(ct,v3);          /* v3: 0  1  0  1  0  1  0  1 */
    dbl(v3,v3);           /* v3: 0  2  0  2  0  2  0  2 */
    hlv(ct,v4);           /* v4: 0  0  1  1  2  2  3  3 */
    mod2(v4,v4);          /* v4: 0  0  1  1  0  0  1  1 */
    add(v3,v4,v1);        /* v1: 0  2  1  3  0  2  1  3 */
    assign(v1,oph);      /* oph: 0  2  1  3  0  2  1  3 */

    assign(one,v7);  /* phase encode increment size */
    status(A);
    sp4off();
    obspower(tpwr1);

    sp2on();
    delay(0.00005);  /* 50 us trigger to indicate start of volume acq */
    sp2off();

    delay(d1);

    if ((image & 1) == 1) xgate(ticks);
    if ((image & 2) == 2) sp3on(); else sp3off();

    initval(one,v9);  /* steady state flag */
    initval(dummy_ssc,v10);  /* steady state counter */

    /* Compressed phase-encode loop *************************/
    peloop(seqcon[3],nv2,v11,v12);    
     peloop(seqcon[2],nv,v5,v6);

      poffset(pss[0],gss);

      /* Relaxation delay ***********************************/       
      ifzero(v10);
        assign(zero,v9);
      endif(v10);
      sub(v10,one,v10);  /* steady state stuff (part 1) */

      delay(predelay);
       
      /* RF pulse *******************************************/ 
      sp1on(); /* put out trigger on sp1 every spin excitation */
      obl_shapedgradient("","","gss",pw_gss,0.0,0.0,gss,1,NOWAIT);
      delay(agss*grate+rftep);
      sp4on();
      shapedpulse(p1pat,p1,v1,rof1,rof1);
      sp4off();
      delay(pw_gss_d);
      sp1off();

      /* play out the slice direction flow compensation */
      obl_shapedgradient("","","slice",wave_time_slice,
          0.0,0.0,gslicefc,1,WAIT);

      /* do the slice and phase direction phase encoding */
      if (pefc[0] == 'y') {
        pe2_shapedgradient("gdephase_up",risetref,0.0,-sgpe*nv/2.0,
            sgpe2*nv2/2.0-sgpe2/2.0,sgpe,-sgpe2,v6,v12);
        delay(delta1);
        pe2_shapedgradient("gdephase_dn",risetref,0.0,-sgpe*nv/2.0,
            sgpe2*nv2/2.0-sgpe2/2.0,sgpe,-sgpe2,v6,v12);
        pe2_shapedgradient("gdephase_up",risetref,0.0,sgpe*nv/2.0,
            -sgpe2*nv2/2.0+sgpe2/2.0,-sgpe,sgpe2,v6,v12);
        delay(delta2);
        pe2_shapedgradient("gdephase_dn",risetref,0.0,sgpe*nv/2.0,
            -sgpe2*nv2/2.0+sgpe2/2.0,-sgpe,sgpe2,v6,v12);
      } else {
        pe2_shapedgradient("gdephase_up",risetref,0.0,sgpe*nv/2.0,
            -sgpe2*nv2/2.0+sgpe2/2.0,-sgpe,sgpe2,v6,v12);
        delay(tref-risetref);
        pe2_shapedgradient("gdephase_dn",risetref,0.0,sgpe*nv/2.0,
            -sgpe2*nv2/2.0+sgpe2/2.0,-sgpe,sgpe2,v6,v12);
      }

      delay(tA);

      poffset(pro,gro);

      /* play out the read direction flow compensation */
      obl_shapedgradient("read","","",wave_time_read,
          greadfc,0.0,0.0,1,NOWAIT);
      delay(wave_time_read-at-pw_gro);
      /* Acquire echo ***************************************/
      delay(tep);
      ifzero(v9);   /* steady state stuff (part 2) */
        acquire(np,1.0/sw);
      elsenz(v9);
        delay(at);
      endif(v9);

      /* rewinder stuff */
      pe2_shapedgradient("gdephase_up",risetref, 0.0,-sgpe*nv/2.0,
          gcrush+sgpe2*nv2/2.0-sgpe2/2.0,sgpe,-sgpe2,v6,v12);
      delay(tref-risetref);
      pe2_shapedgradient("gdephase_dn",risetref, 0.0,-sgpe*nv/2.0,
          gcrush+sgpe2*nv2/2.0-sgpe2/2.0,sgpe,-sgpe2,v6,v12);      

      zero_all_gradients(); /* double check that all gradients are zero */

      ifzero(v9);   /* steady state stuff (part 3) */
      elsenz(v9);
        sub(v6,one,v6);
      endif(v9);

      /*******************************************************
      * Optional RF spoiling.  REMINDER: Do not place any events
      * after this block or before poffset at the start of peloop.
      *******************************************************/
      if ((rfspoil[0] == 'y' || rfspoil[0] == 'Y') && nv > 0) {
        voffset(0,v6);
        delay(RFSPOILDELAY - OFFSETDELAY);
      }

     endpeloop(seqcon[2],v6);
    endpeloop(seqcon[3],v12);
}
