/*==========================================================================*/
double create_venogram_slice_flow_comp(obj_grad,fstem,obj_duration,obj_ramp,
   max_grad,max_slew,seq_time)
/*==========================================================================*/
/*   
     Routine to calculate the optimum flow compensation scheme for the
     slice direction of a gradient echo sequence. The routine takes an
     object slice selection gradient (usually the second half of the slice
     select gradient waveform) as an argument, and also the maximum
     gradient strength allowed (in units of g/cm), and the maximum allowable
     slew rate (max_slew in units of usec per g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. The waveform is calculated to have an integer number
     of steps of resolution GRAD_UPDATE_TIME microseconds.

                                                  Delta2
                                                  ->  <-
                                              2*delta
                                            ->      <-
                                  obj_ramp          __
                                   -> <-           /  \
          _ _ _ _  _ _ ______________             /gmax\
     ____/              obj_grad     \           /      \_____________
                       <------------> \         /   ->   <-
                        obj_duration   \  gmax /     delta
                                        \_____/
                                    ->   <-
                                     delta
                                         <--->
                                         Delta1
     INPUTS:
     =======

     obj_grad           gradient strength of the gradient to be flow
                        compensated in g/cm (normally this would be the
                        second half of the slice select gradient).

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "gssfc")

     obj_duration       duration of flat part of the gradient waveform to
                        be compensated (normally this would be the flat top
                        portion of the second half of the slice select
                        gradient).

     obj_ramp           duration of ramp down part of the gradient waveform
                        to be compensated (normally this would be the flat top
                        portion of the second half of the slice select
                        gradient).

     max_grad           maximum gradient strength allowed for the flow
                        compensation gradients.
     
     max_slew           maximum gradient slew rate of the flow compensation
                        gradient waveform in sec per g/cm (i.e., grise/gmax).

     &seq_time          seq_time gets replaced with the duration of the
                        waveform (in secs).

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of max_grad in case it is adjusted down by the code.
     seq_time is also replaced with the actual sequence_time (the value
     0.0 can be supplied as an argument).
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
     adapted from create_ge_slice_flow_comp in March 2000
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_ramp;
double max_grad,max_slew,*seq_time;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points;
  double delta, delta1, delta2;
  double standard_flow_comp_duration();
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  if (obj_grad == 0.0) {
    printf("Error: zero gradient value passed to create_ge_slice_flow_comp\n");
    exit(0);
  }  

  delta2=-1.0;
  while (delta2 == -1.0) {
    delta2=standard_flow_comp_duration(obj_grad,obj_duration,obj_ramp,
      fabs(max_grad),max_slew);
    /* keep reducing the maximum gradient until delay delta2 becomes +ive */
    if (delta2 == -1.0) max_grad=max_grad*0.9;
  }
  delta1=delta2+obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  /* round delta1 and delta2 off to the nearest GRAD_UPDATE_TIME */
  delta1=(float)((int)(delta1*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;
  delta2=(float)((int)(delta2*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;

  /* round off flow comp ramp time (delta) to the nearest GRAD_UPDATE_TIME */
  delta=(float)((int)(fabs(max_grad)*max_slew*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=delta*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* firstly write out the longer negative flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta1*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac)); 
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac), 
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* secondly write out the shorter positive flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta2*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac)); 
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac), 
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255); 

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  *seq_time=(4.0*delta+delta1+delta2);

  return(max_grad);
 
}

/*==========================================================================*/
double create_venogram_read_flow_comp(obj_grad,fstem,obj_duration,obj_duration2,
   obj_ramp,max_grad,max_slew,seq_time)
/*==========================================================================*/
/*   
     Routine to calculate the optimum flow compensation scheme for the
     read direction of a gradient echo sequence. The routine takes an
     object read direction gradient (usually the first half of the read out
     acquisition gradient waveform) as an argument, and also the maximum
     gradient strength allowed (in units of g/cm), and the maximum allowable
     slew rate (max_slew in units of usec per g/cm). The routine writes
     out a waveform file into the ~/vnmrsys/shapelib directory with the
     name {fstem}.GRD. The waveform is calculated to have an integer number
     of steps of resolution GRAD_UPDATE_TIME microseconds.

                  Delta2
                  ->  <-
                     2*delta
                    ->      <-
                    __           obj_ramp
                   /  \           -> <-
                  /gmax\             ___________ _ _ _ _ _ _
           ______/      \           /  obj_grad             \___
               ->   <-   \         / 
                delta     \  gmax /  <---------><----------> 
                           \_____/  obj_duration obj_duration2
                      ->   <-
                       delta
                            <--->
                            Delta1


     INPUTS:
     =======

     obj_grad           gradient strength of the gradient to be flow
                        compensated in g/cm (normally this would be the
                        first half of the read out gradient). The ramp
                        down portion of the object gradient is also given
                        the same ramp time.

     fstem              string variable containing the filestem of the
                        desired waveform (e.g., "grofc")

     obj_duration       duration of flat part of the gradient waveform to
                        be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     obj_duration2      duration of flat part of the second half of the
                        object gradient (normally this would be the flat top
                        portion of the second half of the read out gradient).

     obj_ramp           duration of ramp up part of the gradient waveform
                        to be compensated (normally this would be the flat top
                        portion of the first half of the read out gradient).

     max_grad           maximum gradient strength allowed for the flow
                        compensation gradients.
     
     max_slew           maximum gradient slew rate of the flow compensation
                        gradient waveform in sec per g/cm (i.e., grise/gmax).

     &seq_time          seq_time gets replaced with the duration of the
                        waveform (in secs).

     OUTPUTS:
     ========
 
     The routine returns a double floating point variable containing the
     actual value of max_grad in case it is adjusted down by the code.
     seq_time is also replaced with the actual sequence_time (the value
     If the waveform will not fit into the requested time window then the
     negative duration by which the waveform overflows is returned.
     0.0 can be supplied as an argument).
 
     MODIFICATION HISTORY:
     =====================
 
     written by: P. Jezzard, Oxford, November 1997
     adapted from create_ge_slice_flow_comp in March 2000
 
                                                                            */
/*==========================================================================*/
double obj_grad,obj_duration,obj_duration2,obj_ramp;
double max_grad,max_slew,*seq_time;
char fstem[];
{
  double num_ramp_steps, temp_dac;
  double flat_top_points,rnd_up_duration;
  double delta, delta1, delta2;
  double standard_flow_comp_duration();
  char full_fname[MAXSTR], fname_ext[MAXSTR], sub_dir_name[MAXSTR];
  char *home_dir;
  int i, num_clock_blocks;
  FILE *fp;

  if (obj_grad == 0.0) {
    printf("Error: zero gradient value passed to create_ge_read_flow_comp\n");
    exit(0);
  }

  delta2=-1.0;
  while (delta2 == -1.0) {
    delta2=standard_flow_comp_duration(obj_grad,obj_duration,obj_ramp,
      fabs(max_grad),max_slew);
    /* keep reducing the maximum gradient until delay delta2 becomes +ive */
    if (delta2 == -1.0) max_grad=max_grad*0.9;
  }
  delta1=delta2+obj_grad/max_grad*(obj_duration+obj_ramp/2.0);

  /* round delta1 and delta2 off to the nearest GRAD_UPDATE_TIME */
  delta1=(float)((int)(delta1*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;
  delta2=(float)((int)(delta2*1.0e6/GRAD_UPDATE_TIME+0.5))*
        GRAD_UPDATE_TIME/1.0e6;

  /* round off flow comp ramp time (delta) to the nearest GRAD_UPDATE_TIME */
  delta=(float)((int)(fabs(max_grad)*max_slew*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  /* round off obj_ramp time to the nearest GRAD_UPDATE_TIME */
  obj_ramp=(float)((int)(obj_ramp*1.0e6/GRAD_UPDATE_TIME+0.5))*
     GRAD_UPDATE_TIME/1.0e6;
     
  num_ramp_steps=delta*1.0e6/GRAD_UPDATE_TIME;

  home_dir = (char *) getenv("HOME");
  strcpy(full_fname, home_dir);
  if (full_fname == "") {
    printf("Error determining HOME environment variable\n");
    exit(0);
  }
  sprintf(sub_dir_name,"/vnmrsys/shapelib/");
  strcat(full_fname, sub_dir_name);
  strcat(full_fname, fstem);

  sprintf(fname_ext,".GRD");
  strcat(full_fname, fname_ext);
     
  if ((fp=fopen(full_fname, "w")) == NULL) {
    printf("Error opening output data file %s\n", full_fname);
    exit(0);
  }  

  temp_dac=0.0;

  /* round off the flat top durations to give integer GRAD_UPDATE_TIME */
  rnd_up_duration=(float)((int)((obj_duration+obj_duration2)*1.0e6/
       GRAD_UPDATE_TIME+0.5))*GRAD_UPDATE_TIME/1.0e6;

  /* write out the shorter positive flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta2*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* secondly write out the longer negative flow comp pulse */

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=delta1*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=MAX_DAC_VAL-(float)(i+1)/num_ramp_steps*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(-1.0*obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* lastly write out the read out waveform */

  num_ramp_steps=obj_ramp*1.0e6/GRAD_UPDATE_TIME;

  /* attack portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=(float)(i+1)/num_ramp_steps*fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  /* flat top portion of trapezoid */
  flat_top_points=rnd_up_duration*1e6/GRAD_UPDATE_TIME;
  /* correct for a bug in the waveform memory calculation which doesn't
     allow for more than 255 clock cycles of a given DAC level */
  num_clock_blocks=(int)((flat_top_points+1.e-6)/255.0)+1;
  for (i=0; i<(num_clock_blocks-1); i++) {
    fprintf(fp, "%d   255\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
  fprintf(fp, "%d   %d\n",(int)(obj_grad/fabs(obj_grad)*temp_dac),
      (int)(flat_top_points+1.e-6)-(num_clock_blocks-1)*255);

  /* descent portion of trapezoid */
  for (i=0; i<(int)(num_ramp_steps+1.e-6); i++) {
    temp_dac=0.0;  /* default */
    temp_dac=fabs(obj_grad)/max_grad*MAX_DAC_VAL-(float)(i+1)/num_ramp_steps* 
           fabs(obj_grad)/max_grad*MAX_DAC_VAL;
    fprintf(fp, "%d   1\n",(int)(obj_grad/fabs(obj_grad)*temp_dac));
  }
    
  fprintf(fp, "\n");  /* end file with a carriage return */
    
  fclose(fp);
 
  *seq_time=(4.0*delta+delta1+delta2+2.0*obj_ramp)+rnd_up_duration;

  return(max_grad);
 
}
